/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef PCL_TRANSFORM_H_
#define PCL_TRANSFORM_H_

#include <Eigen3/Geometry>

// Define some libeigen based transformation method
namespace pcl
{
  /** Get the unique 3D rotation that will rotate z_axis into (0,0,1)
   *  and y_direction into a vector with x=0 (or into (0,1,0) should y_direction be orthogonal to z_axis) */
  inline void getTransformationFromTwoUnitVectors(const Eigen3::Vector3f& y_direction, const Eigen3::Vector3f& z_axis, Eigen3::Affine3f& transformation);

  /** Same as above */
  inline Eigen3::Affine3f getTransformationFromTwoUnitVectors(const Eigen3::Vector3f& y_direction, const Eigen3::Vector3f& z_axis);

  /** Same as above, but defining a complete tranformation, not only a rotation */
  inline void getTransformationFromTwoUnitVectorsAndOrigin(const Eigen3::Vector3f& y_direction, const Eigen3::Vector3f& z_axis,
                                                           const Eigen3::Vector3f& origin, Eigen3::Affine3f& transformation);

  /** Output operator for Tranform3f */
  inline std::ostream& operator<<(std::ostream& os, const Eigen3::Affine3f& m) {
    os << "{("<<m(0,0)<<","<<m(0,1)<<","<<m(0,2)<<","<<m(0,3)
       << ")("<<m(1,0)<<","<<m(1,1)<<","<<m(1,2)<<","<<m(1,2)
       << ")("<<m(2,0)<<","<<m(2,1)<<","<<m(2,2)<<","<<m(2,3)
       << ")("<<m(3,0)<<","<<m(3,1)<<","<<m(3,2)<<","<<m(3,3)<<")}";
    return os;
  }

  /** Get only the rotation part of the transformation */
  inline Eigen3::Affine3f getRotation(const Eigen3::Affine3f& transformation);

  /** Get only the translation part of the transformation */
  inline Eigen3::Vector3f getTranslation(const Eigen3::Affine3f& transformation);

  /** Get the inverse of an Eigen3::Affine3f object */
  inline void getInverse(const Eigen3::Affine3f& transformation, Eigen3::Affine3f& inverse_transformation);

  /** Same as above */
  inline Eigen3::Affine3f getInverse(const Eigen3::Affine3f& transformation);

  /** Transform a point with members x,y,z */
  template <typename PointType>
  inline PointType transformXYZ(const Eigen3::Affine3f& tranformation, const PointType& point);
  
  //! Transforms each point in the given point cloud according to the given transformation
  template <typename PointCloudType>
  inline void getTransformedPointCloud(const PointCloudType& input, const Eigen3::Affine3f& transformation, PointCloudType& output);
  
  //! Extract the Euler angles (XYZ-convention) from the given transformation
  inline void getEulerAngles(const Eigen3::Affine3f& t, float& roll, float& pitch, float& yaw);
  //! Creates a transformation from the given translation and Euler angles (XYZ-convention)
  inline void getTransformation(float x, float y, float z, float roll, float pitch, float yaw, Eigen3::Affine3f& t);
  //! Same as above
  inline Eigen3::Affine3f getTransformation(float x, float y, float z, float roll, float pitch, float yaw);

}  // namespasce end

#include "pcl/common/transform.hpp"

#endif  //#ifndef PCL_NORMS_H_
