/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

/**
\author Bastian Steder
**/


#ifndef PCL_RANGE_IMAGE_BORDER_EXTRACTOR_H_
#define PCL_RANGE_IMAGE_BORDER_EXTRACTOR_H_

//#include <pcl/common/piecewise_linear_function.h>

namespace pcl
{

// FORWARD DECLARATIONS:
class RangeImage;
class BorderDescription;
template <typename PointType>
class PointCloud;

/** @b Extract obstacle borders from range images, meaning positions where there is a transition from foreground to background.
  * \author Bastian Steder
  */
class RangeImageBorderExtractor
{
  public:
    // =====PUBLIC STRUCTS=====
    //! Stores some information extracted from the neighborhood of a point
    struct LocalSurface
    {
      Eigen3::Vector3f normal;
      Eigen3::Vector3f neighborhood_mean;
      Eigen3::Vector3f eigen_values;
      Eigen3::Vector3f normal_no_jumps;
      Eigen3::Vector3f neighborhood_mean_no_jumps;
      Eigen3::Vector3f eigen_values_no_jumps;
      float max_neighbor_distance_squared;
    };
    //! Stores the indices of the shadow border corresponding to obstacle borders
    struct ShadowBorderIndices {
      ShadowBorderIndices() : left(-1), right(-1), top(-1), bottom(-1) {}
      int left, right, top, bottom;
    };
    //! Parameters used in this class
    struct Parameters
    {
      Parameters() : pixel_radius_borders(3), pixel_radius_plane_extraction(2), pixel_radius_border_direction(2), 
                     minimum_border_probability(0.8), pixel_radius_principal_curvature(2) {}
      int pixel_radius_borders;
      int pixel_radius_plane_extraction;
      int pixel_radius_border_direction;
      float minimum_border_probability;
      int pixel_radius_principal_curvature;
    };
    
    // =====STATIC METHODS=====
    //! Take the information from BorderTraits to calculate the local direction of the border
    static float getObstacleBorderAngle(const BorderTraits& border_traits);
    
    // =====CONSTRUCTOR & DESTRUCTOR=====
    /** Constructor */
    RangeImageBorderExtractor(const RangeImage* range_image=NULL);
    /** Destructor */
    ~RangeImageBorderExtractor();
    
    // =====METHODS=====
    //! Setter for the range image pointer
    void setRangeImage(const RangeImage* range_image);
    
    //! Erase all data calculated for the current range image
    void clearData();
     
    /** Get the 2D directions in the range image from the border directions - probably mainly useful for visualization */
    float* getAnglesImageForBorderDirections();
    /** Get the 2D directions in the range image from the surface change directions - probably mainly useful for visualization */
    float* getAnglesImageForSurfaceChangeDirections();
    
    /** Get the border probability for a pixel regarding the given neighbor direction.
     *  It returns the probability of the pixel being an obstacle border or the negative probability if it is a shadow border.
     *  If min_border_probability is given, the function might return early (and return 0) if it's clear that the probability
     *  is lower than min_border_probability. */
    //float getBorderProbability(const RangeImage& range_image, int x, int y, int neighbor_offset_x, int neighbor_offset_y,
                               //float max_impact_angle, float max_surface_change_angle, float min_border_probability=0.0f) const;
    
    // =====GETTER=====
    Parameters& getParameters() { return parameters_; }
    const RangeImage& getRangeImage() const { return *range_image_; }
    float* getBorderScoresLeft()   { extractBorderScoreImages(); return border_scores_left_; }
    float* getBorderScoresRight()  { extractBorderScoreImages(); return border_scores_right_; }
    float* getBorderScoresTop()    { extractBorderScoreImages(); return border_scores_top_; }
    float* getBorderScoresBottom() { extractBorderScoreImages(); return border_scores_bottom_; }
    LocalSurface** getSurfaceStructure() { extractLocalSurfaceStructure(); return surface_structure_; }
    PointCloud<BorderDescription>& getBorderDescriptions() { classifyBorders(); return *border_descriptions_; }
    ShadowBorderIndices** getShadowBorderInformations() { findAndEvaluateShadowBorders(); return shadow_border_informations_; }
    Eigen3::Vector3f** getBorderDirections() { calculateBorderDirections(); return border_directions_; }
    float* getSurfaceChangeScores() { calculateSurfaceChanges(); return surface_change_scores_; }
    Eigen3::Vector3f* getSurfaceChangeDirections() { calculateSurfaceChanges(); return surface_change_directions_; }
    
    
  protected:
    // =====PROTECTED MEMBER VARIABLES=====
    Parameters parameters_;
    const RangeImage* range_image_;
    float* border_scores_left_, * border_scores_right_, * border_scores_top_, * border_scores_bottom_;
    LocalSurface** surface_structure_;
    PointCloud<BorderDescription>* border_descriptions_;
    ShadowBorderIndices** shadow_border_informations_;
    Eigen3::Vector3f** border_directions_;
    
    float* surface_change_scores_;
    Eigen3::Vector3f* surface_change_directions_;
    
    
    // =====PROTECTED METHODS=====
    //! Calculates a border score based on how distant the neighbor is, compared to the closest neighbors
    inline float getNeighborDistanceChangeScore(const LocalSurface& local_surface, int x, int y, int offset_x, int offset_y, int pixel_radius=1) const;
    
    //! Calculates a border score based on how much the neighbor is away from the local surface plane
    inline float getNormalBasedBorderScore(const LocalSurface& local_surface, int x, int y, int offset_x, int offset_y) const;
    
    //! Find the best corresponding shadow border and lower score according to the shadow borders value
    inline bool changeScoreAccordingToShadowBorderValue(int x, int y, int offset_x, int offset_y, float* border_scores,
                                                 float* border_scores_other_direction, int& shadow_border_idx) const;
    
    //! Returns a new score for the given pixel that is >= the original value, based on the neighbors values
    inline float updatedScoreAccordingToNeighborValues(int x, int y, const float* border_scores) const;
    //! Does the above for all pixels
    float* updatedScoresAccordingToNeighborValues(const float* border_scores) const;
    //! Replaces all border score values with updates according to the above function
    void updateScoresAccordingToNeighborValues();
    
    //! Checks if a potential border point has a corresponding shadow border
    inline bool checkPotentialBorder(int x, int y, int offset_x, int offset_y,
                              float* border_scores_left, float* border_scores_right, int& shadow_border_idx) const;
    //! Checks if a potential border point is a maximum reagarding the border score
    inline bool checkIfMaximum(int x, int y, int offset_x, int offset_y, float* border_scores, int shadow_border_idx) const;
    
    //! Find the best corresponding shadow border and lower score according to the shadow borders value
    void findAndEvaluateShadowBorders();
    
    //! Extract local plane information in every point (see getSurfaceStructure())
    void extractLocalSurfaceStructure();
    
    //! Get images representing the probability that the corresponding pixels are borders in that direction (see getBorderScores...())
    void extractBorderScoreImages();
    
    /** Classifies the pixels in the range image according to the different classes defined below in enum BorderClass.
     *  minImpactAngle (in radians) defines how flat the angle at which a surface was seen can be. */
    void classifyBorders();
    
    //! Calculate the 3D direction of the border just using the border traits at this position (facing away from the obstacle)
    inline void calculateBorderDirection(int x, int y);
    
    //! Call the above function for every point and average the result over parameters_.pixel_radius_border_direction
    void calculateBorderDirections();
    
    //! Calculate a 3d direction from a border point by projecting the direction in the range image - returns false if direction could not be calculated
    inline bool get3dDirection(const BorderDescription& border_description, Eigen3::Vector3f& direction, const LocalSurface* local_surface=NULL);
    
    //! Calculate the main principal curvature (the largest eigenvalue and corresponding eigenvector for the normals in the area) in the given point
    bool calculateMainPrincipalCurvature(int x, int y, int radius, int step_size, float& magnitude, Eigen3::Vector3f& main_direction) const;
    
    /** Uses either the border or principal curvature to define a score how much the surface changes in a point (1 for a border)
      * and what the main direction of that change is */
    void calculateSurfaceChanges();

    /** Puts a blur on the surface change images */
    void blurSurfaceChanges();
    
    //! Transform the surface change directions to face in the positive coordinate system quadrant (meaning the 2d projection angles wil be in [0, 90]deg.
    //void unifySurfaceChangeDirections();
    
    //float getImpactAngleBorderProbability(const RangeImage& range_image, int x1, int y1, int x2, int y2, float max_impact_angle) const;
    //float getSurfaceChangeBorderProbability(const RangeImage& range_image, int x, int y, int offset_x, int offset_y, float max_surface_change_angle) const;
    //float getNormalChangeBorderProbability(const RangeImage& range_image, int radius, int x1, int y1, int x2, int y2) const;
    //float getBorderProbability(const RangeImage& range_image, int x, int y, int offset_x, int offset_y,
                               //const float* acuteness_value_image_x, const float* acuteness_value_image_y) const;
    //float getDistanceChangeScore(const RangeImage& range_image, int x, int y, int offset_x, int offset_y, int max_pixel_distance=0) const;
    //float getPrecalculatedAcuteness(int x, int y, int offset_x, int offset_y, int width, int height,
                                    //const float* acuteness_value_image_x, const float* acuteness_value_image_y) const;
    //float getAverageAcuteness(const RangeImage& range_image, int x, int y, int offset_x, int offset_y, float acuteness_in_pixel_distance) const;
    //void updateStatistic(float acuteness, float acuteness_change, bool is_border) const;
    //void outputStatistic() const;
};

}  // namespace end

#include "pcl/features/range_image_border_extractor.hpp"  // Definitions of templated and inline functions

#endif  //#ifndef PCL_RANGE_IMAGE_BORDER_EXTRACTOR_H_
