/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: voxel_grid.h 32774 2010-09-20 12:08:49Z lorenz $
 *
 */

#ifndef PCL_FILTERS_VOXEL_GRID_MAP_H_
#define PCL_FILTERS_VOXEL_GRID_MAP_H_

#include "pcl/common/common.h"
#include "pcl/filters/filter.h"

namespace pcl
{
    void getMinMax3D (const sensor_msgs::PointCloud2ConstPtr &cloud, int x_idx, int y_idx, int z_idx, Eigen3::Vector4f &min_pt, Eigen3::Vector4f &max_pt);

  void getMinMax3D (const sensor_msgs::PointCloud2ConstPtr &cloud, int x_idx, int y_idx, int z_idx, const std::string &distance_field_name, float min_distance, float max_distance, Eigen3::Vector4f &min_pt, Eigen3::Vector4f &max_pt, bool limit_negative = false);

  template <typename PointT> inline void getMinMax3D (const typename pcl::PointCloud<PointT>::ConstPtr &cloud, const std::string &distance_field_name, float min_distance, float max_distance, Eigen3::Vector4f &min_pt, Eigen3::Vector4f &max_pt, bool limit_negative = false);

  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief Helper functor structure for copying data between an Eigen3::VectorXf and a PointT. */
  template <typename PointT>
    struct NdCopyEigenPointFunctor
  {
    typedef typename traits::POD<PointT>::type Pod;
    
    NdCopyEigenPointFunctor (const Eigen3::VectorXf &p1, PointT &p2)
      : p1_ (p1),
        p2_ (reinterpret_cast<Pod&>(p2)),
        f_idx_ (0) { }

    template<typename Key> inline void operator() () { boost::fusion::at_key<Key> (p2_) = p1_[f_idx_++]; }

    private:
      const Eigen3::VectorXf &p1_;
      Pod &p2_;
      int f_idx_;
  };

  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief Helper functor structure for copying data between an Eigen3::VectorXf and a PointT. */
  template <typename PointT>
    struct NdCopyPointEigenFunctor
  {
    typedef typename traits::POD<PointT>::type Pod;
    
    NdCopyPointEigenFunctor (const PointT &p1, Eigen3::VectorXf &p2)
      : p1_ (reinterpret_cast<const Pod&>(p1)), p2_ (p2), f_idx_ (0) { }

    template<typename Key> inline void operator() () { p2_[f_idx_++] = boost::fusion::at_key<Key> (p1_); }

    private:
      const Pod &p1_;
      Eigen3::VectorXf &p2_;
      int f_idx_;
  };

  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief @b VoxelGrid assembles a local 3D grid over a given PointCloud, and downsamples + filters the data.
    * \author Radu Bogdan Rusu, Bastian Steder
    */
  template <typename PointT>
  class VoxelGrid: virtual public Filter<PointT>
  {
    using Filter<PointT>::filter_name_;
    using Filter<PointT>::getClassName;
    using Filter<PointT>::input_;
    using Filter<PointT>::indices_;
    using Filter<PointT>::filter_limit_negative_;
    using Filter<PointT>::filter_limit_min_;
    using Filter<PointT>::filter_limit_max_;
    using Filter<PointT>::filter_field_name_;

    typedef typename Filter<PointT>::PointCloud PointCloud;
    typedef typename PointCloud::Ptr PointCloudPtr;
    typedef typename PointCloud::ConstPtr PointCloudConstPtr;

    public:
      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Empty constructor. */
      VoxelGrid () : downsample_all_data_ (true)
      {
        leaf_size_.setZero ();
        filter_name_ = "VoxelGrid";
      }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Set the voxel grid leaf size.
        * \param leaf_size the voxel grid leaf size
        */
      inline void setLeafSize (const Eigen3::Vector4f &leaf_size) { leaf_size_ = leaf_size; }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Set the voxel grid leaf size.
        * \param lx the leaf size for X
        * \param ly the leaf size for Y
        * \param lz the leaf size for Z
        */
      inline void
        setLeafSize (float lx, float ly, float lz)
      {
        leaf_size_[0] = lx; leaf_size_[1] = ly; leaf_size_[2] = lz;
      }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Get the voxel grid leaf size. */
      inline Eigen3::Vector3f getLeafSize () { return (leaf_size_.head<3> ()); }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Set to true if all fields need to be downsampled, or false if just XYZ.
        * \param downsample the new value (true/false)
        */
      inline void setDownsampleAllData (bool downsample) { downsample_all_data_ = downsample; }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Get the state of the internal downsampling parameter (true if all fields need to be downsampled, false
       * if just XYZ). */
      inline bool getDownsampleAllData () { return (downsample_all_data_); }

    protected:
      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Simple structure to hold an nD centroid and the number of points in a leaf. */
      struct Leaf
      {
        Leaf() : nr_points(0) {}
        Eigen3::VectorXf centroid;    // @todo we do not support FLOAT64 just yet due to memory issues. Need to fix this.
        int nr_points;
      };

      /** \brief The 3D grid leaves. */
      std::map<size_t, Leaf> leaves_;

      /** \brief The size of a leaf. */
      Eigen3::Vector4f leaf_size_;

      /** \brief Set to true if all fields need to be downsampled, or false if just XYZ. */
      bool downsample_all_data_;

      typedef typename pcl::traits::fieldList<PointT>::type FieldList;

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Downsample a Point Cloud using a voxelized grid approach
        * \param output the resultant point cloud message
        */
      void applyFilter (PointCloud &output);
  };

  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief @b VoxelGrid assembles a local 3D grid over a given PointCloud, and downsamples + filters the data.
    * \author Radu Bogdan Rusu, Bastian Steder
    */
  template <>
  class VoxelGrid<sensor_msgs::PointCloud2> : virtual public Filter<sensor_msgs::PointCloud2>
  {
    using Filter<sensor_msgs::PointCloud2>::filter_name_;
    using Filter<sensor_msgs::PointCloud2>::getClassName;

    typedef sensor_msgs::PointCloud2 PointCloud;
    typedef PointCloud::Ptr PointCloudPtr;
    typedef PointCloud::ConstPtr PointCloudConstPtr;

    public:
      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Empty constructor. */
      VoxelGrid () : downsample_all_data_ (true)
      {
        leaf_size_.setZero ();
        filter_name_ = "VoxelGrid";
      }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Set the voxel grid leaf size.
        * \param leaf_size the voxel grid leaf size
        */
      inline void setLeafSize (const Eigen3::Vector4f &leaf_size) { leaf_size_ = leaf_size; }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Set the voxel grid leaf size.
        * \param lx the leaf size for X
        * \param ly the leaf size for Y
        * \param lz the leaf size for Z
        */
      inline void
        setLeafSize (float lx, float ly, float lz)
      {
        leaf_size_[0] = lx; leaf_size_[1] = ly; leaf_size_[2] = lz;
      }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Get the voxel grid leaf size. */
      inline Eigen3::Vector3f getLeafSize () { return (leaf_size_.head<3> ()); }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Set to true if all fields need to be downsampled, or false if just XYZ.
        * \param downsample the new value (true/false)
        */
      inline void setDownsampleAllData (bool downsample) { downsample_all_data_ = downsample; }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Get the state of the internal downsampling parameter (true if all fields need to be downsampled, false
       * if just XYZ). */
      inline bool getDownsampleAllData () { return (downsample_all_data_); }

    protected:
      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Simple structure to hold an nD centroid and the number of points in a leaf. */
      struct Leaf
      {
        Leaf() : nr_points(0) { }
        Eigen3::VectorXf centroid;    // @todo we do not support FLOAT64 just yet due to memory issues. Need to fix this.
        int nr_points;
      };

      /** \brief The 3D grid leaves. */
      std::map<size_t, Leaf> leaves_;

      /** \brief The size of a leaf. */
      Eigen3::Vector4f leaf_size_;

      /** \brief Set to true if all fields need to be downsampled, or false if just XYZ. */
      bool downsample_all_data_;

      void applyFilter (PointCloud &output);
  };
}

#include "pcl/filters/voxel_grid.hpp"

#endif  //#ifndef PCL_FILTERS_VOXEL_GRID_MAP_H_
