/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: pcd_io.h 32297 2010-09-02 06:18:50Z rusu $
 *
 */

#ifndef PCL_IO_PCD_IO_H_
#define PCL_IO_PCD_IO_H_

#include <pcl/pcl_base.h>
#include "pcl/io/io.h"

namespace pcl
{
  namespace io
  {
    int loadPCDFileHeader (const std::string &file_name, sensor_msgs::PointCloud2 &cloud);
    int loadPCDFile (const std::string &file_name, sensor_msgs::PointCloud2 &cloud);
    int savePCDFileASCII (const std::string &file_name, const sensor_msgs::PointCloud2 &cloud, int precision = 5);
    int savePCDFileBinary (const std::string &file_name, const sensor_msgs::PointCloud2 &cloud);

    int savePCDFile (std::string file_name, const sensor_msgs::PointCloud2 &cloud, bool binary_mode = false);

    std::string generateHeader (const sensor_msgs::PointCloud2 &cloud);

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /** \brief Load a PCD file into a templated PointCloud type
      * \param file_name the name of the file to load
      * \param cloud the resultant templated point cloud
      */
    template<typename PointT> int
      loadPCDFile (const std::string &file_name, pcl::PointCloud<PointT> &cloud)
    {
      sensor_msgs::PointCloud2 cloud_blob;
      int res = loadPCDFile (file_name, cloud_blob);
      // Exit in case of error
      if (res == -1)
        return (res);
      pcl::fromROSMsg (cloud_blob, cloud);
      return (res);
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /** \brief Templated version for saving point cloud data to a PCD file containing a specific given cloud format in
      * binary format
      * \param file_name the output file name
      * \param cloud the point cloud data message
      */
    template<typename PointT> int
      savePCDFileBinary (const std::string &file_name, const pcl::PointCloud<PointT> &cloud)
    {
      sensor_msgs::PointCloud2 cloud_blob;
      pcl::toROSMsg (cloud, cloud_blob);
      return (savePCDFileBinary (file_name, cloud_blob));
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /** \brief Templated version for saving point cloud data to a PCD file containing a specific given cloud format in
      * ASCII format
      * \param file_name the output file name
      * \param cloud the point cloud data message
      * \param precision the desired output precision
      */
    template<typename PointT> int
      savePCDFileASCII (const std::string &file_name, const pcl::PointCloud<PointT> &cloud, int precision = 5)
    {
      sensor_msgs::PointCloud2 cloud_blob;
      pcl::toROSMsg (cloud, cloud_blob);
      return (savePCDFileASCII (file_name, cloud_blob, precision));
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /** \brief Templated version for saving point cloud data to a PCD file containing a specific given cloud format
      * \param file_name the output file name
      * \param cloud the point cloud data message
      * \param binary_mode true for binary mode, false for ascii
      */
    template<typename PointT> int
      savePCDFile (const std::string &file_name, const pcl::PointCloud<PointT> &cloud, bool binary_mode = false)
    {
      sensor_msgs::PointCloud2 cloud_blob;
      pcl::toROSMsg (cloud, cloud_blob);
      if (binary_mode)
        return (savePCDFileBinary (file_name, cloud_blob));
      else
        return (savePCDFileASCII (file_name, cloud_blob, 5));
    }


    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /** \brief Templated version for saving point cloud data to a PCD file containing a specific given cloud format in
      * binary format using a set of given point indices
      * \param file_name the output file name
      * \param cloud the point cloud data message
      * \param indices the set of indices to save
      */
    template<typename PointT> int
      savePCDFileBinary (const std::string &file_name, const pcl::PointCloud<PointT> &cloud, const std::vector<int> &indices)
    {
      // Copy indices to a new point cloud
      pcl::PointCloud<PointT> cloud_out;
      copyPointCloud (cloud, indices, cloud_out);
      // Save the data
      sensor_msgs::PointCloud2 cloud_blob;
      pcl::toROSMsg (cloud_out, cloud_blob);
      return (savePCDFileBinary (file_name, cloud_blob));
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /** \brief Templated version for saving point cloud data to a PCD file containing a specific given cloud format in
      * ASCII format using a set of given point indices
      * \param file_name the output file name
      * \param cloud the point cloud data message
      * \param indices the set of indices to save
      * \param precision the desired output precision
      */
    template<typename PointT> int
      savePCDFileASCII (const std::string &file_name, const pcl::PointCloud<PointT> &cloud,
                        const std::vector<int> &indices, int precision = 5)
    {
      // Copy indices to a new point cloud
      pcl::PointCloud<PointT> cloud_out;
      copyPointCloud (cloud, indices, cloud_out);
      // Save the data
      sensor_msgs::PointCloud2 cloud_blob;
      pcl::toROSMsg (cloud_out, cloud_blob);
      return (savePCDFileASCII (file_name, cloud_blob, precision));
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /** \brief Templated version for saving point cloud data to a PCD file containing a specific given cloud format
      * \param file_name the output file name
      * \param cloud the point cloud data message
      * \param indices the set of indices to save
      * \param binary_mode true for binary mode, false for ascii
      */
    template<typename PointT> int
      savePCDFile (const std::string &file_name, const pcl::PointCloud<PointT> &cloud,
                   const std::vector<int> &indices, bool binary_mode = false)
    {
      // Copy indices to a new point cloud
      pcl::PointCloud<PointT> cloud_out;
      copyPointCloud (cloud, indices, cloud_out);
      // Save the data
      sensor_msgs::PointCloud2 cloud_blob;
      pcl::toROSMsg (cloud_out, cloud_blob);
      if (binary_mode)
        return (savePCDFileBinary (file_name, cloud_blob));
      else
        return (savePCDFileASCII (file_name, cloud_blob, 5));
    }

  }

  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief Point Cloud Data (PCD) file format reader.
    * \author Radu Bogdan Rusu
    */
  class PCDReader: virtual public PCLBase<sensor_msgs::PointCloud2>
  {
    public:
      typedef sensor_msgs::PointCloud2 PointCloud;
      typedef PointCloud::Ptr PointCloudPtr;
      typedef PointCloud::ConstPtr PointCloudConstPtr;

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Empty constructor. */
      PCDReader () : output_ () {};

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Empty destructor. */
      virtual ~PCDReader () {};

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Get a pointer to the input point cloud dataset. */
      inline PointCloudConstPtr getInputCloud () { return (boost::make_shared <PointCloud> (*output_)); }

      bool read (const std::string &file_name, PointCloud &output);

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Read the actual data from disk.
        * \param file_name the name of the file containing the actual PointCloud data
        * \param output the resultant PointCloud message read from disk
        */
      template<typename PointT> inline bool
        read (const std::string &file_name, pcl::PointCloud<PointT> &output)
      {
        // Load the data
        if (io::loadPCDFile<PointT> (file_name, output) == -1)
        {
          ROS_ERROR ("[pcl::PCDReader::read] Failed to load a valid PointCloud from %s.", file_name.c_str ());
          return (false);
        }
        return (true);
      }

    protected:
      /** \brief The name of the file that contains the PointCloud data. */
      std::string file_name_;

      /** \brief The output point cloud dataset containing the points loaded from the file. */
      PointCloudPtr output_;
  };

  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief Point Cloud Data (PCD) file format writer.
    * \author Radu Bogdan Rusu
    */
  class PCDWriter: virtual public PCLBase<sensor_msgs::PointCloud2>
  {
    public:
      typedef sensor_msgs::PointCloud2 PointCloud;
      typedef PointCloud::Ptr PointCloudPtr;
      typedef PointCloud::ConstPtr PointCloudConstPtr;

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Empty constructor. */
      PCDWriter () : file_name_ (""), binary_mode_ (true), input_ () {};

      void write (const std::string &file_name, const PointCloudConstPtr &cloud, bool binary);
      void write (const std::string &file_name, const PointCloud &cloud, bool binary);

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Write the actual data to disk.
        * \param file_name the name of the output file that will contain the actual PointCloud data
        * \param cloud the PointCloud message to be written to disk
        * \param binary set to true if the file is to be written in a binary PCD format, false for ASCII
        */
      template<typename PointT> inline void
        write (const std::string &file_name, const pcl::PointCloud<PointT> &cloud, bool binary)
      {
        // Save the data
        io::savePCDFile<PointT> (file_name, cloud, binary);
      }

    protected:
      /** \brief The name of the file that contains the PointCloud data. */
      std::string file_name_;

      /** \brief Set to true if the output files should be saved in binary mode (true). */
      bool binary_mode_;

      /** \brief The input point cloud dataset containing the points we need to work with. */
      PointCloudConstPtr input_;
  };
}

#endif  //#ifndef PCL_IO_PCD_IO_H_
