/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: kdtree_flann.h 32573 2010-03-03 03:27:09Z rusu $
 *
 */

#ifndef PCL_KDTREE_KDTREE_FLANN_H_
#define PCL_KDTREE_KDTREE_FLANN_H_

#include <cstdio>
#include "pcl/kdtree/kdtree.h"
#include <boost/thread/mutex.hpp>
#include <flann/flann.h>

namespace pcl
{
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief @b KdTreeFLANN is a generic type of 3D spatial locator using kD-tree structures. The class is making use of
    * the FLANN (Fast Library for Approximate Nearest Neighbor) project by Marius Muja and David Lowe.
    *
    * @note libFLANN is not thread safe, so we need mutices in places to make KdTreeFLANN thread safe.
    * \author Radu Bogdan Rusu
    */
  template <typename PointT>
  class KdTreeFLANN : public KdTree<PointT>
  {
    using KdTree<PointT>::input_;
    using KdTree<PointT>::indices_;
    using KdTree<PointT>::epsilon_;
    using KdTree<PointT>::point_representation_;

    typedef typename KdTree<PointT>::PointCloud PointCloud;
    typedef typename KdTree<PointT>::PointCloudConstPtr PointCloudConstPtr;

    typedef boost::shared_ptr <std::vector<int> > IndicesPtr;
    typedef boost::shared_ptr <const std::vector<int> > IndicesConstPtr;

    public:
      // Boost shared pointers
      typedef boost::shared_ptr<KdTreeFLANN<PointT> > Ptr;
      typedef boost::shared_ptr<const KdTreeFLANN<PointT> > ConstPtr;

      //-----STATIC METHODS-----
      /** \brief Set the norm that is used for neighbor queries.
        * \param metric_type - 	EUCLIDEAN, MANHATTAN, MINKOWSKI, MAXDIST, HIK, JM, CS, KL
        *                      (see flann/general.h for the definition)
        * \param order - only needed for Minkowski metric
        */
      inline static void setDistanceMetric(flann_distance_t metric_type=EUCLIDEAN, int order=0);

      /** \brief Constructor for KdTree.
        * \note ATTENTION: This method breaks the 1-1 mapping between the indices returned using \a getNeighborsIndices
        * and the ones from the \a cloud message ! When using this method, make sure to get the underlying point data
        * using the \a getPoint method
        * param cloud the point cloud data array
        * param indices the point cloud indices
        */
//      KdTreeFLANN (const PointCloud &cloud, const std::vector<int> &indices);
//
      //! \brief Default Constructor for KdTreeFLANN.
      inline KdTreeFLANN ();

      /** \brief Destructor for KdTreeFLANN. Deletes all allocated data arrays and destroys the kd-tree structures. */
      virtual ~KdTreeFLANN ();

      /** \brief Provide a pointer to the input dataset.
        * \param cloud the const boost shared pointer to a PointCloud message
        * \param indices the point indices subset that is to be used from \a cloud - if NULL the whole cloud is used
        */
      inline void setInputCloud (const PointCloudConstPtr &cloud, const IndicesConstPtr &indices=IndicesConstPtr());

      /** \brief Search for k-nearest neighbors for the given query point.
        * \param point the given query point
        * \param k the number of neighbors to search for
        * \param k_indices the resultant point indices (must be resized to \a k a priori!)
        * \param k_distances the resultant point distances (must be resized to \a k a priori!)
        */
      inline bool nearestKSearch (const PointT &point, int k, std::vector<int> &k_indices, std::vector<float> &k_distances);

      /** \brief Search for k-nearest neighbors for the given query point.
        * \param cloud the point cloud data
        * \param index the index in \a cloud representing the query point
        * \param k the number of neighbors to search for
        * \param k_indices the resultant point indices (must be resized to \a k a priori!)
        * \param k_distances the resultant point distances (must be resized to \a k a priori!)
        */
      inline bool nearestKSearch (const PointCloud &cloud, int index, int k, std::vector<int> &k_indices, std::vector<float> &k_distances);

      /** \brief Search for k-nearest neighbors for the given query point (zero-copy).
        * \param index the index representing the query point in the dataset given by \a setInputCloud
        *        if indices were given in setInputCloud, index will be the position in the indices vector
        * \param k the number of neighbors to search for
        * \param k_indices the resultant point indices (must be resized to \a k a priori!)
        * \param k_distances the resultant point distances (must be resized to \a k a priori!)
        */
      inline bool nearestKSearch (int index, int k, std::vector<int> &k_indices, std::vector<float> &k_distances);

      /** \brief Search for all the nearest neighbors of the query point in a given radius.
        * \param p_q the given query point
        * \param radius the radius of the sphere bounding all of p_q's neighbors
        * \param k_indices the resultant point indices
        * \param k_distances the resultant point distances
        * \param max_nn if given, bounds the maximum returned neighbors to this value
        */
      inline bool radiusSearch (const PointT &point, double radius, std::vector<int> &k_indices,
                                std::vector<float> &k_distances, int max_nn = INT_MAX) const;

      /** \brief Search for all the nearest neighbors of the query point in a given radius.
        * \param cloud the point cloud data
        * \param index the index in \a cloud representing the query point
        * \param radius the radius of the sphere bounding all of p_q's neighbors
        * \param k_indices the resultant point indices
        * \param k_distances the resultant point distances
        * \param max_nn if given, bounds the maximum returned neighbors to this value
        */
      inline bool radiusSearch (const PointCloud &cloud, int index, double radius, std::vector<int> &k_indices,
                                std::vector<float> &k_distances, int max_nn = INT_MAX) const;

      /** \brief Search for all the nearest neighbors of the query point in a given radius (zero-copy).
        * \param index the index representing the query point in the dataset given by \a setInputCloud
        *        if indices were given in setInputCloud, index will be the position in the indices vector
        * \param radius the radius of the sphere bounding all of p_q's neighbors
        * \param k_indices the resultant point indices
        * \param k_distances the resultant point distances
        * \param max_nn if given, bounds the maximum returned neighbors to this value
        */
      inline bool radiusSearch (int index, double radius, std::vector<int> &k_indices,
                                std::vector<float> &k_distances, int max_nn = INT_MAX) const;


    private:
      /** \brief Internal cleanup method. */
      void cleanup ();

      /** \brief Simple initialization method for internal parameters.
        * \param nr_points the number of points in the given dataset
        */
      inline bool initParameters ();

      /** \brief Simple initialization method for internal data buffers. */
      void initData ();

      /** \brief Converts a ROS PointCloud message to the internal FLANN point array representation. Returns the number
        * of points.
        * \param ros_cloud the ROS PointCloud message
        */
      void convertCloudToArray (const PointCloud &ros_cloud);

      /** \brief Converts a ROS PointCloud message with a given set of indices to the internal FLANN point array
       * representation. Returns the number of points.
       * \note ATTENTION: This method breaks the 1-1 mapping between the indices returned using \a getNeighborsIndices
       * and the ones from the \a ros_cloud message ! When using this method, make sure to get the underlying point data
       * using the \a getPoint method
       * \param ros_cloud the ROS PointCloud message
       * \param indices the point cloud indices
       */
      void convertCloudToArray (const PointCloud &ros_cloud, const std::vector<int> &indices);

    private:
      /** \brief Class getName method. */
      virtual std::string getName () const { return ("KdTreeFLANN"); }

      mutable boost::mutex m_lock_;

      /** \brief A FL-ANN type index reference. */
      FLANN_INDEX index_id_;

      /** \brief A pointer to a FL-ANN parameter structure. */
      //FLANNParameters flann_param_;
      mutable FLANNParameters flann_param_;  // TODO - should not really be mutable - wait for fix in FLANN

      /** \brief Internal pointer to data. */
      float* cloud_;
      
      /** \brief mapping between internal and external indices. */
      vector<int> index_mapping_;

      /** \brief Tree dimensionality (i.e. the number of dimensions per point). */
      int dim_;
  };

}

#include <pcl/kdtree/kdtree_flann.hpp>

#endif
