#ifndef PCL_POINT_REPRESENTATION_H_
#define PCL_POINT_REPRESENTATION_H_

#include "pcl/point_types.h"

namespace pcl
{
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief @b PointRepresentation provides a set of methods for converting a point structs/object into an
   *  n-dimensional vector.
   *  @note This is an abstract class.  Subclasses must set nr_dimensions_ to the appropriate value in the constructor 
   *  and provide an implemention of the pure virtual copyToFloatArray method.
   */
  template <typename PointT>
    class PointRepresentation
  {
    protected:
      /** \brief The number of dimensions in this point's vector (i.e. the "k" in "k-D") */
      int nr_dimensions_;
      /** \brief A vector containing the rescale factor to apply to each dimension. */
      std::vector<float> alpha_;
      
    public:
      /** \brief Empty constructor */
      PointRepresentation() : nr_dimensions_ (0), alpha_ (0) {}
      
      /** \brief Copy point data from input point to a float array. This method must be overriden in all subclasses. 
       *  \param p The input point
       *  \param out A pointer to a float array.
       */
      virtual void copyToFloatArray (const PointT &p, float * out) const = 0;
      
      /** \brief Verify that the input point is valid.
       *  \param p The point to validate
       */
      virtual bool 
        isValid (const PointT &p) const
      {
        float temp[nr_dimensions_];
        copyToFloatArray (p, temp);
        bool is_valid = true;
        for (int i = 0; i < nr_dimensions_; ++i)
        {
          if (!std::isfinite (temp[i]))
          {
            is_valid = false;
            break;
          }
        }
        return (is_valid);
      }
      
      /** \brief Convert input point into a vector representation, rescaling by \a alpha.
       *  \param p
       *  \param out The output vector.  Can be of any type that implements the [] operator.
       */
      template <typename OutputType> void
        vectorize (const PointT &p, OutputType &out) const
      {
        float temp[nr_dimensions_];
        copyToFloatArray (p, temp);
        if (alpha_.empty ())
        {
          for (int i = 0; i < nr_dimensions_; ++i)
            out[i] = temp[i];
        }
        else
        {
          for (int i = 0; i < nr_dimensions_; ++i)
            out[i] = temp[i] * alpha_[i];
        }
      }
      
      /** \brief Set the rescale values to use when vectorizing points
       *  \param rescale_array The array/vector of rescale values.  Can be of any type that implements the [] operator.
       */
      //template <typename InputType>
      //void setRescaleValues (const InputType &rescale_array)
      void 
        setRescaleValues (const float * rescale_array)
      {
        alpha_.resize (nr_dimensions_);
        for (int i = 0; i < nr_dimensions_; ++i)
          alpha_[i] = rescale_array[i];
      }
      
      /** \brief Return the number of dimensions in the point's vector representation. */
      inline int getNumberOfDimensions () const { return (nr_dimensions_); }
  };

  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief @DefaultPointRepresentation extends PointRepresentation to define default behavior for common point types.
   */
  template <typename PointDefault>
  class DefaultPointRepresentation : public PointRepresentation <PointDefault>
  {
    using PointRepresentation <PointDefault>::nr_dimensions_;

    public:
      DefaultPointRepresentation ()
      {
        // If point type is unknown, assume it's a struct/array of floats, and compute the number of dimensions
        nr_dimensions_ = sizeof (PointDefault) / sizeof (float);
        // Limit the default representation to the first 3 elements
        if (nr_dimensions_ > 3) nr_dimensions_ = 3;
      }

      virtual void 
        copyToFloatArray (const PointDefault &p, float * out) const
      {
        // If point type is unknown, treat it as a struct/array of floats
        const float * ptr = (float *)&p;
        for (int i = 0; i < nr_dimensions_; ++i)
        {
          out[i] = ptr[i];
        }
      }
  };

  template <>
  class DefaultPointRepresentation <PointXYZ> : public  PointRepresentation <PointXYZ>
  {
    public:
      DefaultPointRepresentation ()
      {
        nr_dimensions_ = 3;
      }

      virtual void 
        copyToFloatArray (const PointXYZ &p, float * out) const
      {
        out[0] = p.x;
        out[1] = p.y;
        out[2] = p.z;
      }
  };

  template <>
  class DefaultPointRepresentation <PointXYZI> : public  PointRepresentation <PointXYZI>
  {
    public:
      DefaultPointRepresentation ()
      {
        nr_dimensions_ = 3;
      }

      virtual void 
        copyToFloatArray (const PointXYZI &p, float * out) const
      {
        out[0] = p.x;
        out[1] = p.y;
        out[2] = p.z;
        // By default, p.intensity is not part of the PointXYZI vectorization
      }
  };

  template <>
  class DefaultPointRepresentation <PointNormal> : public  PointRepresentation <PointNormal>
  {
    public:
      DefaultPointRepresentation ()
      {
        nr_dimensions_ = 3;
      }

      virtual void 
        copyToFloatArray (const PointNormal &p, float * out) const
      {
        out[0] = p.x;
        out[1] = p.y;
        out[2] = p.z;
      }
  };

  template <>
  class DefaultPointRepresentation <PFHSignature125> : public  PointRepresentation <PFHSignature125>
  {
    public:
      DefaultPointRepresentation ()
      {
        nr_dimensions_ = 125;
      }

      virtual void 
        copyToFloatArray (const PFHSignature125 &p, float * out) const
      {
        for (int i = 0; i < nr_dimensions_; ++i)
          out[i] = p.histogram[i];
      }
  };

  template <>
  class DefaultPointRepresentation <FPFHSignature33> : public  PointRepresentation <FPFHSignature33>
  {
    public:
      DefaultPointRepresentation ()
      {
        nr_dimensions_ = 33;
      }

      virtual void 
        copyToFloatArray (const FPFHSignature33 &p, float * out) const
      {
        for (int i = 0; i < nr_dimensions_; ++i)
          out[i] = p.histogram[i];
      }
  };
}

#endif // #ifndef PCL_POINT_REPRESENTATION_H_
