/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: concave_hull.h 31050 2010-07-20 21:28:57Z rusu $
 *
 */

#ifndef PCL_CONCAVE_HULL_2D_H_
#define PCL_CONCAVE_HULL_2D_H_

// PCL includes
#include <pcl/pcl_base.h>
#include "pcl/io/io.h"

#include "pcl/features/normal_3d.h"

extern "C"
{
#include <qhull/qhull_a.h>
}

namespace pcl
{
  void concaveHull2D (std::vector<coordT> &points, std::vector<Eigen3::Vector4f, Eigen3::aligned_allocator<Eigen3::Vector4f> > &hull)
  {
    // True if qhull should free points in qh_freeqhull() or reallocation
    boolT ismalloc = 0;
    // option flags for qhull, see qh_opt.htm, output volume (FA)
    char flags[] = "qhull d Tcv FA";
    // 0 if no error from qhull
    int exitcode = qh_new_qhull (2, points.size (), &points[0], ismalloc, flags, NULL, NULL);
    
    if (!exitcode)
    {
      // qh facet_list' contains the hull
      // If you want a Voronoi diagram ('v') and do not request output (i.e., outfile=NULL), call qh_setvoronoi_all() after qh_new_qhull().
      qh_setvoronoi_all ();

      facetT *facet;            // set by FORALLfacets
      FORALLfacets
      {
      }
    }

    double totvol = qh totvol;    // copy volume

    // Free memory
    qh_freeqhull (!qh_ALL);
    int curlong, totlong;                 // memory remaining after qh_memfreeshort
    qh_memfreeshort (&curlong, &totlong); // free short memory and memory allocator
    if (curlong || totlong)
      ROS_ERROR ("qhull internal warning (main): did not free %d bytes of long memory (%d pieces)\n", totlong, curlong);


  }

  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief Compute a 2D concave hull in 3D space using alpha shapes
    * \param cloud the point cloud message
    * \param indices the point indices to use from the cloud (they must form a planar model)
    * \param coeff the normalized planar model coefficients
    * \param hull the resultant concave hull model
    */
   template <typename PointT> void concaveHull2Dfrom3D (const pcl::PointCloud<PointT> &cloud, const std::vector<int> &indices, const Eigen3::Vector4f &coeff, pcl::PointCloud<PointT> &hull)
   {
   }

  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief @b ConcaveHull2D represents a 2D ConcaveHull implementation.
    * \author Radu Bogdan Rusu
    */
  template <typename PointInT, typename PointOutT>
  class ConcaveHull2D: virtual public PCLBase<PointInT>
  {
    using PCLBase<PointInT>::input_;
    using PCLBase<PointInT>::indices_;
    using PCLBase<PointInT>::initCompute;
    using PCLBase<PointInT>::deinitCompute;

    public:
      typedef pcl::PointCloud<PointInT> PointCloudIn;
      typedef typename PointCloudIn::Ptr PointCloudInPtr;
      typedef typename PointCloudIn::ConstPtr PointCloudInConstPtr;

      typedef pcl::PointCloud<PointOutT> PointCloudOut;

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Base method for surface reconstruction for all points given in <setInputCloud (), setIndices ()>
        * \param output the resultant reconstructed surface model
        */
      inline void reconstruct (PointCloudOut &output)
      {
        if (!initCompute ()) return;

        // Perform the actual surface reconstruction
        performReconstruction (output);

        output.header   = input_->header;
        output.width    = output.points.size ();
        output.height   = 1;
        output.is_dense = false;

        deinitCompute ();
      }

    protected:
      /** \brief A pointer to the vector of model coefficients. */
      ModelCoefficientsConstPtr model_;

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Abstract surface reconstruction method. */
      inline void performReconstruction (PointCloudOut &output)
      {
/*        // Compute the plane coefficients (cheaper than waiting for them on the wire)
        Eigen3::Vector4f model_coefficients;
        float curvature;
        computePointNormal<PointInT> (*input_, *indices_, model_coefficients, curvature);
        // Check if the plane coefficients were correctly estimated
        if (isnan (model_coefficients[0]) || isnan (model_coefficients[1]) || isnan (model_coefficients[2]) || isnan (model_coefficients[3]))
        {
          ROS_WARN ("[pcl::%s::performReconstruction] Plane coefficients could not be estimated (nan)!", getClassName ().c_str ());
          return;
        }
        //convexHull2Dfrom3D (*input_, *indices_, model_->values, output);
        convexHull2Dfrom3D (*input_, *indices_, model_coefficients, output);*/

      }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Abstract class get name method. */
      std::string getClassName () const { return ("ConcaveHull2D"); }
  };
}

//#include "pcl/surface/convex_hull.hpp"

#endif  //#ifndef PCL_CONCAVE_HULL_2D_H_
