/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: common.hpp 32774 2010-09-20 12:08:49Z lorenz $
 *
 */

#include <pcl/point_types.h>

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Compute the smallest angle between two vectors in the [ 0, PI ) interval in 3D.
  * \param v1 the first 3D vector (represented as a \a Eigen3::Vector4f)
  * \param v2 the second 3D vector (represented as a \a Eigen3::Vector4f)
  */
inline double
  pcl::getAngle3D (const Eigen3::Vector4f &v1, const Eigen3::Vector4f &v2)
{
  // Compute the actual angle
  double rad = v1.dot (v2) / sqrt (v1.squaredNorm () * v2.squaredNorm ());
  if (rad < -1.0) rad = -1.0;
  if (rad >  1.0) rad = 1.0;
  return (acos (rad));
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Compute both the mean and the standard deviation of an array of values
  * \param values the array of values
  * \param mean the resultant mean of the distribution
  * \param stddev the resultant standard deviation of the distribution
  */
inline void
  pcl::getMeanStd (const std::vector<float> &values, double &mean, double &stddev)
{
  double sum = 0, sq_sum = 0;

  for (size_t i = 0; i < values.size (); ++i)
  {
    sum += values[i];
    sq_sum += values[i] * values[i];
  }
  mean = sum / values.size ();
  double variance = (double)(sq_sum - sum * sum / values.size ()) / (values.size () - 1);
  stddev = sqrt (variance);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Get a set of points residing in a box given its bounds
  * \param cloud the point cloud data message
  * \param min_pt the minimum bounds
  * \param max_pt the maximum bounds
  * \param indices the resultant set of point indices residing in the box
  */
template <typename PointT> inline void
  pcl::getPointsInBox (const pcl::PointCloud<PointT> &cloud, Eigen3::Vector4f &min_pt, Eigen3::Vector4f &max_pt,
                       std::vector<int> &indices)
{
  indices.resize (cloud.points.size ());
  int l = 0;
  for (size_t i = 0; i < cloud.points.size (); ++i)
  {
    // Check if the point is invalid
    if (isnan (cloud.points[i].x) || isnan (cloud.points[i].y) || isnan (cloud.points[i].z))
      continue;
    // Check if the point is inside bounds
    if (cloud.points[i].x < min_pt[0] || cloud.points[i].y < min_pt[1] || cloud.points[i].z < min_pt[2])
      continue;
    if (cloud.points[i].x > max_pt[0] || cloud.points[i].y > max_pt[1] || cloud.points[i].z > max_pt[2])
      continue;
    indices[l++] = i;
  }
  indices.resize (l);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Get the minimum and maximum values on each of the 3 (x-y-z) dimensions in a given pointcloud
  * \param cloud the point cloud data message
  * \param min_pt the resultant minimum bounds
  * \param max_pt the resultant maximum bounds
  */
template <typename PointT> inline void
  pcl::getMinMax3D (const pcl::PointCloud<PointT> &cloud, PointT &min_pt, PointT &max_pt)
{
  Eigen3::Array4f min_p, max_p;
  min_p.setConstant (FLT_MAX);
  max_p.setConstant (-FLT_MAX);

  for (size_t i = 0; i < cloud.points.size (); ++i)
  {
    // Check if the point is invalid
    if (isnan (cloud.points[i].x) || isnan (cloud.points[i].y) || isnan (cloud.points[i].z))
      continue;
    pcl::Array4fMap pt = cloud.points[i].getArray4fMap ();
    min_p = min_p.min (pt);
    max_p = max_p.max (pt);
  }
  min_pt.x = min_p[0]; min_pt.y = min_p[1]; min_pt.z = min_p[2];
  max_pt.x = max_p[0]; max_pt.y = max_p[1]; max_pt.z = max_p[2];
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Get the minimum and maximum values on each of the 3 (x-y-z) dimensions in a given pointcloud
  * \param cloud the point cloud data message
  * \param min_pt the resultant minimum bounds
  * \param max_pt the resultant maximum bounds
  */
template <typename PointT> inline void
  pcl::getMinMax3D (const pcl::PointCloud<PointT> &cloud, Eigen3::Vector4f &min_pt, Eigen3::Vector4f &max_pt)
{
  Eigen3::Array4f min_p, max_p;
  min_p.setConstant (FLT_MAX);
  max_p.setConstant (-FLT_MAX);

  for (size_t i = 0; i < cloud.points.size (); ++i)
  {
    // Check if the point is invalid
    if (isnan (cloud.points[i].x) || isnan (cloud.points[i].y) || isnan (cloud.points[i].z))
      continue;
    pcl::Array4fMap pt = cloud.points[i].getArray4fMap ();
    min_p = min_p.min (pt);
    max_p = max_p.max (pt);
  }
  min_pt = min_p;
  max_pt = max_p;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Get the minimum and maximum values on each of the 3 (x-y-z) dimensions in a given pointcloud
  * \param cloud the point cloud data message
  * \param indices the vector of point indices to use from \a cloud
  * \param min_pt the resultant minimum bounds
  * \param max_pt the resultant maximum bounds
  */
template <typename PointT> inline void
  pcl::getMinMax3D (const pcl::PointCloud<PointT> &cloud, const pcl::PointIndices &indices,
                    Eigen3::Vector4f &min_pt, Eigen3::Vector4f &max_pt)
{
  Eigen3::Array4f min_p, max_p;
  min_p.setConstant (FLT_MAX);
  max_p.setConstant (-FLT_MAX);

  for (size_t i = 0; i < indices.indices.size (); ++i)
  {
    // Check if the point is invalid
    if (isnan (cloud.points[indices.indices[i]].x) || isnan (cloud.points[indices.indices[i]].y) || isnan (cloud.points[indices.indices[i]].z))
      continue;
    pcl::Array4fMap pt = cloud.points[indices.indices[i]].getArray4fMap ();
    min_p = min_p.min (pt);
    max_p = max_p.max (pt);
  }
  min_pt = min_p;
  max_pt = max_p;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Get the minimum and maximum values on each of the 3 (x-y-z) dimensions in a given pointcloud
  * \param cloud the point cloud data message
  * \param indices the vector of point indices to use from \a cloud
  * \param min_pt the resultant minimum bounds
  * \param max_pt the resultant maximum bounds
  */
template <typename PointT> inline void
  pcl::getMinMax3D (const pcl::PointCloud<PointT> &cloud, const std::vector<int> &indices,
                    Eigen3::Vector4f &min_pt, Eigen3::Vector4f &max_pt)
{
  min_pt.setConstant (FLT_MAX);
  max_pt.setConstant (-FLT_MAX);

  for (size_t i = 0; i < indices.size (); ++i)
  {
    // Check if the point is invalid
    if (isnan (cloud.points[indices[i]].x) || isnan (cloud.points[indices[i]].y) || isnan (cloud.points[indices[i]].z))
      continue;
    pcl::Array4fMap pt = cloud.points[indices[i]].getArray4fMap ();
    min_pt = min_pt.array ().min (pt);
    max_pt = max_pt.array ().max (pt);
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Compute the radius of a circumscribed circle for a triangle formed of three points pa, pb, and pc
  * \param pa the first point
  * \param pb the second point
  * \param pc the third point
  */
template <typename PointT> inline double 
  pcl::getCircumcircleRadius (const PointT &pa, const PointT &pb, const PointT &pc)
{
  Eigen3::Vector4f p1 (pa.x, pa.y, pa.z, 0);
  Eigen3::Vector4f p2 (pb.x, pb.y, pb.z, 0);
  Eigen3::Vector4f p3 (pc.x, pc.y, pc.z, 0);

  double p2p1 = (p2 - p1).norm (), p3p2 = (p3 - p2).norm (), p1p3 = (p1 - p3).norm ();
  // Calculate the area of the triangle using Heron's formula (http://en.wikipedia.org/wiki/Heron's_formula)
  double semiperimeter = (p2p1 + p3p2 + p1p3) / 2.0;
  double area = sqrt (semiperimeter * (semiperimeter - p2p1) * (semiperimeter - p3p2) * (semiperimeter - p1p3));
  // Compute the radius of the circumscribed circle
  return ((p2p1 * p3p2 * p1p3) / (4.0 * area));
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Get the minimum and maximum values on a point histogram
  * \param histogram the point representing a multi-dimensional histogram
  * \param len the length of the histogram
  * \param min_p the resultant minimum 
  * \param max_p the resultant maximum 
  */
template <typename PointT> inline void 
  pcl::getMinMax (const PointT &histogram, int len, float &min_p, float &max_p)
{
  min_p = FLT_MAX;
  max_p = -FLT_MAX;

  for (int i = 0; i < len; ++i)
  {
    min_p = (histogram[i] > min_p) ? min_p : histogram[i]; 
    max_p = (histogram[i] < max_p) ? max_p : histogram[i]; 
  }
}

