/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: boundary.hpp 32774 2010-09-20 12:08:49Z lorenz $
 *
 */
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Compute the angle in the [ 0, 2*PI ) interval of a point (direction) with a reference (0, 0) in 2D.
  * \param point a 2D point
  */
inline float
  pcl::getAngle2D (const float point[2])
{
  float rad;
  if (point[0] == 0)
    rad = (point[1] < 0) ? -M_PI / 2.0 : M_PI / 2.0;
  else
  {
    rad = atan (point[1] / point[0]);
    if (point[0] < 0)
      rad += M_PI;
  }
  if (rad < 0)
    rad += 2 * M_PI;

  return (rad);
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Get a u-v-n coordinate system that lies on a plane defined by its normal
  * \param p_coeff the plane coefficients (containing the plane normal)
  * \param u the resultant u direction
  * \param v the resultant v direction
  */
template <typename PointInT, typename PointNT, typename PointOutT> inline void
  pcl::BoundaryEstimation<PointInT, PointNT, PointOutT>::getCoordinateSystemOnPlane (const PointNT &p_coeff, Eigen3::Vector3f &u, Eigen3::Vector3f &v)
{
  pcl::Vector3fMap p_coeff_v = p_coeff.getNormalVector3fMap ();

  // Initialize normalized u vector with "random" values not parallel with normal
  u.setZero ();
  if (fabs (p_coeff.normal[0]) != 1)
    u[0] = 1;
  else
    u[2] = 1;

  // Compute the v vector and normalize it
  v = p_coeff_v.cross (u);
  v.normalize ();

  // Recompute u and normalize it
  u = v.cross (p_coeff_v);
  u.normalize ();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Check whether a point is a boundary point in a planar patch of projected points given by indices.
  * \note A coordinate system u-v-n must be computed a-priori using \a getCoordinateSystemOnPlane
  * \param cloud a pointer to the input point cloud
  * \param q_idx the index of the query point in \a cloud
  * \param indices the estimated point neighbors of the query point
  * \param u the u direction
  * \param v the v direction
  * \param angle_threshold the threshold angle (default \f$\pi / 2.0\f$)
  */
template <typename PointInT, typename PointNT, typename PointOutT> inline bool
  pcl::BoundaryEstimation<PointInT, PointNT, PointOutT>::isBoundaryPoint (const pcl::PointCloud<PointInT> &cloud, int q_idx, 
                                                                          const std::vector<int> &indices, 
                                                                          const Eigen3::Vector3f &u, const Eigen3::Vector3f &v, 
                                                                          float angle_threshold)
{
  return isBoundaryPoint (cloud, cloud.points[q_idx], indices, u, v, angle_threshold);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Check whether a point is a boundary point in a planar patch of projected points given by indices.
  * \note A coordinate system u-v-n must be computed a-priori using \a getCoordinateSystemOnPlane
  * \param cloud a pointer to the input point cloud
  * \param q_point a pointer to the querry point
  * \param indices the estimated point neighbors of the query point
  * \param u the u direction
  * \param v the v direction
  * \param angle_threshold the threshold angle (default \f$\pi / 2.0\f$)
  */
template <typename PointInT, typename PointNT, typename PointOutT> inline bool
  pcl::BoundaryEstimation<PointInT, PointNT, PointOutT>::isBoundaryPoint (const pcl::PointCloud<PointInT> &cloud, const PointInT &q_point, 
                                                                          const std::vector<int> &indices, 
                                                                          const Eigen3::Vector3f &u, const Eigen3::Vector3f &v, 
                                                                          float angle_threshold)
{
  if (indices.size () < 3)
    return (false);
  float uvn_nn[2];
  Eigen3::Vector3f delta;
  delta.setZero ();
  // Compute the angles between each neighbouring point and the query point itself
  std::vector<float> angles;
  angles.reserve (indices.size ());
  for (size_t i = 0; i < indices.size (); ++i)
  {
    delta[0] = cloud.points[indices[i]].x - q_point.x;
    delta[1] = cloud.points[indices[i]].y - q_point.y;
    delta[2] = cloud.points[indices[i]].z - q_point.z;

    uvn_nn[0] = u.dot (delta);
    uvn_nn[1] = v.dot (delta);

    if (uvn_nn[0] == 0 && uvn_nn[1] == 0)
      continue;
    angles.push_back (getAngle2D (uvn_nn));
  }
  std::sort (angles.begin (), angles.end ());

  // Compute the maximal angle difference between two consecutive angles
  float max_dif = FLT_MIN, dif;
  for (size_t i = 0; i < angles.size () - 1; ++i)
  {
    dif = angles[i + 1] - angles[i];
    if (max_dif < dif)
      max_dif = dif;
  }
  // Get the angle difference between the last and the first
  dif = 2 * M_PI - angles[angles.size () - 1] + angles[0];
  if (max_dif < dif)
    max_dif = dif;

  // Check results
  if (max_dif > angle_threshold)
    return (true);
  else
    return (false);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Estimate whether a set of points is lying on surface boundaries using an angle criterion for all points
  * given in <setInputCloud (), setIndices ()> using the surface in setSearchSurface () and the spatial locator in
  * setSearchMethod ()
  * \param output the resultant point cloud model dataset that contains boundary point estimates
  */
template <typename PointInT, typename PointNT, typename PointOutT> inline void
  pcl::BoundaryEstimation<PointInT, PointNT, PointOutT>::computeFeature (PointCloudOut &output)
{
  // Check if input was set
  if (!normals_)
  {
    ROS_ERROR ("[pcl::%s::computeFeature] No input dataset containing normals was given!", getClassName ().c_str ());
    return;
  }
  if (normals_->points.size () != surface_->points.size ())
  {
    ROS_ERROR ("[pcl::%s::computeFeature] The number of points in the input dataset differs from the number of points in the dataset containing the normals!", getClassName ().c_str ());
    return;
  }

  // Allocate enough space to hold the results
  // \note This resize is irrelevant for a radiusSearch ().
  std::vector<int> nn_indices (k_);
  std::vector<float> nn_dists (k_);

  Eigen3::Vector3f u, v;

  // Iterating over the entire index vector
  for (size_t idx = 0; idx < indices_->size (); ++idx)
  {
    searchForNeighbors ((*indices_)[idx], search_parameter_, nn_indices, nn_dists);

    // Obtain a coordinate system on the least-squares plane
    getCoordinateSystemOnPlane (normals_->points[(*indices_)[idx]], u, v);

    // Estimate whether the point is lying on a boundary surface or not
    output.points[idx].boundary_point = isBoundaryPoint (*surface_, input_->points[(*indices_)[idx]], nn_indices, u, v, angle_threshold_);
  }
}

