/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: feature.hpp 32872 2010-09-26 18:00:20Z rusu $
 *
 */

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Compute the 3D (X-Y-Z) centroid of a set of points using their indices and return it as a 3D vector.
  * \param cloud the input point cloud
  * \param indices the point cloud indices that need to be used
  * \param centroid the output centroid
  */
template <typename PointT> inline void
  pcl::compute3DCentroid (const pcl::PointCloud<PointT> &cloud, const std::vector<int> &indices,
                          Eigen3::Vector4f &centroid)
{
  // Initialize to 0
  centroid.setZero ();
  if (indices.empty ()) 
    return;
  // For each point in the cloud
  int cp = 0;
  for (size_t i = 0; i < indices.size (); ++i)
  {
    // Check if the point is invalid
    if (isnan (cloud.points[indices[i]].x) || isnan (cloud.points[indices[i]].y) || isnan (cloud.points[indices[i]].z))
      continue;

    centroid += cloud.points[indices[i]].getVector4fMap ();
    cp++;
  }
  centroid[3] = 0;
  centroid /= cp;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Compute the 3D (X-Y-Z) centroid of a set of points and return it as a 3D vector.
  * \param cloud the input point cloud
  * \param centroid the output centroid
  */
template <typename PointT> inline void
  pcl::compute3DCentroid (const pcl::PointCloud<PointT> &cloud, Eigen3::Vector4f &centroid)
{
  // Initialize to 0
  centroid.setZero ();
  if (cloud.points.empty ()) 
    return;
  // For each point in the cloud
  int cp = 0;
  for (size_t i = 0; i < cloud.points.size (); ++i)
  {
    // Check if the point is invalid
    if (isnan (cloud.points[i].x) || isnan (cloud.points[i].y) || isnan (cloud.points[i].z))
      continue;

    centroid += cloud.points[i].getVector4fMap ();
    cp++;
  }
  centroid[3] = 0;
  centroid /= cp;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief General, all purpose nD centroid estimation for a set of points using their indices.
  * \param cloud the input point cloud
  * \param indices the point cloud indices that need to be used
  * \param centroid the output centroid
  */
template <typename PointT> inline void
  pcl::computeNDCentroid (const pcl::PointCloud<PointT> &cloud, const std::vector<int> &indices,
                          Eigen3::VectorXf &centroid)
{
  typedef typename pcl::traits::fieldList<PointT>::type FieldList;

  // Get the size of the fields
  centroid.setZero (boost::mpl::size<FieldList>::value);

  if (indices.empty ()) 
    return;
  // Iterate over each point
  int nr_points = indices.size ();
  for (int i = 0; i < nr_points; ++i)
  {
    // Iterate over each dimension
    pcl::for_each_type <FieldList> (NdCentroidFunctor <PointT> (cloud.points[indices[i]], centroid));
  }
  centroid /= nr_points;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief General, all purpose nD centroid estimation for a set of points using their indices.
  * \param cloud the input point cloud
  * \param centroid the output centroid
  */
template <typename PointT> inline void
  pcl::computeNDCentroid (const pcl::PointCloud<PointT> &cloud, Eigen3::VectorXf &centroid)
{
  typedef typename pcl::traits::fieldList<PointT>::type FieldList;

  // Get the size of the fields
  centroid.setZero (boost::mpl::size<FieldList>::value);

  if (cloud.points.empty ())
    return;
  // Iterate over each point
  int size = cloud.points.size ();
  for (int i = 0; i < size; ++i)
  {
    // Iterate over each dimension
    pcl::for_each_type <FieldList> (NdCentroidFunctor <PointT> (cloud.points[i], centroid));
  }
  centroid /= size;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Compute the 3x3 covariance matrix of a given set of points.
  * The result is returned as a Eigen3::Matrix3f.
  * Note: the covariance matrix is not normalized with the number of
  * points. For a normalized covariance, please use
  * computeNormalizedCovarianceMatrix.
  * \param cloud the input point cloud
  * \param centroid the centroid of the set of points in the cloud
  * \param covariance_matrix the resultant 3x3 covariance matrix
  */
template <typename PointT> inline void
  pcl::computeCovarianceMatrix (const pcl::PointCloud<PointT> &cloud,
                                const Eigen3::Vector4f &centroid, Eigen3::Matrix3f &covariance_matrix)
{
  // Initialize to 0
  covariance_matrix.setZero ();

  if (cloud.points.empty ())
    return;
  // For each point in the cloud
  for (size_t i = 0; i < cloud.points.size (); ++i)
  {
    // Check if the point is invalid
    if (isnan (cloud.points[i].x) || isnan (cloud.points[i].y) || isnan (cloud.points[i].z))
      continue;

    Eigen3::Vector4f pt = cloud.points[i].getVector4fMap () - centroid;

    covariance_matrix (1, 1) += pt.y () * pt.y ();
    covariance_matrix (1, 2) += pt.y () * pt.z ();

    covariance_matrix (2, 2) += pt.z () * pt.z ();

    pt *= pt.x ();
    covariance_matrix (0, 0) += pt.x ();// * pt.x ();
    covariance_matrix (0, 1) += pt.y ();// * pt.x ();
    covariance_matrix (0, 2) += pt.z ();// * pt.x ();

    //covariance_matrix (1, 0) += demean_xy;
    //covariance_matrix (2, 0) += demean_xz;
    //covariance_matrix (2, 1) += demean_yz;
  }
  covariance_matrix (1, 0) = covariance_matrix (0, 1);
  covariance_matrix (2, 0) = covariance_matrix (0, 2);
  covariance_matrix (2, 1) = covariance_matrix (1, 2);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Compute the normalized 3x3 covariance matrix of a given set of points.
  * The result is returned as a Eigen3::Matrix3f.
  * \param cloud the input point cloud
  * \param centroid the centroid of the set of points in the cloud
  * \param covariance_matrix the resultant 3x3 covariance matrix
  */
template <typename PointT> inline void
  pcl::computeCovarianceMatrixNormalized (const pcl::PointCloud<PointT> &cloud,
                                          const Eigen3::Vector4f &centroid, Eigen3::Matrix3f &covariance_matrix)
{
  computeCovarianceMatrix (cloud, centroid, covariance_matrix);
  covariance_matrix /= cloud.points.size ();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Compute the 3x3 covariance matrix of a given set of points using their indices.
  * The result is returned as a Eigen3::Matrix3f.
  * Note: the covariance matrix is not normalized with the number of
  * points. For a normalized covariance, please use
  * computeNormalizedCovarianceMatrix.
  * \param cloud the input point cloud
  * \param indices the point cloud indices that need to be used
  * \param centroid the centroid of the set of points in the cloud
  * \param covariance_matrix the resultant 3x3 covariance matrix
  */
template <typename PointT> inline void
  pcl::computeCovarianceMatrix (const pcl::PointCloud<PointT> &cloud, const std::vector<int> &indices,
                                const Eigen3::Vector4f &centroid, Eigen3::Matrix3f &covariance_matrix)
{
  // Initialize to 0
  covariance_matrix.setZero ();

  if (indices.empty ())
    return;
  // For each point in the cloud
  for (size_t i = 0; i < indices.size (); ++i)
  {
    // Check if the point is invalid
    if (isnan (cloud.points[indices[i]].x) || isnan (cloud.points[indices[i]].y) || isnan (cloud.points[indices[i]].z))
      continue;

    Eigen3::Vector4f pt = cloud.points[indices[i]].getVector4fMap () - centroid;

    covariance_matrix (1, 1) += pt.y () * pt.y ();
    covariance_matrix (1, 2) += pt.y () * pt.z ();

    covariance_matrix (2, 2) += pt.z () * pt.z ();

    pt *= pt.x ();
    covariance_matrix (0, 0) += pt.x ();// * pt.x ();
    covariance_matrix (0, 1) += pt.y ();// * pt.x ();
    covariance_matrix (0, 2) += pt.z ();// * pt.x ();

    //covariance_matrix (1, 0) += demean_xy;
    //covariance_matrix (2, 0) += demean_xz;
    //covariance_matrix (2, 1) += demean_yz;
    
  }
  covariance_matrix (1, 0) = covariance_matrix (0, 1);
  covariance_matrix (2, 0) = covariance_matrix (0, 2);
  covariance_matrix (2, 1) = covariance_matrix (1, 2);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Compute the normalized 3x3 covariance matrix of a given set of points using their indices
  * The result is returned as a Eigen3::Matrix3f.
  * \param cloud the input point cloud
  * \param centroid the centroid of the set of points in the cloud
  * \param covariance_matrix the resultant 3x3 covariance matrix
  */
template <typename PointT> inline void
  pcl::computeCovarianceMatrixNormalized (const pcl::PointCloud<PointT> &cloud, const std::vector<int> &indices,
                                          const Eigen3::Vector4f &centroid, Eigen3::Matrix3f &covariance_matrix)
{
  computeCovarianceMatrix (cloud, indices, centroid, covariance_matrix);
  covariance_matrix /= indices.size ();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Solve the eigenvalues and eigenvectors of a given 3x3 covariance matrix, and estimate the least-squares
  * plane normal and surface curvature.
  * \param covariance_matrix the 3x3 covariance matrix
  * \param point a point lying on the least-squares plane (SSE aligned)
  * \param plane_parameters the resultant plane parameters as: a, b, c, d (ax + by + cz + d = 0)
  * \param curvature the estimated surface curvature as a measure of
  * \f[
  * \lambda_0 / (\lambda_0 + \lambda_1 + \lambda_2)
  * \f]
  */
inline void
  pcl::solvePlaneParameters (const Eigen3::Matrix3f &covariance_matrix, const Eigen3::Vector4f &point,
                             Eigen3::Vector4f &plane_parameters, float &curvature)
{
  // Avoid getting hung on Eigen's optimizers
  for (int i = 0; i < 3; ++i)
    for (int j = 0; j < 3; ++j)
      if (!std::isfinite (covariance_matrix (i, j)))
      {
        //ROS_WARN ("[pcl::solvePlaneParameteres] Covariance matrix has NaN/Inf values!");
        plane_parameters.setConstant (std::numeric_limits<float>::quiet_NaN ());
        curvature = std::numeric_limits<float>::quiet_NaN ();
        return;
      }

  // Extract the eigenvalues and eigenvectors
  //Eigen3::SelfAdjointEigenSolver<Eigen3::Matrix3f> ei_symm (covariance_matrix);
  //EIGEN_ALIGN16 Eigen3::Vector3f eigen_values  = ei_symm.eigenvalues ();
  //EIGEN_ALIGN16 Eigen3::Matrix3f eigen_vectors = ei_symm.eigenvectors ();
  EIGEN_ALIGN16 Eigen3::Vector3f eigen_values;
  EIGEN_ALIGN16 Eigen3::Matrix3f eigen_vectors;
  pcl::eigen33 (covariance_matrix, eigen_vectors, eigen_values);

  // Normalize the surface normal (eigenvector corresponding to the smallest eigenvalue)
  // Note: Remember to take care of the eigen_vectors ordering
  //float norm = 1.0 / eigen_vectors.col (0).norm ();

  //plane_parameters[0] = eigen_vectors (0, 0) * norm;
  //plane_parameters[1] = eigen_vectors (1, 0) * norm;
  //plane_parameters[2] = eigen_vectors (2, 0) * norm;

  // The normalization is not necessary, since the eigenvectors from libeigen are already normalized
  plane_parameters[0] = eigen_vectors (0, 0);
  plane_parameters[1] = eigen_vectors (1, 0);
  plane_parameters[2] = eigen_vectors (2, 0);
  plane_parameters[3] = 0;

  // Hessian form (D = nc . p_plane (centroid here) + p)
  plane_parameters[3] = -1 * plane_parameters.dot (point);

  // Compute the curvature surface change
  float eig_sum = eigen_values.sum ();
  if (eig_sum != 0)
    curvature = fabs ( eigen_values[0] / eig_sum );
  else
    curvature = 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Solve the eigenvalues and eigenvectors of a given 3x3 covariance matrix, and estimate the least-squares
  * plane normal and surface curvature.
  * \param covariance_matrix the 3x3 covariance matrix
  * \param nx the resultant X component of the plane normal
  * \param ny the resultant Y component of the plane normal
  * \param nz the resultant Z component of the plane normal
  * \param curvature the estimated surface curvature as a measure of
  * \f[
  * \lambda_0 / (\lambda_0 + \lambda_1 + \lambda_2)
  * \f]
  */
inline void
  pcl::solvePlaneParameters (const Eigen3::Matrix3f &covariance_matrix,
                             float &nx, float &ny, float &nz, float &curvature)
{
  // Avoid getting hung on Eigen's optimizers
  for (int i = 0; i < 3; ++i)
    for (int j = 0; j < 3; ++j)
      if (!std::isfinite (covariance_matrix (i, j)))
      {
        //ROS_WARN ("[pcl::solvePlaneParameteres] Covariance matrix has NaN/Inf values!");
        nx = ny = nz = curvature = std::numeric_limits<float>::quiet_NaN ();
        return;
      }
  // Extract the eigenvalues and eigenvectors
  //Eigen3::SelfAdjointEigenSolver<Eigen3::Matrix3f> ei_symm (covariance_matrix);
  //EIGEN_ALIGN16 Eigen3::Vector3f eigen_values  = ei_symm.eigenvalues ();
  //EIGEN_ALIGN16 Eigen3::Matrix3f eigen_vectors = ei_symm.eigenvectors ();
  EIGEN_ALIGN16 Eigen3::Vector3f eigen_values;
  EIGEN_ALIGN16 Eigen3::Matrix3f eigen_vectors;
  pcl::eigen33 (covariance_matrix, eigen_vectors, eigen_values);

  // Normalize the surface normal (eigenvector corresponding to the smallest eigenvalue)
  // Note: Remember to take care of the eigen_vectors ordering
  //float norm = 1.0 / eigen_vectors.col (0).norm ();

  //nx = eigen_vectors (0, 0) * norm;
  //ny = eigen_vectors (1, 0) * norm;
  //nz = eigen_vectors (2, 0) * norm;
  
  // The normalization is not necessary, since the eigenvectors from libeigen are already normalized
  nx = eigen_vectors (0, 0);
  ny = eigen_vectors (1, 0);
  nz = eigen_vectors (2, 0);

  // Compute the curvature surface change
  float eig_sum = eigen_values.sum ();
  if (eig_sum != 0)
    curvature = fabs ( eigen_values[0] / eig_sum );
  else
    curvature = 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Base method for feature estimation for all points given in <setInputCloud (), setIndices ()> using
  * the surface in setSearchSurface () and the spatial locator in setSearchMethod ()
  * \param output the resultant point cloud model dataset containing the estimated features
  */
template <typename PointInT, typename PointOutT> inline void
  pcl::Feature<PointInT, PointOutT>::compute (PointCloudOut &output)
{
  if (!initCompute ()) return;

  // Check if a space search locator was given
  if (!tree_)
  {
    ROS_ERROR ("[pcl::%s::compute] No spatial search method was given!", getClassName ().c_str ());
    return;
  }

  // If no search surface has been defined, use the input dataset as the search surface itself
  if (!surface_)
  {
    fake_surface_ = true;
    surface_ = input_;
  }

  // Send the surface dataset to the spatial locator
  tree_->setInputCloud (surface_);

  // Do a fast check to see if the search parameters are well defined
  if (search_radius_ != 0.0)
  {
    if (k_ != 0)
    {
      ROS_ERROR ("[pcl::%s::compute] Both radius (%f) and K (%d) defined! Set one of them to zero first and then re-run compute ().", getClassName ().c_str (), search_radius_, k_);
      return;
    }
    else                  // Use the radiusSearch () function
    {
      search_parameter_ = search_radius_;
      if (surface_ == input_)       // if the two surfaces are the same
      {
        // Declare the search locator definition
        bool (KdTree::*radiusSearch)(int index, double radius, std::vector<int> &k_indices,
                                     std::vector<float> &k_distances, int max_nn) const = &KdTree::radiusSearch;
        search_method_ = boost::bind (radiusSearch, boost::ref (tree_), _1, _2, _3, _4, INT_MAX);
      }
      else
      {
        // Declare the search locator definition
        bool (KdTree::*radiusSearchSurface)(const PointCloudIn &cloud, int index, double radius, std::vector<int> &k_indices,
                                            std::vector<float> &k_distances, int max_nn) const = &KdTree::radiusSearch;
        search_method_surface_ = boost::bind (radiusSearchSurface, boost::ref (tree_), _1, _2, _3, _4, _5, INT_MAX);
      }
    }
  }
  else
  {
    if (k_ != 0)         // Use the nearestKSearch () function
    {
      search_parameter_ = k_;
      if (surface_ == input_)       // if the two surfaces are the same
      {
        // Declare the search locator definition
        bool (KdTree::*nearestKSearch)(int index, int k, std::vector<int> &k_indices, std::vector<float> &k_distances) = &KdTree::nearestKSearch;
        search_method_ = boost::bind (nearestKSearch, boost::ref (tree_), _1, _2, _3, _4);
      }
      else
      {
        // Declare the search locator definition
        bool (KdTree::*nearestKSearchSurface)(const PointCloudIn &cloud, int index, int k, std::vector<int> &k_indices, std::vector<float> &k_distances) = &KdTree::nearestKSearch;
        search_method_surface_ = boost::bind (nearestKSearchSurface, boost::ref (tree_), _1, _2, _3, _4, _5);
      }
    }
    else
    {
      ROS_ERROR ("[pcl::%s::compute] Neither radius nor K defined! Set one of them to a positive number first and then re-run compute ().", getClassName ().c_str ());
      return;
    }
  }

  // Resize the output dataset
  if (output.points.size () != indices_->size ())
    output.points.resize (indices_->size ());
  // Copy the header
  output.header   = input_->header;
  // Check if the output is dense or not
  if (indices_->size () != input_->points.size ())
  {
    output.width    = indices_->size ();
    output.height   = 1;
    output.is_dense = false;
  }
  else
  {
    output.width    = input_->width;
    output.height   = input_->height;
    output.is_dense = input_->is_dense;
  }

  // Perform the actual feature computation
  computeFeature (output);

  deinitCompute ();

  // Reset the surface
  if (fake_surface_)
  {
    surface_.reset ();
    fake_surface_ = false;
  }
}
