/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: intensity_spin.hpp 32774 2010-09-20 12:08:49Z lorenz $
 *
 */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Estimate the intensity-domain spin image descriptor for a given point based on its spatial neighborhood of 
  * 3D points and their intensities. 
  * \param cloud the dataset containing the Cartesian coordinates and intensity values of the points
  * \param radius the radius of the feature
  * \param sigma the standard deviation of the Gaussian smoothing kernel to use during the soft histogram update
  * \param indices the indices of the points that comprise the query point's neighborhood
  * \param squared_distances the squared distances from the query point to each point in the neighborhood
  * \param intensity_spin_image the resultant intensity-domain spin image descriptor
  */
template <typename PointInT, typename PointOutT> inline void
  pcl::IntensitySpinEstimation<PointInT, PointOutT>::computeIntensitySpinImage (
      const pcl::PointCloud<PointInT> &cloud, float radius, float sigma, const std::vector<int> &indices, 
      const std::vector<float> &squared_distances, Eigen3::MatrixXf &intensity_spin_image)
{
  // Determine the number of bins to use based on the size of intensity_spin_image
  int nr_distance_bins = intensity_spin_image.cols ();
  int nr_intensity_bins = intensity_spin_image.rows ();

  // Find the min and max intensity values in the given neighborhood
  float min_intensity = std::numeric_limits<float>::max ();
  float max_intensity = std::numeric_limits<float>::min ();
  for (size_t idx = 0; idx < indices.size (); ++idx)
  {
    min_intensity = min (min_intensity, cloud.points[indices[idx]].intensity);
    max_intensity = max (max_intensity, cloud.points[indices[idx]].intensity);
  }

  // Compute the intensity spin image
  intensity_spin_image.setZero ();
  for (size_t idx = 0; idx < indices.size (); ++idx)
  {
    // Normalize distance and intensity values to: 0.0 <= d,i < nr_distance_bins,nr_intensity_bins
    const float eps = std::numeric_limits<float>::epsilon();
    float d = nr_distance_bins * sqrt (squared_distances[idx]) / (radius + eps);
    float i = nr_intensity_bins * 
      (cloud.points[indices[idx]].intensity - min_intensity) / (max_intensity - min_intensity + eps);

    if (sigma == 0)
    {
      // If sigma is zero, update the histogram with no smoothing kernel
      int d_idx = (int) d;
      int i_idx = (int) i;
      intensity_spin_image (i_idx, d_idx) += 1;
    }
    else
    {
      // Compute the bin indices that need to be updated (+/- 3 standard deviations)
      int d_idx_min = max ((int) floor (d - 3*sigma), 0);
      int d_idx_max = min ((int) ceil  (d + 3*sigma), nr_distance_bins - 1);
      int i_idx_min = max ((int) floor (i - 3*sigma), 0);
      int i_idx_max = min ((int) ceil  (i + 3*sigma), nr_intensity_bins - 1);
    
      // Update the appropriate bins of the histogram 
      for (int i_idx = i_idx_min; i_idx <= i_idx_max; ++i_idx)  
      {
        for (int d_idx = d_idx_min; d_idx <= d_idx_max; ++d_idx)
        {
          // Compute a "soft" update weight based on the distance between the point and the bin
          float w = exp (-pow (d - d_idx, 2) / (2.0*sigma_*sigma_) 
                         -pow (i - i_idx, 2) / (2.0*sigma_*sigma_));
          intensity_spin_image (i_idx, d_idx) += w;
        }
      }
    }
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Estimate the intensity-domain descriptors at a set of points given by <setInputCloud (), setIndices ()> 
  * using the surface in setSearchSurface (), and the spatial locator in setSearchMethod ().
  * \param output the resultant point cloud model dataset that contains the intensity-domain spin image features
  */
template <typename PointInT, typename PointOutT> inline void
  pcl::IntensitySpinEstimation<PointInT, PointOutT>::computeFeature (PointCloudOut &output)
{
  // Make sure a search radius is set
  if (search_radius_ == 0.0)
  {
    ROS_ERROR ("[pcl::%s::computeFeature] The search radius must be set before computing the feature!",
               getClassName ().c_str ());
    return;
  }

  // Make sure the spin image has valid dimensions
  if (nr_intensity_bins_ <= 0)
  {
    ROS_ERROR ("[pcl::%s::computeFeature] The number of intensity bins must be greater than zero!",
               getClassName ().c_str ());
    return;
  }
  if (nr_distance_bins_ <= 0)
  {
    ROS_ERROR ("[pcl::%s::computeFeature] The number of distance bins must be greater than zero!",
               getClassName ().c_str ());
    return;
  }

  Eigen3::MatrixXf intensity_spin_image (nr_intensity_bins_, nr_distance_bins_);
  std::vector<int> nn_indices;
  std::vector<float> nn_dist_sqr;
 
  // Iterating over the entire index vector
  for (size_t idx = 0; idx < indices_->size (); ++idx)
  {
    // Find neighbors within the search radius
    tree_->radiusSearch ((*indices_)[idx], search_radius_, nn_indices, nn_dist_sqr);

    // Compute the intensity spin image
    computeIntensitySpinImage (*input_, search_radius_, sigma_, nn_indices, nn_dist_sqr, intensity_spin_image);

    // Copy into the resultant cloud
    for (int bin = 0; bin < intensity_spin_image.size (); ++bin)
      output.points[idx].histogram[bin] = intensity_spin_image (bin);
  }
}

