/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: rift.hpp 32774 2010-09-20 12:08:49Z lorenz $
 *
 */

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Estimate the Rotation Invariant Feature Transform (RIFT) descriptors for a given point based on its spatial 
  * neighborhood of 3D points and the corresponding intensity gradient vector field
  * \param cloud the dataset containing the Cartesian coordinates of the points
  * \param gradient the dataset containing the intensity gradient at each point in \a cloud
  * \param p_idx the index of the query point in \a cloud (i.e. the center of the neighborhood)
  * \param radius the radius of the RIFT feature
  * \param indices the indices of the points that comprise \a p_idx's neighborhood in \a cloud
  * \param squared_distances the squared distances from the query point to each point in the neighborhood
  * \param rift_descriptor the resultant RIFT descriptor
  */
template <typename PointInT, typename GradientT, typename PointOutT> inline void
  pcl::RIFTEstimation<PointInT, GradientT, PointOutT>::computeRIFT (
      const pcl::PointCloud<PointInT> &cloud, const pcl::PointCloud<GradientT> &gradient, int p_idx, float radius,
      const std::vector<int> &indices, const std::vector<float> &sqr_distances, Eigen3::MatrixXf &rift_descriptor)
{
  ROS_ASSERT (indices.size () > 0);

  // Determine the number of bins to use based on the size of rift_descriptor
  int nr_distance_bins = rift_descriptor.cols ();
  int nr_gradient_bins = rift_descriptor.rows ();

  // Get the center point
  pcl::Vector3fMap p0 = cloud.points[p_idx].getVector3fMap ();

  // Compute the RIFT descriptor
  rift_descriptor.setZero ();
  for (size_t idx = 0; idx < indices.size (); ++idx)
  {
    // Compute the gradient magnitude and orientation (relative to the center point)
    ///Eigen3::Map<Eigen3::Vector3f> point (& (cloud.points[indices[idx]].x));
    pcl::Vector3fMap point = cloud.points[indices[idx]].getVector3fMap ();
    Eigen3::Map<Eigen3::Vector3f> gradient_vector (& (gradient.points[indices[idx]].gradient[0]));

    float gradient_magnitude = gradient_vector.norm ();
    float gradient_angle_from_center = acos (gradient_vector.dot ((point - p0).normalized ()) / gradient_magnitude);
    if (isnan (gradient_angle_from_center))
    {
      gradient_angle_from_center = 0.0;
    }

    // Normalize distance and angle values to: 0.0 <= d,g < nr_distances_bins,nr_gradient_bins
    const float eps = std::numeric_limits<float>::epsilon ();
    float d = nr_distance_bins * sqrt (sqr_distances[idx]) / (radius + eps);
    float g = nr_gradient_bins * gradient_angle_from_center / (M_PI + eps);

    // Compute the bin indices that need to be updated
    int d_idx_min = max ((int) ceil (d - 1), 0);
    int d_idx_max = min ((int) floor (d + 1), nr_distance_bins - 1);
    int g_idx_min = (int) ceil (g - 1);
    int g_idx_max = (int) floor (g + 1);

    // Update the appropriate bins of the histogram 
    for (int g_idx = g_idx_min; g_idx <= g_idx_max; ++g_idx)  
    {
      // Because gradient orientation is cyclical, out-of-bounds values must wrap around
      int g_idx_wrapped = ((g_idx + nr_gradient_bins) % nr_gradient_bins); 

      for (int d_idx = d_idx_min; d_idx <= d_idx_max; ++d_idx)
      {
        // To avoid boundary effects, use linear interpolation when updating each bin 
        float w = (1 - fabs (d - d_idx)) * (1 - fabs (g - g_idx));

        rift_descriptor (g_idx_wrapped * nr_distance_bins + d_idx) += w * gradient_magnitude;
      }
    }
  }

  // Normalize the RIFT descriptor to unit magnitude
  rift_descriptor.normalize ();
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Estimate the Rotation Invariant Feature Transform (RIFT) descriptors at a set of points given by
  * <setInputCloud (), setIndices ()> using the surface in setSearchSurface (), the gradient in 
  * setInputGradient (), and the spatial locator in setSearchMethod ()
  * \param output the resultant point cloud model dataset that contains the RIFT feature estimates
  */
template <typename PointInT, typename GradientT, typename PointOutT> inline void
  pcl::RIFTEstimation<PointInT, GradientT, PointOutT>::computeFeature (PointCloudOut &output)
{
  // Make sure a search radius is set
  if (search_radius_ == 0.0)
  {
    ROS_ERROR ("[pcl::%s::computeFeature] The search radius must be set before computing the feature!",
               getClassName ().c_str ());
    return;
  }

  // Make sure the RIFT descriptor has valid dimensions
  if (nr_gradient_bins_ <= 0)
  {
    ROS_ERROR ("[pcl::%s::computeFeature] The number of gradient bins must be greater than zero!",
               getClassName ().c_str ());
    return;
  }
  if (nr_distance_bins_ <= 0)
  {
    ROS_ERROR ("[pcl::%s::computeFeature] The number of distance bins must be greater than zero!",
               getClassName ().c_str ());
    return;
  }

  // Check for valid input gradient
  if (!gradient_)
  {
    ROS_ERROR ("[pcl::%s::computeFeature] No input gradient was given!", getClassName ().c_str ());
    return;
  }
  if (gradient_->points.size () != surface_->points.size ())
  {
    ROS_ERROR ("[pcl::%s::computeFeature] The number of points in the input dataset differs from the number of points in the gradient!", getClassName ().c_str ());
    return;
  }

  Eigen3::MatrixXf rift_descriptor (nr_gradient_bins_, nr_distance_bins_);
  std::vector<int> nn_indices;
  std::vector<float> nn_dist_sqr;
 
  // Iterating over the entire index vector
  for (size_t idx = 0; idx < indices_->size (); ++idx)
  {
    // Find neighbors within the search radius
    tree_->radiusSearch ((*indices_)[idx], search_radius_, nn_indices, nn_dist_sqr);

    // Compute the RIFT descriptor
    computeRIFT (*surface_, *gradient_, (*indices_)[idx], search_radius_, nn_indices, nn_dist_sqr, rift_descriptor);

    // Copy into the resultant cloud
    for (int bin = 0; bin < rift_descriptor.size (); ++bin)
      output.points[idx].histogram[bin] = rift_descriptor (bin);
  }
}
