/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: voxel_grid.cpp 32774 2010-09-20 12:08:49Z lorenz $
 *
 */

#include "pcl/filters/voxel_grid.h"

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Get the minimum and maximum values on each of the 3 (x-y-z) dimensions
  * in a given pointcloud, without considering points outside of a distance threshold from the laser origin
  * \param cloud the point cloud data message
  * \param x_idx the field index of the X coordinate
  * \param y_idx the field index of the Y coordinate
  * \param z_idx the field index of the Z coordinate
  * \param min_pt the resultant minimum bounds
  * \param max_pt the resultant maximum bounds
  */
void
  pcl::getMinMax3D (const sensor_msgs::PointCloud2ConstPtr &cloud, int x_idx, int y_idx, int z_idx,
                    Eigen3::Vector4f &min_pt, Eigen3::Vector4f &max_pt)
{
  // @todo fix this
  ROS_ASSERT (cloud->fields[x_idx].datatype == sensor_msgs::PointField::FLOAT32);
  ROS_ASSERT (cloud->fields[y_idx].datatype == sensor_msgs::PointField::FLOAT32);
  ROS_ASSERT (cloud->fields[z_idx].datatype == sensor_msgs::PointField::FLOAT32);

  Eigen3::Array4f min_p, max_p;
  min_p.setConstant (FLT_MAX);
  max_p.setConstant (-FLT_MAX);

  int nr_points = cloud->width * cloud->height;

  Eigen3::Array4f pt = Eigen3::Array4f::Zero ();
  Eigen3::Array4i xyz_offset (cloud->fields[x_idx].offset,
                             cloud->fields[y_idx].offset,
                             cloud->fields[z_idx].offset,
                             0);

  for (int cp = 0; cp < nr_points; ++cp)
  {
    // Unoptimized memcpys: assume fields x, y, z are in random order
    memcpy (&pt[0], &cloud->data[xyz_offset[0]], sizeof (float));
    memcpy (&pt[1], &cloud->data[xyz_offset[1]], sizeof (float));
    memcpy (&pt[2], &cloud->data[xyz_offset[2]], sizeof (float));
    // Check if the point is invalid
    if (isnan (pt[0]) || isnan (pt[1]) || isnan (pt[2]))
    {
      xyz_offset += cloud->point_step;
      continue;
    }
    xyz_offset += cloud->point_step;
    min_p = min_p.min (pt);
    max_p = max_p.max (pt);
  }
  min_pt = min_p;
  max_pt = max_p;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Get the minimum and maximum values on each of the 3 (x-y-z) dimensions
  * in a given pointcloud, without considering points outside of a distance threshold from the laser origin
  * \param cloud the point cloud data message
  * \param x_idx the field index of the X coordinate
  * \param y_idx the field index of the Y coordinate
  * \param z_idx the field index of the Z coordinate
  * \param distance_field_name the field name that contains the distance values
  * \param min_distance the minimum distance a point will be considered from
  * \param max_distance the maximum distance a point will be considered to
  * \param min_pt the resultant minimum bounds
  * \param max_pt the resultant maximum bounds
  * \param limit_negative if set to true, then all points outside of the interval (min_distance;max_distace) are considered
  */
void
  pcl::getMinMax3D (const sensor_msgs::PointCloud2ConstPtr &cloud, int x_idx, int y_idx, int z_idx,
                    const std::string &distance_field_name, float min_distance, float max_distance,
                    Eigen3::Vector4f &min_pt, Eigen3::Vector4f &max_pt, bool limit_negative)
{
  // @todo fix this
  ROS_ASSERT (cloud->fields[x_idx].datatype == sensor_msgs::PointField::FLOAT32);
  ROS_ASSERT (cloud->fields[y_idx].datatype == sensor_msgs::PointField::FLOAT32);
  ROS_ASSERT (cloud->fields[z_idx].datatype == sensor_msgs::PointField::FLOAT32);

  Eigen3::Array4f min_p, max_p;
  min_p.setConstant (FLT_MAX);
  max_p.setConstant (-FLT_MAX);

  // Get the distance field index
  int distance_idx = pcl::getFieldIndex (*cloud, distance_field_name);

  // @todo fix this
  ROS_ASSERT (cloud->fields[distance_idx].datatype == sensor_msgs::PointField::FLOAT32);

  int nr_points = cloud->width * cloud->height;

  Eigen3::Array4f pt = Eigen3::Array4f::Zero ();
  Eigen3::Array4i xyz_offset (cloud->fields[x_idx].offset,
                             cloud->fields[y_idx].offset,
                             cloud->fields[z_idx].offset,
                             0);
  float distance_value = 0;
  for (int cp = 0; cp < nr_points; ++cp)
  {
    int point_offset = cp * cloud->point_step;

    // Get the distance value
    memcpy (&distance_value, &cloud->data[point_offset + cloud->fields[distance_idx].offset], sizeof (float));

    if (limit_negative)
    {
      // Use a threshold for cutting out points which inside the interval
      if ((distance_value < max_distance) && (distance_value > min_distance))
      {
        xyz_offset += cloud->point_step;
        continue;
      }
    }
    else
    {
      // Use a threshold for cutting out points which are too close/far away
      if ((distance_value > max_distance) || (distance_value < min_distance))
      {
        xyz_offset += cloud->point_step;
        continue;
      }
    }

    // Unoptimized memcpys: assume fields x, y, z are in random order
    memcpy (&pt[0], &cloud->data[xyz_offset[0]], sizeof (float));
    memcpy (&pt[1], &cloud->data[xyz_offset[1]], sizeof (float));
    memcpy (&pt[2], &cloud->data[xyz_offset[2]], sizeof (float));
    // Check if the point is invalid
    if (isnan (pt[0]) || isnan (pt[1]) || isnan (pt[2]))
    {
      xyz_offset += cloud->point_step;
      continue;
    }
    xyz_offset += cloud->point_step;
    min_p = min_p.min (pt);
    max_p = max_p.max (pt);
  }
  min_pt = min_p;
  max_pt = max_p;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Downsample a Point Cloud using a voxelized grid approach
  * \param output the resultant point cloud message
  */
void
  pcl::VoxelGrid<sensor_msgs::PointCloud2>::applyFilter (PointCloud &output)
{
  // If fields x/y/z are not present, we cannot downsample
  if (x_idx_ == -1 || y_idx_ == -1 || z_idx_ == -1)
  {
    ROS_ERROR ("[pcl::%s::applyFilter] Input dataset doesn't have x-y-z coordinates!", getClassName ().c_str ());
    return;
  }
  // Avoid NANs
  if (leaf_size_[3] == 0)
    leaf_size_[3] = 1;

  int nr_points  = input_->width * input_->height;

  // Copy the header (and thus the frame_id) + allocate enough space for points
  output.header       = input_->header;
  output.height       = 1;                    // downsampling breaks the organized structure
  if (downsample_all_data_)
  {
    output.fields       = input_->fields;
    output.point_step   = input_->point_step;
  }
  else
  {
    output.fields.resize (4);

    output.fields[0] = input_->fields[x_idx_];
    output.fields[0].offset = 0;

    output.fields[1] = input_->fields[y_idx_];
    output.fields[1].offset = 4;

    output.fields[2] = input_->fields[z_idx_];
    output.fields[2].offset = 8;

    output.fields[3].name = "rgba";
    output.fields[3].offset = 12;
    output.fields[3].datatype = sensor_msgs::PointField::FLOAT32;

    output.point_step = 16;
  }
  output.is_bigendian = input_->is_bigendian;
  output.row_step     = input_->row_step;
  output.is_dense     = false;

  Eigen3::Vector4f min_p, max_p;
  Eigen3::Vector4i min_b, max_b, div_b;
  // Get the minimum and maximum dimensions
  if (!filter_field_name_.empty ()) // If we don't want to process the entire cloud...
    getMinMax3D (input_, x_idx_, y_idx_, z_idx_, filter_field_name_, filter_limit_min_, filter_limit_max_, min_p, max_p, filter_limit_negative_);
  else
    getMinMax3D (input_, x_idx_, y_idx_, z_idx_, min_p, max_p);

  // Use multiplications instead of divisions
  Eigen3::Array4f ones = Eigen3::Array4f::Ones ();
  Eigen3::Array4f leaf_size = ones / leaf_size_.array ();

  // @todo fix floor
  //min_b = (min_p.cwise () * leaf_size).cast<int> ();
  //max_b = (max_p.cwise () * leaf_size).cast<int> ();
  // Compute the minimum and maximum bounding box values
  min_b[0] = (int)(floor (min_p[0] * leaf_size[0]));
  max_b[0] = (int)(floor (max_p[0] * leaf_size[0]));
  min_b[1] = (int)(floor (min_p[1] * leaf_size[1]));
  max_b[1] = (int)(floor (max_p[1] * leaf_size[1]));
  min_b[2] = (int)(floor (min_p[2] * leaf_size[2]));
  max_b[2] = (int)(floor (max_p[2] * leaf_size[2]));

  // Compute the number of divisions needed along all axis
  div_b = max_b - min_b + Eigen3::Vector4i::Ones ();
  div_b[3] = 0;

  // Clear the leaves
  leaves_.clear();

  // Create the first xyz_offset, and set up the division multiplier
  Eigen3::Array4i xyz_offset (input_->fields[x_idx_].offset,
                             input_->fields[y_idx_].offset,
                             input_->fields[z_idx_].offset,
                             0);
  Eigen3::Vector4i divb_mul   = Eigen3::Vector4i (1, div_b[0], div_b[0] * div_b[1], 0);
  Eigen3::Vector4f pt  = Eigen3::Vector4f::Zero ();
  Eigen3::Vector4i ijk = Eigen3::Vector4i::Zero ();

  int centroid_size = 4;
  if (downsample_all_data_)
    centroid_size = input_->fields.size ();

  // If we don't want to process the entire cloud, but rather filter points far away from the viewpoint first...
  if (!filter_field_name_.empty ())
  {
    // Get the distance field index
    int distance_idx = pcl::getFieldIndex (*input_, filter_field_name_);

    // @todo fixme
    if (input_->fields[distance_idx].datatype != sensor_msgs::PointField::FLOAT32)
    {
      ROS_ERROR ("[pcl::%s::downsample] Distance filtering requested, but distances are not float/double in the dataset! Only FLOAT32/FLOAT64 distances are supported right now.", getClassName ().c_str ());
      return;
    }

    // First pass: go over all points and insert them into the right leaf
    float distance_value = 0;
    for (int cp = 0; cp < nr_points; ++cp)
    {
      int point_offset = cp * input_->point_step;
      // Get the distance value
      memcpy (&distance_value, &input_->data[point_offset + input_->fields[distance_idx].offset], sizeof (float));

      if (filter_limit_negative_)
      {
        // Use a threshold for cutting out points which inside the interval
        if (distance_value < filter_limit_max_ && distance_value > filter_limit_min_)
        {
          xyz_offset += input_->point_step;
          continue;
        }
      }
      else
      {
        // Use a threshold for cutting out points which are too close/far away
        if (distance_value > filter_limit_max_ || distance_value < filter_limit_min_)
        {
          xyz_offset += input_->point_step;
          continue;
        }
      }

      // Unoptimized memcpys: assume fields x, y, z are in random order
      memcpy (&pt[0], &input_->data[xyz_offset[0]], sizeof (float));
      memcpy (&pt[1], &input_->data[xyz_offset[1]], sizeof (float));
      memcpy (&pt[2], &input_->data[xyz_offset[2]], sizeof (float));

      // Check if the point is invalid
      if (isnan (pt[0]) || isnan (pt[1]) || isnan (pt[2]))
      {
        xyz_offset += input_->point_step;
        continue;
      }

      // @todo fix floor
      //ijk = (pt.cwise () / leaf_size_).cast<int> ();
      ijk[0] = (int)(floor (pt[0] * leaf_size[0]));
      ijk[1] = (int)(floor (pt[1] * leaf_size[1]));
      ijk[2] = (int)(floor (pt[2] * leaf_size[2]));

      // Compute the centroid leaf index
      int idx = (ijk - min_b).dot (divb_mul);
      Leaf& leaf = leaves_[idx];

      if (leaf.nr_points == 0)
      {
        leaf.centroid.resize (centroid_size);
        leaf.centroid.setZero ();
      }

      // Do we need to process all the fields?
      if (!downsample_all_data_)
      {
        leaf.centroid.head<4> () += pt;
      }
      else
      {
        // Copy all the fields
        Eigen3::VectorXf centroid = Eigen3::VectorXf::Zero (centroid_size);
        for (int d = 0; d < centroid_size; ++d)
          memcpy (&centroid[d], &input_->data[point_offset + input_->fields[d].offset], field_sizes_[d]);
        leaf.centroid += centroid;
      }
      ++leaf.nr_points;

      xyz_offset += input_->point_step;
    }
  }
  // No distance filtering, process all data
  else
  {
    // First pass: go over all points and insert them into the right leaf
    for (int cp = 0; cp < nr_points; ++cp)
    {
      // Unoptimized memcpys: assume fields x, y, z are in random order
      memcpy (&pt[0], &input_->data[xyz_offset[0]], sizeof (float));
      memcpy (&pt[1], &input_->data[xyz_offset[1]], sizeof (float));
      memcpy (&pt[2], &input_->data[xyz_offset[2]], sizeof (float));

      // Check if the point is invalid
      if (isnan (pt[0]) || isnan (pt[1]) || isnan (pt[2]))
      {
        xyz_offset += input_->point_step;
        continue;
      }

      // @todo fix floor
      //ijk = (pt.cwise () / leaf_size_).cast<int> ();
      ijk[0] = (int)(floor (pt[0] * leaf_size[0]));
      ijk[1] = (int)(floor (pt[1] * leaf_size[1]));
      ijk[2] = (int)(floor (pt[2] * leaf_size[2]));

      // Compute the centroid leaf index
      int idx = (ijk - min_b).dot (divb_mul);


      Leaf& leaf = leaves_[idx];

      if (leaf.nr_points == 0)
      {
        leaf.centroid.resize (centroid_size);
        leaf.centroid.setZero ();
      }

      // Do we need to process all the fields?
      if (!downsample_all_data_)
      {
        leaf.centroid.head<4> () += pt;
      }
      else
      {
        // Copy all the fields
        Eigen3::VectorXf centroid = Eigen3::VectorXf::Zero (centroid_size);
        int point_offset = cp * input_->point_step;

        for (int d = 0; d < centroid_size; ++d)
          memcpy (&centroid[d], &input_->data[point_offset + input_->fields[d].offset], field_sizes_[d]);
        leaf.centroid += centroid;
      }
      ++leaf.nr_points;

      xyz_offset += input_->point_step;
    }
  }

  // Second pass: go over all leaves and compute centroids
  int nr_p = 0;
  // If we downsample each field, the {x,y,z}_idx_ offsets should correspond in input_ and output
  if (downsample_all_data_)
    xyz_offset = Eigen3::Array4i (output.fields[x_idx_].offset,
                                 output.fields[y_idx_].offset,
                                 output.fields[z_idx_].offset,
                                 0);
  else
    // If not, we must have created a new xyzw cloud
    xyz_offset = Eigen3::Array4i (0, 4, 8, 12);

  Eigen3::VectorXf centroid;
  output.width = leaves_.size();
  output.row_step = output.point_step * output.width;
  output.data.resize (leaves_.size()* output.point_step);

  for (std::map<size_t, Leaf>::const_iterator it=leaves_.begin(); it!=leaves_.end(); ++it)
  {
    const Leaf& leaf = it->second;
    // Normalize the centroid
    float norm_pts = 1.0f / leaf.nr_points;
    centroid = leaf.centroid * norm_pts;

    // Do we need to process all the fields?
    if (!downsample_all_data_)
    {
      // Copy the data
      memcpy (&output.data[xyz_offset[0]], &centroid[0], sizeof (float));
      memcpy (&output.data[xyz_offset[1]], &centroid[1], sizeof (float));
      memcpy (&output.data[xyz_offset[2]], &centroid[2], sizeof (float));
      xyz_offset += output.point_step;
    }
    else
    {
      int point_offset = nr_p * output.point_step;
      // Copy all the fields
      for (size_t d = 0; d < output.fields.size (); ++d)
        memcpy (&output.data[point_offset + output.fields[d].offset], &centroid[d], field_sizes_[d]);
    }
    ++nr_p;
  }
}


