/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: voxel_grid.hpp 32774 2010-09-20 12:08:49Z lorenz $
 *
 */

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Get the minimum and maximum values on each of the 3 (x-y-z) dimensions
  * in a given pointcloud, without considering points outside of a distance threshold from the laser origin
  * \param cloud the point cloud data message
  * \param distance_field_name the field name that contains the distance values
  * \param min_distance the minimum distance a point will be considered from
  * \param max_distance the maximum distance a point will be considered to
  * \param min_pt the resultant minimum bounds
  * \param max_pt the resultant maximum bounds
  * \param limit_negative if set to true, then all points outside of the interval (min_distance;max_distace) are considered
  */
template <typename PointT> inline void
  pcl::getMinMax3D (const typename pcl::PointCloud<PointT>::ConstPtr &cloud,
                    const std::string &distance_field_name, float min_distance, float max_distance,
                    Eigen3::Vector4f &min_pt, Eigen3::Vector4f &max_pt, bool limit_negative)
{
  Eigen3::Array4f min_p, max_p;
  min_p.setConstant (FLT_MAX);
  max_p.setConstant (-FLT_MAX);

  // Get the fields list and the distance field index
  std::vector<sensor_msgs::PointField> fields;
  int distance_idx = pcl::getFieldIndex (*cloud, distance_field_name, fields);

  float distance_value;
  for (size_t i = 0; i < cloud->points.size (); ++i)
  {
    // Get the distance value
    uint8_t* pt_data = (uint8_t*)&cloud->points[i];
    memcpy (&distance_value, pt_data + fields[distance_idx].offset, sizeof (float));

    if (limit_negative)
    {
      // Use a threshold for cutting out points which inside the interval
      if ((distance_value < max_distance) && (distance_value > min_distance))
        continue;
    }
    else
    {
      // Use a threshold for cutting out points which are too close/far away
      if ((distance_value > max_distance) || (distance_value < min_distance))
        continue;
    }

    // Check if the point is invalid
    if (isnan (cloud->points[i].x) || isnan (cloud->points[i].y) || isnan (cloud->points[i].z))
      continue;
    // Create the point structure and get the min/max
    pcl::Array4fMap pt = cloud->points[i].getArray4fMap ();
    min_p = min_p.min (pt);
    max_p = max_p.max (pt);
  }
  min_pt = min_p;
  max_pt = max_p;
}

  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief Helper functor structure for copying data between an Eigen3::VectorXf and a PointT. */
  template <typename PointT>
    struct NdCopyEigenPointFunctor
  {
    NdCopyEigenPointFunctor (const Eigen3::VectorXf &p1, PointT &p2) : p1_ (p1), p2_ (p2), f_idx_ (0) { }

    template<typename Key> inline void operator() () { boost::fusion::at_key<Key> (p2_) = p1_[f_idx_++]; }

    private:
      const Eigen3::VectorXf &p1_;
      PointT &p2_;
      int f_idx_;
  };

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Downsample a Point Cloud using a voxelized grid approach
  * \param output the resultant point cloud message
  */
template <typename PointT> void
  pcl::VoxelGrid<PointT>::applyFilter (PointCloud &output)
{
  // Has the input dataset been set already?
  if (!input_)
  {
    ROS_WARN ("[pcl::%s::applyFilter] No input dataset given!", getClassName ().c_str ());
    return;
  }
  // Avoid NANs
  if (leaf_size_[3] == 0)
    leaf_size_[3] = 1;

  // Copy the header (and thus the frame_id) + allocate enough space for points
  output.header       = input_->header;
  output.height       = 1;                    // downsampling breaks the organized structure
  output.is_dense     = false;

  Eigen3::Vector4f min_p, max_p;
  // Get the minimum and maximum dimensions
  if (!filter_field_name_.empty ()) // If we don't want to process the entire cloud...
    getMinMax3D<PointT>(input_, filter_field_name_, filter_limit_min_, filter_limit_max_, min_p, max_p, filter_limit_negative_);
  else
    getMinMax3D<PointT>(*input_, min_p, max_p);

  // Use multiplications instead of divisions
  Eigen3::Array4f ones = Eigen3::Array4f::Ones ();
  Eigen3::Array4f leaf_size = ones / leaf_size_.array ();

  Eigen3::Vector4i min_b, max_b, div_b;
  // @todo fix floor
  //min_b = (min_p.cwise () * leaf_size).cast<int> ();
  //max_b = (max_p.cwise () * leaf_size).cast<int> ();
  // Compute the minimum and maximum bounding box values
  min_b[0] = (int)(floor (min_p[0] * leaf_size[0]));
  max_b[0] = (int)(floor (max_p[0] * leaf_size[0]));
  min_b[1] = (int)(floor (min_p[1] * leaf_size[1]));
  max_b[1] = (int)(floor (max_p[1] * leaf_size[1]));
  min_b[2] = (int)(floor (min_p[2] * leaf_size[2]));
  max_b[2] = (int)(floor (max_p[2] * leaf_size[2]));

  // Compute the number of divisions needed along all axis
  div_b = max_b - min_b + Eigen3::Vector4i::Ones ();
  div_b[3] = 0;

  // Clear the leaves
  leaves_.clear();

  // Set up the division multiplier
  Eigen3::Vector4i divb_mul = Eigen3::Vector4i (1, div_b[0], div_b[0] * div_b[1], 0);
  Eigen3::Vector4i ijk = Eigen3::Vector4i::Zero ();

  int centroid_size = 4;
  if (downsample_all_data_)
    centroid_size = boost::mpl::size<FieldList>::value;

  // If we don't want to process the entire cloud, but rather filter points far away from the viewpoint first...
  if (!filter_field_name_.empty ())
  {
    // Get the distance field index
    std::vector<sensor_msgs::PointField> fields;
    int distance_idx = pcl::getFieldIndex (*input_, filter_field_name_, fields);
    if (distance_idx == -1)
      ROS_WARN ("[pcl::%s::applyFilter] Invalid filter field name. Index is %d.", getClassName ().c_str (), distance_idx);

    // First pass: go over all points and insert them into the right leaf
    for (size_t cp = 0; cp < input_->points.size (); ++cp)
    {
      // Check if the point is invalid
      if (isnan (input_->points[cp].x) || isnan (input_->points[cp].y) || isnan (input_->points[cp].z))
        continue;

      // Get the distance value
      uint8_t* pt_data = (uint8_t*)&input_->points[cp];
      float distance_value = 0;
      memcpy (&distance_value, pt_data + fields[distance_idx].offset, sizeof (float));

      if (filter_limit_negative_)
      {
        // Use a threshold for cutting out points which inside the interval
        if ((distance_value < filter_limit_max_) && (distance_value > filter_limit_min_))
          continue;
      }
      else
      {
        // Use a threshold for cutting out points which are too close/far away
        if ((distance_value > filter_limit_max_) || (distance_value < filter_limit_min_))
          continue;
      }

      Eigen3::Vector4f pt (input_->points[cp].x, input_->points[cp].y, input_->points[cp].z, 0);

      // @todo fix floor
      //ijk = (pt.cwise () / leaf_size_).cast<int> ();
      ijk[0] = (int)(floor (pt[0] * leaf_size[0]));
      ijk[1] = (int)(floor (pt[1] * leaf_size[1]));
      ijk[2] = (int)(floor (pt[2] * leaf_size[2]));

      // Compute the centroid leaf index
      int idx = (ijk - min_b).dot (divb_mul);
      Leaf& leaf = leaves_[idx];
      if (leaf.nr_points == 0)
      {
        leaf.centroid.resize (centroid_size);
        leaf.centroid.setZero ();
      }

      // Do we need to process all the fields?
      if (!downsample_all_data_)
      {
        leaf.centroid.template head<4> () += pt;
      }
      else
      {
        // Copy all the fields
        Eigen3::VectorXf centroid = Eigen3::VectorXf::Zero (centroid_size);
        pcl::for_each_type <FieldList> (NdCopyPointEigenFunctor <PointT> (input_->points[cp], centroid));
        leaf.centroid += centroid;
      }
      ++leaf.nr_points;
    }
  }
  // No distance filtering, process all data
  else
  {
    // First pass: go over all points and insert them into the right leaf
    for (size_t cp = 0; cp < input_->points.size (); ++cp)
    {
      // Check if the point is invalid
      if (isnan (input_->points[cp].x) || isnan (input_->points[cp].y) || isnan (input_->points[cp].z))
        continue;

      Eigen3::Vector4f pt (input_->points[cp].x, input_->points[cp].y, input_->points[cp].z, 0);

      // @todo fix floor
      //ijk = (pt.cwise () / leaf_size_).cast<int> ();
      ijk[0] = (int)(floor (pt[0] * leaf_size[0]));
      ijk[1] = (int)(floor (pt[1] * leaf_size[1]));
      ijk[2] = (int)(floor (pt[2] * leaf_size[2]));

      // Compute the centroid leaf index
      int idx = (ijk - min_b).dot (divb_mul);
      Leaf& leaf = leaves_[idx];
      if (leaf.nr_points == 0)
      {
        leaf.centroid.resize (centroid_size);
        leaf.centroid.setZero ();
      }

      // Do we need to process all the fields?
      if (!downsample_all_data_)
      {
        leaf.centroid.template head<4> () += pt;
      }
      else
      {
        // Copy all the fields
        Eigen3::VectorXf centroid = Eigen3::VectorXf::Zero (centroid_size);
        pcl::for_each_type <FieldList> (NdCopyPointEigenFunctor <PointT> (input_->points[cp], centroid));
        leaf.centroid += centroid;
      }
      ++leaf.nr_points;
    }
  }

  // Second pass: go over all leaves and compute centroids
  output.points.clear ();
  output.points.reserve (leaves_.size());
  for (typename std::map<size_t, Leaf>::const_iterator it=leaves_.begin(); it!=leaves_.end(); ++it)
  {
    const Leaf& leaf = it->second;
    output.points.push_back(PointT());
    PointT& point = output.points.back();

    // Normalize the centroid
    float norm_pts = 1.0f / leaf.nr_points;
    Eigen3::VectorXf centroid = leaf.centroid * norm_pts;

    // Do we need to process all the fields?
    if (!downsample_all_data_)
    {
      point.x = centroid[0];
      point.y = centroid[1];
      point.z = centroid[2];
    }
    else
      pcl::for_each_type <FieldList> (pcl::NdCopyEigenPointFunctor <PointT> (centroid, point));
  }
  output.width = output.points.size();
}

