/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: pcd_read.cpp 33162 2010-03-10 07:41:56Z rusu $
 *
 */

/** \author Radu Bogdan Rusu */

#include <fstream>
#include <fcntl.h>
#include <string>
#include <stdlib.h>
#include <sys/mman.h>
#include <boost/algorithm/string.hpp>
#include <boost/filesystem.hpp>
#include <boost/numeric/conversion/cast.hpp>
#include "pcl/io/io.h"
#include "pcl/io/pcd_io.h"


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Load a point cloud data header from a PCD v.6 file.
 * Returns -1 on error, 0 on success 
 * \param file_name the name of the file to load
 * \param cloud the resultant point cloud dataset (only the header will be filled)
 */
int
  pcl::io::loadPCDFileHeader (const std::string &file_name, sensor_msgs::PointCloud2 &cloud)
{
  // Check if the file exists
  if (!boost::filesystem::exists (file_name))
    return (-1);

  cloud.width = cloud.height = cloud.point_step = cloud.row_step = 0;
  int nr_points = 0;
  std::ifstream fs;
  std::string line;

  int specified_channel_count = 0;

  // Open file
  fs.open (file_name.c_str ());
  if (!fs.is_open () || fs.fail ())
    return (-1);

  // field_sizes represents the size of one element in a field (e.g., float = 4, char = 1)
  // field_counts represents the number of elements in a field (e.g., x = 1, normal_x = 1, fpfh = 33)
  std::vector<int> field_sizes, field_counts;
  // field_types represents the type of data in a field (e.g., F = float, U = unsigned)
  std::vector<char> field_types;
  std::vector<std::string> st;

  // Read the header and fill it in with wonderful values
  try
  {
    while (!fs.eof ())
    {
      getline (fs, line);
      if (line == "")
        continue;

      // Tokenize the line
      boost::split (st, line, boost::is_any_of (" "), boost::token_compress_on);

      std::string line_type = st.at (0);

      // ---[ Perform checks to see what does this line represent
      if (line_type.substr (0, 1) == "#")
        continue;

      // Get the field indices (check for COLUMNS too for backwards compatibility)
      if ( (line_type.substr (0, 6) == "FIELDS") || (line_type.substr (0, 7) == "COLUMNS") )
      {
        specified_channel_count = st.size () - 1;

        // Allocate enough memory to accommodate all fields
        cloud.fields.resize (specified_channel_count);
        for (int i = 0; i < specified_channel_count; ++i)
        {
          std::string col_type = st.at (i + 1);
          cloud.fields[i].name = col_type;
        }

        // Default the sizes and the types of each field to float32 to avoid crashes while using older PCD files
        int offset = 0;
        for (int i = 0; i < specified_channel_count; ++i, offset += 4)
        {
          cloud.fields[i].offset   = offset;
          cloud.fields[i].datatype = sensor_msgs::PointField::FLOAT32;
          cloud.fields[i].count    = 1;
        }
        cloud.point_step = offset;
        continue;
      }

      // Get the field sizes
      if (line_type.substr (0, 4) == "SIZE")
      {
        specified_channel_count = st.size () - 1;

        // Allocate enough memory to accommodate all fields
        if (specified_channel_count != (int)cloud.fields.size ())
          throw "The number of elements in <SIZE> differs than the number of elements in <FIELDS>!";

        // Resize to accommodate the number of values
        field_sizes.resize (specified_channel_count);

        int offset = 0;
        for (int i = 0; i < specified_channel_count; ++i)
        {
          int col_type = atoi (st.at (i + 1).c_str ());
          cloud.fields[i].offset = offset;                // estimate and save the data offsets
          offset += col_type;
          field_sizes[i] = col_type;                      // save a temporary copy
        }
        cloud.point_step = offset;
        //if (cloud.width != 0)
          //cloud.row_step   = cloud.point_step * cloud.width;
        continue;
      }

      // Get the field types
      if (line_type.substr (0, 4) == "TYPE")
      {
        if (field_sizes.empty ())
          throw "TYPE of FIELDS specified before SIZE in header!";

        specified_channel_count = st.size () - 1;

        // Allocate enough memory to accommodate all fields
        if (specified_channel_count != (int)cloud.fields.size ())
          throw "The number of elements in <TYPE> differs than the number of elements in <FIELDS>!";

        field_types.resize (specified_channel_count);

        for (int i = 0; i < specified_channel_count; ++i)
        {
          field_types[i] = st.at (i + 1).c_str ()[0];
          cloud.fields[i].datatype = getFieldType (field_sizes[i], field_types[i]);
        }
        continue;
      }

      // Get the field counts
      if (line_type.substr (0, 5) == "COUNT")
      {
        if (field_sizes.empty () || field_types.empty ())
          throw "COUNT of FIELDS specified before SIZE or TYPE in header!";

        specified_channel_count = st.size () - 1;

        // Allocate enough memory to accommodate all fields
        if (specified_channel_count != (int)cloud.fields.size ())
          throw "The number of elements in <COUNT> differs than the number of elements in <FIELDS>!";

        field_counts.resize (specified_channel_count);

        int offset = 0;
        for (int i = 0; i < specified_channel_count; ++i)
        {
          int col_count = atoi (st.at (i + 1).c_str ());
          cloud.fields[i].count = col_count;
          offset += col_count * field_sizes[i];
        }
        // Adjust the offset for count (number of elements)
        cloud.point_step = offset;
        continue;
      }

      // Get the width of the data (organized point cloud dataset)
      if (line_type.substr (0, 5) == "WIDTH")
      {
        cloud.width = atoi (st.at (1).c_str ());
        if (cloud.point_step != 0)
          cloud.row_step = cloud.point_step * cloud.width;      // row_step only makes sense for organized datasets
        continue;
      }

      // Get the height of the data (organized point cloud dataset)
      if (line_type.substr (0, 6) == "HEIGHT")
      {
        cloud.height = atoi (st.at (1).c_str ());
        if (cloud.height == 1)
          cloud.is_dense = false;
        else
          cloud.is_dense = true;
        continue;
      }

      // Get the number of points
      if (line_type.substr (0, 6) == "POINTS")
      {
        nr_points = atoi (st.at (1).c_str ());
        // Need to allocate: N * point_step
        cloud.data.resize (nr_points * cloud.point_step);
        continue;
      }

      break;
    }
  }
  catch (const char *exception)
  {
    ROS_ERROR ("[pcl::io::loadPCDFile] %s", exception);
    return (-1);
  }

  // Compatibility with older PCD file versions
  if (cloud.width == 0 && cloud.height == 0)
  {
    cloud.width  = nr_points;
    cloud.height = 1;
    cloud.row_step = cloud.point_step * cloud.width;      // row_step only makes sense for organized datasets
  }
  assert (cloud.row_step != 0);       // If row_step = 0, either point_step was not set or width doesn't exist

  // if both height/width are not given, assume an unorganized dataset
  if (cloud.height == 0)
  {
    cloud.height = 1;
    ROS_WARN ("[pcl::io::loadPCDFile] no HEIGHT given, setting to 1 (unorganized).");
    if (cloud.width == 0)
      cloud.width  = nr_points;
    cloud.is_dense = false;
  }
  else
  {
    if (cloud.width == 0)
    {
      ROS_ERROR ("[pcl::io::loadPCDFile] HEIGHT given (%d) but no WIDTH!", cloud.height);
      return (-1);
    }
  }

  if (int(cloud.width * cloud.height) != nr_points)
  {
    ROS_ERROR ("[pcl::io::loadPCDFile] HEIGHT (%d) x WIDTH (%d) != number of points (%d)", cloud.height, cloud.width, nr_points);
    return (-1);
  }

  // Close file
  fs.close ();

  return (0);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Load point cloud data from a PCD v.6 file containing n-D points.
 * Returns -1 on error, > 0 on success (denoting the underlying PCD file type: 0=ascii, 1=binary)
 * \note All lines besides:
 * - the ones beginning with # (treated as comments)
 * - FIELDS ...
 * - SIZE ...
 * - TYPE ...
 * - COUNT ...
 * - WIDTH ...
 * - HEIGHT ...
 * - POINTS ...
 * - DATA ascii/binary
 * ...are intepreted as data points.
 * \param file_name the name of the file to load
 * \param cloud the resultant point cloud dataset
 */
int
  pcl::io::loadPCDFile (const std::string &file_name, sensor_msgs::PointCloud2 &cloud)
{
  // Check if the file exists
  if (!boost::filesystem::exists (file_name))
    return (-1);

  cloud.width = cloud.height = cloud.point_step = cloud.row_step = 0;
  bool binary_data = false;
  int nr_points = 0;
  std::ifstream fs;
  std::string line;

  int specified_channel_count = 0;

  int idx = 0;

  // Open file
  fs.open (file_name.c_str ());
  if (!fs.is_open () || fs.fail ())
    return (-1);

  // field_sizes represents the size of one element in a field (e.g., float = 4, char = 1)
  // field_counts represents the number of elements in a field (e.g., x = 1, normal_x = 1, fpfh = 33)
  std::vector<int> field_sizes, field_counts;
  // field_types represents the type of data in a field (e.g., F = float, U = unsigned)
  std::vector<char> field_types;
  std::vector<std::string> st;

  // Read the header and fill it in with wonderful values
  try
  {
    while (!fs.eof ())
    {
      getline (fs, line);
      if (line == "")
        continue;

      // Tokenize the line
      boost::split (st, line, boost::is_any_of (" "), boost::token_compress_on);

      std::string line_type = st.at (0);

      // ---[ Perform checks to see what does this line represent
      if (line_type.substr (0, 1) == "#")
        continue;

      // Get the field indices (check for COLUMNS too for backwards compatibility)
      if ( (line_type.substr (0, 6) == "FIELDS") || (line_type.substr (0, 7) == "COLUMNS") )
      {
        specified_channel_count = st.size () - 1;

        // Allocate enough memory to accommodate all fields
        cloud.fields.resize (specified_channel_count);
        for (int i = 0; i < specified_channel_count; ++i)
        {
          std::string col_type = st.at (i + 1);
          cloud.fields[i].name = col_type;
        }

        // Default the sizes and the types of each field to float32 to avoid crashes while using older PCD files
        int offset = 0;
        for (int i = 0; i < specified_channel_count; ++i, offset += 4)
        {
          cloud.fields[i].offset   = offset;
          cloud.fields[i].datatype = sensor_msgs::PointField::FLOAT32;
          cloud.fields[i].count    = 1;
        }
        cloud.point_step = offset;
        continue;
      }

      // Get the field sizes
      if (line_type.substr (0, 4) == "SIZE")
      {
        specified_channel_count = st.size () - 1;

        // Allocate enough memory to accommodate all fields
        if (specified_channel_count != (int)cloud.fields.size ())
          throw "The number of elements in <SIZE> differs than the number of elements in <FIELDS>!";

        // Resize to accommodate the number of values
        field_sizes.resize (specified_channel_count);

        int offset = 0;
        for (int i = 0; i < specified_channel_count; ++i)
        {
          int col_type = atoi (st.at (i + 1).c_str ());
          cloud.fields[i].offset = offset;                // estimate and save the data offsets
          offset += col_type;
          field_sizes[i] = col_type;                      // save a temporary copy
        }
        cloud.point_step = offset;
        //if (cloud.width != 0)
          //cloud.row_step   = cloud.point_step * cloud.width;
        continue;
      }

      // Get the field types
      if (line_type.substr (0, 4) == "TYPE")
      {
        if (field_sizes.empty ())
          throw "TYPE of FIELDS specified before SIZE in header!";

        specified_channel_count = st.size () - 1;

        // Allocate enough memory to accommodate all fields
        if (specified_channel_count != (int)cloud.fields.size ())
          throw "The number of elements in <TYPE> differs than the number of elements in <FIELDS>!";

        field_types.resize (specified_channel_count);

        for (int i = 0; i < specified_channel_count; ++i)
        {
          field_types[i] = st.at (i + 1).c_str ()[0];
          cloud.fields[i].datatype = getFieldType (field_sizes[i], field_types[i]);
        }
        continue;
      }

      // Get the field counts
      if (line_type.substr (0, 5) == "COUNT")
      {
        if (field_sizes.empty () || field_types.empty ())
          throw "COUNT of FIELDS specified before SIZE or TYPE in header!";

        specified_channel_count = st.size () - 1;

        // Allocate enough memory to accommodate all fields
        if (specified_channel_count != (int)cloud.fields.size ())
          throw "The number of elements in <COUNT> differs than the number of elements in <FIELDS>!";

        field_counts.resize (specified_channel_count);

        int offset = 0;
        for (int i = 0; i < specified_channel_count; ++i)
        {
          int col_count = atoi (st.at (i + 1).c_str ());
          cloud.fields[i].count = col_count;
          offset += col_count * field_sizes[i];
        }
        // Adjust the offset for count (number of elements)
        cloud.point_step = offset;
        continue;
      }

      // Get the width of the data (organized point cloud dataset)
      if (line_type.substr (0, 5) == "WIDTH")
      {
        cloud.width = atoi (st.at (1).c_str ());
        if (cloud.point_step != 0)
          cloud.row_step = cloud.point_step * cloud.width;      // row_step only makes sense for organized datasets
        continue;
      }

      // Get the height of the data (organized point cloud dataset)
      if (line_type.substr (0, 6) == "HEIGHT")
      {
        cloud.height = atoi (st.at (1).c_str ());
        if (cloud.height == 1)
          cloud.is_dense = false;
        else
          cloud.is_dense = true;
        continue;
      }

      // Get the number of points
      if (line_type.substr (0, 6) == "POINTS")
      {
        nr_points = atoi (st.at (1).c_str ());
        // Need to allocate: N * point_step
        cloud.data.resize (nr_points * cloud.point_step);
        continue;
      }

      // Check DATA type
      if (line_type.substr (0, 4) == "DATA")
      {
        if (st.at (1).substr (0, 6) == "binary")
        {
          binary_data = true;
          break;
        }
        continue;
      }

      // Nothing of the above? We must have points then
      // Convert the first token to float and use it as the first point coordinate
      if (idx >= nr_points)
      {
        ROS_WARN ("[pcl::io::loadPCDFile] input file %s has more points than advertised (%d)!", file_name.c_str (), nr_points);
        break;
      }

      // Copy data
      for (size_t d = 0; d < cloud.fields.size (); ++d)
      {
        for (size_t c = 0; c < cloud.fields[d].count; ++c)
        {
          switch (cloud.fields[d].datatype)
          {
            case sensor_msgs::PointField::INT8:
            {
              char value = (char)atoi (st.at (d + c).c_str ());
              memcpy (&cloud.data[idx * cloud.point_step + cloud.fields[d].offset + c * sizeof (char)], &value, sizeof (char));
              break;
            }
            case sensor_msgs::PointField::UINT8:
            {
              unsigned char value = atoi (st.at (d + c).c_str ());
              memcpy (&cloud.data[idx * cloud.point_step + cloud.fields[d].offset + c * sizeof (unsigned char)], &value, sizeof (unsigned char));
              break;
            }
            case sensor_msgs::PointField::INT16:
            {
              short value = atoi (st.at (d + c).c_str ());
              memcpy (&cloud.data[idx * cloud.point_step + cloud.fields[d].offset + c * sizeof (short)], &value, sizeof (short));
              break;
            }
            case sensor_msgs::PointField::UINT16:
            {
              unsigned short value = atoi (st.at (d + c).c_str ());
              memcpy (&cloud.data[idx * cloud.point_step + cloud.fields[d].offset + c * sizeof (unsigned short)], &value, sizeof (unsigned short));
              break;
            }
            case sensor_msgs::PointField::INT32:
            {
              int value = atoi (st.at (d + c).c_str ());
              memcpy (&cloud.data[idx * cloud.point_step + cloud.fields[d].offset + c * sizeof (int)], &value, sizeof (int));
              break;
            }
            case sensor_msgs::PointField::UINT32:
            {
              unsigned int value = atoi (st.at (d + c).c_str ());
              memcpy (&cloud.data[idx * cloud.point_step + cloud.fields[d].offset + c * sizeof (unsigned int)], &value, sizeof (unsigned int));
              break;
            }
            case sensor_msgs::PointField::FLOAT32:
            {
              float value = atof (st.at (d + c).c_str ());
              memcpy (&cloud.data[idx * cloud.point_step + cloud.fields[d].offset + c * sizeof (float)], &value, sizeof (float));
              break;
            }
            case sensor_msgs::PointField::FLOAT64:
            {
              double value = atof (st.at (d + c).c_str ());
              memcpy (&cloud.data[idx * cloud.point_step + cloud.fields[d].offset + c * sizeof (double)], &value, sizeof (double));
              break;
            }
            default:
              ROS_WARN ("Incorrect field data type specified (%d)!",cloud.fields[d].datatype);
              break;
          }
        }
      }

      idx++;
    }
  }
  catch (const char *exception)
  {
    ROS_ERROR ("[pcl::io::loadPCDFile] %s", exception);
    return (-1);
  }

  // Compatibility with older PCD file versions
  if (cloud.width == 0 && cloud.height == 0)
  {
    cloud.width  = nr_points;
    cloud.height = 1;
    cloud.row_step = cloud.point_step * cloud.width;      // row_step only makes sense for organized datasets
  }
  assert (cloud.row_step != 0);       // If row_step = 0, either point_step was not set or width doesn't exist

  // if both height/width are not given, assume an unorganized dataset
  if (cloud.height == 0)
  {
    cloud.height = 1;
    ROS_WARN ("[pcl::io::loadPCDFile] no HEIGHT given, setting to 1 (unorganized).");
    if (cloud.width == 0)
      cloud.width  = nr_points;
    cloud.is_dense = false;
  }
  else
  {
    if (cloud.width == 0)
    {
      ROS_ERROR ("[pcl::io::loadPCDFile] HEIGHT given (%d) but no WIDTH!", cloud.height);
      return (-1);
    }
  }

  if (int(cloud.width * cloud.height) != nr_points)
  {
    ROS_ERROR ("[pcl::io::loadPCDFile] HEIGHT (%d) x WIDTH (%d) != number of points (%d)", cloud.height, cloud.width, nr_points);
    return (-1);
  }

  // @todo fixme
  cloud.is_bigendian = false;

  // Close file
  fs.close ();

  /// ---[ Binary mode only
  /// We must re-open the file and read with mmap () for binary
  if (binary_data)
  {
    // Open for reading
    int fd = open (file_name.c_str (), O_RDONLY);
    if (fd == -1)
      return (-1);

    // Prepare the map
    char *map = (char*)mmap (0, cloud.data.size (), PROT_READ, MAP_SHARED, fd, getpagesize ());
    if (map == MAP_FAILED)
    {
      close (fd);
      return (-1);
    }

    // Copy the data
    memcpy (&cloud.data[0], &map[0], cloud.data.size ());

    // Unmap the pages of memory
    if (munmap (map, cloud.data.size ()) == -1)
    {
      close (fd);
      return (-1);
    }
    close (fd);
  }

  if ( (idx != nr_points) && (!binary_data) )
  {
    ROS_ERROR ("[pcl::io::loadPCDFile] Number of points read (%d) is different than expected (%d)", idx, nr_points);
    return (-1);
  }

  return (binary_data);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Generate the header of a PCD v.5 file format
  * \param cloud the point cloud data message
  */
std::string
  pcl::io::generateHeader (const sensor_msgs::PointCloud2 &cloud)
{
  std::ostringstream oss;

  oss << "# .PCD v.6 - Point Cloud Data file format"
         "\nFIELDS ";

  oss << getFieldsList (cloud);
  oss << "\nSIZE ";

  // Write the SIZE of each field
  for (size_t d = 0; d < cloud.fields.size () - 1; ++d)
    oss << pcl::getFieldSize (cloud.fields[d].datatype) << " ";
  oss << pcl::getFieldSize (cloud.fields[cloud.fields.size () - 1].datatype) << "\nTYPE ";

  // Write the TYPE of each field
  for (size_t d = 0; d < cloud.fields.size () - 1; ++d)
    oss << pcl::getFieldType (cloud.fields[d].datatype) << " ";
  oss << pcl::getFieldType (cloud.fields[cloud.fields.size () - 1].datatype) << "\nCOUNT ";
  
  // Write the TYPE of each field
  for (size_t d = 0; d < cloud.fields.size () - 1; ++d)
  {
    int count = cloud.fields[d].count;
    if (count == 0) 
      count = 1;          // we simply cannot tolerate 0 counts (coming from older converter code)

    oss << count << " ";
  }
  int count = cloud.fields[cloud.fields.size () - 1].count;
  if (count == 0)
    count = 1;
  oss << count <<
         "\nWIDTH " << cloud.width << "\nHEIGHT " << cloud.height << "\nPOINTS " << cloud.width * cloud.height << "\n";

  return (oss.str ());
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Save point cloud data to a PCD file containing n-D points
  * \param file_name the output file name
  * \param cloud the point cloud data message
  * \param binary_mode true for binary mode, false for ascii
  */
int
  pcl::io::savePCDFile (std::string file_name, const sensor_msgs::PointCloud2 &cloud, bool binary_mode)
{
  if (binary_mode)
    return (pcl::io::savePCDFileBinary (file_name, cloud));
  else
    return (pcl::io::savePCDFileASCII (file_name, cloud, 5));
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Save point cloud data to a PCD file containing n-D points, in ASCII format
  * \param file_name the output file name
  * \param cloud the point cloud data message
  * \param precision the specified output numeric stream precision
  */
int
  pcl::io::savePCDFileASCII (const std::string &file_name, const sensor_msgs::PointCloud2 &cloud, int precision)
{
  if (cloud.data.empty ())
  {
    ROS_ERROR ("[pcl::io::savePCDFileASCII] Input point cloud has no data!");
    return (-1);
  }

  std::ofstream fs;
  fs.precision (precision);
  fs.open (file_name.c_str ());      // Open file

  int nr_points  = cloud.width * cloud.height;
  int point_size = cloud.data.size () / nr_points;

  // Write the header information
  fs << generateHeader (cloud) << "DATA ascii\n";

  // Iterate through the points
  for (int i = 0; i < nr_points; ++i)
  {
    for (size_t d = 0; d < cloud.fields.size (); ++d)
    {
      int count = cloud.fields[d].count;
      if (count == 0) 
        count = 1;          // we simply cannot tolerate 0 counts (coming from older converter code)

      for (int c = 0; c < count; ++c)
      {
        switch (cloud.fields[d].datatype)
        {
          case sensor_msgs::PointField::INT8:
          {
            char value;
            memcpy (&value, &cloud.data[i * point_size + cloud.fields[d].offset + c * sizeof (char)], sizeof (char));
            fs << boost::numeric_cast<int>(value);
            break;
          }
          case sensor_msgs::PointField::UINT8:
          {
            unsigned char value;
            memcpy (&value, &cloud.data[i * point_size + cloud.fields[d].offset + c * sizeof (unsigned char)], sizeof (unsigned char));
            fs << boost::numeric_cast<int>(value);
            break;
          }
          case sensor_msgs::PointField::INT16:
          {
            short value;
            memcpy (&value, &cloud.data[i * point_size + cloud.fields[d].offset + c * sizeof (short)], sizeof (short));
            fs << boost::numeric_cast<int>(value);
            break;
          }
          case sensor_msgs::PointField::UINT16:
          {
            unsigned short value;
            memcpy (&value, &cloud.data[i * point_size + cloud.fields[d].offset + c * sizeof (unsigned short)], sizeof (unsigned short));
            fs << boost::numeric_cast<int>(value);
            break;
          }
          case sensor_msgs::PointField::INT32:
          {
            int value;
            memcpy (&value, &cloud.data[i * point_size + cloud.fields[d].offset + c * sizeof (int)], sizeof (int));
            fs << value;
            break;
          }
          case sensor_msgs::PointField::UINT32:
          {
            unsigned int value;
            memcpy (&value, &cloud.data[i * point_size + cloud.fields[d].offset + c * sizeof (unsigned int)], sizeof (unsigned int));
            fs << value;
            break;
          }
          case sensor_msgs::PointField::FLOAT32:
          {
            float value;
            memcpy (&value, &cloud.data[i * point_size + cloud.fields[d].offset + c * sizeof (float)], sizeof (float));
            fs << value;
            break;
          }
          case sensor_msgs::PointField::FLOAT64:
          {
            double value;
            memcpy (&value, &cloud.data[i * point_size + cloud.fields[d].offset + c * sizeof (double)], sizeof (double));
            fs << value;
            break;
          }
          default:
            ROS_WARN ("Incorrect field data type specified (%d)!",cloud.fields[d].datatype);
            break;
        }

        if (d < cloud.fields.size () - 1 || c < (int)cloud.fields[d].count - 1)
          fs << " ";
      }
    }
    fs << std::endl;
  }
  fs.close ();              // Close file
  return (0);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Save point cloud data to a PCD file containing n-D points, in BINARY format
  * \param file_name the output file name
  * \param cloud the point cloud data message
  */
int
  pcl::io::savePCDFileBinary (const std::string &file_name, const sensor_msgs::PointCloud2 &cloud)
{
  if (cloud.data.empty ())
  {
    ROS_ERROR ("[pcl::io::savePCDFileBinary] Input point cloud has no data!");
    return (-1);
  }
  int data_idx = 0;
  std::ofstream fs;
  // Open file
  fs.open (file_name.c_str ());

  // Write the header information
  fs << generateHeader (cloud) << "DATA binary\n";

  data_idx = fs.tellp ();
  // Close file
  fs.close ();

  if (data_idx > getpagesize ())
  {
    ROS_ERROR ("[pcl::io::savePCDFileBinary] Header size (%d) is bigger than page size (%d)! Reduce the number of channels or save in ASCII format.", data_idx, getpagesize ());
    return (-1);
  }

  // Open for writing
  int fd = open (file_name.c_str (), O_RDWR);
  if (fd < 0)
  {
    ROS_ERROR ("[pcl::io::savePCDFileBinary] Error during open ()!");
    return (-1);
  }

  // Stretch the file size to the size of the data
  int result = lseek (fd, getpagesize () + cloud.data.size () - 1, SEEK_SET);
  if (result < 0)
  {
    close (fd);
    ROS_ERROR ("[pcl::io::savePCDFileBinary] Error during lseek ()!");
    return (-1);
  }
  // Write a bogus entry so that the new file size comes in effect
  result = write (fd, "", 1);
  if (result != 1)
  {
    close (fd);
    ROS_ERROR ("[pcl::io::savePCDFileBinary] Error during write ()!");
    return (-1);
  }

  // Prepare the map
  char *map = (char*)mmap (0, cloud.data.size (), PROT_READ | PROT_WRITE, MAP_SHARED, fd, getpagesize ());
  if (map == MAP_FAILED)
  {
    close (fd);
    ROS_ERROR ("[pcl::io::savePCDFileBinary] Error during mmap ()!");
    return (-1);
  }

  // Copy the data
  memcpy (&map[0], &cloud.data[0], cloud.data.size ());

  // Unmap the pages of memory
  if (munmap (map, (cloud.data.size ())) < 0)
  {
    close (fd);
    ROS_ERROR ("[savePCDFileBinary] Error during munmap ()!");
    return (-1);
  }

  // Close file
  close (fd);
  return (0);
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Read the actual data from disk.
  * \param file_name the name of the file containing the actual PointCloud data
  * \param output the resultant PointCloud message read from disk
  */
bool
  pcl::PCDReader::read (const std::string &file_name, PointCloud &output)
{
  // Load the data
  if (io::loadPCDFile (file_name, output) == -1)
  {
    ROS_ERROR ("[pcl::PCDReader::read] Failed to load a valid PointCloud from %s.", file_name.c_str ());
    return (false);
  }
  return (true);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Write the actual data to disk.
  * \param file_name the name of the output file that will contain the actual PointCloud data
  * \param cloud the PointCloud message to be written to disk
  * \param binary set to true if the file is to be written in a binary PCD format, false for ASCII
  */
void
  pcl::PCDWriter::write (const std::string &file_name, const PointCloudConstPtr &cloud, bool binary)
{
  // Save the data
  io::savePCDFile (file_name, *cloud, binary);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Write the actual data to disk.
  * \param file_name the name of the output file that will contain the actual PointCloud data
  * \param cloud the PointCloud message to be written to disk
  * \param binary set to true if the file is to be written in a binary PCD format, false for ASCII
  */
void
  pcl::PCDWriter::write (const std::string &file_name, const PointCloud &cloud, bool binary)
{
  // Save the data
  io::savePCDFile (file_name, cloud, binary);
}
