/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: kdtree_ann.h 32573 2010-03-03 03:27:09Z rusu $
 *
 */

namespace pcl
{
  template <typename PointT>
  KdTreeANN<PointT>::KdTreeANN () : KdTree<PointT> (), ann_kd_tree_ (NULL), cloud_ (NULL)
  {
    cleanup ();
  }

  //template <typename PointT>
  //KdTreeANN<PointT>::KdTreeANN (const PointCloud &cloud, const std::vector<int> &indices)
  //{
    //if (!initParameters ()) return;

    //ann_mutex_lock_.lock ();
    //// Allocate enough data
    //nr_points_ = convertCloudToArray (cloud, indices);

    //initData ();
    //ann_mutex_lock_.unlock ();
  //}

  template <typename PointT>
  KdTreeANN<PointT>::~KdTreeANN ()
  {
    //std::cout << __PRETTY_FUNCTION__<<" called for "<<this<<".\n";
    cleanup ();
  }

  template <typename PointT>
  inline void KdTreeANN<PointT>::setInputCloud (const PointCloudConstPtr &cloud, const IndicesConstPtr &indices)
  {
    cleanup ();   // Perform an automatic cleanup of structures

    input_ = cloud;
    indices_ = indices;
    if (!initParameters())
    {
      ROS_ERROR ("[pcl::%s::setInputCloud] Error setting the input cloud!", getName ().c_str ());
      return;
    }

    // Allocate enough data
    ROS_ASSERT (input_);
    if (indices != NULL)
      convertCloudToArray (*input_, *indices_);
    else
      convertCloudToArray (*input_);

    initData ();
  }

  //template <typename PointT>
  //inline void KdTreeANN<PointT>::setPointRepresentation (const PointRepresentationConstPtr &point_representation)
  //{
    //point_representation_ = point_representation;

    //// When a new PointRepresentation is set, the internal point cloud data (e.g. dim_, cloud_) needs to be updated
    //setInputCloud(input_, indices_);
    
    ////if (input_ != NULL)
    ////{
      ////if (indices_ != NULL)
      ////{
        ////ann_mutex_lock_.lock ();
        ////if (cloud_ != NULL) annDeallocPts (cloud_);
        ////index_mapping_.clear();
        ////delete ann_kd_tree_;
        ////annClose ();
        ////ann_mutex_lock_.unlock ();

        ////initParameters();
        ////convertCloudToArray (*input_, *indices_);
        ////initData ();
      ////}
      ////else
      ////{
        ////cleanup ();

        ////initParameters();
        ////convertCloudToArray (*input_);
        ////initData ();
      ////}
    ////}
  //}

  template <typename PointT>
  inline bool KdTreeANN<PointT>::nearestKSearch (const PointT &point, int k, std::vector<int> &k_indices, std::vector<float> &k_distances)
  {
    //cout << __PRETTY_FUNCTION__<<" called.\n";

    if (!point_representation_->isValid (point))
      return (false);

    std::vector<float> tmp (dim_);
    point_representation_->vectorize (point, tmp);
    ann_mutex_lock_.lock ();
    ann_kd_tree_->annkSearch (&tmp[0], k, &k_indices[0], &k_distances[0], epsilon_);
    ann_mutex_lock_.unlock ();

    // Do mapping to original point cloud
    for (size_t i = 0; i < k_indices.size (); ++i)
    {
      int& neighbor_index = k_indices[i];
      //cout <<neighbor_index;
      neighbor_index = index_mapping_[neighbor_index];
      //cout <<"=>"<<neighbor_index<<"\n";
    }

    return (true);
  }

  template <typename PointT>
  inline bool KdTreeANN<PointT>::nearestKSearch (const PointCloud &cloud, int index, int k, std::vector<int> &k_indices, std::vector<float> &k_distances)
  {
    if (index >= (int)cloud.points.size ())
      return (false);
    return nearestKSearch(cloud.points[index], k, k_indices, k_distances);
  }

  template <typename PointT>
  inline bool KdTreeANN<PointT>::nearestKSearch (int index, int k, std::vector<int> &k_indices, std::vector<float> &k_distances)
  {
    if (indices_ == NULL)
    {
      if (index >= (int)input_->points.size())
        return (false);
      return nearestKSearch(input_->points[index], k, k_indices, k_distances);
    }
    else
    {
      if (index >= (int)indices_->size())
        return (false);
      return nearestKSearch(input_->points[(*indices_)[index]], k, k_indices, k_distances);
    }
  }

  template <typename PointT>
  inline bool KdTreeANN<PointT>::radiusSearch (const PointT &point, double radius, std::vector<int> &k_indices,
                                               std::vector<float> &k_distances, int max_nn) const
  {
    //cout << __PRETTY_FUNCTION__<<" called.\n";

    if (!point_representation_->isValid (point))
      return (false);
    //cout << PVARN(point);

    std::vector<float> tmp (dim_);
    point_representation_->vectorize (point, tmp);

    radius *= radius;

    ann_mutex_lock_.lock ();
    int neighbors_in_radius = ann_kd_tree_->annkFRSearch (&tmp[0], radius, 0, NULL, NULL, epsilon_);
    ann_mutex_lock_.unlock ();

    neighbors_in_radius  = std::min(neighbors_in_radius, max_nn);
    k_indices.resize (neighbors_in_radius);
    k_distances.resize (neighbors_in_radius);

    if (neighbors_in_radius == 0)
      return (false);

    ann_mutex_lock_.lock ();
    ann_kd_tree_->annkFRSearch (&tmp[0], radius, neighbors_in_radius, &k_indices[0], &k_distances[0], epsilon_);
    ann_mutex_lock_.unlock ();
    
    // Do mapping to original point cloud
    for (size_t i = 0; i < k_indices.size (); ++i)
    {
      int& neighbor_index = k_indices[i];
      //cout << neighbor_index;
      neighbor_index = index_mapping_[neighbor_index];
      //cout <<"=>"<<neighbor_index<<"\n";
      //cout << euclideanDistance(point, input_->points[neighbor_index])<<"("<<sqrtf(k_distances[i])<<") ";
    }

    return (true);
  }

  template <typename PointT>
  inline bool KdTreeANN<PointT>::radiusSearch (const PointCloud &cloud, int index, double radius, std::vector<int> &k_indices,
                                               std::vector<float> &k_distances, int max_nn) const
  {
    if (index >= (int)cloud.points.size ())
      return (false);
    return radiusSearch(cloud.points[index], radius, k_indices, k_distances, max_nn);
  }

  template <typename PointT>
  inline bool KdTreeANN<PointT>::radiusSearch (int index, double radius, std::vector<int> &k_indices,
                                               std::vector<float> &k_distances, int max_nn) const
  {
    if (indices_ == NULL)
    {
      if (index >= (int)input_->points.size())
        return (false);
      return radiusSearch(input_->points[index], radius, k_indices, k_distances, max_nn);
    }
    else
    {
      if (index >= (int)indices_->size())
        return (false);
      return radiusSearch(input_->points[(*indices_)[index]], radius, k_indices, k_distances, max_nn);
    }
  }

  template <typename PointT>
  inline bool KdTreeANN<PointT>::initParameters()
  {
    epsilon_ = 0.0;   // default error bound value
    dim_           = point_representation_->getNumberOfDimensions (); // Number of dimensions
    bucket_size_   = 30;    // default bucket size value

    return (true);
  }

  template <typename PointT>
  inline void KdTreeANN<PointT>::initData ()
  {
    ann_mutex_lock_.lock ();
    ann_kd_tree_ = new ANNkd_tree (cloud_, index_mapping_.size(), dim_, bucket_size_);  // Create the kd_tree representation
    ann_mutex_lock_.unlock ();
  }

  template <typename PointT>
  void KdTreeANN<PointT>::convertCloudToArray (const PointCloud &ros_cloud)
  {
    //std::cout << __PRETTY_FUNCTION__<<" called.\n";

    // No point in doing anything if the array is empty
    if (&ros_cloud == NULL || ros_cloud.points.empty ())
    {
      ann_mutex_lock_.lock ();
      cloud_ = NULL;
      ann_mutex_lock_.unlock ();
      return;
    }

    int original_no_of_points = ros_cloud.points.size();

    ann_mutex_lock_.lock ();
    cloud_ = annAllocPts (original_no_of_points, dim_);
    float** cloud_ptr = cloud_;
    index_mapping_.reserve(original_no_of_points);

    for (int cloud_index = 0; cloud_index < original_no_of_points; ++cloud_index)
    {
      const PointT point = ros_cloud.points[cloud_index];
      // Check if the point is invalid
      if (!point_representation_->isValid (point))
        continue;

      index_mapping_.push_back(cloud_index);

      float*& copied_point = *(cloud_ptr++);
      point_representation_->vectorize (point, copied_point);
    }
    //std::cout << "kdtree contains "<<index_mapping_.size()<<" of the given "<<original_no_of_points
              //<<" points from a point cloud of size "<<ros_cloud.points.size()<<".\n";
    ann_mutex_lock_.unlock ();
  }

  template <typename PointT>
  void KdTreeANN<PointT>::convertCloudToArray (const PointCloud &ros_cloud, const std::vector<int> &indices)
  {
    //std::cout << __PRETTY_FUNCTION__<<" called.\n";

    // No point in doing anything if the array is empty
    if (ros_cloud.points.empty () || indices.size () > ros_cloud.points.size () || indices.empty ())
    {
      ROS_WARN ("[pcl::%s::convertCloudToArray] No valid input given! Points: %zu, indices: %zu.", getName ().c_str (), ros_cloud.points.size (), indices.size ());
      ann_mutex_lock_.lock ();
      cloud_ = NULL;
      ann_mutex_lock_.unlock ();
      return;
    }

    int original_no_of_points = indices.size();

    ann_mutex_lock_.lock ();
    cloud_ = annAllocPts (original_no_of_points, dim_);
    float** cloud_ptr = cloud_;
    index_mapping_.reserve(original_no_of_points);

    for (int indices_index = 0; indices_index < original_no_of_points; ++indices_index)
    {
      int cloud_index = indices[indices_index];
      const PointT point = ros_cloud.points[cloud_index];
      // Check if the point is invalid
      if (!point_representation_->isValid (point))
        continue;

      index_mapping_.push_back(indices_index);  // If the returned index should be for the indices vector
      //index_mapping_.push_back(cloud_index);  // If the returned index should be for the ros cloud

      float*& copied_point = *(cloud_ptr++);
      point_representation_->vectorize (point, copied_point);
    }
    //std::cout << "kdtree contains "<<index_mapping_.size()<<" of the given "<<original_no_of_points
              //<<" points from a point cloud of size "<<ros_cloud.points.size()<<".\n";
    ann_mutex_lock_.unlock ();
  }

  template <typename PointT>
  void KdTreeANN<PointT>::cleanup ()
  {
    //std::cout << __PRETTY_FUNCTION__<<" called for "<<this<<".\n";
    ann_mutex_lock_.lock ();

    // Data array cleanup
    if (cloud_ != NULL)
      annDeallocPts (cloud_);
    cloud_ = NULL;
    index_mapping_.clear();

    // ANN Cleanup
    delete ann_kd_tree_;
    ann_kd_tree_ = NULL;
    annClose ();

    if (indices_)
      indices_.reset ();

    ann_mutex_lock_.unlock ();
  }

}  // namespace end

