/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 */

namespace pcl
{
  //template <typename PointT>
  //KdTreeFLANN<PointT>::KdTreeFLANN (const PointCloud &cloud, const std::vector<int> &indices)
  //{
    //if (!initParameters (indices.size ())) return;

    //// Allocate enough data
    //m_lock_.lock ();
    //convertCloudToArray (cloud, indices);

    //initData ();
    //m_lock_.unlock ();
  //}

  template <typename PointT>
  KdTreeFLANN<PointT>::KdTreeFLANN() : KdTree<PointT> (), index_id_(NULL), flann_param_(FLANNParameters()), cloud_(NULL)
  {
    cleanup ();
  }

  template <typename PointT>
  KdTreeFLANN<PointT>::~KdTreeFLANN ()
  {
    cleanup ();
  }

  template <typename PointT>
  inline void KdTreeFLANN<PointT>::setInputCloud (const PointCloudConstPtr &cloud, const IndicesConstPtr &indices)
  {
    cleanup ();   // Perform an automatic cleanup of structures

    if (!initParameters())
      return;

    input_   = cloud;
    indices_ = indices;
    
    if (input_ == NULL)
      return;

    m_lock_.lock ();
    // Allocate enough data
    //ROS_ASSERT (input_);
    if (indices != NULL)
      convertCloudToArray (*input_, *indices_);
    else
      convertCloudToArray (*input_);

    initData ();
    m_lock_.unlock ();
  }

  template <typename PointT>
  inline bool KdTreeFLANN<PointT>::nearestKSearch (const PointT &point, int k, std::vector<int> &k_indices, std::vector<float> &k_distances)
  {
    //cout << __PRETTY_FUNCTION__<<" called.\n";

    if (!point_representation_->isValid(point))
      return (false);

    std::vector<float> tmp(dim_);
    point_representation_->vectorize ((PointT)point, tmp);

    m_lock_.lock ();
    //flann_find_nearest_neighbors_index (index_id_, tmp_, 1, &k_indices[0], &k_distances[0], k, flann_param_.checks, &flann_param_);
    flann_find_nearest_neighbors_index(index_id_, &tmp[0], 1, &k_indices[0], &k_distances[0], k, &flann_param_);
    m_lock_.unlock ();

    // Do mapping to original point cloud
    for (unsigned int i=0; i<k_indices.size(); ++i)
    {
      int& neighbor_index = k_indices[i];
      //cout <<neighbor_index;
      neighbor_index = index_mapping_[neighbor_index];
      //cout <<"=>"<<neighbor_index<<"\n";
    }

    return (true);
  }

  template <typename PointT>
  inline bool KdTreeFLANN<PointT>::nearestKSearch (const PointCloud &cloud, int index, int k, std::vector<int> &k_indices, std::vector<float> &k_distances)
  {
    if (index >= (int)cloud.points.size ())
      return (false);
    return nearestKSearch(cloud.points[index], k, k_indices, k_distances);
  }

  template <typename PointT>
  inline bool KdTreeFLANN<PointT>::nearestKSearch (int index, int k, std::vector<int> &k_indices, std::vector<float> &k_distances)
  {
    if (indices_ == NULL)
    {
      if (index >= (int)input_->points.size())
        return (false);
      return nearestKSearch(input_->points[index], k, k_indices, k_distances);
    }
    else
    {
      if (index >= (int)indices_->size())
        return (false);
      return nearestKSearch(input_->points[(*indices_)[index]], k, k_indices, k_distances);
    }
  }

  template <typename PointT>
  inline bool KdTreeFLANN<PointT>::radiusSearch (const PointT &point, double radius, std::vector<int> &k_indices,
                                                 std::vector<float> &k_distances, int max_nn) const
  {
    //cout << __PRETTY_FUNCTION__<<" called.\n";

    if (!point_representation_->isValid (point))
      return (false);
 
    std::vector<float> tmp(dim_);
    point_representation_->vectorize ((PointT)point, tmp);

    radius *= radius;

    int neighbors_in_radius_ = flann_param_.checks;

    if (neighbors_in_radius_  > max_nn) neighbors_in_radius_ = max_nn;
    k_indices.resize (neighbors_in_radius_);
    k_distances.resize (neighbors_in_radius_);

    m_lock_.lock ();
    int neighbors_found = flann_radius_search (index_id_, &tmp[0], &k_indices[0], &k_distances[0], neighbors_in_radius_, radius, &flann_param_);
    m_lock_.unlock ();

    if (neighbors_found == 0)
      return (false);

    k_indices.resize (neighbors_found);
    k_distances.resize (neighbors_found);
    
    // Do mapping to original point cloud
    for (unsigned int i=0; i<k_indices.size(); ++i)
    {
      int& neighbor_index = k_indices[i];
      //cout << neighbor_index;
      neighbor_index = index_mapping_[neighbor_index];
      //cout <<"=>"<<neighbor_index<<"\n";
    }

    return (true);
  }

  template <typename PointT>
  inline bool KdTreeFLANN<PointT>::radiusSearch (const PointCloud &cloud, int index, double radius, std::vector<int> &k_indices,
                                                 std::vector<float> &k_distances, int max_nn) const
  {
    if (index >= (int)cloud.points.size ())
      return (false);
    return radiusSearch(cloud.points[index], radius, k_indices, k_distances, max_nn);
  }

  template <typename PointT>
  inline bool KdTreeFLANN<PointT>::radiusSearch (int index, double radius, std::vector<int> &k_indices,
                                                 std::vector<float> &k_distances, int max_nn) const
  {
    if (indices_ == NULL)
    {
      if (index >= (int)input_->points.size())
        return (false);
      return radiusSearch(input_->points[index], radius, k_indices, k_distances, max_nn);
    }
    else
    {
      if (index >= (int)indices_->size())
        return (false);
      return radiusSearch(input_->points[(*indices_)[index]], radius, k_indices, k_distances, max_nn);
    }
  }

  template <typename PointT>
  void KdTreeFLANN<PointT>::cleanup ()
  {
    m_lock_.lock ();
    // Data array cleanup
    free (cloud_);
    cloud_ = NULL;
    index_mapping_.clear();

    // FLANN Cleanup
    flann_free_index (index_id_, &flann_param_);

    if (indices_)
      indices_.reset ();

    m_lock_.unlock ();
  }

  template <typename PointT>
  inline bool KdTreeFLANN<PointT>::initParameters ()
  {
    epsilon_ = 0.0;   // default error bound value
    dim_ = point_representation_->getNumberOfDimensions (); // Number of dimensions - default is 3 = xyz
    // Create the kd_tree representation
    flann_param_.algorithm = KDTREE;
    flann_param_.log_level = LOG_NONE;
    //flann_param_.log_destination = NULL;

    flann_param_.trees = 1;
    flann_param_.target_precision = -1;
    flann_param_.checks = 200;
    flann_param_.random_seed = 1337;

    return (true);
  }

  template <typename PointT>
  void KdTreeFLANN<PointT>::initData ()
  {
    float speedup;
    index_id_ = flann_build_index (cloud_, index_mapping_.size(), dim_, &speedup, &flann_param_);
  }

  template <typename PointT>
  void KdTreeFLANN<PointT>::convertCloudToArray (const PointCloud &ros_cloud)
  {
    //cout << __PRETTY_FUNCTION__<<" called.\n";
    // No point in doing anything if the array is empty
    if (ros_cloud.points.empty ())
    {
      cloud_ = NULL;
      return;
    }

    int original_no_of_points = ros_cloud.points.size();

    cloud_ = (float*)malloc (original_no_of_points * dim_ * sizeof(float));
    float* cloud_ptr = cloud_;
    index_mapping_.reserve(original_no_of_points);

    for (int cloud_index = 0; cloud_index < original_no_of_points; ++cloud_index)
    {
      const PointT point = ros_cloud.points[cloud_index];
      // Check if the point is invalid
      if (!point_representation_->isValid(point))
        continue;

      index_mapping_.push_back(cloud_index);

      point_representation_->vectorize(point, cloud_ptr);
      cloud_ptr += dim_;
    }
    //std::cout << "kdtree contains "<<index_mapping_.size()<<" of the given "<<original_no_of_points
              //<<" points from a point cloud of size "<<ros_cloud.points.size()<<".\n";
  }

  template <typename PointT>
  void KdTreeFLANN<PointT>::convertCloudToArray (const PointCloud &ros_cloud, const std::vector<int> &indices)
  {
    // No point in doing anything if the array is empty
    if (ros_cloud.points.empty ())
    {
      cloud_ = NULL;
      return;
    }

    int original_no_of_points = indices.size();

    cloud_ = (float*)malloc (original_no_of_points * dim_ * sizeof (float));
    float* cloud_ptr = cloud_;
    index_mapping_.reserve(original_no_of_points);

    for (int indices_index = 0; indices_index < original_no_of_points; ++indices_index)
    {
      int cloud_index = indices[indices_index];
      const PointT point = ros_cloud.points[cloud_index];
      // Check if the point is invalid
      if (!point_representation_->isValid(point))
        continue;

      index_mapping_.push_back(indices_index);  // If the returned index should be for the indices vector
      //index_mapping_.push_back(cloud_index);  // If the returned index should be for the ros cloud
      
      point_representation_->vectorize(point, cloud_ptr);
      cloud_ptr += dim_;
    }
    //std::cout << "kdtree contains "<<index_mapping_.size()<<" of the given "<<original_no_of_points
              //<<" points from a point cloud of size "<<ros_cloud.points.size()<<".\n";
  }
  
  template <typename PointT>
  void KdTreeFLANN<PointT>::setDistanceMetric(flann_distance_t metric_type, int order)
  {
    flann_set_distance_type(metric_type, order);
  }


}  // end namespace
