/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 *
 */

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Subtract a centroid from a point cloud and return the de-meaned representation
  * \param cloud_in the input point cloud
  * \param centroid the centroid of the point cloud
  * \param cloud_out the resultant output point cloud
  */
template <typename PointT> inline void
  pcl::demeanPointCloud (const pcl::PointCloud<PointT> &cloud_in, const Eigen3::Vector4f &centroid,
                         pcl::PointCloud<PointT> &cloud_out)
{
  cloud_out = cloud_in;

  // Subtract the centroid from cloud_in
  for (size_t i = 0; i < cloud_in.points.size (); ++i)
  {
    pcl::Vector4fMap p = cloud_out.points[i].getVector4fMap ();
    p -= centroid;
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Subtract a centroid from a point cloud and return the de-meaned representation
  * \param cloud_in the input point cloud
  * \param indices the set of point indices to use from the input point cloud
  * \param centroid the centroid of the point cloud
  * \param cloud_out the resultant output point cloud
  */
template <typename PointT> inline void
  pcl::demeanPointCloud (const pcl::PointCloud<PointT> &cloud_in, const std::vector<int> &indices,
                         const Eigen3::Vector4f &centroid, pcl::PointCloud<PointT> &cloud_out)
{
  cloud_out.header = cloud_in.header;
  if (indices.size () == cloud_in.points.size ())
  {
    cloud_out.width    = cloud_in.width;
    cloud_out.height   = cloud_in.height;
    cloud_out.is_dense = cloud_in.is_dense;
  }
  else
  {
    cloud_out.width    = indices.size ();
    cloud_out.height   = 1;
    cloud_out.is_dense = false;
  }
  cloud_out.points.resize (indices.size ());

  // Subtract the centroid from cloud_in
  for (size_t i = 0; i < cloud_in.points.size (); ++i)
  {
    pcl::Vector4fMap p_in  = cloud_in.points[indices[i]].getVector4fMap ();
    pcl::Vector4fMap p_out = cloud_out.points[i].getVector4fMap ();
    p_out = p_in - centroid;
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Subtract a centroid from a point cloud and return the de-meaned representation as an Eigen matrix
  * \param cloud_in the input point cloud
  * \param centroid the centroid of the point cloud
  * \param cloud_out the resultant output XYZ0 dimensions of \a cloud_in as an Eigen matrix (4 rows, N pts columns)
  */
template <typename PointT> inline void
  pcl::demeanPointCloud (const pcl::PointCloud<PointT> &cloud_in, const Eigen3::Vector4f &centroid,
                         Eigen3::MatrixXf &cloud_out)
{
  size_t npts = cloud_in.points.size ();

  cloud_out = Eigen3::MatrixXf::Zero (4, npts);        // keep the data aligned

  for (size_t i = 0; i < npts; ++i)
  {
    pcl::Vector4fMap p = cloud_in.points[i].getVector4fMap ();
    cloud_out.block<4, 1> (0, i) = p - centroid;       // One column at a time
  }
  // Make sure we zero the 4th dimension out (1 row, N columns)
  cloud_out.block (3, 0, 1, npts).setZero ();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Subtract a centroid from a point cloud and return the de-meaned representation as an Eigen matrix
  * \param cloud_in the input point cloud
  * \param indices the set of point indices to use from the input point cloud
  * \param centroid the centroid of the point cloud
  * \param cloud_out the resultant output XYZ0 dimensions of \a cloud_in as an Eigen matrix (4 rows, N pts columns)
  */
template <typename PointT> inline void
  pcl::demeanPointCloud (const pcl::PointCloud<PointT> &cloud_in, const std::vector<int> &indices,
                         const Eigen3::Vector4f &centroid, Eigen3::MatrixXf &cloud_out)
{
  size_t npts = indices.size ();

  cloud_out = Eigen3::MatrixXf::Zero (4, npts);        // keep the data aligned

  for (size_t i = 0; i < npts; ++i)
  {
    pcl::Vector4fMap p = cloud_in.points[indices[i]].getVector4fMap ();
    cloud_out.block<4, 1> (0, i) = p - centroid;       // One column at a time
  }
  // Make sure we zero the 4th dimension out (1 row, N columns)
  cloud_out.block (3, 0, 1, npts).setZero ();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Apply an affine transform defined by an Eigen Transform
  * \param cloud_in the input point cloud
  * \param cloud_out the resultant output point cloud
  * \param transform an affine transformation (typically a rigid transformation)
  * \note The density of the point cloud is lost, since density implies that the origin is the point of view
  * \note Can be used with cloud_in equal to cloud_out
  */
template <typename PointT> inline void
  pcl::transformPointCloud (const pcl::PointCloud<PointT> &cloud_in, pcl::PointCloud<PointT> &cloud_out,
                            const Eigen3::Affine3f &transform)
{
  if (&cloud_in != &cloud_out)
  {
    // Note: could be replaced by cloud_out = cloud_in
    cloud_out.header   = cloud_in.header;
    cloud_out.width    = cloud_in.width;
    cloud_out.height   = cloud_in.height;
    cloud_out.is_dense = cloud_in.is_dense;
    cloud_out.points.reserve (cloud_out.points.size ());
    cloud_out.points.assign (cloud_in.points.begin (), cloud_in.points.end ());
  }

  for (size_t i = 0; i < cloud_out.points.size (); ++i)
  {
    if (std::isnan (cloud_in.points[i].x) || std::isnan (cloud_in.points[i].y) || std::isnan (cloud_in.points[i].z))
      continue;
    Eigen3::Map<Eigen3::Vector3f> p_in (&cloud_in.points[i].x, 3, 1);
    Eigen3::Map<Eigen3::Vector3f> p_out (&cloud_out.points[i].x, 3, 1);
    p_out = transform * p_in;
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Transform a point cloud and rotate its normals using an Eigen transform.
  * \param cloud_in the input point cloud
  * \param cloud_out the resultant output point cloud
  * \param transform an affine transformation (typically a rigid transformation)
  * \note The density of the point cloud is lost, since density implies that the origin is the point of view
  * \note Can be used with cloud_in equal to cloud_out
  */
template <typename PointT> inline void
  pcl::transformPointCloudWithNormals (const pcl::PointCloud<PointT> &cloud_in, pcl::PointCloud<PointT> &cloud_out,
                                       const Eigen3::Affine3f &transform)
{
  if (&cloud_in != &cloud_out)
  {
    // Note: could be replaced by cloud_out = cloud_in
    cloud_out.header   = cloud_in.header;
    cloud_out.width    = cloud_in.width;
    cloud_out.height   = cloud_in.height;
    cloud_out.is_dense = cloud_in.is_dense;
    cloud_out.points.reserve (cloud_out.points.size ());
    cloud_out.points.assign (cloud_in.points.begin (), cloud_in.points.end ());
  }

  for (size_t i = 0; i < cloud_out.points.size (); ++i)
  {
    if (std::isnan (cloud_in.points[i].x) || std::isnan (cloud_in.points[i].y) || std::isnan (cloud_in.points[i].z))
      continue;
    Eigen3::Map<Eigen3::Vector3f> p_in (&cloud_in.points[i].x, 3, 1);
    Eigen3::Map<Eigen3::Vector3f> p_out (&cloud_out.points[i].x, 3, 1);
    p_out = transform * p_in;

    // Rotate normals
    Eigen3::Map<Eigen3::Vector3f> n_in (&cloud_in.points[i].normal[0], 3, 1);
    Eigen3::Map<Eigen3::Vector3f> n_out (&cloud_out.points[i].normal[0], 3, 1);
    n_out = transform * n_in;
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Apply an affine transform defined by an Eigen Transform
  * \param cloud_in the input point cloud
  * \param cloud_out the resultant output point cloud
  * \param transform an affine transformation (typically a rigid transformation)
  * \note The density of the point cloud is lost, since density implies that the origin is the point of view
  * \note Can be used with cloud_in equal to cloud_out
  */
template <typename PointT> inline void
  pcl::transformPointCloud (const pcl::PointCloud<PointT> &cloud_in, pcl::PointCloud<PointT> &cloud_out,
                            const Eigen3::Matrix4f &transform)
{
  if (&cloud_in != &cloud_out)
  {
    // Note: could be replaced by cloud_out = cloud_in
    cloud_out.header   = cloud_in.header;
    cloud_out.width    = cloud_in.width;
    cloud_out.height   = cloud_in.height;
    cloud_out.is_dense = cloud_in.is_dense;
    cloud_out.points.reserve (cloud_out.points.size ());
    cloud_out.points.assign (cloud_in.points.begin (), cloud_in.points.end ());
  }

  Eigen3::Matrix3f rot   = transform.block<3, 3> (0, 0);
  Eigen3::Vector3f trans = transform.block<3, 1> (0, 3);
  for (size_t i = 0; i < cloud_out.points.size (); ++i)
  {
    if (std::isnan (cloud_in.points[i].x) || std::isnan (cloud_in.points[i].y) || std::isnan (cloud_in.points[i].z))
      continue;
    Eigen3::Map<Eigen3::Vector3f> p_in (&cloud_in.points[i].x, 3, 1);
    Eigen3::Map<Eigen3::Vector3f> p_out (&cloud_out.points[i].x, 3, 1);
    p_out = rot * p_in + trans;
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Transform a point cloud and rotate its normals using an Eigen transform.
  * \param cloud_in the input point cloud
  * \param cloud_out the resultant output point cloud
  * \param transform an affine transformation (typically a rigid transformation)
  * \note The density of the point cloud is lost, since density implies that the origin is the point of view
  * \note Can be used with cloud_in equal to cloud_out
  */
template <typename PointT> inline void
  pcl::transformPointCloudWithNormals (const pcl::PointCloud<PointT> &cloud_in, pcl::PointCloud<PointT> &cloud_out,
                                       const Eigen3::Matrix4f &transform)
{
  if (&cloud_in != &cloud_out)
  {
    // Note: could be replaced by cloud_out = cloud_in
    cloud_out.header   = cloud_in.header;
    cloud_out.width    = cloud_in.width;
    cloud_out.height   = cloud_in.height;
    cloud_out.is_dense = cloud_in.is_dense;
    cloud_out.points.reserve (cloud_out.points.size ());
    cloud_out.points.assign (cloud_in.points.begin (), cloud_in.points.end ());
  }

  Eigen3::Matrix3f rot   = transform.block<3, 3> (0, 0);
  Eigen3::Vector3f trans = transform.block<3, 1> (0, 3);
  for (size_t i = 0; i < cloud_out.points.size (); ++i)
  {
    if (std::isnan (cloud_in.points[i].x) || std::isnan (cloud_in.points[i].y) || std::isnan (cloud_in.points[i].z))
      continue;
    Eigen3::Map<Eigen3::Vector3f> p_in (&cloud_in.points[i].x, 3, 1);
    Eigen3::Map<Eigen3::Vector3f> p_out (&cloud_out.points[i].x, 3, 1);
    p_out = rot * p_in + trans;

    // Rotate normals
    Eigen3::Map<Eigen3::Vector3f> n_in (&cloud_in.points[i].normal[0], 3, 1);
    Eigen3::Map<Eigen3::Vector3f> n_out (&cloud_out.points[i].normal[0], 3, 1);
    n_out = rot * n_in;
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Apply a rigid transform defined by a 3D offset and a quaternion
  * \param cloud_in the input point cloud
  * \param cloud_out the resultant output point cloud
  * \param offset the translation component of the rigid transformation
  * \param rotation the rotation component of the rigid transformation
  * \note density of the point cloud is lost, since density implies that the origin is the point of view
  */
template <typename PointT> inline void
  pcl::transformPointCloud (const pcl::PointCloud<PointT> &cloud_in, pcl::PointCloud<PointT> &cloud_out,
                            const Eigen3::Vector3f &offset, const Eigen3::Quaternionf &rotation)
{
  Eigen3::Translation3f translation (offset);
  // Assemble an Eigen Transform
  Eigen3::Affine3f t;
  t = translation * rotation;
  transformPointCloud (cloud_in, cloud_out, t);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Transform a point cloud and rotate its normals using an Eigen transform.
  * \param cloud_in the input point cloud
  * \param cloud_out the resultant output point cloud
  * \param offset the translation component of the rigid transformation
  * \param rotation the rotation component of the rigid transformation
  * \note density of the point cloud is lost, since density implies that the origin is the point of view
  */
template <typename PointT> inline void
  pcl::transformPointCloudWithNormals (const pcl::PointCloud<PointT> &cloud_in, pcl::PointCloud<PointT> &cloud_out,
                                       const Eigen3::Vector3f &offset, const Eigen3::Quaternionf &rotation)
{
  Eigen3::Translation3f translation (offset);
  // Assemble an Eigen Transform
  Eigen3::Affine3f t;
  t = translation * rotation;
  transformPointCloudWithNormals (cloud_in, cloud_out, t);
}
