/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: sac_model_circle.hpp 32883 2010-09-27 07:31:00Z rusu $
 *
 */

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Get 3 random non-collinear points as data samples and return them as point indices.
  * \param iterations the internal number of iterations used by SAC methods
  * \param samples the resultant model samples
  * \note assumes unique points!
  */
template <typename PointT> void
  pcl::SampleConsensusModelCircle2D<PointT>::getSamples (int &iterations, std::vector<int> &samples)
{
  // We're assuming that indices_ have already been set in the constructor
  ROS_ASSERT (indices_->size () != 0);

  samples.resize (3);
  double trand = indices_->size () / (RAND_MAX + 1.0);

  // Get a random number between 1 and max_indices
  int idx = (int)(rand () * trand);
  // Get the index
  samples[0] = (*indices_)[idx];

  // Get a second point which is different than the first
  do
  {
    idx = (int)(rand () * trand);
    samples[1] = (*indices_)[idx];
    //iterations++;
  } while (samples[1] == samples[0]);
  //iterations--;

  // Get the values at the two points
  Eigen3::Array2d p0 (input_->points[samples[0]].x, input_->points[samples[0]].y);
  Eigen3::Array2d p1 (input_->points[samples[1]].x, input_->points[samples[1]].y);

  // Compute the segment values (in 2d) between p1 and p0
  p1 -= p0;

  Eigen3::Array2d dy1dy2;
  int iter = 0;
  do
  {
    // Get the third point, different from the first two
    do
    {
      idx = (int)(rand () * trand);
      samples[2] = (*indices_)[idx];
      //iterations++;
    } while ( (samples[2] == samples[1]) || (samples[2] == samples[0]) );
    //iterations--;

    Eigen3::Array2d p2 (input_->points[samples[2]].x, input_->points[samples[2]].y);

    // Compute the segment values (in 2d) between p2 and p0
    p2 -= p0;

    dy1dy2 = p1 / p2;
    ++iter;
    if (iter > MAX_ITERATIONS_COLLINEAR )
    {
      ROS_DEBUG ("[pcl::SampleConsensusModelCircle2D::getSamples] WARNING: Could not select 3 non collinear points in %d iterations!", MAX_ITERATIONS_COLLINEAR);
      break;
    }
    //iterations++;
  }
  while ( (dy1dy2[0] == dy1dy2[1]) );
  //iterations--;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Check whether the given index samples can form a valid 2D circle model, compute the model coefficients
  * from these samples and store them in model_coefficients. The circle coefficients are: x, y, R.
  * \param samples the point indices found as possible good candidates for creating a valid model
  * \param model_coefficients the resultant model coefficients
  */
template <typename PointT> bool
  pcl::SampleConsensusModelCircle2D<PointT>::computeModelCoefficients (const std::vector<int> &samples, Eigen3::VectorXf &model_coefficients)
{
  // Need 3 samples
  ROS_ASSERT (samples.size () == 3);

  model_coefficients.resize (3);

  Eigen3::Vector2d p0 (input_->points[samples[0]].x, input_->points[samples[0]].y);
  Eigen3::Vector2d p1 (input_->points[samples[1]].x, input_->points[samples[1]].y);
  Eigen3::Vector2d p2 (input_->points[samples[2]].x, input_->points[samples[2]].y);

  Eigen3::Vector2d u = (p0 + p1) / 2.0;
  Eigen3::Vector2d v = (p1 + p2) / 2.0;

  Eigen3::Vector2d p1p0dif = p1 - p0;
  Eigen3::Vector2d p2p1dif = p2 - p1;
  Eigen3::Vector2d uvdif   = u - v;

  Eigen3::Vector2d m (- p1p0dif[0] / p1p0dif[1], - p2p1dif[0] / p2p1dif[1]);

  // Center (x, y)
  model_coefficients[0] = (m[0] * u[0] -  m[1] * v[0]  - uvdif[1] )             / (m[0] - m[1]);
  model_coefficients[1] = (m[0] * m[1] * uvdif[0] +  m[0] * v[1] - m[1] * u[1]) / (m[0] - m[1]);

  // Radius
  model_coefficients[2] = sqrt ((model_coefficients[0] - p0[0]) * (model_coefficients[0] - p0[0]) +
                                (model_coefficients[1] - p0[1]) * (model_coefficients[1] - p0[1]));
  return (true);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Compute all distances from the cloud data to a given 2D circle model.
  * \param model_coefficients the coefficients of a 2D circle model that we need to compute distances to
  * \param distances the resultant estimated distances
  */
template <typename PointT> void
  pcl::SampleConsensusModelCircle2D<PointT>::getDistancesToModel (const Eigen3::VectorXf &model_coefficients, std::vector<double> &distances)
{
  // Needs a valid model coefficients
  ROS_ASSERT (model_coefficients.size () == 4);

  // Check if the model is valid given the user constraints
  if (!isModelValid (model_coefficients))
  {
    distances.resize (0);
    return;
  }
  distances.resize (indices_->size ());

  // Iterate through the 3d points and calculate the distances from them to the sphere
  for (size_t i = 0; i < indices_->size (); ++i)
    // Calculate the distance from the point to the circle as the difference between
    // dist(point,circle_origin) and circle_radius
    distances[i] = fabs (sqrt (
                               ( input_->points[(*indices_)[i]].x - model_coefficients[0] ) *
                               ( input_->points[(*indices_)[i]].x - model_coefficients[0] ) +

                               ( input_->points[(*indices_)[i]].y - model_coefficients[1] ) *
                               ( input_->points[(*indices_)[i]].y - model_coefficients[1] )
                              ) - model_coefficients[2]);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Compute all distances from the cloud data to a given 2D circle model.
  * \param model_coefficients the coefficients of a 2D circle model that we need to compute distances to
  * \param threshold a maximum admissible distance threshold for determining the inliers from the outliers
  * \param inliers the resultant model inliers
  */
template <typename PointT> void
  pcl::SampleConsensusModelCircle2D<PointT>::selectWithinDistance (const Eigen3::VectorXf &model_coefficients, double threshold, std::vector<int> &inliers)
{
  // Needs a valid model coefficients
  ROS_ASSERT (model_coefficients.size () == 3);

  // Check if the model is valid given the user constraints
  if (!isModelValid (model_coefficients))
  {
    inliers.resize (0);
    return;
  }
  int nr_p = 0;
  inliers.resize (indices_->size ());

  // Iterate through the 3d points and calculate the distances from them to the sphere
  for (size_t i = 0; i < indices_->size (); ++i)
  {
    // Calculate the distance from the point to the sphere as the difference between
    // dist(point,sphere_origin) and sphere_radius
    float distance = fabs (sqrt (
                                 ( input_->points[(*indices_)[i]].x - model_coefficients[0] ) *
                                 ( input_->points[(*indices_)[i]].x - model_coefficients[0] ) +

                                 ( input_->points[(*indices_)[i]].y - model_coefficients[1] ) *
                                 ( input_->points[(*indices_)[i]].y - model_coefficients[1] )
                                ) - model_coefficients[2]);
    if (distance < threshold)
    {
      // Returns the indices of the points whose distances are smaller than the threshold
      inliers[nr_p] = (*indices_)[i];
      nr_p++;
    }
  }
  inliers.resize (nr_p);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Recompute the 2d circle coefficients using the given inlier set and return them to the user.
  * @note: these are the coefficients of the 2d circle model after refinement (eg. after SVD)
  * \param inliers the data inliers found as supporting the model
  * \param model_coefficients the initial guess for the optimization
  * \param optimized_coefficients the resultant recomputed coefficients after non-linear optimization
  */
template <typename PointT> void
  pcl::SampleConsensusModelCircle2D<PointT>::optimizeModelCoefficients (
      const std::vector<int> &inliers, const Eigen3::VectorXf &model_coefficients, Eigen3::VectorXf &optimized_coefficients)
{
  boost::mutex::scoped_lock lock (tmp_mutex_);

  int n_unknowns = 3;      // 3 unknowns
  // Needs a set of valid model coefficients
  ROS_ASSERT (model_coefficients.size () == n_unknowns);

  if (inliers.empty ())
  {
    ROS_DEBUG ("[pcl::SampleConsensusModelCircle2D::optimizeModelCoefficients] Inliers vector empty! Returning the same coefficients.");
    optimized_coefficients = model_coefficients;
    return;
  }

  // Need at least 3 samples
  if (inliers.size () <= 3)
  {
    optimized_coefficients = model_coefficients;
    return;
  }

  tmp_inliers_ = &inliers;

  int m = inliers.size ();

  double *fvec = new double[m];

  int iwa[n_unknowns];

  int lwa = m * n_unknowns + 5 * n_unknowns + m;
  double *wa = new double[lwa];

  // Set the initial solution
  double x[n_unknowns];
  for (int d = 0; d < n_unknowns; ++d)
    x[d] = model_coefficients[d];   // initial guess

  // Set tol to the square root of the machine. Unless high solutions are required, these are the recommended settings.
  double tol = sqrt (dpmpar (1));

  // Optimize using forward-difference approximation LM
  int info = lmdif1 (&pcl::SampleConsensusModelCircle2D<PointT>::functionToOptimize, this, m, n_unknowns, x, fvec, tol, iwa, wa, lwa);

  // Compute the L2 norm of the residuals
  ROS_DEBUG ("[pcl::SampleConsensusModelCircle2D::optimizeModelCoefficients] LM solver finished with exit code %i, having a residual norm of %g. \nInitial solution: %g %g %g \nFinal solution: %g %g %g",
             info, enorm (m, fvec), model_coefficients[0], model_coefficients[1], model_coefficients[2], x[0], x[1], x[2]);

  optimized_coefficients = Eigen3::Vector3f (x[0], x[1], x[2]);

  free (wa); free (fvec);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////(
/** \brief Cost function to be minimized
  * \param p a pointer to our data structure array
  * \param m the number of functions
  * \param n the number of variables
  * \param x a pointer to the variables array
  * \param fvec a pointer to the resultant functions evaluations
  * \param iflag set to -1 inside the function to terminate execution
  */
template <typename PointT> int
  pcl::SampleConsensusModelCircle2D<PointT>::functionToOptimize (void *p, int m, int n, const double *x, double *fvec, int iflag)
{
  SampleConsensusModelCircle2D *model = (SampleConsensusModelCircle2D*)p;

  for (int i = 0; i < m; ++i)
  {
    // Compute the difference between the center of the circle and the datapoint X_i
    double xt = model->input_->points[(*model->tmp_inliers_)[i]].x - x[0];
    double yt = model->input_->points[(*model->tmp_inliers_)[i]].y - x[1];

    // g = sqrt ((x-a)^2 + (y-b)^2) - R
    fvec[i] = sqrt (xt * xt + yt * yt) - x[2];
  }
  return (0);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Create a new point cloud with inliers projected onto the 2d circle model.
  * \param inliers the data inliers that we want to project on the 2d circle model
  * \param model_coefficients the coefficients of a 2d circle model
  * \param projected_points the resultant projected points
  * \param copy_data_fields set to true if we need to copy the other data fields
  * \todo implement this.
  */
template <typename PointT> void
  pcl::SampleConsensusModelCircle2D<PointT>::projectPoints (
      const std::vector<int> &inliers, const Eigen3::VectorXf &model_coefficients,
      PointCloud &projected_points, bool copy_data_fields)
{
  // Needs a valid set of model coefficients
  ROS_ASSERT (model_coefficients.size () == 3);

  projected_points.header   = input_->header;

  // Copy all the data fields from the input cloud to the projected one?
  if (copy_data_fields)
  {
    // Allocate enough space and copy the basics
    projected_points.points.resize (input_->points.size ());
    projected_points.width    = input_->width;
    projected_points.height   = input_->height;
    projected_points.is_dense = input_->is_dense;

    typedef typename pcl::traits::fieldList<PointT>::type FieldList;
    // Iterate over each point
    for (size_t i = 0; i < projected_points.points.size (); ++i)
      // Iterate over each dimension
      pcl::for_each_type <FieldList> (NdConcatenateFunctor <PointT, PointT> (input_->points[i], projected_points.points[i]));

    // Iterate through the 3d points and calculate the distances from them to the plane
    for (size_t i = 0; i < inliers.size (); ++i)
    {
      float dx = input_->points[inliers[i]].x - model_coefficients[0];
      float dy = input_->points[inliers[i]].y - model_coefficients[1];
      float a = sqrt ( (model_coefficients[2] * model_coefficients[2]) / (dx * dx + dy * dy) );

      projected_points.points[inliers[i]].x = a * dx + model_coefficients[0];
      projected_points.points[inliers[i]].y = a * dy + model_coefficients[1];
    }
  }
  else
  {
    // Allocate enough space and copy the basics
    projected_points.points.resize (inliers.size ());
    projected_points.width    = inliers.size ();
    projected_points.height   = 1;
    projected_points.is_dense = false;

    // Iterate through the 3d points and calculate the distances from them to the plane
    for (size_t i = 0; i < inliers.size (); ++i)
    {
      float dx = input_->points[inliers[i]].x - model_coefficients[0];
      float dy = input_->points[inliers[i]].y - model_coefficients[1];
      float a = sqrt ( (model_coefficients[2] * model_coefficients[2]) / (dx * dx + dy * dy) );

      projected_points.points[i].x = a * dx + model_coefficients[0];
      projected_points.points[i].y = a * dy + model_coefficients[1];
    }
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Verify whether a subset of indices verifies the given 2d circle model coefficients.
  * \param indices the data indices that need to be tested against the 2d circle model
  * \param model_coefficients the 2d circle model coefficients
  * \param threshold a maximum admissible distance threshold for determining the inliers from the outliers
  */
template <typename PointT> bool
  pcl::SampleConsensusModelCircle2D<PointT>::doSamplesVerifyModel (
      const std::set<int> &indices, const Eigen3::VectorXf &model_coefficients, double threshold)
{
  // Needs a valid model coefficients
  ROS_ASSERT (model_coefficients.size () == 3);

  for (std::set<int>::iterator it = indices.begin (); it != indices.end (); ++it)
    // Calculate the distance from the point to the sphere as the difference between
    //dist(point,sphere_origin) and sphere_radius
    if (fabs (sqrt (
                    ( input_->points[*it].x - model_coefficients[0] ) *
                    ( input_->points[*it].x - model_coefficients[0] ) +
                    ( input_->points[*it].y - model_coefficients[1] ) *
                    ( input_->points[*it].y - model_coefficients[1] )
                   ) - model_coefficients[2]) > threshold)
      return (false);

  return (true);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Check whether a model is valid given the user constraints.
  * \param model_coefficients the set of model coefficients
  */
template <typename PointT> inline bool 
  pcl::SampleConsensusModelCircle2D<PointT>::isModelValid (const Eigen3::VectorXf &model_coefficients)
{
  // Needs a valid model coefficients
  ROS_ASSERT (model_coefficients.size () == 3);

  if (radius_min_ != -DBL_MAX && model_coefficients[2] < radius_min_)
    return (false);
  if (radius_max_ != DBL_MAX && model_coefficients[2] > radius_max_)
    return (false);

  return (true);
}

