/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: sac_model_line.hpp 32883 2010-09-27 07:31:00Z rusu $
 *
 */

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Get 2 random points as data samples and return them as point indices.
  * \param iterations the internal number of iterations used by SAC methods
  * \param samples the resultant model samples
  * \note assumes unique points!
  */
template <typename PointT> void
  pcl::SampleConsensusModelLine<PointT>::getSamples (int &iterations, std::vector<int> &samples)
{
  // We're assuming that indices_ have already been set in the constructor
  ROS_ASSERT (indices_->size () != 0);

  samples.resize (2);
  double trand = indices_->size () / (RAND_MAX + 1.0);

  // Get a random number between 1 and max_indices
  int idx = (int)(rand () * trand);
  // Get the index
  samples[0] = (*indices_)[idx];

  // Get a second point which is different than the first
  int iter = 0;
  do
  {
    idx = (int)(rand () * trand);
    samples[1] = (*indices_)[idx];
    ++iter;

    if (iter > MAX_ITERATIONS_UNIQUE )
    {
      ROS_DEBUG ("[pcl::SampleConsensusModelLine::getSamples] WARNING: Could not select 2 unique points in %d iterations (%zu indices)!", MAX_ITERATIONS_UNIQUE, indices_->size ());
      break;
    }
    //iterations++;
  }
  while (samples[1] == samples[0]);
  //iterations--;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Check whether the given index samples can form a valid line model, compute the model coefficients from
  * these samples and store them internally in model_coefficients_. The line coefficients are represented by a
  * point and a line direction
  * \param samples the point indices found as possible good candidates for creating a valid model
  * \param model_coefficients the resultant model coefficients
  */
template <typename PointT> bool
  pcl::SampleConsensusModelLine<PointT>::computeModelCoefficients (
      const std::vector<int> &samples, Eigen3::VectorXf &model_coefficients)
{
  // Need 2 samples
  ROS_ASSERT (samples.size () == 2);

  model_coefficients.resize (6);
  model_coefficients[0] = input_->points[samples[0]].x;
  model_coefficients[1] = input_->points[samples[0]].y;
  model_coefficients[2] = input_->points[samples[0]].z;

  model_coefficients[3] = input_->points[samples[1]].x - model_coefficients[0];
  model_coefficients[4] = input_->points[samples[1]].y - model_coefficients[1];
  model_coefficients[5] = input_->points[samples[1]].z - model_coefficients[2];

  model_coefficients.template tail<3> ().normalize ();
  return (true);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Compute all squared distances from the cloud data to a given line model.
  * \param model_coefficients the coefficients of a line model that we need to compute distances to
  * \param distances the resultant estimated squared distances
  */
template <typename PointT> void
  pcl::SampleConsensusModelLine<PointT>::getDistancesToModel (
      const Eigen3::VectorXf &model_coefficients, std::vector<double> &distances)
{
  // Needs a valid set of model coefficients
  ROS_ASSERT (model_coefficients.size () == 6);

  distances.resize (indices_->size ());

  // Obtain the line point and direction
  Eigen3::Vector4f line_pt  (model_coefficients[0], model_coefficients[1], model_coefficients[2], 0);
  Eigen3::Vector4f line_dir (model_coefficients[3], model_coefficients[4], model_coefficients[5], 0);
  Eigen3::Vector4f line_p2 = line_pt + line_dir;

  // Iterate through the 3d points and calculate the distances from them to the line
  for (size_t i = 0; i < indices_->size (); ++i)
  {
    // Calculate the distance from the point to the line
    // D = ||(P2-P1) x (P1-P0)|| / ||P2-P1|| = norm (cross (p2-p1, p2-p0)) / norm(p2-p1)
    Eigen3::Vector4f pt (input_->points[(*indices_)[i]].x, input_->points[(*indices_)[i]].y, input_->points[(*indices_)[i]].z, 0);
    Eigen3::Vector4f pp = line_p2 - pt;

    Eigen3::Vector4f c = pp.cross3 (line_dir);
    c[3] = 0;

    // Need to estimate sqrt here to keep MSAC and friends general
    distances[i] = sqrt (c.dot (c)) / line_dir.dot (line_dir);
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Select all the points which respect the given model coefficients as inliers.
  * \param model_coefficients the coefficients of a line model that we need to compute distances to
  * \param threshold a maximum admissible distance threshold for determining the inliers from the outliers
  * \param inliers the resultant model inliers
  */
template <typename PointT> void
  pcl::SampleConsensusModelLine<PointT>::selectWithinDistance (
      const Eigen3::VectorXf &model_coefficients, double threshold, std::vector<int> &inliers)
{
  // Needs a valid set of model coefficients
  ROS_ASSERT (model_coefficients.size () == 6);

  double sqr_threshold = threshold * threshold;

  int nr_p = 0;
  inliers.resize (indices_->size ());

  // Obtain the line point and direction
  Eigen3::Vector4f line_pt  (model_coefficients[0], model_coefficients[1], model_coefficients[2], 0);
  Eigen3::Vector4f line_dir (model_coefficients[3], model_coefficients[4], model_coefficients[5], 0);
  Eigen3::Vector4f line_p2 = line_pt + line_dir;

  // Iterate through the 3d points and calculate the distances from them to the line
  for (size_t i = 0; i < indices_->size (); ++i)
  {
    // Calculate the distance from the point to the line
    // D = ||(P2-P1) x (P1-P0)|| / ||P2-P1|| = norm (cross (p2-p1, p2-p0)) / norm(p2-p1)
    Eigen3::Vector4f pt (input_->points[(*indices_)[i]].x, input_->points[(*indices_)[i]].y, input_->points[(*indices_)[i]].z, 0);
    Eigen3::Vector4f pp = line_p2 - pt;

    Eigen3::Vector4f c = pp.cross3 (line_dir);
    c[3] = 0;
    //distances[i] = sqrt (c.dot (c)) / line_dir.dot (line_dir);
    double sqr_distance = c.dot (c) / line_dir.dot (line_dir);

    if (sqr_distance < sqr_threshold)
    {
      // Returns the indices of the points whose squared distances are smaller than the threshold
      inliers[nr_p] = (*indices_)[i];
      nr_p++;
    }
  }
  inliers.resize (nr_p);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Recompute the line coefficients using the given inlier set and return them to the user.
  * @note: these are the coefficients of the line model after refinement (eg. after SVD)
  * \param inliers the data inliers found as supporting the model
  * \param model_coefficients the initial guess for the model coefficients
  * \param optimized_coefficients the resultant recomputed coefficients after optimization
  */
template <typename PointT>
void
  pcl::SampleConsensusModelLine<PointT>::optimizeModelCoefficients (
      const std::vector<int> &inliers, const Eigen3::VectorXf &model_coefficients, Eigen3::VectorXf &optimized_coefficients)
{
  // Needs a valid set of model coefficients
  ROS_ASSERT (model_coefficients.size () == 6);

  if (inliers.empty ())
  {
    ROS_DEBUG ("[pcl::SampleConsensusModelLine::optimizeModelCoefficients] Inliers vector empty! Returning the same coefficients.");
    optimized_coefficients = model_coefficients;
    return;
  }

  // Need at least 2 points to estimate a line
  ROS_ASSERT (inliers.size () > 2);

  optimized_coefficients.resize (6);

  // Compute the 3x3 covariance matrix
  Eigen3::Vector4f centroid;
  compute3DCentroid (*input_, inliers, centroid);
  Eigen3::Matrix3f covariance_matrix;
  computeCovarianceMatrix (*input_, inliers, centroid, covariance_matrix);
  optimized_coefficients[0] = centroid[0];
  optimized_coefficients[1] = centroid[1];
  optimized_coefficients[2] = centroid[2];

  // Extract the eigenvalues and eigenvectors
  //Eigen3::SelfAdjointEigenSolver<Eigen3::Matrix3f> ei_symm (covariance_matrix);
  //EIGEN_ALIGN16 Eigen3::Vector3f eigen_values  = ei_symm.eigenvalues ();
  //EIGEN_ALIGN16 Eigen3::Matrix3f eigen_vectors = ei_symm.eigenvectors ();
  EIGEN_ALIGN16 Eigen3::Vector3f eigen_values;
  EIGEN_ALIGN16 Eigen3::Matrix3f eigen_vectors;
  pcl::eigen33 (covariance_matrix, eigen_vectors, eigen_values);

  optimized_coefficients.template tail<3> () = eigen_vectors.col (2).normalized ();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Create a new point cloud with inliers projected onto the line model.
  * \param inliers the data inliers that we want to project on the line model
  * \param model_coefficients the *normalized* coefficients of a line model
  * \param projected_points the resultant projected points
  * \param copy_data_fields set to true if we need to copy the other data fields
  */
template <typename PointT> void
  pcl::SampleConsensusModelLine<PointT>::projectPoints (
      const std::vector<int> &inliers, const Eigen3::VectorXf &model_coefficients, PointCloud &projected_points, bool copy_data_fields)
{
  // Needs a valid model coefficients
  ROS_ASSERT (model_coefficients.size () == 6);

  // Obtain the line point and direction
  Eigen3::Vector4f line_pt  (model_coefficients[0], model_coefficients[1], model_coefficients[2], 0);
  Eigen3::Vector4f line_dir (model_coefficients[3], model_coefficients[4], model_coefficients[5], 0);

  projected_points.header = input_->header;

  // Copy all the data fields from the input cloud to the projected one?
  if (copy_data_fields)
  {
    // Allocate enough space and copy the basics
    projected_points.points.resize (input_->points.size ());
    projected_points.width    = input_->width;
    projected_points.height   = input_->height;
    projected_points.is_dense = input_->is_dense;

    typedef typename pcl::traits::fieldList<PointT>::type FieldList;
    // Iterate over each point
    for (size_t i = 0; i < projected_points.points.size (); ++i)
      // Iterate over each dimension
      pcl::for_each_type <FieldList> (NdConcatenateFunctor <PointT, PointT> (input_->points[i], projected_points.points[i]));

    // Iterate through the 3d points and calculate the distances from them to the line
    for (size_t i = 0; i < inliers.size (); ++i)
    {
      Eigen3::Vector4f pt (input_->points[inliers[i]].x, input_->points[inliers[i]].y, input_->points[inliers[i]].z, 0);
      // double k = (DOT_PROD_3D (points[i], p21) - dotA_B) / dotB_B;
      double k = (pt.dot (line_dir) - line_pt.dot (line_dir)) / line_dir.dot (line_dir);

      Eigen3::Vector4f pp = line_pt + k * line_dir;
      // Calculate the projection of the point on the line (pointProj = A + k * B)
      projected_points.points[inliers[i]].x = pp[0];
      projected_points.points[inliers[i]].y = pp[1];
      projected_points.points[inliers[i]].z = pp[2];
    }
  }
  else
  {
    // Allocate enough space and copy the basics
    projected_points.points.resize (inliers.size ());
    projected_points.width    = inliers.size ();
    projected_points.height   = 1;
    projected_points.is_dense = false;

    // Iterate through the 3d points and calculate the distances from them to the line
    for (size_t i = 0; i < inliers.size (); ++i)
    {
      Eigen3::Vector4f pt (input_->points[inliers[i]].x, input_->points[inliers[i]].y, input_->points[inliers[i]].z, 0);
      // double k = (DOT_PROD_3D (points[i], p21) - dotA_B) / dotB_B;
      double k = (pt.dot (line_dir) - line_pt.dot (line_dir)) / line_dir.dot (line_dir);

      Eigen3::Vector4f pp = line_pt + k * line_dir;
      // Calculate the projection of the point on the line (pointProj = A + k * B)
      projected_points.points[i].x = pp[0];
      projected_points.points[i].y = pp[1];
      projected_points.points[i].z = pp[2];
    }
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Verify whether a subset of indices verifies the given plane model coefficients.
  * \param indices the data indices that need to be tested against the plane model
  * \param model_coefficients the plane model coefficients
  * \param threshold a maximum admissible distance threshold for determining the inliers from the outliers
  */
template <typename PointT> bool
  pcl::SampleConsensusModelLine<PointT>::doSamplesVerifyModel (
      const std::set<int> &indices, const Eigen3::VectorXf &model_coefficients, double threshold)
{
  // Needs a valid set of model coefficients
  ROS_ASSERT (model_coefficients.size () == 6);

  // Obtain the line point and direction
  Eigen3::Vector4f line_pt  (model_coefficients[0], model_coefficients[1], model_coefficients[2], 0);
  Eigen3::Vector4f line_dir (model_coefficients[3], model_coefficients[4], model_coefficients[5], 0);
  Eigen3::Vector4f line_p2 = line_pt + line_dir;

  double sqr_threshold = threshold * threshold;
  // Iterate through the 3d points and calculate the distances from them to the line
  for (std::set<int>::iterator it = indices.begin (); it != indices.end (); ++it)
  {
    // Calculate the distance from the point to the line
    // D = ||(P2-P1) x (P1-P0)|| / ||P2-P1|| = norm (cross (p2-p1, p2-p0)) / norm(p2-p1)
    Eigen3::Vector4f pt (input_->points[*it].x, input_->points[*it].y, input_->points[*it].z, 0);
    Eigen3::Vector4f pp = line_p2 - pt;

    Eigen3::Vector4f c = pp.cross3 (line_dir);
    c[3] = 0;
    double sqr_distance = c.dot (c) / line_dir.dot (line_dir);

    if (sqr_distance > sqr_threshold)
      return (false);
  }

  return (true);
}

