/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: sac_model_parallel_line.hpp 32877 2010-09-26 19:24:17Z blodow $
 *
 */

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Select all the points which respect the given model coefficients as inliers.
  * \param model_coefficients the coefficients of a line model that we need to compute distances to
  * \param threshold a maximum admissible distance threshold for determining the inliers from the outliers
  * \param inliers the resultant model inliers
  */
template <typename PointT> inline void
  pcl::SampleConsensusModelParallelLine<PointT>::selectWithinDistance (
      const Eigen3::VectorXf &model_coefficients, double threshold, std::vector<int> &inliers)
{
  // Needs a valid model coefficients
  ROS_ASSERT (model_coefficients.size () == 6);

  // Check if the model is valid given the user constraints
  if (!isModelValid (model_coefficients))
  {
    inliers.resize (0);
    return;
  }

  SampleConsensusModelLine<PointT>::selectWithinDistance (model_coefficients, threshold, inliers);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Compute all squared distances from the cloud data to a given line model.
  * \param model_coefficients the coefficients of a line model that we need to compute distances to
  * \param distances the resultant estimated squared distances
  */
template <typename PointT> inline void
  pcl::SampleConsensusModelParallelLine<PointT>::getDistancesToModel (
      const Eigen3::VectorXf &model_coefficients, std::vector<double> &distances)
{
  // Needs a valid model coefficients
  ROS_ASSERT (model_coefficients.size () == 6);

  // Check if the model is valid given the user constraints
  if (!isModelValid (model_coefficients))
  {
    distances.resize (0);
    return;
  }

  SampleConsensusModelLine<PointT>::getDistancesToModel (model_coefficients, distances);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Check whether a model is valid given the user constraints.
  * \param model_coefficients the set of model coefficients
  */
template <typename PointT> inline bool
  pcl::SampleConsensusModelParallelLine<PointT>::isModelValid (const Eigen3::VectorXf &model_coefficients)
{
  // Needs a valid model coefficients
  ROS_ASSERT (model_coefficients.size () == 6);

  // Check against template, if given
  if (eps_angle_ > 0.0)
  {
    // Obtain the line direction
    Eigen3::Vector4f line_dir (model_coefficients[3], model_coefficients[4], model_coefficients[5], 0);

    Eigen3::Vector4f axis (axis_[0], axis_[1], axis_[2], 0);
    double angle_diff = fabs (getAngle3D (axis, line_dir));
    angle_diff = fmin (angle_diff, M_PI - angle_diff);
    // Check whether the current plane model satisfies our angle threshold criterion with respect to the given axis
    if (angle_diff > eps_angle_)
      return (false);
  }

  return (true);
}

