/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: sac_model_plane.hpp 32883 2010-09-27 07:31:00Z rusu $
 *
 */

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Get 3 random non-collinear points as data samples and return them as point indices.
  * \param iterations the internal number of iterations used by SAC methods
  * \param samples the resultant model samples
  * \note assumes unique points!
  */
template <typename PointT> void
  pcl::SampleConsensusModelPlane<PointT>::getSamples (int &iterations, std::vector<int> &samples)
{
  // We're assuming that indices_ have already been set in the constructor
  ROS_ASSERT (indices_->size () != 0);

  samples.resize (3);
  double trand = indices_->size () / (RAND_MAX + 1.0);

  // Get a random number between 1 and max_indices
  int idx = (int)(rand () * trand);
  // Get the index
  samples[0] = (*indices_)[idx];

  // Get a second point which is different than the first
  do
  {
    idx = (int)(rand () * trand);
    samples[1] = (*indices_)[idx];
    //iterations++;
  } while (samples[1] == samples[0]);
  //iterations--;

  // Get the values at the two points
  pcl::Array4fMap p0 = input_->points[samples[0]].getArray4fMap ();
  pcl::Array4fMap p1 = input_->points[samples[1]].getArray4fMap ();

  // Compute the segment values (in 3d) between p1 and p0
  Eigen3::Array4f p1p0 = p1 - p0;

  Eigen3::Array4f dy1dy2;
  int iter = 0;
  do
  {
    // Get the third point, different from the first two
    do
    {
      idx = (int)(rand () * trand);
      samples[2] = (*indices_)[idx];
      //iterations++;
    } while ( (samples[2] == samples[1]) || (samples[2] == samples[0]) );
    //iterations--;

    pcl::Array4fMap p2 = input_->points[samples[2]].getArray4fMap ();

    // Compute the segment values (in 3d) between p2 and p0
    Eigen3::Array4f p2p0 = p2 - p0;

    dy1dy2 = p1p0 / p2p0;
    ++iter;
    if (iter > MAX_ITERATIONS_COLLINEAR )
    {
      ROS_DEBUG ("[pcl::SampleConsensusModelPlane::getSamples] WARNING: Could not select 3 non collinear points in %d iterations!", MAX_ITERATIONS_COLLINEAR);
      break;
    }
    //iterations++;
  }
  while ( (dy1dy2[0] == dy1dy2[1]) && (dy1dy2[2] == dy1dy2[1]) );
  //iterations--;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Check whether the given index samples can form a valid plane model, compute the model coefficients from
  * these samples and store them internally in model_coefficients_. The plane coefficients are:
  * a, b, c, d (ax+by+cz+d=0)
  * \param samples the point indices found as possible good candidates for creating a valid model
  * \param model_coefficients the resultant model coefficients
  */
template <typename PointT> bool
  pcl::SampleConsensusModelPlane<PointT>::computeModelCoefficients (
      const std::vector<int> &samples, Eigen3::VectorXf &model_coefficients)
{
  // Need 3 samples
  ROS_ASSERT (samples.size () == 3);

  pcl::Array4fMap p0 = input_->points[samples[0]].getArray4fMap ();
  pcl::Array4fMap p1 = input_->points[samples[1]].getArray4fMap ();
  pcl::Array4fMap p2 = input_->points[samples[2]].getArray4fMap ();

  // Compute the segment values (in 3d) between p1 and p0
  Eigen3::Array4f p1p0 = p1 - p0;
  // Compute the segment values (in 3d) between p2 and p0
  Eigen3::Array4f p2p0 = p2 - p0;

  // Avoid some crashes by checking for collinearity here
  Eigen3::Array4f dy1dy2 = p1p0 / p2p0;
  if ( (dy1dy2[0] == dy1dy2[1]) && (dy1dy2[2] == dy1dy2[1]) )          // Check for collinearity
    return (false);

  // Compute the plane coefficients from the 3 given points in a straightforward manner
  // calculate the plane normal n = (p2-p1) x (p3-p1) = cross (p2-p1, p3-p1)
  model_coefficients.resize (4);
  model_coefficients[0] = p1p0[1] * p2p0[2] - p1p0[2] * p2p0[1];
  model_coefficients[1] = p1p0[2] * p2p0[0] - p1p0[0] * p2p0[2];
  model_coefficients[2] = p1p0[0] * p2p0[1] - p1p0[1] * p2p0[0];
  model_coefficients[3] = 0;

  // Normalize
  model_coefficients.normalize ();

  // ... + d = 0
  model_coefficients[3] = -1 * (model_coefficients.template head<4>().dot (p0.matrix ()));

  return (true);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Compute all distances from the cloud data to a given plane model.
  * \param model_coefficients the coefficients of a plane model that we need to compute distances to
  * \param distances the resultant estimated distances
  */
template <typename PointT> void
  pcl::SampleConsensusModelPlane<PointT>::getDistancesToModel (
      const Eigen3::VectorXf &model_coefficients, std::vector<double> &distances)
{
  // Needs a valid set of model coefficients
  ROS_ASSERT (model_coefficients.size () == 4);

  distances.resize (indices_->size ());

  // Iterate through the 3d points and calculate the distances from them to the plane
  for (size_t i = 0; i < indices_->size (); ++i)
  {
    // Calculate the distance from the point to the plane normal as the dot product
    // D = (P-A).N/|N|
    /*distances[i] = fabs (model_coefficients[0] * input_->points[(*indices_)[i]].x +
                         model_coefficients[1] * input_->points[(*indices_)[i]].y +
                         model_coefficients[2] * input_->points[(*indices_)[i]].z +
                         model_coefficients[3]);*/
    pcl::Vector4fMap pt = input_->points[(*indices_)[i]].getVector4fMap ();
    pt[3] = 1;
    distances[i] = fabs (model_coefficients.dot (pt));
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Select all the points which respect the given model coefficients as inliers.
  * \param model_coefficients the coefficients of a plane model that we need to compute distances to
  * \param threshold a maximum admissible distance threshold for determining the inliers from the outliers
  * \param inliers the resultant model inliers
  */
template <typename PointT> void
  pcl::SampleConsensusModelPlane<PointT>::selectWithinDistance (
      const Eigen3::VectorXf &model_coefficients, double threshold, std::vector<int> &inliers)
{
  // Needs a valid set of model coefficients
  ROS_ASSERT (model_coefficients.size () == 4);

  int nr_p = 0;
  inliers.resize (indices_->size ());

  // Iterate through the 3d points and calculate the distances from them to the plane
  for (size_t i = 0; i < indices_->size (); ++i)
  {
    // Calculate the distance from the point to the plane normal as the dot product
    // D = (P-A).N/|N|
    pcl::Vector4fMap pt = input_->points[(*indices_)[i]].getVector4fMap ();
    pt[3] = 1;
    if (fabs (model_coefficients.dot (pt)) < threshold)
    {
      // Returns the indices of the points whose distances are smaller than the threshold
      inliers[nr_p] = (*indices_)[i];
      nr_p++;
    }
  }
  inliers.resize (nr_p);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Recompute the plane coefficients using the given inlier set and return them to the user.
  * @note: these are the coefficients of the plane model after refinement (eg. after SVD)
  * \param inliers the data inliers found as supporting the model
  * \param model_coefficients the initial guess for the model coefficients
  * \param optimized_coefficients the resultant recomputed coefficients after non-linear optimization
  */
template <typename PointT> void
  pcl::SampleConsensusModelPlane<PointT>::optimizeModelCoefficients (
      const std::vector<int> &inliers, const Eigen3::VectorXf &model_coefficients, Eigen3::VectorXf &optimized_coefficients)
{
  // Needs a valid set of model coefficients
  ROS_ASSERT (model_coefficients.size () == 4);

  if (inliers.empty ())
  {
    ROS_DEBUG ("[pcl::SampleConsensusModelPlane::optimizeModelCoefficients] Inliers vector empty! Returning the same coefficients.");
    optimized_coefficients = model_coefficients;
    return;
  }

  // Need at least 3 points to estimate a plane
  if (inliers.size () < 4)
  {
    ROS_ERROR ("[pcl::SampleConsensusModelPlane::optimizeModelCoefficients] Not enough inliers found to support a model! Returning the same coefficients.");
    optimized_coefficients = model_coefficients;
    return;
  }

  Eigen3::Vector4f plane_parameters;
  float curvature;

  // Use Least-Squares to fit the plane through all the given sample points and find out its coefficients
  computePointNormal (*input_, inliers, plane_parameters, curvature);
  optimized_coefficients = plane_parameters;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Create a new point cloud with inliers projected onto the plane model.
  * \param inliers the data inliers that we want to project on the plane model
  * \param model_coefficients the *normalized* coefficients of a plane model
  * \param projected_points the resultant projected points
  * \param copy_data_fields set to true if we need to copy the other data fields
  */
template <typename PointT> void
  pcl::SampleConsensusModelPlane<PointT>::projectPoints (
      const std::vector<int> &inliers, const Eigen3::VectorXf &model_coefficients, PointCloud &projected_points, bool copy_data_fields)
{
  // Needs a valid set of model coefficients
  ROS_ASSERT (model_coefficients.size () == 4);

  projected_points.header = input_->header;

  Eigen3::Vector4f mc (model_coefficients[0], model_coefficients[1], model_coefficients[2], 0);
  // Copy all the data fields from the input cloud to the projected one?
  if (copy_data_fields)
  {
    // Allocate enough space and copy the basics
    projected_points.points.resize (input_->points.size ());
    projected_points.width    = input_->width;
    projected_points.height   = input_->height;
    projected_points.is_dense = input_->is_dense;

    typedef typename pcl::traits::fieldList<PointT>::type FieldList;
    // Iterate over each point
    for (size_t i = 0; i < projected_points.points.size (); ++i)
      // Iterate over each dimension
      pcl::for_each_type <FieldList> (NdConcatenateFunctor <PointT, PointT> (input_->points[i], projected_points.points[i]));

    // Iterate through the 3d points and calculate the distances from them to the plane
    for (size_t i = 0; i < inliers.size (); ++i)
    {
      // Calculate the distance from the point to the plane
      pcl::Vector4fMap p = input_->points[inliers[i]].getVector4fMap ();
      p[3] = 1;
      float distance_to_plane = model_coefficients.dot (p);

      pcl::Vector4fMap pp = projected_points.points[inliers[i]].getVector4fMap ();
      pp = p - mc * distance_to_plane;        // mc[3] = 0, therefore the 3rd coordinate is safe
    }
  }
  else
  {
    // Allocate enough space and copy the basics
    projected_points.points.resize (inliers.size ());
    projected_points.width    = inliers.size ();
    projected_points.height   = 1;
    projected_points.is_dense = false;

    // Iterate through the 3d points and calculate the distances from them to the plane
    for (size_t i = 0; i < inliers.size (); ++i)
    {
      // Calculate the distance from the point to the plane
      pcl::Vector4fMap p = input_->points[inliers[i]].getVector4fMap ();
      p[3] = 1;
      float distance_to_plane = model_coefficients.dot (p);

      pcl::Vector4fMap pp = projected_points.points[i].getVector4fMap ();
      pp = p - mc * distance_to_plane;        // mc[3] = 0, therefore the 3rd coordinate is safe
    }
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Verify whether a subset of indices verifies the given plane model coefficients.
  * \param indices the data indices that need to be tested against the plane model
  * \param model_coefficients the plane model coefficients
  * \param threshold a maximum admissible distance threshold for determining the inliers from the outliers
  */
template <typename PointT> bool
  pcl::SampleConsensusModelPlane<PointT>::doSamplesVerifyModel (
      const std::set<int> &indices, const Eigen3::VectorXf &model_coefficients, double threshold)
{
  // Needs a valid set of model coefficients
  ROS_ASSERT (model_coefficients.size () == 4);

  for (std::set<int>::iterator it = indices.begin (); it != indices.end (); ++it)
  {
    pcl::Vector4fMap pt = input_->points[*it].getVector4fMap ();
    pt[3] = 1;
    if (fabs (model_coefficients.dot (pt)) > threshold)
      return (false);
  }

  return (true);
}

