/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: extract_clusters.hpp 30785 2010-07-13 20:20:51Z rusu $
 *
 */

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Decompose a region of space into clusters based on the Euclidean distance between points
  * \param cloud the point cloud message
  * \param tree the spatial locator (e.g., kd-tree) used for nearest neighbors searching
  * \note the tree has to be created as a spatial locator on \a cloud
  * \param tolerance the spatial cluster tolerance as a measure in L2 Euclidean space
  * \param clusters the resultant clusters containing point indices (as a vector of PointIndices)
  * \param min_pts_per_cluster minimum number of points that a cluster may contain (default = 1)
  */
template <typename PointT> void
  pcl::extractEuclideanClusters (const pcl::PointCloud<PointT> &cloud, 
                                 const boost::shared_ptr<pcl::KdTree<PointT> > &tree,
                                 float tolerance, std::vector<pcl::PointIndices> &clusters,
                                 unsigned int min_pts_per_cluster = 1, 
                                 unsigned int max_pts_per_cluster = std::numeric_limits<int>::max ())
{
  ROS_ASSERT (tree->getInputCloud ()->points.size () == cloud.points.size ());
  // Create a bool vector of processed point indices, and initialize it to false
  std::vector<bool> processed (cloud.points.size (), false);

  std::vector<int> nn_indices;
  std::vector<float> nn_distances;
  // Process all points in the indices vector
  for (size_t i = 0; i < cloud.points.size (); ++i)
  {
    if (processed[i])
      continue;

    std::vector<int> seed_queue;
    int sq_idx = 0;
    seed_queue.push_back (i);

    processed[i] = true;

    while (sq_idx < (int)seed_queue.size ())
    {
      // Search for sq_idx
      if (!tree->radiusSearch (seed_queue[sq_idx], tolerance, nn_indices, nn_distances))
      {
        sq_idx++;
        continue;
      }

      for (size_t j = 1; j < nn_indices.size (); ++j)             // nn_indices[0] should be sq_idx
      {
        if (processed[nn_indices[j]])                             // Has this point been processed before ?
          continue;

        // Perform a simple Euclidean clustering
        seed_queue.push_back (nn_indices[j]);
        processed[nn_indices[j]] = true;
      }

      sq_idx++;
    }

    // If this queue is satisfactory, add to the clusters
    if (seed_queue.size () >= min_pts_per_cluster && seed_queue.size () <= max_pts_per_cluster)
    {
      pcl::PointIndices r;
      r.indices.resize (seed_queue.size ());
      for (size_t j = 0; j < seed_queue.size (); ++j)
        r.indices[j] = seed_queue[j];

      sort (r.indices.begin (), r.indices.end ());
      r.indices.erase (unique (r.indices.begin (), r.indices.end ()), r.indices.end ());

      r.header = cloud.header;
      clusters.push_back (r);   // We could avoid a copy by working directly in the vector
    }
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Decompose a region of space into clusters based on the Euclidean distance between points
  * \param cloud the point cloud message
  * \param indices a list of point indices to use from \a cloud
  * \param tree the spatial locator (e.g., kd-tree) used for nearest neighbors searching
  * \note the tree has to be created as a spatial locator on \a cloud and \a indices
  * \param tolerance the spatial cluster tolerance as a measure in L2 Euclidean space
  * \param clusters the resultant clusters containing point indices (as a vector of PointIndices)
  * \param min_pts_per_cluster minimum number of points that a cluster may contain (default = 1)
  */
template <typename PointT> void
  pcl::extractEuclideanClusters (const pcl::PointCloud<PointT> &cloud, 
                                 const std::vector<int> &indices,
                                 const boost::shared_ptr<pcl::KdTree<PointT> > &tree,
                                 float tolerance, std::vector<pcl::PointIndices> &clusters,
                                 unsigned int min_pts_per_cluster = 1, 
                                 unsigned int max_pts_per_cluster = std::numeric_limits<int>::max ())
{
  // \note If the tree was created over <cloud, indices>, we guarantee a 1-1 mapping between what the tree returns
  //and indices[i]
  ROS_ASSERT (tree->getInputCloud ()->points.size () == cloud.points.size ());
  ROS_ASSERT (tree->getIndices ()->size () == indices.size ());

  // Create a bool vector of processed point indices, and initialize it to false
  std::vector<bool> processed (indices.size (), false);

  std::vector<int> nn_indices;
  std::vector<float> nn_distances;
  // Process all points in the indices vector
  for (size_t i = 0; i < indices.size (); ++i)
  {
    if (processed[i])
      continue;

    std::vector<int> seed_queue;
    int sq_idx = 0;
    seed_queue.push_back (i);

    processed[i] = true;

    while (sq_idx < (int)seed_queue.size ())
    {
      // Search for sq_idx
      if (!tree->radiusSearch (seed_queue[sq_idx], tolerance, nn_indices, nn_distances))
      {
        sq_idx++;
        continue;
      }

      for (size_t j = 1; j < nn_indices.size (); ++j)             // nn_indices[0] should be sq_idx
      {
        if (processed[nn_indices[j]])                             // Has this point been processed before ?
          continue;

        // Perform a simple Euclidean clustering
        seed_queue.push_back (nn_indices[j]);
        processed[nn_indices[j]] = true;
      }

      sq_idx++;
    }

    // If this queue is satisfactory, add to the clusters
    if (seed_queue.size () >= min_pts_per_cluster && seed_queue.size () <= max_pts_per_cluster)
    {
      pcl::PointIndices r;
      r.indices.resize (seed_queue.size ());
      for (size_t j = 0; j < seed_queue.size (); ++j)
        // This is the only place where indices come into play
        r.indices[j] = indices[seed_queue[j]];

      //r.indices.assign(seed_queue.begin(), seed_queue.end());
      sort (r.indices.begin (), r.indices.end ());
      r.indices.erase (unique (r.indices.begin (), r.indices.end ()), r.indices.end ());

      r.header = cloud.header;
      clusters.push_back (r);   // We could avoid a copy by working directly in the vector
    }
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Decompose a region of space into clusters based on the euclidean distance between points, and the normal
  * angular deviation
  * \param cloud the point cloud message
  * \param normals the point cloud message containing normal information
  * \param tree the spatial locator (e.g., kd-tree) used for nearest neighbors searching
  * \note the tree has to be created as a spatial locator on \a cloud
  * \param tolerance the spatial cluster tolerance as a measure in the L2 Euclidean space
  * \param clusters the resultant clusters containing point indices (as a vector of PointIndices)
  * \param eps_angle the maximum allowed difference between normals in degrees for cluster/region growing
  * \param min_pts_per_cluster minimum number of points that a cluster may contain (default = 1)
  */
template <typename PointT, typename Normal> void
  pcl::extractEuclideanClusters (const pcl::PointCloud<PointT> &cloud, 
                                 const pcl::PointCloud<Normal> &normals,
                                 float tolerance, 
                                 const boost::shared_ptr<pcl::KdTree<PointT> > &tree,
                                 std::vector<pcl::PointIndices> &clusters, double eps_angle, 
                                 unsigned int min_pts_per_cluster = 1, 
                                 unsigned int max_pts_per_cluster = std::numeric_limits<int>::max ())
{
  ROS_ASSERT (tree->getInputCloud ()->points.size () == cloud.points.size ());
  ROS_ASSERT (cloud.points.size () == normals.points.size ());

  // Create a bool vector of processed point indices, and initialize it to false
  std::vector<bool> processed (cloud.points.size (), false);

  std::vector<int> nn_indices;
  std::vector<float> nn_distances;
  // Process all points in the indices vector
  for (size_t i = 0; i < cloud.points.size (); ++i)
  {
    if (processed[i])
      continue;

    std::vector<unsigned int> seed_queue;
    int sq_idx = 0;
    seed_queue.push_back (i);

    processed[i] = true;

    while (sq_idx < (int)seed_queue.size ())
    {
      // Search for sq_idx
      if (!tree->radiusSearch (seed_queue[sq_idx], tolerance, nn_indices, nn_distances))
      {
        sq_idx++;
        continue;
      }

      for (size_t j = 1; j < nn_indices.size (); ++j)             // nn_indices[0] should be sq_idx
      {
        if (processed[nn_indices[j]])                         // Has this point been processed before ?
          continue;

        processed[nn_indices[j]] = true;
        // [-1;1]
        double dot_p = normals.points[i].normal[0] * normals.points[nn_indices[j]].normal[0] +
                       normals.points[i].normal[1] * normals.points[nn_indices[j]].normal[1] +
                       normals.points[i].normal[2] * normals.points[nn_indices[j]].normal[2];
        if ( fabs (acos (dot_p)) < eps_angle )
        {
          processed[nn_indices[j]] = true;
          seed_queue.push_back (nn_indices[j]);
        }
      }

      sq_idx++;
    }

    // If this queue is satisfactory, add to the clusters
    if (seed_queue.size () >= min_pts_per_cluster && seed_queue.size () <= max_pts_per_cluster)
    {
      pcl::PointIndices r;
      r.indices.resize (seed_queue.size ());
      for (size_t j = 0; j < seed_queue.size (); ++j)
        r.indices[j] = seed_queue[j];

      sort (r.indices.begin (), r.indices.end ());
      r.indices.erase (unique (r.indices.begin (), r.indices.end ()), r.indices.end ());

      r.header = cloud.header;
      clusters.push_back (r);   // We could avoid a copy by working directly in the vector
    }
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Decompose a region of space into clusters based on the euclidean distance between points, and the normal
  * angular deviation
  * \param cloud the point cloud message
  * \param normals the point cloud message containing normal information
  * \param indices a list of point indices to use from \a cloud
  * \param tree the spatial locator (e.g., kd-tree) used for nearest neighbors searching
  * \note the tree has to be created as a spatial locator on \a cloud
  * \param tolerance the spatial cluster tolerance as a measure in the L2 Euclidean space
  * \param clusters the resultant clusters containing point indices (as PointIndices)
  * \param eps_angle the maximum allowed difference between normals in degrees for cluster/region growing
  * \param min_pts_per_cluster minimum number of points that a cluster may contain (default = 1)
  */
template <typename PointT, typename Normal> void
  pcl::extractEuclideanClusters (const pcl::PointCloud<PointT> &cloud, const pcl::PointCloud<Normal> &normals,
                                 const std::vector<int> &indices, 
                                 const boost::shared_ptr<pcl::KdTree<PointT> > &tree,
                                 float tolerance, std::vector<pcl::PointIndices> &clusters,
                                 double eps_angle, 
                                 unsigned int min_pts_per_cluster = 1, 
                                 unsigned int max_pts_per_cluster = std::numeric_limits<int>::max ())
{
  // \note If the tree was created over <cloud, indices>, we guarantee a 1-1 mapping between what the tree returns
  //and indices[i]
  ROS_ASSERT (tree->getInputCloud ()->points.size () == cloud.points.size ());
  ROS_ASSERT (tree->getIndices ()->size () == indices.size ());
  ROS_ASSERT (cloud.points.size () == normals.points.size ());
  // Create a bool vector of processed point indices, and initialize it to false
  std::vector<bool> processed (indices.size (), false);

  std::vector<int> nn_indices;
  std::vector<float> nn_distances;
  // Process all points in the indices vector
  for (size_t i = 0; i < indices.size (); ++i)
  {
    if (processed[i])
      continue;

    std::vector<int> seed_queue;
    int sq_idx = 0;
    seed_queue.push_back (i);

    processed[i] = true;

    while (sq_idx < (int)seed_queue.size ())
    {
      // Search for sq_idx
      if (!tree->radiusSearch (seed_queue[sq_idx], tolerance, nn_indices, nn_distances))
      {
        sq_idx++;
        continue;
      }

      for (size_t j = 1; j < nn_indices.size (); ++j)             // nn_indices[0] should be sq_idx
      {
        if (processed[nn_indices[j]])                             // Has this point been processed before ?
          continue;

        processed[nn_indices[j]] = true;
        // [-1;1]
        double dot_p =
          normals.points[indices[i]].normal[0] * normals.points[indices[nn_indices[j]]].normal[0] +
          normals.points[indices[i]].normal[1] * normals.points[indices[nn_indices[j]]].normal[1] +
          normals.points[indices[i]].normal[2] * normals.points[indices[nn_indices[j]]].normal[2];
        if ( fabs (acos (dot_p)) < eps_angle )
        {
          processed[nn_indices[j]] = true;
          seed_queue.push_back (nn_indices[j]);
        }
      }

      sq_idx++;
    }

    // If this queue is satisfactory, add to the clusters
    if (seed_queue.size () >= min_pts_per_cluster && seed_queue.size () <= max_pts_per_cluster)
    {
      pcl::PointIndices r;
      r.indices.resize (seed_queue.size ());
      for (size_t j = 0; j < seed_queue.size (); ++j)
        r.indices[j] = indices[seed_queue[j]];

      sort (r.indices.begin (), r.indices.end ());
      r.indices.erase (unique (r.indices.begin (), r.indices.end ()), r.indices.end ());

      r.header = cloud.header;
      clusters.push_back (r);
    }
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Cluster extraction in a PointCloud given by <setInputCloud (), setIndices ()>
  * \param clusters the resultant point clusters
  */
template <typename PointT> void 
  pcl::EuclideanClusterExtraction<PointT>::extract (std::vector<pcl::PointIndices> &clusters)
{
  if (!initCompute ()) 
    return;

  // Initialize the spatial locator
  initTree (spatial_locator_, tree_);
  // Send the input dataset to the spatial locator

  tree_->setInputCloud (input_, indices_);
  extractEuclideanClusters (*input_, *indices_, tree_, cluster_tolerance_, clusters, min_pts_per_cluster_, max_pts_per_cluster_);
  //tree_->setInputCloud (input_);
  //extractEuclideanClusters (*input_, tree_, cluster_tolerance_, clusters, min_pts_per_cluster_, max_pts_per_cluster_);

  deinitCompute ();
}


