/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 *
 */

/**

\author Bastian Steder

@b convert_alu_pc_to_ros_pc loads a binary point cloud file as used in software by the University of Freiburg and saves it as a pcd file.

**/

#include <iostream>
#include <fstream>
using namespace std;

#include "pcl/io/pcd_io.h"
#include "pcl/point_types.h"

bool printEveryPoint = false;

bool checkFileHeader(std::istream& file, const std::string headerKeyWord) {
  unsigned int posInFile = file.tellg();
  unsigned int headerKeyWordLength = headerKeyWord.length();

  char headerKeyWordFromFileC[headerKeyWordLength+1]; headerKeyWordFromFileC[headerKeyWordLength] = 0;
  file.read(headerKeyWordFromFileC, headerKeyWordLength);
  string headerKeyWordFromFile = headerKeyWordFromFileC;
  
  file.seekg(posInFile);  // Go back to former pos in file
  
  if (headerKeyWord == headerKeyWordFromFile) {
    //cout << "Found correct header \""<<headerKeyWordFromFile<<"\".\n";
    return true;
  }
  
  //cout << "Did not find correct header: \""<<headerKeyWordFromFile<<"\" instead of \""<<headerKeyWord<<"\".\n";
  return false;
}

/* ---[ */
int main (int argc, char** argv)
{
  if (argc < 2) {
    ROS_ERROR ("Need file name as command line argument.");
    return -1;
  }
  
  string fileName = argv[1];
  
  ROS_INFO ("Trying to open \"%s\".\n", fileName.c_str());
  std::ifstream file(fileName.c_str());
  if (!file) {
    ROS_ERROR ("Was not able to open file \"%s\".\n", fileName.c_str());
    return -1;
  }
  string outputFileName = fileName + ".pcd";
  
  std::string headerLine;
  std::string headerKeyWord = "PointCloudColoredWithSensorPosesT<float>";
  if (checkFileHeader(file, headerKeyWord)) {
    ROS_INFO_STREAM("File is of type \""<<headerKeyWord<<"\".\n");
    getline(file, headerLine);
    //ROS_INFO ("Header is \"%s\".\n", headerLine.c_str());

    pcl::PointCloud<pcl::PointXYZRGBA> pointCloud;
    pcl::PointXYZRGBA currentPoint;
    pcl::PointXYZ sensorPosSum, sensorPos;
    sensorPosSum.x = sensorPosSum.y = sensorPosSum.z = 0.0f;
    float& pointX=currentPoint.x, & pointY=currentPoint.y, & pointZ=currentPoint.z;
    unsigned char* rgbaPtr = (unsigned char*) &currentPoint.rgba;
    unsigned char& colorR=*rgbaPtr, & colorG=*(rgbaPtr+1), & colorB=*(rgbaPtr+2), & colorA=*(rgbaPtr+3);
    
    u_int32_t numPoints;
    file.read((char*) &numPoints, sizeof(numPoints));
    ROS_INFO_STREAM("Point cloud has size "<<numPoints<<".\n");
    
    for (unsigned int i=0; i<numPoints; ++i) {
      u_int32_t vectorSize;
      file.read((char*) &vectorSize, sizeof(vectorSize));
      if (vectorSize != 3) {
        ROS_ERROR_STREAM("Vector size is not 3!.\n");
        return -1;
      }

      float freiburgX, freiburgY, freiburgZ;
      // Read current point and convert to different coordinate system
      file.read((char*) &freiburgX, sizeof(freiburgZ));
      file.read((char*) &freiburgY, sizeof(freiburgX));
      file.read((char*) &freiburgZ, sizeof(freiburgZ));
      pointX = -freiburgY;
      pointY = -freiburgZ;
      pointZ =  freiburgX;
      
      file.read((char*) &vectorSize, sizeof(vectorSize));
      if (vectorSize != 3) {
        ROS_ERROR_STREAM("Vector size is not 3!.\n");
        return -1;
      }
      
      // Read current sensor pos and convert to different coordinate system
      file.read((char*) &freiburgX, sizeof(freiburgZ));
      file.read((char*) &freiburgY, sizeof(freiburgX));
      file.read((char*) &freiburgZ, sizeof(freiburgZ));
      float sensorX = -freiburgY,
            sensorY = -freiburgZ,
            sensorZ =  freiburgX;
      sensorPosSum.x+=sensorX;  sensorPosSum.y+=sensorY;  sensorPosSum.z+=sensorZ;
      bool hasColor;
      file.read((char*) &hasColor, sizeof(hasColor));
      colorR = colorG = colorB = colorA = 0;
      if (hasColor) {
        colorA = 255;
        file.read((char*) &colorR, sizeof(colorR));
        file.read((char*) &colorG, sizeof(colorG));
        file.read((char*) &colorB, sizeof(colorB));
      }

      if (printEveryPoint) {
        cout << "Point=("<<pointX<<","<<pointY<<","<<pointZ<<")";
        cout << ", SensorPos=("<<sensorX<<","<<sensorY<<","<<sensorZ<<")";
        if (hasColor) {
          cout << ", color=("<<(int)colorR<<","<<(int)colorG<<","<<(int)colorB<<")";
        }
        else {
          cout << " with no color";
        }
        cout << ".\n";
      }
      
      currentPoint.x = pointX;
      currentPoint.y = pointY;
      currentPoint.z = pointZ;
      
      pointCloud.points.push_back(currentPoint);
    }
    
    // Move sensorPos mean to 0,0,0
    sensorPos.x=sensorPosSum.x/numPoints; sensorPos.y=sensorPosSum.y/numPoints; sensorPos.z=sensorPosSum.z/numPoints; 
    for (unsigned int i=0; i<numPoints; ++i) {
      pcl::PointXYZRGBA& currentPoint = pointCloud.points[i];
      currentPoint.x-=sensorPos.x;  currentPoint.y-=sensorPos.y;  currentPoint.z-=sensorPos.z;
    }
    
    pointCloud.width = pointCloud.points.size();
    pointCloud.height = 1;
    ROS_INFO_STREAM(pointCloud);
    pcl::io::savePCDFileASCII (outputFileName, pointCloud);
    
    return 0;
  }
  
  headerKeyWord = "PointCloudWithSensorPosesT<float>";
  if (checkFileHeader(file, headerKeyWord)) {
    ROS_INFO_STREAM("File is of type \""<<headerKeyWord<<"\".\n");
    getline(file, headerLine);
    //ROS_INFO ("Header is \"%s\".\n", headerLine.c_str());

    pcl::PointCloud<pcl::PointXYZ> pointCloud;
    pcl::PointXYZ currentPoint;
    pcl::PointXYZ sensorPosSum, sensorPos;
    sensorPosSum.x = sensorPosSum.y = sensorPosSum.z = 0.0f;
    float& pointX=currentPoint.x, & pointY=currentPoint.y, & pointZ=currentPoint.z;
    
    u_int32_t numPoints;
    file.read((char*) &numPoints, sizeof(numPoints));
    ROS_INFO_STREAM("Point cloud has size "<<numPoints<<".\n");
    
    for (unsigned int i=0; i<numPoints; ++i) {
      float freiburgX, freiburgY, freiburgZ;
      // Read current point and convert to different coordinate system
      file.read((char*) &freiburgX, sizeof(freiburgZ));
      file.read((char*) &freiburgY, sizeof(freiburgX));
      file.read((char*) &freiburgZ, sizeof(freiburgZ));
      pointX = -freiburgY;
      pointY = -freiburgZ;
      pointZ =  freiburgX;
      
      // Read current sensor pos and convert to different coordinate system
      file.read((char*) &freiburgX, sizeof(freiburgZ));
      file.read((char*) &freiburgY, sizeof(freiburgX));
      file.read((char*) &freiburgZ, sizeof(freiburgZ));
      float sensorX = -freiburgY,
            sensorY = -freiburgZ,
            sensorZ =  freiburgX;
      sensorPosSum.x+=sensorX;  sensorPosSum.y+=sensorY;  sensorPosSum.z+=sensorZ;

      if (printEveryPoint) {
        cout << "Point=("<<pointX<<","<<pointY<<","<<pointZ<<")";
        cout << ", SensorPos=("<<sensorX<<","<<sensorY<<","<<sensorZ<<")";
        cout << ".\n";
      }
      
      currentPoint.x = pointX;
      currentPoint.y = pointY;
      currentPoint.z = pointZ;
      
      pointCloud.points.push_back(currentPoint);
    }
    
    // Move sensorPos mean to 0,0,0
    sensorPos.x=sensorPosSum.x/numPoints; sensorPos.y=sensorPosSum.y/numPoints; sensorPos.z=sensorPosSum.z/numPoints; 
    for (unsigned int i=0; i<numPoints; ++i) {
      pcl::PointXYZ& currentPoint = pointCloud.points[i];
      currentPoint.x-=sensorPos.x;  currentPoint.y-=sensorPos.y;  currentPoint.z-=sensorPos.z;
    }
    
    pointCloud.width = pointCloud.points.size();
    pointCloud.height = 1;
    ROS_INFO_STREAM(pointCloud);
    pcl::io::savePCDFileASCII (outputFileName, pointCloud);
    
    return 0;
  }

  headerKeyWord = "PointCloudT<float>";
  if (checkFileHeader(file, headerKeyWord)) {
    ROS_INFO_STREAM("File is of type \""<<headerKeyWord<<"\".\n");
    ROS_ERROR_STREAM("Sorry, not implemented yet.\n");
    
    return 0;
  }
  
  ROS_ERROR ("Could not determine file type. Doing nothing.\n");

  return (-1);
}
/* ]--- */
