/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: feature.h 32052 2010-08-27 02:19:30Z rusu $
 *
 */

#ifndef PCL_ROS_FEATURE_H_
#define PCL_ROS_FEATURE_H_

// PCL includes
#include <pcl/features/feature.h>
#include <pcl/PointIndices.h>

#include "pcl_ros/pcl_nodelet.h"

// Dynamic reconfigure
#include <dynamic_reconfigure/server.h>
#include "pcl_ros/FeatureConfig.h"

namespace pcl_ros
{
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief @b Feature represents the base feature class. Some generic 3D operations that are applicable to all features
    * are defined here as static methods.
    * \author Radu Bogdan Rusu
    */
  template <typename PointInT, typename PointOutT>
  class Feature : virtual public pcl::Feature<PointInT, PointOutT>, virtual public PCLNodelet<PointInT>
  {
    typedef pcl::Feature<PointInT, PointOutT> FeatureBaseClass;
    typedef PCLNodelet<PointInT> PCLNodeletBaseClass;

     public:
      using PCLNodelet<PointInT>::getName;
      using PCLNodelet<PointInT>::isValid;
      using PCLNodelet<PointInT>::sub_input_;
      using PCLNodelet<PointInT>::pub_output_;
      using PCLNodelet<PointInT>::approximate_sync_;
     
      typedef typename pcl::KdTree<PointInT> KdTree;
      typedef typename pcl::KdTree<PointInT>::Ptr KdTreePtr;

      typedef sensor_msgs::PointCloud2 PointCloud2;

      typedef pcl::PointCloud<PointInT> PointCloudIn;
      typedef typename PointCloudIn::Ptr PointCloudInPtr;
      typedef typename PointCloudIn::ConstPtr PointCloudInConstPtr;

      typedef pcl::PointCloud<PointOutT> PointCloudOut;

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Empty constructor. */
      Feature () : use_surface_ (false), spatial_locator_type_ (-1) {};

    protected:
      // Members derived from the base class
      using PCLNodeletBaseClass::pnh_;
      using PCLNodeletBaseClass::max_queue_size_;
      using PCLNodeletBaseClass::sub_input_filter_;
      using PCLNodeletBaseClass::sub_indices_filter_;
      using PCLNodeletBaseClass::sync_input_indices_a_;
      using PCLNodeletBaseClass::sync_input_indices_e_;

      using FeatureBaseClass::input_;
      using FeatureBaseClass::indices_;
      using FeatureBaseClass::surface_;
      using FeatureBaseClass::tree_;
      using FeatureBaseClass::k_;
      using FeatureBaseClass::search_radius_;
      using FeatureBaseClass::use_indices_;

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      // ROS nodelet attributes

      /** \brief The surface PointCloud subscriber filter. */
      message_filters::Subscriber<sensor_msgs::PointCloud2> sub_surface_filter_;

      /** \brief Set to true if the nodelet needs to listen for incoming point clouds representing the search surface. */
      bool use_surface_;

      /** \brief Parameter for the spatial locator tree. By convention, the values represent:
        * 0: ANN (Approximate Nearest Neigbor library) kd-tree
        * 1: FLANN (Fast Library for Approximate Nearest Neighbors) kd-tree
        * 2: Organized spatial dataset index
        */
      int spatial_locator_type_;

      /** \brief Synchronized input, surface, and point indices.*/
      boost::shared_ptr <message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, sensor_msgs::PointCloud2, pcl::PointIndices> > > sync_input_surface_indices_a_;
      boost::shared_ptr <message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, sensor_msgs::PointCloud2, pcl::PointIndices> > > sync_input_surface_indices_e_;

      /** \brief Synchronized input, and surface.*/
      boost::shared_ptr <message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<sensor_msgs::PointCloud2, sensor_msgs::PointCloud2> > > sync_input_surface_a_;
      boost::shared_ptr <message_filters::Synchronizer<message_filters::sync_policies::ExactTime<sensor_msgs::PointCloud2, sensor_msgs::PointCloud2> > > sync_input_surface_e_;

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Compute the feature and publish it. Internal method. */
      void computeAndPublish ();

    private:
      /** \brief Pointer to a dynamic reconfigure service. */
      boost::shared_ptr <dynamic_reconfigure::Server<FeatureConfig> > srv_;

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Nodelet initialization routine. */
      virtual void onInit ();

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Dynamic reconfigure callback
        * \param config the config object
        * \param level the dynamic reconfigure level
        */
      void config_callback (FeatureConfig &config, uint32_t level);

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Input point cloud callback. Used when \a use_indices and \a use_surface are set.
        * \param cloud the pointer to the input point cloud
        * \param cloud_surface the pointer to the surface point cloud
        * \param indices the pointer to the input point cloud indices
        */
      void input_surface_indices_callback (const PointCloud2::ConstPtr &cloud, const PointCloud2::ConstPtr &cloud_surface, const pcl::PointIndicesConstPtr &indices);

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Input point cloud callback. Used when \a use_indices is set.
        * \param cloud the pointer to the input point cloud
        * \param indices the pointer to the input point cloud indices
        */
      void input_indices_callback (const PointCloud2::ConstPtr &cloud, const pcl::PointIndicesConstPtr &indices);

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Input point cloud callback. Used when \a use_surface is set.
        * \param cloud the pointer to the input point cloud
        * \param cloud_surface the pointer to the surface point cloud
        */
      void input_surface_callback (const PointCloud2::ConstPtr &cloud, const PointCloud2::ConstPtr &cloud_surface);

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Input point cloud callback. Used only when \a use_indices and \a use_surface are not set.
        * \param cloud the pointer to the input point cloud
        */
      void input_callback (const PointCloudInConstPtr &cloud);
  };

  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  template <typename PointInT, typename PointNT, typename PointOutT>
  class FeatureFromNormals : virtual public Feature<PointInT, PointOutT>, virtual public pcl::FeatureFromNormals<PointInT, PointNT, PointOutT>
  {
    typedef pcl::FeatureFromNormals<PointInT, PointNT, PointOutT> FeatureFromNormalsBaseClass;
    typedef Feature<PointInT, PointOutT> FeatureBaseClass;


    typedef typename Feature<PointInT, PointOutT>::PointCloudIn PointCloudIn;
    typedef typename PointCloudIn::Ptr PointCloudInPtr;
    typedef typename PointCloudIn::ConstPtr PointCloudInConstPtr;

    typedef typename pcl::PointCloud<PointNT> PointCloudN;
    typedef typename PointCloudN::Ptr PointCloudNPtr;
    typedef typename PointCloudN::ConstPtr PointCloudNConstPtr;

    public:
      using Feature<PointInT, PointOutT>::getName;

      typedef sensor_msgs::PointCloud2 PointCloud2;

      typedef typename Feature<PointInT, PointOutT>::PointCloudOut PointCloudOut;

    protected:
      // Members derived from the base class
      using FeatureFromNormalsBaseClass::normals_;

      using FeatureBaseClass::pnh_;
      using FeatureBaseClass::isValid;
      using FeatureBaseClass::max_queue_size_;
      using FeatureBaseClass::sub_input_filter_;
      using FeatureBaseClass::sub_indices_filter_;
      using FeatureBaseClass::sync_input_indices_a_;
      using FeatureBaseClass::sync_input_indices_e_;
      using FeatureBaseClass::sub_surface_filter_;
      using FeatureBaseClass::input_;
      using FeatureBaseClass::indices_;
      using FeatureBaseClass::use_indices_;
      using FeatureBaseClass::surface_;
      using FeatureBaseClass::pub_output_;
      using FeatureBaseClass::max_queue_size_;
      using FeatureBaseClass::spatial_locator_type_;
      using FeatureBaseClass::use_surface_;
      using FeatureBaseClass::k_;
      using FeatureBaseClass::search_radius_;
      using FeatureBaseClass::computeAndPublish;
      using FeatureBaseClass::approximate_sync_;

    private:
      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      // ROS nodelet attributes
      /** \brief The normals PointCloud subscriber filter. */
      message_filters::Subscriber<sensor_msgs::PointCloud2> sub_normals_filter_;

      /** \brief Synchronized input, and normals.*/
      boost::shared_ptr <message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, PointCloud2> > > sync_input_normals_a_;
      boost::shared_ptr <message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, PointCloud2> > > sync_input_normals_e_;

      /** \brief Synchronized input, normals, and indices.*/
      boost::shared_ptr<message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, PointCloud2, pcl::PointIndices> > > sync_input_normals_indices_a_;
      boost::shared_ptr<message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, PointCloud2, pcl::PointIndices> > > sync_input_normals_indices_e_;

      /** \brief Synchronized input, normals, and surface.*/
      boost::shared_ptr<message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, PointCloud2, PointCloud2> > > sync_input_normals_surface_a_;
      boost::shared_ptr<message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, PointCloud2, PointCloud2> > > sync_input_normals_surface_e_;

      /** \brief Synchronized input, normals, surface, and point indices.*/
      boost::shared_ptr<message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, PointCloud2, PointCloud2, pcl::PointIndices> > > sync_input_normals_surface_indices_a_;
      boost::shared_ptr<message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, PointCloud2, PointCloud2, pcl::PointIndices> > > sync_input_normals_surface_indices_e_;

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Nodelet initialization routine. */
      virtual void onInit ();

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Input point cloud callback. Used when \a use_indices and \a use_surface are set.
        * \param cloud the pointer to the input point cloud
        * \param cloud_normals the pointer to the input point cloud normals
        * \param cloud_surface the pointer to the surface point cloud
        * \param indices the pointer to the input point cloud indices
        */
      void input_normals_surface_indices_callback (const PointCloud2::ConstPtr &cloud, const PointCloud2::ConstPtr &cloud_normals, const PointCloud2::ConstPtr &cloud_surface, const pcl::PointIndicesConstPtr &indices);

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Input point cloud callback. Used when \a use_indices is set.
        * \param cloud the pointer to the input point cloud
        * \param cloud_normals the pointer to the input point cloud normals
        * \param indices the pointer to the input point cloud indices
        */
      void input_normals_indices_callback (const PointCloud2::ConstPtr &cloud, const PointCloud2::ConstPtr &cloud_normals, const pcl::PointIndicesConstPtr &indices);

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Input point cloud callback. Used when \a use_surface is set.
        * \param cloud the pointer to the input point cloud
        * \param cloud_normals the pointer to the input point cloud normals
        * \param cloud_surface the pointer to the surface point cloud
        */
      void input_normals_surface_callback (const PointCloud2::ConstPtr &cloud, const PointCloud2::ConstPtr &cloud_normals, const PointCloud2::ConstPtr &cloud_surface);

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Input point cloud callback. Used only when \a use_indices and \a use_surface are not set.
        * \param cloud the pointer to the input point cloud
        * \param cloud_normals the pointer to the input point cloud normals
        */
      void input_normals_callback (const PointCloud2::ConstPtr &cloud, const PointCloud2::ConstPtr &cloud_normals);
  };

}

#include "pcl_ros/features/feature.hpp"

#endif  //#ifndef PCL_ROS_FEATURE_H_
