/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: filter.h 32052 2010-08-27 02:19:30Z rusu $
 *
 */

#ifndef PCL_ROS_FILTER_H_
#define PCL_ROS_FILTER_H_

// PCL includes
#include <pcl/filters/filter.h>
#include <pcl/ros/conversions.h>

#include "pcl_ros/pcl_nodelet.h"

// Dynamic reconfigure
#include <dynamic_reconfigure/server.h>
#include "pcl_ros/FilterConfig.h"

namespace pcl_ros
{
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief @b Filter represents the base filter class. Some generic 3D operations that are applicable to all filters
    * are defined here as static methods.
    * \author Radu Bogdan Rusu
    */
  template <typename PointT>
  class Filter : virtual public pcl::Filter<PointT>, virtual public PCLNodelet<PointT>
  {
    typedef pcl::Filter<PointT> FilterBaseClass;

    public:
      // Members derived from the base class
      using PCLNodelet<PointT>::getName;
      using PCLNodelet<PointT>::isValid;
      using PCLNodelet<PointT>::pnh_;
      using PCLNodelet<PointT>::max_queue_size_;
      using PCLNodelet<PointT>::sub_input_;
      using PCLNodelet<PointT>::pub_output_;
      using PCLNodelet<PointT>::sub_input_filter_;
      using PCLNodelet<PointT>::sub_indices_filter_;
      using PCLNodelet<PointT>::sync_input_indices_e_;
      using PCLNodelet<PointT>::sync_input_indices_a_;
      using PCLNodelet<PointT>::approximate_sync_;
      using FilterBaseClass::filter_field_name_;
      using FilterBaseClass::filter_limit_min_;
      using FilterBaseClass::filter_limit_max_;
      using FilterBaseClass::getFilterLimits;
      using FilterBaseClass::setFilterLimits;
      using FilterBaseClass::getFilterFieldName;
      using FilterBaseClass::setFilterFieldName;
      using FilterBaseClass::input_;
      using FilterBaseClass::indices_;
      using FilterBaseClass::use_indices_;

      typedef sensor_msgs::PointCloud2 PointCloud2;

      typedef pcl::PointCloud<PointT> PointCloud;
      typedef typename PointCloud::Ptr PointCloudPtr;
      typedef typename PointCloud::ConstPtr PointCloudConstPtr;

    protected:
      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Child initialization routine.
        * \param nh ROS node handle
        * \param has_service set to true if the child has a Dynamic Reconfigure service
        */
      virtual bool child_init (ros::NodeHandle &nh, bool &has_service) { has_service = false; return (true); }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Apply the filter and publish the results. Internal method. */
      void computeAndPublish ();

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Nodelet initialization routine. */
      virtual void onInit ();

    private:

      /** \brief Pointer to a dynamic reconfigure service. */
      boost::shared_ptr <dynamic_reconfigure::Server<pcl_ros::FilterConfig> > srv_;

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Dynamic reconfigure callback
        * \param config the config object
        * \param level the dynamic reconfigure level
        */
      void config_callback (pcl_ros::FilterConfig &config, uint32_t level);

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Input point cloud callback. Used only when \a use_indices is not set.
        * \param cloud the pointer to the input point cloud
        */
      void input_callback (const PointCloudConstPtr &cloud);

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Input point cloud callback. Used when \a use_indices is set.
        * \param cloud the pointer to the input point cloud
        * \param indices the pointer to the input point cloud indices
        */
      void input_indices_callback (const PointCloud2::ConstPtr &cloud, const pcl::PointIndicesConstPtr &indices);
  };

  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief @b Filter represents the base filter class. Some generic 3D operations that are applicable to all filters
    * are defined here as static methods.
    * \author Radu Bogdan Rusu
    */
  template <>
  class Filter<sensor_msgs::PointCloud2> : virtual public pcl::Filter<sensor_msgs::PointCloud2>, virtual public PCLNodelet<sensor_msgs::PointCloud2>
  {
    typedef pcl::Filter<sensor_msgs::PointCloud2> FilterBaseClass;

    public:
      using PCLNodelet<sensor_msgs::PointCloud2>::getName;
      using PCLNodelet<sensor_msgs::PointCloud2>::sub_input_;
      using PCLNodelet<sensor_msgs::PointCloud2>::pub_output_;
      using PCLNodelet<sensor_msgs::PointCloud2>::approximate_sync_;
      using PCLNodelet<sensor_msgs::PointCloud2>::sync_input_indices_e_;
      using PCLNodelet<sensor_msgs::PointCloud2>::sync_input_indices_a_;

      typedef sensor_msgs::PointCloud2 PointCloud2;

    protected:
      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Child initialization routine.
        * \param nh ROS node handle
        * \param has_service set to true if the child has a Dynamic Reconfigure service
        */
      virtual bool child_init (ros::NodeHandle &nh, bool &has_service) { has_service = false; return (true); }

      void computeAndPublish ();

      virtual void onInit ();

    private:
      /** \brief Pointer to a dynamic reconfigure service. */
      boost::shared_ptr<dynamic_reconfigure::Server<pcl_ros::FilterConfig> > srv_;

      void config_callback (pcl_ros::FilterConfig &config, uint32_t level);
      void input_callback (const PointCloud2::ConstPtr &cloud);
      void input_indices_callback (const PointCloud2::ConstPtr &cloud, const pcl::PointIndicesConstPtr &indices);
  };
}

#include "pcl_ros/filters/filter.hpp"

#endif  //#ifndef PCL_ROS_FILTER_H_
