/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: project_inliers.h 32941 2010-09-28 23:47:56Z rusu $
 *
 */

#ifndef PCL_ROS_FILTERS_PROJECT_INLIERS_H_
#define PCL_ROS_FILTERS_PROJECT_INLIERS_H_

// PCL includes
#include <pcl/filters/project_inliers.h>
#include "pcl_ros/filters/filter.h"

#include <message_filters/subscriber.h>

namespace pcl_ros
{
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief @b ProjectInliers uses a model and a set of inlier indices from a PointCloud to project them into a
    * separate PointCloud.
    * \note setFilterFieldName (), setFilterLimits (), and setFilterLimitNegative () are ignored.
    * \author Radu Bogdan Rusu
    */
  template <typename PointT>
  class ProjectInliers: virtual public Filter<PointT>, virtual public pcl::ProjectInliers<PointT>
  {
    using Filter<PointT>::getName;
    using Filter<PointT>::isValid;
    using Filter<PointT>::pnh_;
    using Filter<PointT>::sub_input_filter_;
    using Filter<PointT>::sub_indices_filter_;
    using Filter<PointT>::input_;
    using Filter<PointT>::indices_;
    using Filter<PointT>::pub_output_;
    using Filter<PointT>::max_queue_size_;
    using Filter<PointT>::model_type_;
    using Filter<PointT>::copy_all_data_;
    using Filter<PointT>::model_;
    using Filter<PointT>::computeAndPublish;
    using Filter<PointT>::approximate_sync_;

    typedef sensor_msgs::PointCloud2 PointCloud2;
    
    typedef typename Filter<PointT>::PointCloud PointCloud;
    
    private:
      /** \brief The message filter subscriber for model coefficients. */
      message_filters::Subscriber<pcl::ModelCoefficients> sub_model_;

      /** \brief Synchronized input, indices, and model coefficients.*/
      boost::shared_ptr<message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, pcl::PointIndices, pcl::ModelCoefficients> > > sync_input_indices_model_e_;
      boost::shared_ptr<message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, pcl::PointIndices, pcl::ModelCoefficients> > > sync_input_indices_model_a_;

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Nodelet initialization routine. */
      virtual void onInit ();

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Input point cloud callback.
        * \param cloud the pointer to the input point cloud
        * \param indices the pointer to the input point cloud indices
        * \param model the model coefficients
        */
      void input_indices_model_callback (const PointCloud2::ConstPtr &cloud, const pcl::PointIndicesConstPtr &indices, const pcl::ModelCoefficientsConstPtr &model);
  };

  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief @b ProjectInliers uses a model and a set of inlier indices from a PointCloud to project them into a
    * separate PointCloud.
    * \note setFilterFieldName (), setFilterLimits (), and setFilterLimitNegative () are ignored.
    * \author Radu Bogdan Rusu
    */
  template <>
  class ProjectInliers<sensor_msgs::PointCloud2> : virtual public Filter<sensor_msgs::PointCloud2>, virtual public pcl::ProjectInliers<sensor_msgs::PointCloud2>
  {
    using Filter<sensor_msgs::PointCloud2>::getName;

    typedef sensor_msgs::PointCloud2 PointCloud2;
    
    typedef Filter<sensor_msgs::PointCloud2>::PointCloud PointCloud;

    private:
      /** \brief The message filter subscriber for model coefficients. */
      message_filters::Subscriber<pcl::ModelCoefficients> sub_model_;

      /** \brief Synchronized input, indices, and model coefficients.*/
      boost::shared_ptr<message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, pcl::PointIndices, pcl::ModelCoefficients> > > sync_input_indices_model_e_;
      boost::shared_ptr<message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, pcl::PointIndices, pcl::ModelCoefficients> > > sync_input_indices_model_a_;

      virtual void onInit ();

      void input_indices_model_callback (const PointCloud2::ConstPtr &cloud, const pcl::PointIndicesConstPtr &indices, const pcl::ModelCoefficientsConstPtr &model);
  };
}

#include "pcl_ros/filters/project_inliers.hpp"

#endif  //#ifndef PCL_FILTERS_PROJECT_INLIERS_H_
