/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: pcd_io.h 32942 2010-09-28 23:53:38Z rusu $
 *
 */

#ifndef PCL_ROS_IO_PCD_IO_H_
#define PCL_ROS_IO_PCD_IO_H_

#include <pcl/io/pcd_io.h>
#include "pcl_ros/pcl_nodelet.h"

namespace pcl_ros
{
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief Point Cloud Data (PCD) file format reader.
    * \author Radu Bogdan Rusu
    */
  class PCDReader : virtual public pcl::PCDReader, virtual public PCLNodelet<sensor_msgs::PointCloud2>
  {
    using PCLNodelet<sensor_msgs::PointCloud2>::getName;

    public:
      typedef sensor_msgs::PointCloud2 PointCloud;
      typedef PointCloud::Ptr PointCloudPtr;
      typedef PointCloud::ConstPtr PointCloudConstPtr;

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Empty constructor. */
      PCDReader () : publish_rate_ (0), tf_frame_ ("/base_link") {};

      virtual void onInit ();

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Set the publishing rate in seconds.
        * \param publish_rate the publishing rate in seconds
        */
      inline void setPublishRate (double publish_rate) { publish_rate_ = publish_rate; }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Get the publishing rate in seconds. */
      inline double getPublishRate () { return (publish_rate_); }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Set the TF frame the PointCloud will be published in.
        * \param tf_frame the TF frame the PointCloud will be published in
        */
      inline void setTFframe (std::string tf_frame) { tf_frame_ = tf_frame; }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Get the TF frame the PointCloud is published in. */
      inline std::string getTFframe () { return (tf_frame_); }

    protected:
      /** \brief The publishing interval in seconds. Set to 0 to only publish once (default). */
      double publish_rate_;

      /** \brief The TF frame the data should be published in ("/base_link" by default). */
      std::string tf_frame_;
  };

  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief Point Cloud Data (PCD) file format writer.
    * \author Radu Bogdan Rusu
    */
  class PCDWriter : virtual public pcl::PCDWriter, virtual public PCLNodelet<sensor_msgs::PointCloud2>
  {
    using PCLNodelet<sensor_msgs::PointCloud2>::getName;
    using PCLNodelet<sensor_msgs::PointCloud2>::input_;

    public:
      typedef sensor_msgs::PointCloud2 PointCloud;
      typedef PointCloud::Ptr PointCloudPtr;
      typedef PointCloud::ConstPtr PointCloudConstPtr;

      virtual void onInit ();
      void input_callback (const PointCloudConstPtr &cloud);

      /** \brief The input PointCloud subscriber. */
      pcl_ros::Subscriber<sensor_msgs::PointCloud2> sub_input_;
  };
}

#endif  //#ifndef PCL_ROS_IO_PCD_IO_H_
