/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: pcl_nodelet.h 33238 2010-03-11 00:46:58Z rusu $
 *
 */

/**

\author Radu Bogdan Rusu

**/

#ifndef PCL_NODELET_H_
#define PCL_NODELET_H_

// PCL includes
#include <pcl/pcl_base.h>
#include <pcl/PointIndices.h>
#include <pcl/ModelCoefficients.h>
#include <pcl/ros/conversions.h>

#include "pcl_ros/subscriber.h"
#include "pcl_ros/publisher.h"

// ROS Nodelet includes
#include <nodelet/nodelet.h>
#include <message_filters/subscriber.h>
#include <message_filters/synchronizer.h>
#include <message_filters/sync_policies/exact_time.h>
#include <message_filters/sync_policies/approximate_time.h>

// Include TF
#include <pcl_tf/transforms.h>

namespace pcl_ros
{
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief @b PCLNodelet represents the base PCL Nodelet class. All PCL nodelets should inherit from this class. */
  template <typename PointT>
  class PCLNodelet : virtual public nodelet::Nodelet, virtual public pcl::PCLBase<PointT>
  {
    using pcl::PCLBase<PointT>::use_indices_;

    public:
      typedef sensor_msgs::PointCloud2 PointCloud2;

      typedef pcl::PointCloud<PointT> PointCloud;
      typedef typename PointCloud::Ptr PointCloudPtr;
      typedef typename PointCloud::ConstPtr PointCloudConstPtr;

      typedef pcl::PointIndices::Ptr PointIndicesPtr;
      typedef pcl::PointIndices::ConstPtr PointIndicesConstPtr;

      typedef pcl::ModelCoefficients::Ptr ModelCoefficientsPtr;
      typedef pcl::ModelCoefficients::ConstPtr ModelCoefficientsConstPtr;

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Empty constructor. */
      PCLNodelet () : max_queue_size_ (3), approximate_sync_ (false) {};

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      // ROS nodelet attributes
      /** \brief The message filter subscriber for PointCloud2. */
      message_filters::Subscriber<PointCloud2> sub_input_filter_;

      /** \brief The message filter subscriber for PointIndices. */
      message_filters::Subscriber<pcl::PointIndices> sub_indices_filter_;

      /** \brief The output PointCloud publisher. */
      ros::Publisher pub_output_;

      /** \brief The maximum queue size (default: 3). */
      int max_queue_size_;

      /** \brief True if we use an approximate time synchronizer versus an exact one (false by default). */
      bool approximate_sync_;

      /** \brief The ROS NodeHandle used for parameters, publish/subscribe, etc. */
      boost::shared_ptr<ros::NodeHandle> pnh_;

      /** \brief The input PointCloud subscriber. */
      pcl_ros::Subscriber<PointT> sub_input_;

      /** \brief Synchronized input, and indices.*/
      boost::shared_ptr<message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, pcl::PointIndices> > >       sync_input_indices_e_;
      boost::shared_ptr<message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, pcl::PointIndices> > > sync_input_indices_a_;

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Test whether a given PointCloud message is "valid" (i.e., has points, and width and height are non-zero).
        * \param cloud the point cloud to test
        * \param topic_name an optional topic name (only used for printing, defaults to "input")
        */
      inline bool
        isValid (const PointCloud2::ConstPtr &cloud, const std::string &topic_name = "input")
      {
        if (cloud->width * cloud->height == 0 || cloud->data.size () == 0)
        {
          NODELET_WARN ("Empty PointCloud (data = %zu, width = %d, height = %d) with stamp %f, and frame %s on topic %s received!", cloud->data.size (), cloud->width, cloud->height, cloud->header.stamp.toSec (), cloud->header.frame_id.c_str (), pnh_->resolveName (topic_name).c_str ());

          return (false);
        }
        return (true);
      }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Test whether a given PointCloud message is "valid" (i.e., has points, and width and height are non-zero).
        * \param cloud the point cloud to test
        * \param topic_name an optional topic name (only used for printing, defaults to "input")
        */
      inline bool
        isValid (const PointCloudConstPtr &cloud, const std::string &topic_name = "input")
      {
        if (cloud->width * cloud->height == 0 || cloud->points.size () == 0)
        {
          NODELET_WARN ("Empty PointCloud (points = %zu, width = %d, height = %d) with stamp %f, and frame %s on topic %s received!", cloud->points.size (), cloud->width, cloud->height, cloud->header.stamp.toSec (), cloud->header.frame_id.c_str (), pnh_->resolveName (topic_name).c_str ());

          return (false);
        }
        return (true);
      }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Test whether a given PointIndices message is "valid" (i.e., has values).
        * \param indices the point indices message to test
        * \param topic_name an optional topic name (only used for printing, defaults to "indices")
        */
      inline bool
        isValid (const PointIndicesConstPtr &indices, const std::string &topic_name = "indices")
      {
        if (indices->indices.size () == 0)
        {
          NODELET_WARN ("Empty indices (values = %zu) with stamp %f, and frame %s on topic %s received!", indices->indices.size (), indices->header.stamp.toSec (), indices->header.frame_id.c_str (), pnh_->resolveName (topic_name).c_str ());
          return (false);
        }
        return (true);
      }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Test whether a given ModelCoefficients message is "valid" (i.e., has values).
        * \param indices the point indices message to test
        * \param topic_name an optional topic name (only used for printing, defaults to "model")
        */
      inline bool
        isValid (const ModelCoefficientsConstPtr &model, const std::string &topic_name = "model")
      {
        if (model->values.size () == 0)
        {
          NODELET_WARN ("Empty model (values = %zu) with stamp %f, and frame %s on topic %s received!", model->values.size (), model->header.stamp.toSec (), model->header.frame_id.c_str (), pnh_->resolveName (topic_name).c_str ());
          return (false);
        }
        return (true);
      }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Nodelet initialization routine. */
      virtual void
        onInit ()
      {
        pnh_.reset (new ros::NodeHandle (getMTPrivateNodeHandle ()));
        // Parameters that we care about only at startup
        pnh_->getParam ("max_queue_size", max_queue_size_);
        
        // ---[ Optional parameters
        pnh_->getParam ("use_indices", use_indices_);
        pnh_->getParam ("approximate_sync", approximate_sync_);

//        sub_input_filter_.subscribe (private_nh_, "input", 1);
////        point_cloud_transport::PointCloudTransport input_transport (private_nh_);
//        // If we're supposed to look for PointIndices (indices)
//        if (use_indices_)
//        {
//          // Subscribe to the input using a filter
//          sub_indices_filter_.subscribe (private_nh_, "indices", 1);
//
////          sync_input_indices_ = boost::make_shared <message_filters::TimeSynchronizer<PointCloud2, PointIndices> > (3);
////          sync_input_indices_->connectInput (sub_input_filter_, sub_indices_filter_);
////          sync_input_indices_->registerCallback (bind (&PCLNodelet<PointT>::input_indices_callback, this, _1, _2));
//        }
////        else
//          // Subscribe in an old fashion to input only (no filters)
////          sub_input_.subscribe (private_nh_, "input", 1,  bind (&PCLNodelet<PointT>::input_callback, this, _1));

        NODELET_DEBUG ("[onInit] Nodelet successfully created with the following parameters:\n"
            " - approximate_sync : %s\n"
            " - use_indices      : %s\n"
            " - max_queue_size   : %d",
            (approximate_sync_) ? "true" : "false",
            (use_indices_) ? "true" : "false", max_queue_size_);

//        bool has_service;
//        child_init (private_nh, has_service);
      }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Child initialization routine.
        * \param nh ROS node handle
        * \param has_service set to true if the child has a Dynamic Reconfigure service
        */
//      virtual bool child_init (ros::NodeHandle &nh, bool &has_service) { has_service = false; return (true); }
  
      // Need to make this public, as nodelet::Nodelet::getName is protected
      inline const std::string& getName () const { return (nodelet::Nodelet::getName ()); }
    protected:
      tf::TransformListener tf_listener_;
  };

  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  template <>
  class PCLNodelet<sensor_msgs::PointCloud2> : virtual public nodelet::Nodelet, virtual public pcl::PCLBase<sensor_msgs::PointCloud2>
  { 
    using pcl::PCLBase<sensor_msgs::PointCloud2>::use_indices_;

    public:
      using nodelet::Nodelet::getName;

      typedef sensor_msgs::PointCloud2 PointCloud2;

      typedef pcl::PointIndices::Ptr PointIndicesPtr;
      typedef pcl::PointIndices::ConstPtr PointIndicesConstPtr;

      typedef pcl::ModelCoefficients::Ptr ModelCoefficientsPtr;
      typedef pcl::ModelCoefficients::ConstPtr ModelCoefficientsConstPtr;

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Empty constructor. */
      PCLNodelet () : max_queue_size_ (3), approximate_sync_ (false) {};

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      // ROS nodelet attributes
      /** \brief The message filter subscriber for PointCloud2. */
      message_filters::Subscriber<PointCloud2> sub_input_filter_;

      /** \brief The message filter subscriber for PointIndices. */
      message_filters::Subscriber<pcl::PointIndices> sub_indices_filter_;

      /** \brief The output PointCloud publisher. */
      ros::Publisher pub_output_;

      /** \brief The maximum queue size (default: 3). */
      int max_queue_size_;

      /** \brief True if we use an approximate time synchronizer versus an exact one (false by default). */
      bool approximate_sync_;

      /** \brief The ROS NodeHandle used for parameters, publish/subscribe, etc. */
      boost::shared_ptr<ros::NodeHandle> pnh_;

      /** \brief The input PointCloud subscriber. */
      pcl_ros::Subscriber<PointCloud2> sub_input_;

      /** \brief Synchronized input, and indices.*/
      boost::shared_ptr<message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, pcl::PointIndices> > >       sync_input_indices_e_;
      boost::shared_ptr<message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, pcl::PointIndices> > > sync_input_indices_a_;

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Test whether a given PointCloud message is "valid" (i.e., has points, and width and height are non-zero).
        * \param cloud the point cloud to test
        * \param topic_name an optional topic name (only used for printing, defaults to "input")
        */
      inline bool
        isValid (const PointCloud2::ConstPtr &cloud, const std::string &topic_name = "input")
      {
        if (cloud->width * cloud->height == 0 || cloud->data.size () == 0)
        {
          NODELET_WARN ("Empty PointCloud (points = %zu, width = %d, height = %d) with stamp %f, and frame %s on topic %s received!", cloud->data.size (), cloud->width, cloud->height, cloud->header.stamp.toSec (), cloud->header.frame_id.c_str (), pnh_->resolveName (topic_name).c_str ());

          return (false);
        }
        return (true);
      }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Test whether a given PointIndices message is "valid" (i.e., has values).
        * \param indices the point indices message to test
        * \param topic_name an optional topic name (only used for printing, defaults to "indices")
        */
      inline bool
        isValid (const PointIndicesConstPtr &indices, const std::string &topic_name = "indices")
      {
        if (indices->indices.size () == 0)
        {
          NODELET_WARN ("Empty indices (values = %zu) with stamp %f, and frame %s on topic %s received!", indices->indices.size (), indices->header.stamp.toSec (), indices->header.frame_id.c_str (), pnh_->resolveName (topic_name).c_str ());
          return (false);
        }
        return (true);
      }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Test whether a given ModelCoefficients message is "valid" (i.e., has values).
        * \param indices the point indices message to test
        * \param topic_name an optional topic name (only used for printing, defaults to "model")
        */
      inline bool
        isValid (const ModelCoefficientsConstPtr &model, const std::string &topic_name = "model")
      {
        if (model->values.size () == 0)
        {
          NODELET_WARN ("Empty model (values = %zu) with stamp %f, and frame %s on topic %s received!", model->values.size (), model->header.stamp.toSec (), model->header.frame_id.c_str (), pnh_->resolveName (topic_name).c_str ());
          return (false);
        }
        return (true);
      }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Nodelet initialization routine. */
      virtual void
        onInit ()
      {
        pnh_.reset (new ros::NodeHandle (getMTPrivateNodeHandle ()));
        // Parameters that we care about only at startup
        pnh_->getParam ("max_queue_size", max_queue_size_);

        // ---[ Optional parameters
        pnh_->getParam ("use_indices", use_indices_);
        pnh_->getParam ("approximate_sync", approximate_sync_);

////        sub_input_filter_.subscribe (private_nh_, "input", 1);
//////        point_cloud_transport::PointCloudTransport input_transport (private_nh_);
////        // If we're supposed to look for PointIndices (indices)
////        if (use_indices_)
////        {
////          // Subscribe to the input using a filter
////          sub_indices_filter_.subscribe (private_nh_, "indices", 1);
////
//////          sync_input_indices_ = boost::make_shared <message_filters::TimeSynchronizer<sensor_msgs::PointCloud2, PointIndices> > (3);
//////          sync_input_indices_->connectInput (sub_input_filter_, sub_indices_filter_);
//////          sync_input_indices_->registerCallback (bind (&PCLNodelet<PointT>::input_indices_callback, this, _1, _2));
////        }
//////        else
////          // Subscribe in an old fashion to input only (no filters)
//////          sub_input_.subscribe (private_nh_, "input", 1,  bind (&PCLNodelet<PointT>::input_callback, this, _1));
//
        
        NODELET_DEBUG ("[init] Nodelet successfully created with the following parameters:\n"
            " - approximate_sync : %s\n"
            " - use_indices      : %s\n"
            " - max_queue_size   : %d\n",
            (approximate_sync_) ? "true" : "false",
            (use_indices_) ? "true" : "false", max_queue_size_);

//        child_init (private_nh);
      }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Child initialization routine (virtual).
        * \param nh ROS node handle
        */
//      virtual bool child_init (ros::NodeHandle &nh) { return (true); }

      // Need to make this public, as nodelet::Nodelet::getName is protected
      //inline const std::string& getName () const { std::cerr << "name: " << nodelet::Nodelet::getName () << std::endl; return (nodelet::Nodelet::getName ()); }

    protected:
      tf::TransformListener tf_listener_;
  };
}

#endif  //#ifndef PCL_NODELET_H_
