#ifndef PCL_ROS_POINT_CLOUD_H_
#define PCL_ROS_POINT_CLOUD_H_

#include <pcl/point_cloud.h>
#include <pcl/ros/point_traits.h>
#include <pcl/ros/for_each_type.h>
#include <pcl/ros/conversions.h>
#include <sensor_msgs/PointCloud2.h>

namespace pcl {
namespace detail {

template<typename Stream, typename PointT>
struct FieldStreamer
{
  FieldStreamer(Stream& stream) : stream_(stream) {}

  template<typename U> void operator() ()
  {
    const char* name = traits::name<PointT, U>::value;
    uint32_t name_length = strlen(name);
    stream.next(name_length);
    if (name_length > 0)
      memcpy(stream.advance(name_length), name, name_length);

    uint32_t offset = traits::offset<PointT, U>::value;
    stream.next(offset);

    uint8_t datatype = traits::datatype<PointT, U>::value;
    stream.next(datatype);

    uint32_t count = traits::datatype<PointT, U>::size;
    stream.next(count);
  }

  Stream& stream_;
};

template<typename PointT>
struct FieldsLength
{
  FieldsLength() : length(0) {}

  template<typename U> void operator() ()
  {
    uint32_t name_length = strlen(traits::name::<PointT, U>::value);
    length += name_length + 13;
  }

  uint32_t length;
};

template<typename PointCloudT>
boost::shared_ptr<PointCloudT> createMessage(boost::shared_ptr<MsgFieldMap> mapping)
{
  return boost::make_shared<PointCloudT>(mapping);
}

} // namespace pcl::detail

// Arbitrary boost::function object
template<typename PointCloudT>
ros::Subscriber subscribe(ros::NodeHandle &nh, const std::string &topic, uint32_t queue_size,
                          const boost::function< void(const boost::shared_ptr<const PointCloudT>&)> &callback,
                          const ros::VoidConstPtr &tracked_object = ros::VoidConstPtr(),
                          const ros::TransportHints &transport_hints = ros::TransportHints())
{
  ros::SubscribeOptions ops;
  ops.template init<PointCloudT>(topic, queue_size, callback, detail::createMessage<PointCloudT>);
  ops.tracked_object = tracked_object;
  ops.transport_hints = transport_hints;
  return nh.subscribe(ops);
}

} // namespace pcl

namespace ros {

namespace message_traits {

template<typename T> struct MD5Sum<pcl::PointCloud<T> > : MD5Sum<sensor_msgs::PointCloud2>
{
  // If the definition of sensor_msgs/PointCloud2 changes, we'll get a compile error here.
  ROS_STATIC_ASSERT(MD5Sum<sensor_msgs::PointCloud2>::static_value1 == 0x1158d486dd51d683ULL);
  ROS_STATIC_ASSERT(MD5Sum<sensor_msgs::PointCloud2>::static_value2 == 0xce2f1be655c3c181ULL);
};

template<typename T> struct DataType<pcl::PointCloud<T> > : DataType<sensor_msgs::PointCloud2> {};

template<typename T> struct Definition<pcl::PointCloud<T> > : Definition<sensor_msgs::PointCloud2> {};

template<typename T> struct HasHeader<pcl::PointCloud<T> > : TrueType {};

} // namespace ros::message_traits

namespace serialization {

template<typename T>
struct Serializer<pcl::PointCloud<T> >
{
  template<typename Stream>
  inline static void write(Stream& stream, const pcl::PointCloud<T>& m)
  {
    stream.next(m.header);
    
    // Ease the user's burden on specifying width/height for unorganized datasets
    uint32_t height = m.height, width = m.width;
    if (height == 0 && width == 0) {
      width = m.points.size();
      height = 1;
    }
    stream.next(height);
    stream.next(width);

    // Stream out point field metadata
    typedef typename pcl::traits::fieldList<T>::type FieldList;
    uint32_t fields_size = boost::mpl::size<FieldList>::type;
    stream.next(fields_size);
    pcl::for_each_type<FieldList>(pcl::detail::FieldStreamer<Stream, T>(stream));

    // Assume little-endian...
    uint8_t is_bigendian = false;
    stream.next(is_bigendian);

    // Write out point data as binary blob
    uint32_t point_step = sizeof(T);
    stream.next(point_step);
    uint32_t row_step = point_step * width;
    stream.next(row_step);
    uint32_t data_size = row_step * height;
    stream.next(data_size);
    memcpy(stream.advance(data_size), &m.points[0], data_size);

    uint8_t is_dense = m.is_dense;
    stream.next(is_dense);
  }

  template<typename Stream>
  inline static void read(Stream& stream, pcl::PointCloud<T>& m)
  {
    stream.next(m.header);
    stream.next(m.height);
    stream.next(m.width);

    // Construct field mapping if deserializing for the first time
    /// @todo Check that fields haven't changed!
    if (!m.mapping_) {
      ROS_WARN("It looks like you subscribed to a topic with type PointCloud<T> using ros::NodeHandle::subscribe(). "
               "For faster deserialization, use the pcl::subscribe() methods instead.");
      m.mapping_ = boost::make_shared<pcl::MsgFieldMap>();
    }
    pcl::MsgFieldMap& mapping = *m.mapping_;
    if (mapping.empty()) {
      std::vector<sensor_msgs::PointField> fields;
      stream.next(fields);
      pcl::detail::FieldMapper<T> mapper(fields, mapping);
      pcl::for_each_type< typename pcl::traits::fieldList<T>::type > (mapper);
    }

    uint8_t is_bigendian;
    stream.next(is_bigendian); // ignoring...
    uint32_t point_step, row_step;
    stream.next(point_step);
    stream.next(row_step);

    // Copy point data
    uint32_t num_points = m.width * m.height;
    m.points.resize(num_points);
    uint8_t* m_data = reinterpret_cast<uint8_t*>(&m.points[0]);
    for (uint32_t row = 0; row < m.height; ++row) {
      const uint8_t* stream_data = stream.advance(row_step);
      for (uint32_t col = 0; col < m.width; ++col, stream_data += point_step) {
        BOOST_FOREACH(const pcl::detail::FieldMapping& fm, mapping) {
          memcpy(m_data + fm.struct_offset, stream_data + fm.serialized_offset, fm.size);
        }
        m_data += sizeof(T);
      }
    }

    uint8_t is_dense;
    stream.next(is_dense);
    m.is_dense = is_dense;
  }

  inline static uint32_t serializedLength(const pcl::PointCloud<T>& m)
  {
    uint32_t length = 0;

    length += serializationLength(m.header);
    length += 8; // height/width

    pcl::detail::FieldsLength<T> fl;
    typedef typename pcl::traits::fieldList<T>::type FieldList;
    pcl::for_each_type<FieldList>(fl);
    length += 4; // size of 'fields'
    length += fl.length;

    length += 1; // is_bigendian
    length += 4; // point_step
    length += 4; // row_step
    length += 4; // size of 'data'
    length += m.points.size() * sizeof(T); // data
    length += 1; // is_dense
  }
};

} // namespace ros::serialization

/// @todo Printer specialization in message_operations

} // namespace ros

#endif
