/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: extract_clusters.h 32052 2010-08-27 02:19:30Z rusu $
 *
 */

#ifndef PCL_ROS_EXTRACT_CLUSTERS_H_
#define PCL_ROS_EXTRACT_CLUSTERS_H_

#include <pcl/segmentation/extract_clusters.h>
#include "pcl_ros/pcl_nodelet.h"

// Dynamic reconfigure
#include <dynamic_reconfigure/server.h>
#include "pcl_ros/EuclideanClusterExtractionConfig.h"

namespace pcl_ros
{
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief @b EuclideanClusterExtraction represents a segmentation class for cluster extraction in an Euclidean sense.
    * \author Radu Bogdan Rusu
    */
  template <typename PointT>
  class EuclideanClusterExtraction : virtual public pcl::EuclideanClusterExtraction<PointT>, virtual public PCLNodelet<PointT>
  {
    typedef pcl::EuclideanClusterExtraction<PointT> BaseClass;

    typedef sensor_msgs::PointCloud2 PointCloud2;

    typedef typename BaseClass::PointCloud PointCloud;
    typedef typename PointCloud::Ptr PointCloudPtr;
    typedef typename PointCloud::ConstPtr PointCloudConstPtr;

    public:
      using PCLNodelet<PointT>::getName;
      using PCLNodelet<PointT>::isValid;
      using PCLNodelet<PointT>::pnh_;
      using PCLNodelet<PointT>::max_queue_size_;
      using PCLNodelet<PointT>::sub_input_filter_;
      using PCLNodelet<PointT>::sub_indices_filter_;
      using PCLNodelet<PointT>::sync_input_indices_a_;
      using PCLNodelet<PointT>::sync_input_indices_e_;
      using PCLNodelet<PointT>::use_indices_;
      using PCLNodelet<PointT>::input_;
      using PCLNodelet<PointT>::indices_;
      using PCLNodelet<PointT>::sub_input_;
      using PCLNodelet<PointT>::pub_output_;
      using PCLNodelet<PointT>::approximate_sync_;

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Empty constructor. */
      EuclideanClusterExtraction () : publish_indices_ (false), max_clusters_ (std::numeric_limits<int>::max ()) {};
                                      
    protected:
      // Members derived from the base class
      using BaseClass::min_pts_per_cluster_;
      using BaseClass::max_pts_per_cluster_;
      using BaseClass::cluster_tolerance_;
      using BaseClass::spatial_locator_;
      using BaseClass::extract;

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      // ROS nodelet attributes

      /** \brief Publish indices or convert to PointCloud clusters. Default: false */
      bool publish_indices_;

      /** \brief Maximum number of clusters to publish. */
      int max_clusters_;

      /** \brief Pointer to a dynamic reconfigure service. */
      boost::shared_ptr<dynamic_reconfigure::Server<EuclideanClusterExtractionConfig> > srv_;

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Nodelet initialization routine. */
      void onInit ();

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Perform the segmentation and publish the results. Internal method. */
      void segmentAndPublish ();

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Dynamic reconfigure callback
        * \param config the config object
        * \param level the dynamic reconfigure level
        */
      void config_callback (EuclideanClusterExtractionConfig &config, uint32_t level);

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Input point cloud callback. Used only when \a use_indices is not set.
        * \param cloud the pointer to the input point cloud
        */
      void input_callback (const PointCloudConstPtr &cloud);

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Input point cloud callback. Used when \a use_indices is set.
        * \param cloud the pointer to the input point cloud
        * \param indices the pointer to the input point cloud indices
        */
      void input_indices_callback (const PointCloud2::ConstPtr &cloud, const pcl::PointIndicesConstPtr &indices);
  };
}

#include "pcl_ros/segmentation/extract_clusters.hpp"

#endif  //#ifndef PCL_ROS_EXTRACT_CLUSTERS_H_
