/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: sac_segmentation.h 32052 2010-08-27 02:19:30Z rusu $
 *
 */

#ifndef PCL_ROS_SAC_SEGMENTATION_H_
#define PCL_ROS_SAC_SEGMENTATION_H_

#include "pcl_ros/pcl_nodelet.h"

// PCL includes
#include <pcl/segmentation/sac_segmentation.h>

// Dynamic reconfigure
#include <dynamic_reconfigure/server.h>
#include "pcl_ros/SACSegmentationConfig.h"
#include "pcl_ros/SACSegmentationFromNormalsConfig.h"

namespace pcl_ros
{
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief @b SACSegmentation represents the Nodelet segmentation class for Sample Consensus methods and models, in
    * the sense that it just creates a Nodelet wrapper for generic-purpose SAC-based segmentation.
    * \author Radu Bogdan Rusu
    */
  template <typename PointT>
  class SACSegmentation : virtual public pcl::SACSegmentation<PointT>, virtual public PCLNodelet<PointT>
  {
    typedef pcl::SACSegmentation<PointT> BaseClass;

    typedef sensor_msgs::PointCloud2 PointCloud2;

    typedef typename BaseClass::PointCloud PointCloud;
    typedef typename PointCloud::Ptr PointCloudPtr;
    typedef typename PointCloud::ConstPtr PointCloudConstPtr;

    public:
      // Members derived from the base class
      using PCLNodelet<PointT>::getName;
      using PCLNodelet<PointT>::isValid;

      using pcl::SACSegmentation<PointT>::input_;
      using pcl::SACSegmentation<PointT>::indices_;
      using pcl::SACSegmentation<PointT>::use_indices_;
      using pcl::SACSegmentation<PointT>::getRadiusLimits;
      using pcl::SACSegmentation<PointT>::setRadiusLimits;
      using pcl::SACSegmentation<PointT>::getProbability;
      using pcl::SACSegmentation<PointT>::setProbability;
      using pcl::SACSegmentation<PointT>::getMaxIterations;
      using pcl::SACSegmentation<PointT>::setMaxIterations;
      using pcl::SACSegmentation<PointT>::getOptimizeCoefficients;
      using pcl::SACSegmentation<PointT>::setOptimizeCoefficients;
      using pcl::SACSegmentation<PointT>::getDistanceThreshold;
      using pcl::SACSegmentation<PointT>::setDistanceThreshold;

      using BaseClass::model_type_;
      using BaseClass::threshold_;
      using BaseClass::optimize_coefficients_;
      using BaseClass::method_type_;
      using BaseClass::max_iterations_;
      using BaseClass::eps_angle_;
      using BaseClass::axis_;
      using BaseClass::probability_;
      using BaseClass::segment;
      using PCLNodelet<PointT>::pnh_;
      using PCLNodelet<PointT>::sync_input_indices_a_;
      using PCLNodelet<PointT>::sync_input_indices_e_;
      using PCLNodelet<PointT>::approximate_sync_;
      using PCLNodelet<PointT>::sub_input_;
      using PCLNodelet<PointT>::pub_output_;
      using PCLNodelet<PointT>::sub_input_filter_;
      using PCLNodelet<PointT>::sub_indices_filter_;
      using PCLNodelet<PointT>::max_queue_size_;

    protected:
      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Child initialization routine.
        * \param nh ROS node handle
        * \param has_service set to true if the child has a Dynamic Reconfigure service
        */
      virtual bool child_init (ros::NodeHandle &nh, bool &has_service);

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      // ROS nodelet attributes
      /** \brief The output PointIndices publisher. */
      ros::Publisher pub_indices_;

      /** \brief The output ModelCoefficients publisher. */
      ros::Publisher pub_model_;

      /** \brief Pointer to a dynamic reconfigure service. */
      boost::shared_ptr<dynamic_reconfigure::Server<SACSegmentationConfig> > srv_;

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Nodelet initialization routine. */
      void onInit ();

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Dynamic reconfigure callback
        * \param config the config object
        * \param level the dynamic reconfigure level
        */
      void config_callback (SACSegmentationConfig &config, uint32_t level);

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Input point cloud callback. Used only when \a use_indices is not set.
        * \param cloud the pointer to the input point cloud
        */
      void input_callback (const PointCloudConstPtr &cloud);

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Input point cloud callback. Used when \a use_indices is set.
        * \param cloud the pointer to the input point cloud
        * \param indices the pointer to the input point cloud indices
        */
      void input_indices_callback (const PointCloud2::ConstPtr &cloud, const pcl::PointIndicesConstPtr &indices);
  };

  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** \brief @b SACSegmentationFromNormals represents the PCL nodelet segmentation class for Sample Consensus methods and
    * models that require the use of surface normals for estimation.
    */
  template <typename PointT, typename PointNT>
  class SACSegmentationFromNormals: virtual public SACSegmentation<PointT>, virtual public pcl::SACSegmentationFromNormals<PointT, PointNT>
  {
    typedef pcl::SACSegmentationFromNormals<PointT, PointNT> BaseClass;
    
    typedef sensor_msgs::PointCloud2 PointCloud2;

    typedef typename BaseClass::PointCloud PointCloud;
    typedef typename PointCloud::Ptr PointCloudPtr;
    typedef typename PointCloud::ConstPtr PointCloudConstPtr;

    typedef typename BaseClass::PointCloudN PointCloudN;
    typedef typename PointCloudN::Ptr PointCloudNPtr;
    typedef typename PointCloudN::ConstPtr PointCloudNConstPtr;

    public:
      // Members derived from the base class
      using SACSegmentation<PointT>::getName;
      using SACSegmentation<PointT>::isValid;
      using SACSegmentation<PointT>::pub_indices_;
      using SACSegmentation<PointT>::pub_model_;
      using SACSegmentation<PointT>::pnh_;
      using SACSegmentation<PointT>::approximate_sync_;
      using SACSegmentation<PointT>::sub_input_filter_;
      using SACSegmentation<PointT>::max_queue_size_;
      using SACSegmentation<PointT>::sub_indices_filter_;

      using BaseClass::input_;
      using BaseClass::distance_weight_;
      using BaseClass::normals_;
      using BaseClass::indices_;
      using BaseClass::use_indices_;
      using BaseClass::getRadiusLimits;
      using BaseClass::setRadiusLimits;
      using BaseClass::getMaxIterations;
      using BaseClass::setMaxIterations;
      using BaseClass::getOptimizeCoefficients;
      using BaseClass::setOptimizeCoefficients;
      using BaseClass::getDistanceThreshold;
      using BaseClass::setDistanceThreshold;
      using BaseClass::getProbability;
      using BaseClass::setProbability;
      using BaseClass::axis_;
      using BaseClass::segment;

    protected:
      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      // ROS nodelet attributes
      /** \brief The normals PointCloud subscriber filter. */
      message_filters::Subscriber<PointCloud2> sub_normals_filter_;

      /** \brief The input PointCloud subscriber. */
      ros::Subscriber sub_axis_;

      /** \brief Synchronized input, and normals.*/
      boost::shared_ptr <message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, PointCloud2> > > sync_input_normals_a_;
      boost::shared_ptr <message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, PointCloud2> > > sync_input_normals_e_;

      /** \brief Synchronized input, normals, and indices.*/
      boost::shared_ptr<message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, PointCloud2, pcl::PointIndices> > > sync_input_normals_indices_a_;
      boost::shared_ptr<message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, PointCloud2, pcl::PointIndices> > > sync_input_normals_indices_e_;

      /** \brief Pointer to a dynamic reconfigure service. */
      boost::shared_ptr<dynamic_reconfigure::Server<SACSegmentationFromNormalsConfig> > srv_;

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Child initialization routine.
        * \param nh ROS node handle
        * \param has_service set to true if the child has a Dynamic Reconfigure service
        */
      virtual bool child_init (ros::NodeHandle &nh, bool &has_service);

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Dynamic reconfigure callback
        * \param config the config object
        * \param level the dynamic reconfigure level
        */
      void axis_callback (const pcl::ModelCoefficientsConstPtr &model);

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Dynamic reconfigure callback
        * \param config the config object
        * \param level the dynamic reconfigure level
        */
      void config_callback (SACSegmentationFromNormalsConfig &config, uint32_t level);

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      /** \brief Input point cloud callback. Used only when \a use_indices and \a use_surface are not set.
        * \param cloud the pointer to the input point cloud
        * \param cloud_normals the pointer to the input point cloud normals
        */
      void input_normals_callback (const PointCloud2::ConstPtr &cloud, const PointCloud2::ConstPtr &cloud_normals);
  };
}

#include "pcl_ros/segmentation/sac_segmentation.hpp"

#endif  //#ifndef PCL_ROS_SAC_SEGMENTATION_H_
