/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: feature.hpp 32052 2010-08-27 02:19:30Z rusu $
 *
 */

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Compute the feature and publish it. Internal method. */
template <typename PointInT, typename PointOutT> void
  pcl_ros::Feature<PointInT, PointOutT>::computeAndPublish ()
{
  // Initialize the spatial locator
  initTree (spatial_locator_type_, tree_, k_);

  // Estimate the feature
  PointCloudOut output;
  compute (output);
  if (output.points.size () == 0)
  {
    NODELET_WARN ("[computeAndPublish] Output PointCloud has no data points on topic %s (parameters might not be set correctly)!", pnh_->resolveName ("output").c_str ());
    return;
  }

  // Convert from the templated output to the PointCloud blob
  PointCloud2 output_blob;
  pcl::toROSMsg (output, output_blob);
  // Publish a Boost shared ptr const data
  pub_output_.publish (boost::make_shared <const PointCloud2> (output_blob));
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Nodelet initialization routine. */
template <typename PointInT, typename PointOutT> void
  pcl_ros::Feature<PointInT, PointOutT>::onInit ()
{
  // Call the super onInit ()
  PCLNodelet<PointInT>::onInit ();

  pub_output_ = pnh_->template advertise<PointCloud2> ("output", max_queue_size_);

  // ---[ Mandatory parameters
  if (!pnh_->getParam ("k_search", k_) && !pnh_->getParam ("radius_search", search_radius_))
  {
    NODELET_ERROR ("[onInit] Neither 'k_search' nor 'radius_search' set! Need to set at least one of these parameters before continuing.");
    return;
  }
  if (!pnh_->getParam ("spatial_locator", spatial_locator_type_))
  {
    NODELET_ERROR ("[onInit] Need a 'spatial_locator' parameter to be set before continuing!");
    return;
  }

  // ---[ Optional parameters
  pnh_->getParam ("use_surface", use_surface_);

  // Enable the dynamic reconfigure service
  srv_ = boost::shared_ptr<dynamic_reconfigure::Server<FeatureConfig> >(new dynamic_reconfigure::Server<FeatureConfig> (*pnh_));
//        boost::shared_ptr <dynamic_reconfigure::Server<FeatureConfig> > srv_;
//        srv_ = boost::make_shared <dynamic_reconfigure::Server<FeatureConfig> >(*pnh_);
//        dynamic_reconfigure::Server<FeatureConfig>::CallbackType f =  boost::bind (&Feature<PointInT, PointOutT>::config_callback, this, _1, _2);
//        srv_->setCallback (f);
//        dynamic_reconfigure::Server<FeatureConfig> srv_ (*pnh_);
  dynamic_reconfigure::Server<FeatureConfig>::CallbackType f =  boost::bind (&Feature<PointInT, PointOutT>::config_callback, this, _1, _2);
  srv_->setCallback (f);

  // If we're supposed to look for PointIndices (indices) or PointCloud (surface) messages
  if (use_indices_ || use_surface_)
  {
    // Subscribe to the input using a filter
    sub_input_filter_.subscribe (*pnh_, "input", max_queue_size_);
    if (use_indices_)
    {
      // If indices are enabled, subscribe to the indices
      sub_indices_filter_.subscribe (*pnh_, "indices", max_queue_size_);
      if (use_surface_)     // Use both indices and surface
      {
        // If surface is enabled, subscribe to the surface, connect the input-indices-surface trio and register
        sub_surface_filter_.subscribe (*pnh_, "surface", max_queue_size_);
        if (approximate_sync_)
        {
          sync_input_surface_indices_a_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, PointCloud2, pcl::PointIndices> > >(max_queue_size_);
          sync_input_surface_indices_a_->connectInput (sub_input_filter_, sub_surface_filter_, sub_indices_filter_);
          sync_input_surface_indices_a_->registerCallback (bind (&Feature<PointInT, PointOutT>::input_surface_indices_callback, this, _1, _2, _3));
        }
        else
        {
          sync_input_surface_indices_e_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, PointCloud2, pcl::PointIndices> > >(max_queue_size_);
          sync_input_surface_indices_e_->connectInput (sub_input_filter_, sub_surface_filter_, sub_indices_filter_);
          sync_input_surface_indices_e_->registerCallback (bind (&Feature<PointInT, PointOutT>::input_surface_indices_callback, this, _1, _2, _3));
        }
      }
      else                  // Use only indices
      {
        // surface not enabled, connect the input-indices duo and register
        if (approximate_sync_)
        {
          sync_input_indices_a_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, pcl::PointIndices> > >(max_queue_size_);
          sync_input_indices_a_->connectInput (sub_input_filter_, sub_indices_filter_);
          sync_input_indices_a_->registerCallback (bind (&Feature<PointInT, PointOutT>::input_indices_callback, this, _1, _2));
        }
        else
        {
          sync_input_indices_e_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, pcl::PointIndices> > >(max_queue_size_);
          sync_input_indices_e_->connectInput (sub_input_filter_, sub_indices_filter_);
          sync_input_indices_e_->registerCallback (bind (&Feature<PointInT, PointOutT>::input_indices_callback, this, _1, _2));
        }
      }
    }
    else                    // Use only surface
    {
      // indices not enabled, connect the input-surface duo and register
      sub_surface_filter_.subscribe (*pnh_, "surface", max_queue_size_);
      if (approximate_sync_)
      {
        sync_input_surface_a_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, PointCloud2> > >(max_queue_size_);
        sync_input_surface_a_->connectInput (sub_input_filter_, sub_surface_filter_);
        sync_input_surface_a_->registerCallback (bind (&Feature<PointInT, PointOutT>::input_surface_callback, this, _1, _2));
      }
      else
      {
        sync_input_surface_e_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, PointCloud2> > >(max_queue_size_);
        sync_input_surface_e_->connectInput (sub_input_filter_, sub_surface_filter_);
        sync_input_surface_e_->registerCallback (bind (&Feature<PointInT, PointOutT>::input_surface_callback, this, _1, _2));
      }
    }
  }
  else
    // Subscribe in an old fashion to input only (no filters)
    sub_input_.subscribe (*pnh_, "input", max_queue_size_,  bind (&Feature<PointInT, PointOutT>::input_callback, this, _1));

  NODELET_DEBUG ("[onInit] Nodelet successfully created with the following parameters:\n"
                 " - use_surface    : %s\n"
                 " - k_search       : %d\n"
                 " - radius_search  : %f\n"
                 " - spatial_locator: %d",
                (use_surface_) ? "true" : "false", k_, search_radius_, spatial_locator_type_);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Dynamic reconfigure callback
  * \param config the config object
  * \param level the dynamic reconfigure level
  */
template <typename PointInT, typename PointOutT> void
  pcl_ros::Feature<PointInT, PointOutT>::config_callback (FeatureConfig &config, uint32_t level)
{
  if (k_ != config.k_search)
  {
    k_ = config.k_search;
    NODELET_DEBUG ("[config_callback] Setting the number of K nearest neighbors to use for each point: %d.", k_);
  }
  if (search_radius_ != config.radius_search)
  {
    search_radius_ = config.radius_search;
    NODELET_DEBUG ("[config_callback] Setting the nearest neighbors search radius for each point: %f.", search_radius_);
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Input point cloud callback. Used when \a use_indices and \a use_surface are set.
  * \param cloud the pointer to the input point cloud
  * \param cloud_surface the pointer to the surface point cloud
  * \param indices the pointer to the input point cloud indices
  */
template <typename PointInT, typename PointOutT> void
  pcl_ros::Feature<PointInT, PointOutT>::input_surface_indices_callback (
      const PointCloud2::ConstPtr &cloud, const PointCloud2::ConstPtr &cloud_surface, const pcl::PointIndicesConstPtr &indices)
{
  if (pub_output_.getNumSubscribers () <= 0)
    return;

  if (!isValid (cloud) || !isValid (cloud_surface, "surface") || !isValid (indices))
    return;

  NODELET_DEBUG ("[input_surface_indices_callback]\n"
                 "                                         - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.\n"
                 "                                         - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.\n"
                 "                                         - PointIndices with %zu values, stamp %f, and frame %s on topic %s received.",
             cloud->width * cloud->height, pcl::getFieldsList (*cloud).c_str (), cloud->header.stamp.toSec (), cloud->header.frame_id.c_str (), pnh_->resolveName ("input").c_str (),
             cloud_surface->width * cloud_surface->height, pcl::getFieldsList (*cloud_surface).c_str (), cloud_surface->header.stamp.toSec (), cloud_surface->header.frame_id.c_str (), pnh_->resolveName ("surface").c_str (),
             indices->indices.size (), indices->header.stamp.toSec (), indices->header.frame_id.c_str (), pnh_->resolveName ("indices").c_str ());

  if ((int)(cloud->width * cloud->height) < k_)
  {
    NODELET_ERROR ("[input_surface_indices_callback] Requested number of k-nearest neighbors (%d) is larger than the PointCloud size (%d)!", k_, (int)(cloud->width * cloud->height));
    return;
  }

  PointCloudIn input;
  pcl::fromROSMsg (*cloud, input);                  // input_ = cloud;
  input_ = boost::make_shared <const PointCloudIn> (input);

  PointCloudIn surface;
  pcl::fromROSMsg (*cloud_surface, surface);        // surface_ = cloud_surf;
  surface_ = boost::make_shared <const PointCloudIn> (surface);

  indices_ = boost::make_shared <const std::vector<int> > (indices->indices);
  computeAndPublish ();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Input point cloud callback. Used when \a use_indices is set.
  * \param cloud the pointer to the input point cloud
  * \param indices the pointer to the input point cloud indices
  */
template <typename PointInT, typename PointOutT> void
  pcl_ros::Feature<PointInT, PointOutT>::input_indices_callback (
      const PointCloud2::ConstPtr &cloud, const pcl::PointIndicesConstPtr &indices)
{
  if (pub_output_.getNumSubscribers () <= 0)
    return;

  if (!isValid (cloud) || !isValid (indices))
    return;

  NODELET_DEBUG ("[input_indices_callback]\n"
                 "                                 - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.\n"
                 "                                 - PointIndices with %zu values, stamp %f, and frame %s on topic %s received.",
                 cloud->width * cloud->height, pcl::getFieldsList (*cloud).c_str (), cloud->header.stamp.toSec (), cloud->header.frame_id.c_str (), pnh_->resolveName ("input").c_str (),
                 indices->indices.size (), indices->header.stamp.toSec (), indices->header.frame_id.c_str (), pnh_->resolveName ("indices").c_str ());

  if ((int)(cloud->width * cloud->height) < k_)
  {
    NODELET_ERROR ("[input_indices_callback] Requested number of k-nearest neighbors (%d) is larger than the PointCloud size (%d)!", k_, (int)(cloud->width * cloud->height));
    return;
  }

  PointCloudIn input;
  pcl::fromROSMsg (*cloud, input);            // input_ = cloud;
  input_   = boost::make_shared <const PointCloudIn> (input);

  indices_ = boost::make_shared <const std::vector<int> > (indices->indices);

  computeAndPublish ();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Input point cloud callback. Used when \a use_surface is set.
  * \param cloud the pointer to the input point cloud
  * \param cloud_surface the pointer to the surface point cloud
  */
template <typename PointInT, typename PointOutT> void
  pcl_ros::Feature<PointInT, PointOutT>::input_surface_callback (
      const PointCloud2::ConstPtr &cloud, const PointCloud2::ConstPtr &cloud_surface)
{
  if (pub_output_.getNumSubscribers () <= 0)
    return;

  if (!isValid (cloud) || !isValid (cloud_surface, "surface"))
    return;

  NODELET_DEBUG ("[input_surface_callback]\n"
                 "                                 - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.\n"
                 "                                 - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.",
                 cloud->width * cloud->height, pcl::getFieldsList (*cloud).c_str (), cloud->header.stamp.toSec (), cloud->header.frame_id.c_str (), pnh_->resolveName ("input").c_str (),
                 cloud_surface->width * cloud_surface->height, pcl::getFieldsList (*cloud_surface).c_str (), cloud_surface->header.stamp.toSec (), cloud_surface->header.frame_id.c_str (), pnh_->resolveName ("surface").c_str ());

  if ((int)(cloud->width * cloud->height) < k_)
  {
    NODELET_ERROR ("[input_surface_callback] Requested number of k-nearest neighbors (%d) is larger than the PointCloud size (%d)!", k_, (int)(cloud->width * cloud->height));
    return;
  }

  PointCloudIn input;
  pcl::fromROSMsg (*cloud, input);            // input_ = cloud;
  input_ = boost::make_shared <const PointCloudIn> (input);

  PointCloudIn surface;
  pcl::fromROSMsg (*cloud_surface, surface);         // surface_ = cloud_surf;
  surface_ = boost::make_shared <const PointCloudIn> (surface);

  // Reset the indices pointer
  indices_.reset ();

  computeAndPublish ();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Input point cloud callback. Used only when \a use_indices and \a use_surface are not set.
  * \param cloud the pointer to the input point cloud
  */
template <typename PointInT, typename PointOutT> void
  pcl_ros::Feature<PointInT, PointOutT>::input_callback (const PointCloudInConstPtr &cloud)
{
  if (pub_output_.getNumSubscribers () <= 0)
    return;

  if (!isValid (cloud))
    return;

  NODELET_DEBUG ("[input_callback] PointCloud with %d data points, stamp %f, and frame %s on topic %s received.", cloud->width * cloud->height, cloud->header.stamp.toSec (), cloud->header.frame_id.c_str (), pnh_->resolveName ("input").c_str ());

  if ((int)(cloud->width * cloud->height) < k_)
  {
    NODELET_ERROR ("[input_callback] Requested number of k-nearest neighbors (%d) is larger than the PointCloud size (%d)!", k_, (int)(cloud->width * cloud->height));
    return;
  }

  // Reset the indices and surface pointers
  indices_.reset ();
  surface_.reset ();

  input_ = cloud;

  computeAndPublish ();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Nodelet initialization routine. */
template <typename PointInT, typename PointNT, typename PointOutT> void
  pcl_ros::FeatureFromNormals<PointInT, PointNT, PointOutT>::onInit ()
{
  // Call the super onInit ()
  PCLNodelet<PointInT>::onInit ();

  pub_output_ = pnh_->template advertise<PointCloud2> ("output", max_queue_size_);

  // ---[ Mandatory parameters
  if (!pnh_->getParam ("k_search", k_) && !pnh_->getParam ("radius_search", search_radius_))
  {
    NODELET_ERROR ("[onInit] Neither 'k_search' nor 'radius_search' set! Need to set at least one of these parameters before continuing.");
    return;
  }
  if (!pnh_->getParam ("spatial_locator", spatial_locator_type_))
  {
    NODELET_ERROR ("[onInit] Need a 'spatial_locator' parameter to be set before continuing!");
    return;
  }
  // ---[ Optional parameters
  pnh_->getParam ("use_surface", use_surface_);

  sub_input_filter_.subscribe (*pnh_, "input", max_queue_size_);
  sub_normals_filter_.subscribe (*pnh_, "normals", max_queue_size_);

  // If we're supposed to look for PointIndices (indices) or PointCloud (surface) messages
  if (use_indices_ || use_surface_)
  {
    if (use_indices_)
    {
      // If indices are enabled, subscribe to the indices
      sub_indices_filter_.subscribe (*pnh_, "indices", max_queue_size_);
      if (use_surface_)     // Use both indices and surface
      {
        // If surface is enabled, subscribe to the surface, connect the input-indices-surface trio and register
        sub_surface_filter_.subscribe (*pnh_, "surface", max_queue_size_);
        if (approximate_sync_)
        {
          sync_input_normals_surface_indices_a_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, PointCloud2, PointCloud2, pcl::PointIndices> > > (max_queue_size_);
          sync_input_normals_surface_indices_a_->connectInput (sub_input_filter_, sub_normals_filter_, sub_surface_filter_, sub_indices_filter_);
          sync_input_normals_surface_indices_a_->registerCallback (bind (&FeatureFromNormals<PointInT, PointNT, PointOutT>::input_normals_surface_indices_callback, this, _1, _2, _3, _4));
        }
        else
        {
          sync_input_normals_surface_indices_e_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, PointCloud2, PointCloud2, pcl::PointIndices> > > (max_queue_size_);
          sync_input_normals_surface_indices_e_->connectInput (sub_input_filter_, sub_normals_filter_, sub_surface_filter_, sub_indices_filter_);
          sync_input_normals_surface_indices_e_->registerCallback (bind (&FeatureFromNormals<PointInT, PointNT, PointOutT>::input_normals_surface_indices_callback, this, _1, _2, _3, _4));
        }
      }
      else                  // Use only indices
      {
        if (approximate_sync_)
        {
          sync_input_normals_indices_a_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, PointCloud2, pcl::PointIndices> > > (max_queue_size_);
          // surface not enabled, connect the input-indices duo and register
          sync_input_normals_indices_a_->connectInput (sub_input_filter_, sub_normals_filter_, sub_indices_filter_);
          sync_input_normals_indices_a_->registerCallback (bind (&FeatureFromNormals<PointInT, PointNT, PointOutT>::input_normals_indices_callback, this, _1, _2, _3));
        }
        else
        {
          sync_input_normals_indices_e_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, PointCloud2, pcl::PointIndices> > > (max_queue_size_);
          // surface not enabled, connect the input-indices duo and register
          sync_input_normals_indices_e_->connectInput (sub_input_filter_, sub_normals_filter_, sub_indices_filter_);
          sync_input_normals_indices_e_->registerCallback (bind (&FeatureFromNormals<PointInT, PointNT, PointOutT>::input_normals_indices_callback, this, _1, _2, _3));
        }
      }
    }
    else                    // Use only surface
    {
      // indices not enabled, connect the input-surface duo and register
      sub_surface_filter_.subscribe (*pnh_, "surface", max_queue_size_);

      if (approximate_sync_)
      {
        sync_input_normals_surface_a_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, PointCloud2, PointCloud2> > > (max_queue_size_);
        sync_input_normals_surface_a_->connectInput (sub_input_filter_, sub_normals_filter_, sub_surface_filter_);
        sync_input_normals_surface_a_->registerCallback (bind (&FeatureFromNormals<PointInT, PointNT, PointOutT>::input_normals_surface_callback, this, _1, _2, _3));
      }
      else
      {
        sync_input_normals_surface_e_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, PointCloud2, PointCloud2> > > (max_queue_size_);
        sync_input_normals_surface_e_->connectInput (sub_input_filter_, sub_normals_filter_, sub_surface_filter_);
        sync_input_normals_surface_e_->registerCallback (bind (&FeatureFromNormals<PointInT, PointNT, PointOutT>::input_normals_surface_callback, this, _1, _2, _3));
      }
    }
  }
  else
  {
    if (approximate_sync_)
    {
      sync_input_normals_a_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, PointCloud2> > > (max_queue_size_);
      sync_input_normals_a_->connectInput (sub_input_filter_, sub_normals_filter_);
      sync_input_normals_a_->registerCallback (bind (&FeatureFromNormals<PointInT, PointNT, PointOutT>::input_normals_callback, this, _1, _2));
    }
    else
    {
      sync_input_normals_e_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, PointCloud2> > > (max_queue_size_);
      sync_input_normals_e_->connectInput (sub_input_filter_, sub_normals_filter_);
      sync_input_normals_e_->registerCallback (bind (&FeatureFromNormals<PointInT, PointNT, PointOutT>::input_normals_callback, this, _1, _2));
    }
  }

  NODELET_DEBUG ("[onInit] Nodelet successfully created with the following parameters:\n"
                 " - use_surface    : %s\n"
                 " - k_search       : %d\n"
                 " - radius_search  : %f\n"
                 " - spatial_locator: %d",
                 (use_surface_) ? "true" : "false", k_, search_radius_, Feature<PointInT, PointOutT>::spatial_locator_type_);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Input point cloud callback. Used when \a use_indices and \a use_surface are set.
  * \param cloud the pointer to the input point cloud
  * \param cloud_normals the pointer to the input point cloud normals
  * \param cloud_surface the pointer to the surface point cloud
  * \param indices the pointer to the input point cloud indices
  */
template <typename PointInT, typename PointNT, typename PointOutT> void
  pcl_ros::FeatureFromNormals<PointInT, PointNT, PointOutT>::input_normals_surface_indices_callback (
      const PointCloud2::ConstPtr &cloud, const PointCloud2::ConstPtr &cloud_normals,
      const PointCloud2::ConstPtr &cloud_surface, const pcl::PointIndicesConstPtr &indices)
{
  if (pub_output_.getNumSubscribers () <= 0)
    return;

  if (!isValid (cloud) || !isValid (cloud_normals, "normals") || !isValid (cloud_surface, "surface") || !isValid (indices))
    return;

  NODELET_DEBUG ("[input_normals_surface_indices_callback]\n"
                 "                                                 - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.\n"
                 "                                                 - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.\n"
                 "                                                 - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.\n"
                 "                                                 - PointIndices with %zu values, stamp %f, and frame %s on topic %s received.",
                 cloud->width * cloud->height, pcl::getFieldsList (*cloud).c_str (), cloud->header.stamp.toSec (), cloud->header.frame_id.c_str (), pnh_->resolveName ("input").c_str (),
                 cloud_surface->width * cloud_surface->height, pcl::getFieldsList (*cloud_surface).c_str (), cloud_surface->header.stamp.toSec (), cloud_surface->header.frame_id.c_str (), pnh_->resolveName ("surface").c_str (),
                 cloud_normals->width * cloud_normals->height, pcl::getFieldsList (*cloud_normals).c_str (), cloud_normals->header.stamp.toSec (), cloud_normals->header.frame_id.c_str (), pnh_->resolveName ("normals").c_str (),
                 indices->indices.size (), indices->header.stamp.toSec (), indices->header.frame_id.c_str (), pnh_->resolveName ("indices").c_str ());

  if ((int)(cloud->width * cloud->height) < k_)
  {
    NODELET_ERROR ("[input_normals_surface_indices_callback] Requested number of k-nearest neighbors (%d) is larger than the PointCloud size (%d)!", k_, (int)(cloud->width * cloud->height));
    return;
  }
  PointCloudIn input;
  pcl::fromROSMsg (*cloud, input);                // input_ = cloud;
  input_ = boost::make_shared <const PointCloudIn> (input);

  PointCloudN normals;
  pcl::fromROSMsg (*cloud_normals, normals);      // normals_ = cloud_normals;
  normals_ = boost::make_shared <const PointCloudN> (normals);

  PointCloudIn surface;
  pcl::fromROSMsg (*cloud_surface, surface);      // surface_ = cloud_surf;
  surface_ = boost::make_shared <const PointCloudIn> (surface);

  indices_ = boost::make_shared <const std::vector<int> > (indices->indices);

  computeAndPublish ();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Input point cloud callback. Used when \a use_indices is set.
  * \param cloud the pointer to the input point cloud
  * \param cloud_normals the pointer to the input point cloud normals
  * \param indices the pointer to the input point cloud indices
  */
template <typename PointInT, typename PointNT, typename PointOutT> void
  pcl_ros::FeatureFromNormals<PointInT, PointNT, PointOutT>::input_normals_indices_callback (
      const PointCloud2::ConstPtr &cloud, const PointCloud2::ConstPtr &cloud_normals, const pcl::PointIndicesConstPtr &indices)
{
  if (pub_output_.getNumSubscribers () <= 0)
    return;

  if (!isValid (cloud) || !isValid (cloud_normals, "normals") || !isValid (indices))
    return;

  NODELET_DEBUG ("[input_normals_indices_callback]\n"
                 "                                         - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.\n"
                 "                                         - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.\n"
                 "                                         - PointIndices with %zu values, stamp %f, and frame %s on topic %s received.",
                 cloud->width * cloud->height, pcl::getFieldsList (*cloud).c_str (), cloud->header.stamp.toSec (), cloud->header.frame_id.c_str (), pnh_->resolveName ("input").c_str (),
                 cloud_normals->width * cloud_normals->height, pcl::getFieldsList (*cloud_normals).c_str (), cloud_normals->header.stamp.toSec (), cloud_normals->header.frame_id.c_str (), pnh_->resolveName ("normals").c_str (),
                 indices->indices.size (), indices->header.stamp.toSec (), indices->header.frame_id.c_str (), pnh_->resolveName ("indices").c_str ());

  if ((int)(cloud->width * cloud->height) < k_)
  {
    NODELET_ERROR ("[input_normals_indices_callback] Requested number of k-nearest neighbors (%d) is larger than the PointCloud size (%d)!", k_, (int)(cloud->width * cloud->height));
    return;
  }

  PointCloudIn input;
  pcl::fromROSMsg (*cloud, input);                 // input_ = cloud;
  input_ = boost::make_shared <const PointCloudIn> (input);

  PointCloudN normals;
  pcl::fromROSMsg (*cloud_normals, normals);       // normals_ = cloud_normals;
  normals_ = boost::make_shared <const PointCloudN> (normals);

  indices_ = boost::make_shared <const std::vector<int> > (indices->indices);

  computeAndPublish ();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Input point cloud callback. Used when \a use_surface is set.
  * \param cloud the pointer to the input point cloud
  * \param cloud_normals the pointer to the input point cloud normals
  * \param cloud_surface the pointer to the surface point cloud
  */
template <typename PointInT, typename PointNT, typename PointOutT> void
  pcl_ros::FeatureFromNormals<PointInT, PointNT, PointOutT>::input_normals_surface_callback (
      const PointCloud2::ConstPtr &cloud, const PointCloud2::ConstPtr &cloud_normals, const PointCloud2::ConstPtr &cloud_surface)
{
  if (pub_output_.getNumSubscribers () <= 0)
    return;

  if (!isValid (cloud) || !isValid (cloud_normals, "normals") || !isValid (cloud_surface, "surface"))
    return;

  NODELET_DEBUG ("[input_normals_surface_callback]\n"
                 "                                         - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.\n"
                 "                                         - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.\n"
                 "                                         - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.",
                 cloud->width * cloud->height, pcl::getFieldsList (*cloud).c_str (), cloud->header.stamp.toSec (), cloud->header.frame_id.c_str (), pnh_->resolveName ("input").c_str (),
                 cloud_surface->width * cloud_surface->height, pcl::getFieldsList (*cloud_surface).c_str (), cloud_surface->header.stamp.toSec (), cloud_surface->header.frame_id.c_str (), pnh_->resolveName ("surface").c_str (),
                 cloud_normals->width * cloud_normals->height, pcl::getFieldsList (*cloud_normals).c_str (), cloud_normals->header.stamp.toSec (), cloud_normals->header.frame_id.c_str (), pnh_->resolveName ("normals").c_str ());

  if ((int)(cloud->width * cloud->height) < k_)
  {
    NODELET_ERROR ("[input_normals_surface_callback] Requested number of k-nearest neighbors (%d) is larger than the PointCloud size (%d)!", k_, (int)(cloud->width * cloud->height));
    return;
  }

  PointCloudIn input;
  pcl::fromROSMsg (*cloud, input);                 // input_ = cloud;
  input_ = boost::make_shared <const PointCloudIn> (input);

  PointCloudN normals;
  pcl::fromROSMsg (*cloud_normals, normals);       // normals_ = cloud_normals;
  normals_ = boost::make_shared <const PointCloudN> (normals);

  PointCloudIn surface;
  pcl::fromROSMsg (*cloud_surface, surface);       // surface_ = cloud_surf;
  surface_ = boost::make_shared <const PointCloudIn> (surface);

  // Reset the indices pointer
  indices_.reset ();

  computeAndPublish ();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Input point cloud callback. Used only when \a use_indices and \a use_surface are not set.
  * \param cloud the pointer to the input point cloud
  * \param cloud_normals the pointer to the input point cloud normals
  */
template <typename PointInT, typename PointNT, typename PointOutT> void
  pcl_ros::FeatureFromNormals<PointInT, PointNT, PointOutT>::input_normals_callback (
      const PointCloud2::ConstPtr &cloud, const PointCloud2::ConstPtr &cloud_normals)
{
  if (pub_output_.getNumSubscribers () <= 0)
    return;

  if (!isValid (cloud) || !isValid (cloud_normals, "normals")) 
    return;

  NODELET_DEBUG ("[input_normals_callback]\n"
                 "                                 - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.\n"
                 "                                 - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.",
                 cloud->width * cloud->height, pcl::getFieldsList (*cloud).c_str (), cloud->header.stamp.toSec (), cloud->header.frame_id.c_str (), pnh_->resolveName ("input").c_str (),
                 cloud_normals->width * cloud_normals->height, pcl::getFieldsList (*cloud_normals).c_str (), cloud_normals->header.stamp.toSec (), cloud_normals->header.frame_id.c_str (), pnh_->resolveName ("normals").c_str ());

  if ((int)(cloud->width * cloud->height) < k_)
  {
    NODELET_ERROR ("[input_normals_callback] Requested number of k-nearest neighbors (%d) is larger than the PointCloud size (%d)!", k_, (int)(cloud->width * cloud->height));
    return;
  }

  // Clear the indices and the surface pointers
  indices_.reset ();
  surface_.reset ();

  PointCloudIn input;
  pcl::fromROSMsg (*cloud, input);              // input_ = cloud;
  input_ = boost::make_shared <const PointCloudIn> (input);

  PointCloudN normals;
  pcl::fromROSMsg (*cloud_normals, normals);   // normals_ = cloud_normals;
  normals_ = boost::make_shared <const PointCloudN> (normals);

  computeAndPublish ();
}

