/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: filter.cpp 32052 2010-08-27 02:19:30Z rusu $
 *
 */

/**

\author Radu Bogdan Rusu

**/

#include <pluginlib/class_list_macros.h>
#include <sensor_msgs/PointCloud2.h>
#include "pcl/point_types.h"
#include "pcl_ros/filters/filter.h"
#include "pcl_ros/filters/passthrough.h"
#include "pcl_ros/filters/extract_indices.h"

#include "pcl_ros/filters/radius_outlier_removal.h"
#include "pcl_ros/filters/statistical_outlier_removal.h"
#include "pcl_ros/filters/project_inliers.h"
#include "pcl_ros/filters/voxel_grid.h"

/*//#include <pcl/filters/pixel_grid.h>
//#include <pcl/filters/filter_dimension.h>
*/

typedef pcl_ros::PassThrough<sensor_msgs::PointCloud2> PassThrough;
typedef pcl_ros::ExtractIndices<sensor_msgs::PointCloud2> ExtractIndices;
typedef pcl_ros::RadiusOutlierRemoval<sensor_msgs::PointCloud2> RadiusOutlierRemoval;
typedef pcl_ros::StatisticalOutlierRemoval<sensor_msgs::PointCloud2> StatisticalOutlierRemoval;
typedef pcl_ros::ProjectInliers<sensor_msgs::PointCloud2> ProjectInliers;
typedef pcl_ros::VoxelGrid<sensor_msgs::PointCloud2> VoxelGrid;

/*//typedef pcl::PixelGrid<sensor_msgs::PointCloud2> PixelGrid;
//typedef pcl::FilterDimension FilterDimension;
*/

PLUGINLIB_DECLARE_CLASS (pcl, PassThrough, PassThrough, nodelet::Nodelet);
PLUGINLIB_DECLARE_CLASS (pcl, ExtractIndices, ExtractIndices, nodelet::Nodelet);
PLUGINLIB_DECLARE_CLASS (pcl, RadiusOutlierRemoval, RadiusOutlierRemoval, nodelet::Nodelet);
PLUGINLIB_DECLARE_CLASS (pcl, StatisticalOutlierRemoval, StatisticalOutlierRemoval, nodelet::Nodelet);
PLUGINLIB_DECLARE_CLASS (pcl, ProjectInliers, ProjectInliers, nodelet::Nodelet);
PLUGINLIB_DECLARE_CLASS (pcl, VoxelGrid, VoxelGrid, nodelet::Nodelet);

/*//PLUGINLIB_DECLARE_CLASS (pcl, PixelGrid, PixelGrid, nodelet::Nodelet);
//PLUGINLIB_DECLARE_CLASS (pcl, FilterDimension, FilterDimension, nodelet::Nodelet);
*/

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Apply the filter and publish the results. Internal method. */
void
  pcl_ros::Filter<sensor_msgs::PointCloud2>::computeAndPublish ()
{
  PointCloud output;
  filter (output);

  if (output.width * output.height == 0 || output.data.size () == 0)
  {
    NODELET_WARN ("[computeAndPublish] Output PointCloud has no data points on topic %s (parameters might not be set correctly)!", pnh_->resolveName ("output").c_str ());
    return;
  }
  // Publish a boost shared ptr
  pub_output_.publish (boost::make_shared <const PointCloud> (output));
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Nodelet initialization routine. */
void
  pcl_ros::Filter<sensor_msgs::PointCloud2>::onInit ()
{
  // Call the super onInit ()
  PCLNodelet<sensor_msgs::PointCloud2>::onInit ();

  // Call the child's local init
  bool has_service = false;
  if (!child_init (*pnh_, has_service))
  {
    NODELET_ERROR ("[onInit] Initialization failed.");
    return;
  }

  pub_output_ = pnh_->advertise<sensor_msgs::PointCloud2> ("output", max_queue_size_);

  // Enable the dynamic reconfigure service
  if (!has_service)
  {
    srv_ = boost::make_shared <dynamic_reconfigure::Server<pcl_ros::FilterConfig> > (*pnh_);
    dynamic_reconfigure::Server<pcl_ros::FilterConfig>::CallbackType f =  boost::bind (&Filter<sensor_msgs::PointCloud2>::config_callback, this, _1, _2);
    srv_->setCallback (f);
  }

  // If we're supposed to look for PointIndices (indices)
  if (use_indices_)
  {
    // Subscribe to the input using a filter
    sub_input_filter_.subscribe (*pnh_, "input", max_queue_size_);
    sub_indices_filter_.subscribe (*pnh_, "indices", max_queue_size_);

    if (approximate_sync_)
    {
      sync_input_indices_a_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, pcl::PointIndices> > >(max_queue_size_);
      sync_input_indices_a_->connectInput (sub_input_filter_, sub_indices_filter_);
      sync_input_indices_a_->registerCallback (bind (&Filter<PointCloud2>::input_indices_callback, this, _1, _2));
    }
    else
    {
      sync_input_indices_e_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, pcl::PointIndices> > >(max_queue_size_);
      sync_input_indices_e_->connectInput (sub_input_filter_, sub_indices_filter_);
      sync_input_indices_e_->registerCallback (bind (&Filter<PointCloud2>::input_indices_callback, this, _1, _2));
    }
  }
  else
    // Subscribe in an old fashion to input only (no filters)
    sub_input_.subscribe (*pnh_, "input", max_queue_size_,  bind (&Filter<PointCloud2>::input_callback, this, _1));

  NODELET_DEBUG ("[onInit] Nodelet successfully created with the following parameters:\n"
                 " - filter_field_name : %s\n"
                 " - filter_limit_min  : %f\n"
                 " - filter_limit_max  : %f",
                 filter_field_name_.c_str (), filter_limit_min_, filter_limit_max_);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Dynamic reconfigure callback
  * \param config the config object
  * \param level the dynamic reconfigure level
  */
void
  pcl_ros::Filter<sensor_msgs::PointCloud2>::config_callback (pcl_ros::FilterConfig &config, uint32_t level)
{
  double filter_min, filter_max;
  getFilterLimits (filter_min, filter_max);

  if (filter_min != config.filter_limit_min)
  {
    filter_min = config.filter_limit_min;
    NODELET_DEBUG ("[config_callback] Setting the minimum filtering value a point will be considered from to: %f.", filter_min);
  }
  if (filter_max != config.filter_limit_max)
  {
    filter_max = config.filter_limit_max;
    NODELET_DEBUG ("[config_callback] Setting the maximum filtering value a point will be considered from to: %f.", filter_max);
  }

  setFilterLimits (filter_min, filter_max);

  std::string filter_field = getFilterFieldName ();
  if (filter_field != config.filter_field_name)
  {
    setFilterFieldName (config.filter_field_name);
    NODELET_DEBUG ("[config_callback] Setting the filter field name to: %s.", config.filter_field_name.c_str ());
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Input point cloud callback. Used only when \a use_indices is not set.
  * \param cloud the pointer to the input point cloud
  */
void
  pcl_ros::Filter<sensor_msgs::PointCloud2>::input_callback (const PointCloudConstPtr &cloud)
{
  if (pub_output_.getNumSubscribers () <= 0)
    return;
  
  if (!isValid (cloud))
    return;

  NODELET_DEBUG ("[input_callback] PointCloud with %d data points and frame %s on topic %s received.", cloud->width * cloud->height, cloud->header.frame_id.c_str (), pnh_->resolveName ("input").c_str ());

  // Need setInputCloud () here because we have to extract x/y/z
  setInputCloud (cloud);            // input_ = cloud;
  indices_.reset ();          // Reset the indices pointer
  computeAndPublish ();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Input point cloud callback. Used when \a use_indices is set.
  * \param cloud the pointer to the input point cloud
  * \param indices the pointer to the input point cloud indices
  */
void
  pcl_ros::Filter<sensor_msgs::PointCloud2>::input_indices_callback (const PointCloud2::ConstPtr &cloud,
                                                                     const PointIndicesConstPtr &indices)
{
  if (pub_output_.getNumSubscribers () <= 0)
    return;

  if (!isValid (cloud) || !isValid (indices))
    return;

  NODELET_DEBUG ("[input_indices_callback]\n"
                 "                                 - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.\n"
                 "                                 - PointIndices with %zu values, stamp %f, and frame %s on topic %s received.",
                 cloud->width * cloud->height, pcl::getFieldsList (*cloud).c_str (), cloud->header.stamp.toSec (), cloud->header.frame_id.c_str (), pnh_->resolveName ("input").c_str (),
                 indices->indices.size (), indices->header.stamp.toSec (), indices->header.frame_id.c_str (), pnh_->resolveName ("indices").c_str ());
  // Need setInputCloud () here because we have to extract x/y/z
  setInputCloud (cloud);      // input_ = cloud;
  indices_ = boost::make_shared <std::vector<int> > (indices->indices);
  computeAndPublish ();
}

