/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: pcd_io.cpp 32288 2010-09-02 04:38:52Z rusu $
 *
 */

/** \author Radu Bogdan Rusu */

#include <pcl_ros/io/pcd_io.h>

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Nodelet initialization routine. */
void
  pcl_ros::PCDReader::onInit ()
{
  ros::NodeHandle private_nh = getMTPrivateNodeHandle ();
  // Provide a latched topic
  ros::Publisher pub_output = private_nh.advertise<sensor_msgs::PointCloud2> ("output", max_queue_size_, true);

  private_nh.getParam ("publish_rate", publish_rate_);
  private_nh.getParam ("tf_frame", tf_frame_);

  NODELET_DEBUG ("[onInit] Nodelet successfully created with the following parameters:\n"
                 " - publish_rate : %f\n"
                 " - tf_frame     : %s",
                 publish_rate_, tf_frame_.c_str ());

  PointCloudPtr output_new;
  output_.reset (new PointCloud);
  output_new.reset (new PointCloud);

  // Wait in a loop until someone connects
  do
  {
    ROS_DEBUG_ONCE ("[onInit] Waiting for a client to connect...");
    ros::spinOnce ();
    ros::Duration (0.01).sleep ();
  }
  while (private_nh.ok () && pub_output.getNumSubscribers () == 0);

  std::string file_name;

  while (private_nh.ok ())
  {
    // Get the current filename parameter. If no filename set, loop
    if (!private_nh.getParam ("filename", file_name_) && file_name_.empty ())
    {
      ROS_ERROR_ONCE ("[onInit] Need a 'filename' parameter to be set before continuing!");
      ros::Duration (0.01).sleep ();
      ros::spinOnce ();
      continue;
    }

    // If the filename parameter holds a different value than the last one we read
    if (file_name_.compare (file_name) != 0 && !file_name_.empty ())
    {
      ROS_INFO ("[onInit] New file given: %s", file_name_.c_str ());
      file_name = file_name_;
      PointCloud cloud;
      if (!read (file_name_, cloud))
      {
        NODELET_ERROR ("[onInit] Error reading %s !", file_name_.c_str ());
        return;
      }
      output_.reset (new PointCloud (cloud));
      output_->header.stamp    = ros::Time::now ();
      output_->header.frame_id = tf_frame_;
    }

    // We do not publish empty data
    if (output_->data.size () == 0)
      continue;

    if (publish_rate_ == 0)
    {
      if (output_ != output_new)
      {
        NODELET_DEBUG ("Publishing data once (%d points) on topic %s in frame %s.", output_->width * output_->height, getMTPrivateNodeHandle ().resolveName ("output").c_str (), output_->header.frame_id.c_str ());
        pub_output.publish (output_);
        output_new = output_;
      }
      ros::Duration (0.01).sleep ();
    }
    else
    {
      NODELET_DEBUG ("Publishing data (%d points) on topic %s in frame %s.", output_->width * output_->height, getMTPrivateNodeHandle ().resolveName ("output").c_str (), output_->header.frame_id.c_str ());
      output_->header.stamp = ros::Time::now ();
      pub_output.publish (output_);

      ros::Duration (publish_rate_).sleep ();
    }

    ros::spinOnce ();
    // Update parameters from server
    private_nh.getParam ("publish_rate", publish_rate_);
    private_nh.getParam ("tf_frame", tf_frame_);
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Nodelet initialization routine. */
void
  pcl_ros::PCDWriter::onInit ()
{
  ros::NodeHandle private_nh = getMTPrivateNodeHandle ();
  sub_input_.subscribe (private_nh, "input", 1,  bind (&PCDWriter::input_callback, this, _1));
  // ---[ Optional parameters
  private_nh.getParam ("filename", file_name_);
  private_nh.getParam ("binary_mode", binary_mode_);

  NODELET_DEBUG ("[onInit] Nodelet successfully created with the following parameters:\n"
                 " - filename     : %s\n"
                 " - binary_mode  : %s", file_name_.c_str (), (binary_mode_) ? "true" : "false");

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Input point cloud callback.
  * \param cloud the pointer to the input point cloud
  */
void
  pcl_ros::PCDWriter::input_callback (const PointCloudConstPtr &cloud)
{
  if (!isValid (cloud))
    return;
  
  getMTPrivateNodeHandle ().getParam ("filename", file_name_);

  NODELET_DEBUG ("[input_callback] PointCloud with %d data points and frame %s on topic %s received.", cloud->width * cloud->height, cloud->header.frame_id.c_str (), getMTPrivateNodeHandle ().resolveName ("input").c_str ());
 
  std::string fname;
  if (file_name_.empty ())
    fname = boost::lexical_cast<std::string> (cloud->header.stamp.toSec ()) + ".pcd";
  else
    fname = file_name_;
  write (fname, cloud, binary_mode_);

  NODELET_DEBUG ("[input_callback] Data saved to %s", fname.c_str ());
}
