/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: extract_clusters.hpp 32052 2010-08-27 02:19:30Z rusu $
 *
 */

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Nodelet initialization routine. */
template <typename PointT> void
  pcl_ros::EuclideanClusterExtraction<PointT>::onInit ()
{
  // Call the super onInit ()
  PCLNodelet<PointT>::onInit ();

  // ---[ Mandatory parameters
  if (!pnh_->getParam ("cluster_tolerance", cluster_tolerance_))
  {
    NODELET_ERROR ("[onInit] Need a 'cluster_tolerance' parameter to be set before continuing!"); 
    return;
  }
  if (!pnh_->getParam ("spatial_locator", spatial_locator_))
  {
    NODELET_ERROR ("[onInit] Need a 'spatial_locator' parameter to be set before continuing!");
    return;
  }

  // ---[ Optional parameters
  pnh_->getParam ("max_clusters", max_clusters_);
  //private_nh.getParam ("use_indices", use_indices_);
  pnh_->getParam ("publish_indices", publish_indices_);

  if (publish_indices_)
    pub_output_ = pnh_->template advertise<pcl::PointIndices> ("output", max_queue_size_);
  else
    pub_output_ = pnh_->template advertise<PointCloud2> ("output", max_queue_size_);

  // Enable the dynamic reconfigure service
  srv_ = boost::make_shared <dynamic_reconfigure::Server<EuclideanClusterExtractionConfig> > (*pnh_);
  dynamic_reconfigure::Server<EuclideanClusterExtractionConfig>::CallbackType f =  boost::bind (&EuclideanClusterExtraction<PointT>::config_callback, this, _1, _2);
  srv_->setCallback (f);

  // If we're supposed to look for PointIndices (indices)
  if (use_indices_)
  {
    // Subscribe to the input using a filter
    sub_input_filter_.subscribe (*pnh_, "input", max_queue_size_);
    sub_indices_filter_.subscribe (*pnh_, "indices", max_queue_size_);

    if (approximate_sync_)
    {
      sync_input_indices_a_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, pcl::PointIndices> > > (max_queue_size_);
      sync_input_indices_a_->connectInput (sub_input_filter_, sub_indices_filter_);
      sync_input_indices_a_->registerCallback (bind (&EuclideanClusterExtraction<PointT>::input_indices_callback, this, _1, _2));
    }
    else
    {
      sync_input_indices_e_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, pcl::PointIndices> > > (max_queue_size_);
      sync_input_indices_e_->connectInput (sub_input_filter_, sub_indices_filter_);
      sync_input_indices_e_->registerCallback (bind (&EuclideanClusterExtraction<PointT>::input_indices_callback, this, _1, _2));
    }
  }
  else
    // Subscribe in an old fashion to input only (no filters)
    sub_input_.subscribe (*pnh_, "input", max_queue_size_,  bind (&EuclideanClusterExtraction<PointT>::input_callback, this, _1));

  NODELET_DEBUG ("[onInit] Nodelet successfully created with the following parameters:\n"
                 " - max_queue_size    : %d\n"
                 " - use_indices       : %s\n"
                 " - cluster_tolerance : %f\n"
                 " - spatial_locator   : %d\n"
                 " - max_clusters      : %d\n",
                 max_queue_size_,
                 (use_indices_) ? "true" : "false", cluster_tolerance_, spatial_locator_, max_clusters_);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Perform the segmentation and publish the results. Internal method. */
template <typename PointT> void
  pcl_ros::EuclideanClusterExtraction<PointT>::segmentAndPublish ()
{
  std::vector<pcl::PointIndices> clusters;
  extract (clusters);

  if (publish_indices_)
  {
    for (size_t i = 0; i < clusters.size (); ++i)
    {
      if ((int)i >= max_clusters_)
        break;
      // TODO: HACK!!! We need to change the PointCloud2 message to add for an incremental sequence ID number.
      clusters[i].header.stamp += ros::Duration (i * 0.001);
      pub_output_.publish (boost::make_shared<const pcl::PointIndices> (clusters[i]));
    }

    NODELET_DEBUG ("[segmentAndPublish] Published %zu clusters (PointIndices) on topic %s", clusters.size (), pnh_->resolveName ("output").c_str ());
  }
  else
  {
    for (size_t i = 0; i < clusters.size (); ++i)
    {
      if ((int)i >= max_clusters_)
        break;
      PointCloud output;
      copyPointCloud (*input_, clusters[i].indices, output);

      PointCloud2 output_blob;     // Convert from the templated output to the PointCloud blob
      pcl::toROSMsg (output, output_blob);
      // TODO: HACK!!! We need to change the PointCloud2 message to add for an incremental sequence ID number.
      output_blob.header.stamp += ros::Duration (i * 0.001);
      // Publish a Boost shared ptr const data
      pub_output_.publish (boost::make_shared <const PointCloud2> (output_blob));
      NODELET_DEBUG ("[segmentAndPublish] Published cluster %zu (with %zu values and stamp %f) on topic %s",
                     i, clusters[i].indices.size (), output_blob.header.stamp.toSec (), pnh_->resolveName ("output").c_str ());
    }
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Dynamic reconfigure callback
  * \param config the config object
  * \param level the dynamic reconfigure level
  */
template <typename PointT> void
  pcl_ros::EuclideanClusterExtraction<PointT>::config_callback (EuclideanClusterExtractionConfig &config, uint32_t level)
{
  if (cluster_tolerance_ != config.cluster_tolerance)
  {
    cluster_tolerance_ = config.cluster_tolerance;
    NODELET_DEBUG ("[config_callback] Setting new clustering tolerance to: %f.", cluster_tolerance_);
  }
  if (min_pts_per_cluster_ != config.cluster_min_size)
  {
    min_pts_per_cluster_ = config.cluster_min_size;
    NODELET_DEBUG ("[config_callback] Setting the minimum cluster size to: %d.", min_pts_per_cluster_);
  }
  if (max_pts_per_cluster_ != config.cluster_max_size)
  {
    max_pts_per_cluster_ = config.cluster_max_size;
    NODELET_DEBUG ("[config_callback] Setting the maximum cluster size to: %d.", max_pts_per_cluster_);
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Input point cloud callback. Used only when \a use_indices is not set.
  * \param cloud the pointer to the input point cloud
  */
template <typename PointT> void
  pcl_ros::EuclideanClusterExtraction<PointT>::input_callback (const PointCloudConstPtr &cloud)
{
  if (pub_output_.getNumSubscribers () <= 0)
    return;

  if (!isValid (cloud))
    return;

  NODELET_DEBUG ("[input_callback] PointCloud with %d data points, stamp %f, and frame %s on topic %s received.", cloud->width * cloud->height, cloud->header.stamp.toSec (), cloud->header.frame_id.c_str (), pnh_->resolveName ("input").c_str ());

  input_ = cloud;

  // Reset the indices pointer
  indices_.reset ();
  segmentAndPublish ();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Input point cloud callback. Used when \a use_indices is set.
  * \param cloud the pointer to the input point cloud
  * \param indices the pointer to the input point cloud indices
  */
template <typename PointT> void
  pcl_ros::EuclideanClusterExtraction<PointT>::input_indices_callback (
      const PointCloud2::ConstPtr &cloud, const pcl::PointIndicesConstPtr &indices)
{
  if (pub_output_.getNumSubscribers () <= 0)
    return;

  if (!isValid (cloud) || !isValid (indices)) 
    return;

  NODELET_DEBUG ("[input_indices_callback]\n"
                 "                                 - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.\n"
                 "                                 - PointIndices with %zu values, stamp %f, and frame %s on topic %s received.",
                 cloud->width * cloud->height, pcl::getFieldsList (*cloud).c_str (), cloud->header.stamp.toSec (), cloud->header.frame_id.c_str (), pnh_->resolveName ("input").c_str (),
                 indices->indices.size (), indices->header.stamp.toSec (), indices->header.frame_id.c_str (), pnh_->resolveName ("indices").c_str ());

  PointCloud input;
  pcl::fromROSMsg (*cloud, input);            // input_ = cloud;
  input_ = boost::make_shared <const PointCloud> (input);

  indices_ = boost::make_shared <std::vector<int> > (indices->indices);
  segmentAndPublish ();
}

