/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: extract_polygonal_prism_data.hpp 32939 2010-09-28 23:14:22Z rusu $
 *
 */

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
 /** \brief Nodelet initialization routine. */
template <typename PointT> void
  pcl_ros::ExtractPolygonalPrismData<PointT>::onInit ()
{
  // Call the super onInit ()
  PCLNodelet<PointT>::onInit ();

  sub_hull_filter_.subscribe (*pnh_, "planar_hull", max_queue_size_);
  sub_input_filter_.subscribe (*pnh_, "input", max_queue_size_);

  // Enable the dynamic reconfigure service
  srv_ = boost::make_shared <dynamic_reconfigure::Server<ExtractPolygonalPrismDataConfig> > (*pnh_);
  dynamic_reconfigure::Server<ExtractPolygonalPrismDataConfig>::CallbackType f = boost::bind (&ExtractPolygonalPrismData<PointT>::config_callback, this, _1, _2);
  srv_->setCallback (f);

  if (use_indices_)
  {
    sub_indices_filter_.subscribe (*pnh_, "indices", max_queue_size_);
    if (approximate_sync_)
    {
      sync_input_hull_indices_a_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<sensor_msgs::PointCloud2, sensor_msgs::PointCloud2, pcl::PointIndices> > > (max_queue_size_);
      sync_input_hull_indices_a_->connectInput (sub_input_filter_, sub_hull_filter_, sub_indices_filter_);
      sync_input_hull_indices_a_->registerCallback (bind (&ExtractPolygonalPrismData<PointT>::input_hull_indices_callback, this, _1, _2, _3));
    }
    else
    {
      sync_input_hull_indices_e_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ExactTime<sensor_msgs::PointCloud2, sensor_msgs::PointCloud2, pcl::PointIndices> > > (max_queue_size_);
      sync_input_hull_indices_e_->connectInput (sub_input_filter_, sub_hull_filter_, sub_indices_filter_);
      sync_input_hull_indices_e_->registerCallback (bind (&ExtractPolygonalPrismData<PointT>::input_hull_indices_callback, this, _1, _2, _3));
    }
  }
  else
  {
    if (approximate_sync_)
    {
      sync_input_hull_a_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<sensor_msgs::PointCloud2, sensor_msgs::PointCloud2> > > (max_queue_size_);
      sync_input_hull_a_->connectInput (sub_input_filter_, sub_hull_filter_);
      sync_input_hull_a_->registerCallback (bind (&ExtractPolygonalPrismData<PointT>::input_hull_callback, this, _1, _2));
    }
    else
    {
      sync_input_hull_e_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ExactTime<sensor_msgs::PointCloud2, sensor_msgs::PointCloud2> > > (max_queue_size_);
      sync_input_hull_e_->connectInput (sub_input_filter_, sub_hull_filter_);
      sync_input_hull_e_->registerCallback (bind (&ExtractPolygonalPrismData<PointT>::input_hull_callback, this, _1, _2));
    }
  }

  // Advertise the output topics
  pub_output_ = pnh_->template advertise<pcl::PointIndices> ("output", max_queue_size_);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Dynamic reconfigure callback
  * \param config the config object
  * \param level the dynamic reconfigure level
  */
template <typename PointT> void
  pcl_ros::ExtractPolygonalPrismData<PointT>::config_callback (ExtractPolygonalPrismDataConfig &config, uint32_t level)
{
  double height_min, height_max;
  getHeightLimits (height_min, height_max);
  if (height_min != config.height_min)
  {
    height_min = config.height_min;
    NODELET_DEBUG ("[config_callback] Setting new minimum height to the planar model to: %f.", height_min);
  }
  if (height_max != config.height_max)
  {
    height_max = config.height_max;
    NODELET_DEBUG ("[config_callback] Setting new maximum height to the planar model to: %f.", height_max);
  }
  setHeightLimits (height_min, height_max);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Input point cloud callback. Used when \a use_indices is set.
  * \param cloud the pointer to the input point cloud
  * \param hull the pointer to the planar hull point cloud
  */
template <typename PointT> void
  pcl_ros::ExtractPolygonalPrismData<PointT>::input_hull_callback (
      const sensor_msgs::PointCloud2ConstPtr &cloud, const sensor_msgs::PointCloud2ConstPtr &hull)
{
  if (pub_output_.getNumSubscribers () <= 0)
    return;

  if (!isValid (cloud) || !isValid (hull, "planar_hull")) 
    return;

  NODELET_DEBUG ("[input_hull_callback]\n"
                 "                                 - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.\n"
                 "                                 - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.",
                 cloud->width * cloud->height, pcl::getFieldsList (*cloud).c_str (), cloud->header.stamp.toSec (), cloud->header.frame_id.c_str (), pnh_->resolveName ("input").c_str (),
                 hull->width * hull->height, pcl::getFieldsList (*hull).c_str (), hull->header.stamp.toSec (), hull->header.frame_id.c_str (), pnh_->resolveName ("planar_hull").c_str ());

  PointCloud input;
  pcl::fromROSMsg (*cloud, input);            // input_ = cloud;
  input_ = boost::make_shared <const PointCloud> (input);
  indices_.reset ();      // Reset the indices pointer
  PointCloud planar_hull;
  pcl::fromROSMsg (*hull, planar_hull);

  if (input.header.frame_id != planar_hull.header.frame_id)
  {
    NODELET_DEBUG ("[input_hull_callback] Planar hull has a different TF frame (%s) than the input point cloud (%s)! Using TF to transform.", planar_hull.header.frame_id.c_str (), input.header.frame_id.c_str ());
    try
    {
      pcl::transformPointCloud (input.header.frame_id, planar_hull, planar_hull, tf_listener_);
    }
    catch (tf::ExtrapolationException &e)
    {
      NODELET_ERROR ("%s", e.what ());
    }
  }
  planar_hull_   = boost::make_shared <const PointCloud> (planar_hull);

  pcl::PointIndices inliers;
  segment (inliers);
  pub_output_.publish (boost::make_shared<const pcl::PointIndices> (inliers));
  NODELET_DEBUG ("[input_hull_callback] Publishing %zu indices.", inliers.indices.size ());
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Input point cloud callback. Used when \a use_indices is set.
  * \param cloud the pointer to the input point cloud
  * \param hull the pointer to the planar hull point cloud
  * \param indices the pointer to the input point cloud indices
  */
template <typename PointT> void
  pcl_ros::ExtractPolygonalPrismData<PointT>::input_hull_indices_callback (
      const sensor_msgs::PointCloud2ConstPtr &cloud, const sensor_msgs::PointCloud2ConstPtr &hull, const pcl::PointIndicesConstPtr &indices)
{
  if (pub_output_.getNumSubscribers () <= 0)
    return;

  if (!isValid (cloud) || !isValid (hull, "planar_hull") || !isValid (indices)) 
    return;

  NODELET_DEBUG ("[input_indices_hull_callback]\n"
                 "                                 - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.\n"
                 "                                 - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.\n"
                 "                                 - PointIndices with %zu values, stamp %f, and frame %s on topic %s received.",
                 cloud->width * cloud->height, pcl::getFieldsList (*cloud).c_str (), cloud->header.stamp.toSec (), cloud->header.frame_id.c_str (), pnh_->resolveName ("input").c_str (),
                 hull->width * hull->height, pcl::getFieldsList (*hull).c_str (), hull->header.stamp.toSec (), hull->header.frame_id.c_str (), pnh_->resolveName ("planar_hull").c_str (),
                 indices->indices.size (), indices->header.stamp.toSec (), indices->header.frame_id.c_str (), pnh_->resolveName ("indices").c_str ());

  PointCloud input;
  pcl::fromROSMsg (*cloud, input);            // input_ = cloud;
  input_   = boost::make_shared <const PointCloud> (input);
  indices_ = boost::make_shared <std::vector<int> > (indices->indices);
  PointCloud planar_hull;
  pcl::fromROSMsg (*hull, planar_hull);
  planar_hull_   = boost::make_shared <const PointCloud> (planar_hull);

  pcl::PointIndices inliers;
  segment (inliers);
  pub_output_.publish (boost::make_shared<const pcl::PointIndices> (inliers));
  NODELET_DEBUG ("[input_hull_callback] Publishing %zu indices.", inliers.indices.size ());
}

