/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: sac_segmentation.hpp 32882 2010-09-27 07:30:13Z rusu $
 *
 */

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Child initialization routine.
  * \param nh ROS node handle
  * \param has_service set to true if the child has a Dynamic Reconfigure service
  */
template <typename PointT> bool
  pcl_ros::SACSegmentation<PointT>::child_init (ros::NodeHandle &nh, bool &has_service)
{
  // If we're supposed to look for PointIndices (indices)
  if (use_indices_)
  {
    // Subscribe to the input using a filter
    sub_input_filter_.subscribe (nh, "input", max_queue_size_);
    sub_indices_filter_.subscribe (nh, "indices", max_queue_size_);

    if (approximate_sync_)
    {
      sync_input_indices_a_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, pcl::PointIndices> > > (max_queue_size_);
      sync_input_indices_a_->connectInput (sub_input_filter_, sub_indices_filter_);
      sync_input_indices_a_->registerCallback (bind (&SACSegmentation<PointT>::input_indices_callback, this, _1, _2));
    }
    else
    {
      sync_input_indices_e_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, pcl::PointIndices> > > (max_queue_size_);
      sync_input_indices_e_->connectInput (sub_input_filter_, sub_indices_filter_);
      sync_input_indices_e_->registerCallback (bind (&SACSegmentation<PointT>::input_indices_callback, this, _1, _2));
    }
  }
  else
    // Subscribe in an old fashion to input only (no filters)
    sub_input_.subscribe (nh, "input", max_queue_size_,  bind (&SACSegmentation<PointT>::input_callback, this, _1));
  has_service = false;
  return (true);
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Nodelet initialization routine. */
template <typename PointT> void
  pcl_ros::SACSegmentation<PointT>::onInit ()
{
  // Call the super onInit ()
  PCLNodelet<PointT>::onInit ();

  // Call the child's local init
  bool has_service = false;
  if (!child_init (*pnh_, has_service))
    return;

  // Advertise the output topics
  pub_indices_ = pnh_->template advertise<pcl::PointIndices> ("inliers", max_queue_size_);
  pub_model_   = pnh_->template advertise<pcl::ModelCoefficients> ("model", max_queue_size_);

  // ---[ Mandatory parameters
  if (!pnh_->getParam ("model_type", model_type_))
  {
    NODELET_ERROR ("[onInit] Need a 'model_type' parameter to be set before continuing!");
    return;
  }
  if (!pnh_->getParam ("distance_threshold", threshold_))
  {
    NODELET_ERROR ("[onInit] Need a 'distance_threshold' parameter to be set before continuing!");
    return;
  }

  // ---[ Optional parameters
  pnh_->getParam ("method_type", method_type_);

  pnh_->getParam ("eps_angle", eps_angle_);
  // Convert from degrees to radians
  //eps_angle_ = eps_angle_ * M_PI / 180.0;

  XmlRpc::XmlRpcValue axis;
  pnh_->getParam ("axis", axis);
  switch (axis.getType ())
  {
    case XmlRpc::XmlRpcValue::TypeArray:
    {
      if (axis.size () != 3)
      {
        NODELET_ERROR ("[onInit] Parameter 'axis' given but with a different number of values (%d) than required (3)!", axis.size ());
        return;
      }
      for (int i = 0; i < 3; ++i)
      {
        if (axis[i].getType () != XmlRpc::XmlRpcValue::TypeDouble)
        {
          NODELET_ERROR ("[onInit] Need floating point values for 'axis' parameter.");
          return;
        }
        double value = axis[i]; axis_[i] = value;
      }
      break;
    }
    default:
    {
      axis_ = Eigen3::Vector3f::Zero ();
      break;
    }
  }

  // Initialize the random number generator
  srand (time (0));

  // Enable the dynamic reconfigure service
  if (!has_service)
  {
    srv_ = boost::shared_ptr<dynamic_reconfigure::Server<SACSegmentationConfig> >(new dynamic_reconfigure::Server<SACSegmentationConfig> (*pnh_));
    //srv_ = boost::shared_ptr (new dynamic_reconfigure::Server<SACSegmentationConfig> (pnh__));
    dynamic_reconfigure::Server<SACSegmentationConfig>::CallbackType f =  boost::bind (&SACSegmentation<PointT>::config_callback, this, _1, _2);
    srv_->setCallback (f);
  }

  NODELET_DEBUG ("[onInit] Nodelet successfully created with the following parameters:\n"
                 " - model_type               : %d\n"
                 " - method_type              : %d\n"
                 " - model_threshold          : %f\n"
                 " - axis                     : [%f, %f, %f]\n"
                 " - eps_angle                : %f\n",
                 model_type_, method_type_, threshold_, 
                 axis_[0], axis_[1], axis_[2], eps_angle_);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Dynamic reconfigure callback
  * \param config the config object
  * \param level the dynamic reconfigure level
  */
template <typename PointT> void
  pcl_ros::SACSegmentation<PointT>::config_callback (SACSegmentationConfig &config, uint32_t level)
{
  if (threshold_ != config.distance_threshold)
  {
    threshold_ = config.distance_threshold;
    NODELET_DEBUG ("[config_callback] Setting new distance to model threshold to: %f.", threshold_);
    //sac_->setDistanceThreshold (threshold_); - done in initSAC
  }
  if (max_iterations_ != config.max_iterations)
  {
    max_iterations_ = config.max_iterations;
    NODELET_DEBUG ("[config_callback] Setting new maximum number of iterations to: %d.",  max_iterations_);
    //sac_->setMaxIterations (max_iterations_); - done in initSAC
  }
  if (probability_ != config.probability)
  {
    probability_ = config.probability;
    NODELET_DEBUG ("[config_callback] Setting new probability to: %f.", probability_);
    //sac_->setProbability (probability_); - done in initSAC
  }
  if (optimize_coefficients_ != config.optimize_coefficients)
  {
    optimize_coefficients_ = config.optimize_coefficients;
    NODELET_DEBUG ("[config_callback] Setting coefficient optimization to: %s.", (optimize_coefficients_) ? "true" : "false");
  }

  double radius_min, radius_max;
  getRadiusLimits (radius_min, radius_max);
  if (radius_min != config.radius_min)
  {
    radius_min = config.radius_min;
    NODELET_DEBUG ("[config_callback] Setting minimum allowable model radius to: %f.", radius_min);
  }
  if (radius_max != config.radius_max)
  {
    radius_max = config.radius_max;
    NODELET_DEBUG ("[config_callback] Setting maximum allowable model radius to: %f.", radius_max);
  }
  setRadiusLimits (radius_min, radius_max);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Input point cloud callback. Used only when \a use_indices is not set.
  * \param cloud the pointer to the input point cloud
  */
template <typename PointT> void
  pcl_ros::SACSegmentation<PointT>::input_callback (const PointCloudConstPtr &cloud)
{
  if (pub_indices_.getNumSubscribers () <= 0 && pub_model_.getNumSubscribers () <= 0)
    return;

  if (!isValid (cloud)) 
    return;

  NODELET_DEBUG ("[input_callback] PointCloud with %d data points, stamp %f, and frame %s on topic %s received.", cloud->width * cloud->height, cloud->header.stamp.toSec (), cloud->header.frame_id.c_str (), pnh_->resolveName ("input").c_str ());

  input_ = cloud;

  // Reset the indices pointer
  indices_.reset ();

  pcl::PointIndices inliers;
  pcl::ModelCoefficients model;
  segment (inliers, model);

  if (inliers.indices.size () == 0)
    return;

  pub_indices_.publish (boost::make_shared<const pcl::PointIndices> (inliers));
  pub_model_.publish (boost::make_shared<const pcl::ModelCoefficients> (model));
  NODELET_DEBUG ("[input_callback] Published PointIndices with %zu values on topic %s, and ModelCoefficients with %zu values on topic %s",
                 inliers.indices.size (), pnh_->resolveName ("inliers").c_str (),
                 model.values.size (), pnh_->resolveName ("model").c_str ());
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Input point cloud callback. Used when \a use_indices is set.
  * \param cloud the pointer to the input point cloud
  * \param indices the pointer to the input point cloud indices
  */
template <typename PointT> void
  pcl_ros::SACSegmentation<PointT>::input_indices_callback (
      const PointCloud2::ConstPtr &cloud, const pcl::PointIndicesConstPtr &indices)
{
  if (pub_indices_.getNumSubscribers () <= 0 && pub_model_.getNumSubscribers () <= 0)
    return;

  if (!isValid (cloud) || !isValid (indices)) 
    return;

  NODELET_DEBUG ("[input_indices_callback]\n"
                 "                                 - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.\n"
                 "                                 - PointIndices with %zu values, stamp %f, and frame %s on topic %s received.",
                 cloud->width * cloud->height, pcl::getFieldsList (*cloud).c_str (), cloud->header.stamp.toSec (), cloud->header.frame_id.c_str (), pnh_->resolveName ("input").c_str (),
                 indices->indices.size (), indices->header.stamp.toSec (), indices->header.frame_id.c_str (), pnh_->resolveName ("indices").c_str ());

  PointCloud input;
  pcl::fromROSMsg (*cloud, input);            // input_ = cloud;
  input_ = boost::make_shared <const PointCloud> (input);

  indices_ = boost::make_shared <std::vector<int> > (indices->indices);

  pcl::PointIndices inliers;
  pcl::ModelCoefficients model;
  segment (inliers, model);

  if (inliers.indices.size () == 0)
    return;

  pub_indices_.publish (boost::make_shared<const pcl::PointIndices> (inliers));
  pub_model_.publish (boost::make_shared<const pcl::ModelCoefficients> (model));
  NODELET_DEBUG ("[input_callback] Published PointIndices with %zu values on topic %s, and ModelCoefficients with %zu values on topic %s",
                 inliers.indices.size (), pnh_->resolveName ("inliers").c_str (),
                 model.values.size (), pnh_->resolveName ("model").c_str ());
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Child initialization routine.
  * \param nh ROS node handle
  * \param has_service set to true if the child has a Dynamic Reconfigure service
  */
template <typename PointT, typename PointNT> bool
  pcl_ros::SACSegmentationFromNormals<PointT, PointNT>::child_init (ros::NodeHandle &nh, bool &has_service)
{
//        nh.getParam ("normal_distance_weight", distance_weight_);

  // Enable the dynamic reconfigure service
  has_service = true;
  srv_ = boost::make_shared <dynamic_reconfigure::Server<SACSegmentationFromNormalsConfig> > (nh);
  dynamic_reconfigure::Server<SACSegmentationFromNormalsConfig>::CallbackType f = boost::bind (&SACSegmentationFromNormals<PointT, PointNT>::config_callback, this, _1, _2);
  srv_->setCallback (f);

  // Subscribe to the input and normals using filters
  sub_input_filter_.subscribe (nh, "input", max_queue_size_);
  sub_normals_filter_.subscribe (nh, "normals", max_queue_size_);

  // Subscribe to an axis direction along which the model search is to be constrained (the first 3 model coefficients will be checked)
  sub_axis_ = nh.subscribe ("axis", 1, &SACSegmentationFromNormals<PointT, PointNT>::axis_callback, this);

  // If we're supposed to look for PointIndices (indices)
  if (use_indices_)
  {
    // Subscribe to the input using a filter
    sub_indices_filter_.subscribe (nh, "indices", max_queue_size_);
    // surface not enabled, connect the input-indices duo and register
    if (approximate_sync_)
    {
      sync_input_normals_indices_a_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, PointCloud2, pcl::PointIndices> > > (max_queue_size_);
      sync_input_normals_indices_a_->connectInput (sub_input_filter_, sub_normals_filter_, sub_indices_filter_);
//!      sync_input_normals_indices_a_->registerCallback (bind (&SACSegmentationFromNormals<PointT, PointNT>::input_normals_indices_callback, this, _1, _2, _3));
    }
    else
    {
      sync_input_normals_indices_e_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, PointCloud2, pcl::PointIndices> > > (max_queue_size_);
      sync_input_normals_indices_e_->connectInput (sub_input_filter_, sub_normals_filter_, sub_indices_filter_);
//!      sync_input_normals_indices_e_->registerCallback (bind (&SACSegmentationFromNormals<PointT, PointNT>::input_normals_indices_callback, this, _1, _2, _3));
    }
  }
  else
  {
    if (approximate_sync_)
    {
      sync_input_normals_a_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, PointCloud2> > > (max_queue_size_);
      sync_input_normals_a_->connectInput (sub_input_filter_, sub_normals_filter_);
      sync_input_normals_a_->registerCallback (bind (&SACSegmentationFromNormals<PointT, PointNT>::input_normals_callback, this, _1, _2));
    }
    else
    {
      sync_input_normals_e_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, PointCloud2> > > (max_queue_size_);
      sync_input_normals_e_->connectInput (sub_input_filter_, sub_normals_filter_);
      sync_input_normals_e_->registerCallback (bind (&SACSegmentationFromNormals<PointT, PointNT>::input_normals_callback, this, _1, _2));
    }
  }
  NODELET_DEBUG ("[child_init] Nodelet successfully created with the following parameters:\n"
                 " - normal_distance_weight : %f", distance_weight_);
  return (true);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Dynamic reconfigure callback
  * \param config the config object
  * \param level the dynamic reconfigure level
  */
template <typename PointT, typename PointNT> void
  pcl_ros::SACSegmentationFromNormals<PointT, PointNT>::axis_callback (const pcl::ModelCoefficientsConstPtr &model)
{
  if (model->values.size () < 3)
  {
    NODELET_ERROR ("[axis_callback] Invalid axis direction / model coefficients with %zu values sent on %s!", 
                   model->values.size (), pnh_->resolveName ("axis").c_str ());
    return;
  }
  NODELET_DEBUG ("[axis_callback] Received axis direction: %f %f %f", model->values[0], model->values[1], model->values[2]);
  axis_[0] = model->values[0];
  axis_[1] = model->values[1];
  axis_[2] = model->values[2];
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Dynamic reconfigure callback
  * \param config the config object
  * \param level the dynamic reconfigure level
  */
template <typename PointT, typename PointNT> void
  pcl_ros::SACSegmentationFromNormals<PointT, PointNT>::config_callback (SACSegmentationFromNormalsConfig &config, uint32_t level)
{
  double threshold = getDistanceThreshold ();
  if (threshold != config.distance_threshold)
  {
    threshold = config.distance_threshold;
    NODELET_DEBUG ("[config_callback] Setting new distance to model threshold to: %f.", threshold);
    setDistanceThreshold (threshold);
  }

  int max_iterations = getMaxIterations ();
  if (max_iterations != config.max_iterations)
  {
    max_iterations = config.max_iterations;
    NODELET_DEBUG ("[config_callback] Setting new maximum number of iterations to: %d.", max_iterations);
    setMaxIterations (max_iterations);
  }

  double probability = getProbability ();
  if (probability != config.probability)
  {
    probability = config.probability;
    NODELET_DEBUG ("[config_callback] Setting new probability to: %f.", probability);
    setProbability (probability);
  }

  bool optimize_coefficients = getOptimizeCoefficients ();
  if (optimize_coefficients != config.optimize_coefficients)
  {
    optimize_coefficients = config.optimize_coefficients;
    NODELET_DEBUG ("[config_callback] Setting coefficient optimization to: %s.", (optimize_coefficients) ? "true" : "false");
    setOptimizeCoefficients (optimize_coefficients);
  }

  if (distance_weight_ != config.normal_distance_weight)
  {
    NODELET_DEBUG ("[config_callback] Setting new distance weight to: %f.", distance_weight_);
    distance_weight_ = config.normal_distance_weight;
  }

  double radius_min, radius_max;
  getRadiusLimits (radius_min, radius_max);
  if (radius_min != config.radius_min)
  {
    radius_min = config.radius_min;
    NODELET_DEBUG ("[config_callback] Setting minimum allowable model radius to: %f.", radius_min);
  }
  if (radius_max != config.radius_max)
  {
    radius_max = config.radius_max;
    NODELET_DEBUG ("[config_callback] Setting maximum allowable model radius to: %f.", radius_max);
  }
  setRadiusLimits (radius_min, radius_max);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Input point cloud callback. Used only when \a use_indices and \a use_surface are not set.
  * \param cloud the pointer to the input point cloud
  * \param cloud_normals the pointer to the input point cloud normals
  */
template <typename PointT, typename PointNT> void
  pcl_ros::SACSegmentationFromNormals<PointT, PointNT>::input_normals_callback (
      const PointCloud2::ConstPtr &cloud, const PointCloud2::ConstPtr &cloud_normals)
{
  if (pub_indices_.getNumSubscribers () <= 0 && pub_model_.getNumSubscribers () <= 0)
    return;

  if (!isValid (cloud) || !isValid (cloud_normals, "normals")) 
    return;

  NODELET_DEBUG ("[input_normals_callback]\n"
                 "                                 - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.\n"
                 "                                 - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.",
                 cloud->width * cloud->height, pcl::getFieldsList (*cloud).c_str (), cloud->header.stamp.toSec (), cloud->header.frame_id.c_str (), pnh_->resolveName ("input").c_str (),
                 cloud_normals->width * cloud_normals->height, pcl::getFieldsList (*cloud_normals).c_str (), cloud_normals->header.stamp.toSec (), cloud_normals->header.frame_id.c_str (), pnh_->resolveName ("normals").c_str ());

  // Extra checks for safety
  int cloud_nr_points         = cloud->width * cloud->height;
  int cloud_normals_nr_points = cloud_normals->width * cloud_normals->height;
  if (cloud_nr_points != cloud_normals_nr_points)
  {
    NODELET_ERROR ("[input_normals_callback] Number of points in the input dataset (%d) differs from the number of points in the normals (%d)!", cloud_nr_points, cloud_normals_nr_points);
    return;
  }

  PointCloud input;
  pcl::fromROSMsg (*cloud, input);                 // input_ = cloud;
  input_ = boost::make_shared <const PointCloud> (input);

  PointCloudN normals;
  pcl::fromROSMsg (*cloud_normals, normals);       // normals_ = cloud_normals;
  normals_ = boost::make_shared <const PointCloudN> (normals);

  // Reset the indices pointer
  indices_.reset ();

  pcl::PointIndices inliers;
  pcl::ModelCoefficients model;
  segment (inliers, model);

  if (inliers.indices.size () == 0)
    return;

  pub_indices_.publish (boost::make_shared<const pcl::PointIndices> (inliers));
  pub_model_.publish (boost::make_shared<const pcl::ModelCoefficients> (model));
  NODELET_DEBUG ("[input_normals_callback] Published PointIndices with %zu values on topic %s, and ModelCoefficients with %zu values on topic %s", 
                 inliers.indices.size (), pnh_->resolveName ("inliers").c_str (),
                 model.values.size (), pnh_->resolveName ("model").c_str ());

//        if ((int)(cloud->width * cloud->height) < k_)
//        {
//          ROS_ERROR ("[pcl::%s::input_normals_callback] Requested number of k-nearest neighbors (%d) is larger than the PointCloud size (%d)!", getName ().c_str (), k_, (int)(cloud->width * cloud->height));
//          return;
//        }
//
//        // Clear the indices and the surface pointers
//        indices_.reset ();
//        surface_.reset ();
//
//        PointCloudIn input;
//        pcl::fromROSMsg (*cloud, input);              // input_ = cloud;
//        input_ = boost::make_shared <PointCloudIn> (input);
//
//        PointCloudN normals;
//        pcl::fromROSMsg (*cloud_normals, normals);   // normals_ = cloud_normals;
//        normals_ = boost::make_shared <PointCloudN> (normals);
//
//        // Initialize the spatial locator
//        initTree (spatial_locator_, tree_);
//
//        // Estimate the feature
//        PointCloudOut output;
//        compute (output);
//
//        // Convert from the templated output to the PointCloud blob
//        PointCloud2 output_blob;
//        pcl::toROSMsg (output, output_blob);
//        // Publish a Boost shared ptr const data
//        pub_output_.publish (boost::make_shared <const PointCloud2> (output_blob));
}

