/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: convex_hull.hpp 32052 2010-08-27 02:19:30Z rusu $
 *
 */

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Reconstruct the surface and publish the results. Internal method. */
template <typename PointInT, typename PointOutT> void
  pcl_ros::ConvexHull2D<PointInT, PointOutT>::reconstructAndPublish ()
{
  // Estimate the feature
  PointCloudOut output;
  reconstruct (output);
  if (output.points.size () == 0)
  {
    NODELET_WARN ("[reconstructAndPublish] Output PointCloud has no data points on topic %s (parameters might not be set correctly)!", getMTPrivateNodeHandle ().resolveName ("output").c_str ());
    return;
  }

  // Convert from the templated output to the PointCloud blob
  PointCloud2 output_blob;
  pcl::toROSMsg (output, output_blob);
  // Publish a Boost shared ptr const data
  pub_output_.publish (boost::make_shared <const PointCloud2> (output_blob));
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Nodelet initialization routine. */
template <typename PointInT, typename PointOutT> void
  pcl_ros::ConvexHull2D<PointInT, PointOutT>::onInit ()
{
  ros::NodeHandle private_nh = getMTPrivateNodeHandle ();
  pub_output_ = private_nh.advertise<PointCloud2> ("output", max_queue_size_);

  // ---[ Optional parameters
  private_nh.getParam ("use_indices", use_indices_);

  // If we're supposed to look for PointIndices (indices)
  if (use_indices_)
  {
    // Subscribe to the input using a filter
    sub_input_filter_.subscribe (private_nh, "input", 1);
    // If indices are enabled, subscribe to the indices
    sub_indices_filter_.subscribe (private_nh, "indices", 1);

    if (approximate_sync_)
    {
      sync_input_indices_a_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ApproximateTime<PointCloud2, pcl::PointIndices> > >(max_queue_size_);
      // surface not enabled, connect the input-indices duo and register
      sync_input_indices_a_->connectInput (sub_input_filter_, sub_indices_filter_);
      sync_input_indices_a_->registerCallback (bind (&ConvexHull2D<PointInT, PointOutT>::input_indices_callback, this, _1, _2));
    }
    else
    {
      sync_input_indices_e_ = boost::make_shared <message_filters::Synchronizer<message_filters::sync_policies::ExactTime<PointCloud2, pcl::PointIndices> > >(max_queue_size_);
      // surface not enabled, connect the input-indices duo and register
      sync_input_indices_e_->connectInput (sub_input_filter_, sub_indices_filter_);
      sync_input_indices_e_->registerCallback (bind (&ConvexHull2D<PointInT, PointOutT>::input_indices_callback, this, _1, _2));
    }
  }
  else
    // Subscribe in an old fashion to input only (no filters)
    sub_input_.subscribe (private_nh, "input", 1,  bind (&ConvexHull2D<PointInT, PointOutT>::input_callback, this, _1));

  NODELET_DEBUG ("[onInit] Nodelet successfully created with the following parameters:\n"
                 " - use_indices    : %s",
                 (use_indices_) ? "true" : "false");
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Input point cloud callback. Used when \a use_indices is set.
  * \param cloud the pointer to the input point cloud
  */
template <typename PointInT, typename PointOutT> void
  pcl_ros::ConvexHull2D<PointInT, PointOutT>::input_callback (const PointCloudInConstPtr &cloud)
{
  if (pub_output_.getNumSubscribers () <= 0)
    return;

  if (!isValid (cloud)) 
    return;

  NODELET_DEBUG ("[input_callback] PointCloud with %d data points, stamp %f, and frame %s on topic %s received.", cloud->width * cloud->height, cloud->header.stamp.toSec (), cloud->header.frame_id.c_str (), getMTPrivateNodeHandle ().resolveName ("input").c_str ());

  input_ = cloud;

  reconstructAndPublish ();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Input point cloud callback.
  * \param cloud the pointer to the input point cloud
  * \param indices the pointer to the input point cloud indices
  */
template <typename PointInT, typename PointOutT> void
  pcl_ros::ConvexHull2D<PointInT, PointOutT>::input_indices_callback (
      const PointCloud2::ConstPtr &cloud, const pcl::PointIndicesConstPtr &indices)
{
  if (pub_output_.getNumSubscribers () <= 0)
    return;

  if (!isValid (cloud) || !isValid (indices, "indices")) 
    return;

  NODELET_DEBUG ("[input_indices_model_callback]\n"
                 "                                 - PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.\n"
                 "                                 - PointIndices with %zu values, stamp %f, and frame %s on topic %s received.",
                cloud->width * cloud->height, pcl::getFieldsList (*cloud).c_str (), cloud->header.stamp.toSec (), cloud->header.frame_id.c_str (), getMTPrivateNodeHandle ().resolveName ("input").c_str (),
                indices->indices.size (), indices->header.stamp.toSec (), indices->header.frame_id.c_str (), getMTPrivateNodeHandle ().resolveName ("indices").c_str ());

  // Reset the indices and surface pointers
  PointCloudIn input;
  pcl::fromROSMsg (*cloud, input);            // input_ = cloud;
  input_   = boost::make_shared <const PointCloudIn> (input);
  indices_ = boost::make_shared <std::vector<int> > (indices->indices);

  reconstructAndPublish ();
}
