/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 *
 */

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Transform a point cloud and rotate its normals using an Eigen transform.
  * \param cloud_in the input point cloud
  * \param cloud_out the input point cloud
  * \param transform a rigid transformation from tf
  * \note calls the Eigen version
  */
template <typename PointT> inline void
  pcl::transformPointCloudWithNormals (const pcl::PointCloud <PointT> &cloud_in,
                                       pcl::PointCloud <PointT> &cloud_out, const tf::Transform &transform)
{
  // Bullet (used by tf) and Eigen both store quaternions in x,y,z,w order, despite the ordering
  // of arguments in Eigen's constructor. We could use an Eigen Map to convert without copy, but
  // this only works if Bullet uses floats, that is if BT_USE_DOUBLE_PRECISION is not defined.
  // Rather that risking a mistake, we copy the quaternion, which is a small cost compared to
  // the conversion of the point cloud anyway. Idem for the origin.
  tf::Quaternion q = transform.getRotation ();
  Eigen3::Quaternionf rotation (q.w (), q.x (), q.y (), q.z ());       // internally stored as (x,y,z,w)
  tf::Vector3 v = transform.getOrigin ();
  Eigen3::Vector3f origin (v.x (), v.y (), v.z ());
  //    Eigen3::Translation3f translation(v);
  // Assemble an Eigen Transform
  //Eigen3::Transform3f t;
  //t = translation * rotation;
  transformPointCloudWithNormals (cloud_in, cloud_out, origin, rotation);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Apply a rigid transform defined by a 3D offset and a quaternion
  * \param cloud_in the input point cloud
  * \param cloud_out the input point cloud
  * \param transform a rigid transformation from tf
  * \note calls the Eigen version
  */
template <typename PointT> inline void
  pcl::transformPointCloud (const pcl::PointCloud <PointT> &cloud_in,
                            pcl::PointCloud <PointT> &cloud_out, const tf::Transform &transform)
{
  // Bullet (used by tf) and Eigen both store quaternions in x,y,z,w order, despite the ordering
  // of arguments in Eigen's constructor. We could use an Eigen Map to convert without copy, but
  // this only works if Bullet uses floats, that is if BT_USE_DOUBLE_PRECISION is not defined.
  // Rather that risking a mistake, we copy the quaternion, which is a small cost compared to
  // the conversion of the point cloud anyway. Idem for the origin.
  tf::Quaternion q = transform.getRotation ();
  Eigen3::Quaternionf rotation (q.w (), q.x (), q.y (), q.z ());       // internally stored as (x,y,z,w)
  tf::Vector3 v = transform.getOrigin ();
  Eigen3::Vector3f origin (v.x (), v.y (), v.z ());
  //    Eigen3::Translation3f translation(v);
  // Assemble an Eigen Transform
  //Eigen3::Transform3f t;
  //t = translation * rotation;
  transformPointCloud (cloud_in, cloud_out, origin, rotation);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Transforms a point cloud in a given target TF frame using a TransformListener
  * \param target_frame the target TF frame the point cloud should be transformed to
  * \param cloud_in the input point cloud
  * \param cloud_out the input point cloud
  * \param tf_listener a TF listener object
  */
template <typename PointT> inline void
  pcl::transformPointCloudWithNormals (const std::string &target_frame,
                                       const pcl::PointCloud <PointT> &cloud_in,
                                       pcl::PointCloud <PointT> &cloud_out, 
                                       const tf::TransformListener &tf_listener)
{
  tf::StampedTransform transform;
  tf_listener.lookupTransform (target_frame, cloud_in.header.frame_id, cloud_in.header.stamp, transform);
  transformPointCloudWithNormals (cloud_in, cloud_out, transform);
  cloud_out.header.frame_id = target_frame;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Transforms a point cloud in a given target TF frame using a TransformListener
  * \param target_frame the target TF frame the point cloud should be transformed to
  * \param cloud_in the input point cloud
  * \param cloud_out the input point cloud
  * \param tf_listener a TF listener object
  */
template <typename PointT> inline void
  pcl::transformPointCloud (const std::string &target_frame,
                            const pcl::PointCloud <PointT> &cloud_in,
                            pcl::PointCloud <PointT> &cloud_out, 
                            const tf::TransformListener &tf_listener)
{
  tf::StampedTransform transform;
  tf_listener.lookupTransform (target_frame, cloud_in.header.frame_id, cloud_in.header.stamp, transform);
  transformPointCloud (cloud_in, cloud_out, transform);
  cloud_out.header.frame_id = target_frame;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Transforms a point cloud in a given target TF frame using a TransformListener
  * \param target_frame the target TF frame the point cloud should be transformed to
  * \param target_time the target timestamp
  * \param cloud_in the input point cloud
  * \param fixed_frame fixed TF frame
  * \param cloud_out the input point cloud
  * \param tf_listener a TF listener object
  */
template <typename PointT> inline void
  pcl::transformPointCloud (const std::string &target_frame,
                            const ros::Time & target_time,
                            const pcl::PointCloud <PointT> &cloud_in,
                            const std::string &fixed_frame,
                            pcl::PointCloud <PointT> &cloud_out, 
                            const tf::TransformListener &tf_listener)
{
  tf::StampedTransform transform;
  tf_listener.lookupTransform (target_frame, target_time, cloud_in.header.frame_id, cloud_in.header.stamp, fixed_frame, transform);

  transformPointCloud (cloud_in, cloud_out, transform);
  cloud_out.header.frame_id = target_frame;
  cloud_out.header.stamp = target_time;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Transforms a point cloud in a given target TF frame using a TransformListener
  * \param target_frame the target TF frame the point cloud should be transformed to
  * \param target_time the target timestamp
  * \param cloud_in the input point cloud
  * \param fixed_frame fixed TF frame
  * \param cloud_out the input point cloud
  * \param tf_listener a TF listener object
  */
template <typename PointT> inline void
  pcl::transformPointCloudWithNormals (const std::string &target_frame,
                                       const ros::Time & target_time,
                                       const pcl::PointCloud <PointT> &cloud_in,
                                       const std::string &fixed_frame,
                                       pcl::PointCloud <PointT> &cloud_out, 
                                       const tf::TransformListener &tf_listener)
{
  tf::StampedTransform transform;
  tf_listener.lookupTransform (target_frame, target_time, cloud_in.header.frame_id, cloud_in.header.stamp, fixed_frame, transform);

  transformPointCloudWithNormals (cloud_in, cloud_out, transform);
  cloud_out.header.frame_id = target_frame;
  cloud_out.header.stamp = target_time;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Obtain the transformation matrix from TF into an Eigen form
  * \param bt the TF transformation
  * \param out_mat the Eigen transformation
  */
inline void
  pcl::transformAsMatrix (const tf::Transform& bt, Eigen3::Matrix4f &out_mat)
{
  double mv[12];
  bt.getBasis ().getOpenGLSubMatrix (mv);

  tf::Vector3 origin = bt.getOrigin ();

  out_mat (0, 0) = mv[0]; out_mat (0, 1) = mv[4]; out_mat (0, 2) = mv[8];
  out_mat (1, 0) = mv[1]; out_mat (1, 1) = mv[5]; out_mat (1, 2) = mv[9];
  out_mat (2, 0) = mv[2]; out_mat (2, 1) = mv[6]; out_mat (2, 2) = mv[10];
                                                                     
  out_mat (3, 0) = out_mat (3, 1) = out_mat (3, 2) = 0; out_mat (3, 3) = 1;
  out_mat (0, 3) = origin.x ();
  out_mat (1, 3) = origin.y ();
  out_mat (2, 3) = origin.z ();
}

