/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: segmentation_cylinder.cpp 30719 2010-07-09 20:28:41Z rusu $
 *
 */

/**

\author Radu Bogdan Rusu

@b segmentation_cylinder exemplifies how to run a Sample Consensus segmentation for cylindrical models.

To make the example a bit more practical, the following operations are applied to the input dataset (in order):
- data points further away than 1.5 meters are filtered
- surface normals at each point are estimated
- a planar model (describing the table in our demo dataset) is segmented and saved to disk
- a cylindrical model (describing the mug in our demo dataset) is segmented and saved to disk

**/

#include <pcl/ModelCoefficients.h>

#include <pcl/io/pcd_io.h>
#include <pcl/point_types.h>

#include <pcl/features/normal_3d.h>

#include <pcl/filters/extract_indices.h>
#include <pcl/filters/passthrough.h>

#include <pcl/sample_consensus/method_types.h>
#include <pcl/sample_consensus/model_types.h>
#include <pcl/segmentation/sac_segmentation.h>

typedef pcl::PointXYZ PointT;

/* ---[ */
int
  main (int argc, char** argv)
{
  // All the objects needed
  pcl::PCDReader reader;
  pcl::PassThrough<PointT> pass;
  pcl::NormalEstimation<PointT, pcl::Normal> ne;
  pcl::SACSegmentationFromNormals<PointT, pcl::Normal> seg;
  pcl::PCDWriter writer;
  pcl::ExtractIndices<PointT> extract;
  pcl::ExtractIndices<pcl::Normal> extract_normals;
  pcl::KdTreeANN<PointT>::Ptr tree = boost::make_shared<pcl::KdTreeANN<PointT> > ();

  // Datasets
  pcl::PointCloud<PointT> cloud;
  pcl::PointCloud<PointT> cloud_filtered;
  pcl::PointCloud<pcl::Normal> cloud_normals;
  pcl::ModelCoefficients coefficients_plane, coefficients_cylinder;
  pcl::PointIndices inliers_plane, inliers_cylinder;

  // Read in the cloud data
  reader.read ("table_scene_mug_stereo_textured.pcd", cloud);
  ROS_INFO ("PointCloud has: %d data points.", (int)cloud.points.size ());

  // Build a passthrough filter to remove spurious NaNs
  pass.setInputCloud (boost::make_shared<pcl::PointCloud<PointT> > (cloud));
  pass.setFilterFieldName ("z");
  pass.setFilterLimits (0, 1.5);
  pass.filter (cloud_filtered);
  ROS_INFO ("PointCloud after filtering has: %d data points.", (int)cloud_filtered.points.size ());

  // Estimate point normals
  ne.setSearchMethod (tree);
  ne.setInputCloud (boost::make_shared<pcl::PointCloud<PointT> >(cloud_filtered));
  ne.setKSearch (50);
  ne.compute (cloud_normals);

  // Create the segmentation object for the planar model and set all the parameters
  seg.setOptimizeCoefficients (true);
  seg.setModelType (pcl::SACMODEL_NORMAL_PLANE);
  seg.setNormalDistanceWeight (0.1);
  seg.setMethodType (pcl::SAC_RANSAC);
  seg.setMaxIterations (100);
  seg.setDistanceThreshold (0.03);
  seg.setInputCloud (boost::make_shared<pcl::PointCloud<PointT> >(cloud_filtered));
  seg.setInputNormals (boost::make_shared<pcl::PointCloud<pcl::Normal> >(cloud_normals));
  // Obtain the plane inliers and coefficients
  seg.segment (inliers_plane, coefficients_plane);
  std::cerr << "Plane coefficients: " << coefficients_plane << std::endl;

  // Extract the planar inliers from the input cloud
  extract.setInputCloud (boost::make_shared<pcl::PointCloud<PointT> > (cloud_filtered));
  extract.setIndices (boost::make_shared<pcl::PointIndices> (inliers_plane));
  extract.setNegative (false);

  // Write the planar inliers to disk
  pcl::PointCloud<PointT> cloud_plane;
  extract.filter (cloud_plane);
  ROS_INFO ("PointCloud representing the planar component: %d data points.", (int)cloud_plane.points.size ());
  writer.write ("table_scene_mug_stereo_textured_plane.pcd", cloud_plane, false);

  // Remove the planar inliers, extract the rest
  extract.setNegative (true);
  extract.filter (cloud_filtered);
  extract_normals.setNegative (true);
  extract_normals.setInputCloud (boost::make_shared<pcl::PointCloud<pcl::Normal> > (cloud_normals));
  extract_normals.setIndices (boost::make_shared<pcl::PointIndices> (inliers_plane));
  extract_normals.filter (cloud_normals);

  // Create the segmentation object for cylinder segmentation and set all the parameters
  seg.setOptimizeCoefficients (true);
  seg.setModelType (pcl::SACMODEL_CYLINDER);
  seg.setMethodType (pcl::SAC_RANSAC);
  seg.setNormalDistanceWeight (0.1);
  seg.setMaxIterations (10000);
  seg.setDistanceThreshold (0.05);
  seg.setRadiusLimits (0, 0.1);
  seg.setInputCloud (boost::make_shared<pcl::PointCloud<PointT> >(cloud_filtered));
  seg.setInputNormals (boost::make_shared<pcl::PointCloud<pcl::Normal> >(cloud_normals));

  // Obtain the cylinder inliers and coefficients
  seg.segment (inliers_cylinder, coefficients_cylinder);
  std::cerr << "Cylinder coefficients: " << coefficients_cylinder << std::endl;

  // Write the cylinder inliers to disk
  extract.setInputCloud (boost::make_shared<pcl::PointCloud<PointT> > (cloud_filtered));
  extract.setIndices (boost::make_shared<pcl::PointIndices> (inliers_cylinder));
  extract.setNegative (false);
  pcl::PointCloud<PointT> cloud_cylinder;
  extract.filter (cloud_cylinder);
  ROS_INFO ("PointCloud representing the cylindrical component: %d data points.", (int)cloud_cylinder.points.size ());
  writer.write ("table_scene_mug_stereo_textured_cylinder.pcd", cloud_cylinder, false);

  return (0);
}
/* ]--- */
