/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: pcl_point_cloud_handlers.cpp 32257 2010-09-01 05:25:25Z rusu $
 *
 */

#include <pcl_visualization/pcl_point_cloud_handlers.h>

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Obtain the actual color for the input dataset as vtk scalars.
  * \param scalars the resultant scalars containing the color for the input dataset
  */
void
  pcl_visualization::PointCloudColorHandlerCustom<sensor_msgs::PointCloud2>::getColor (vtkSmartPointer<vtkDataArray> &scalars) const
{
  if (!capable_)
    return;

  if (!scalars)
    scalars = vtkSmartPointer<vtkUnsignedCharArray>::New ();
  scalars->SetNumberOfComponents (3);
  
  // Get a random color
  unsigned char scalar[3] = {r_, g_, b_};

  // Color every point
  for (size_t cp = 0; cp < cloud_->width * cloud_->height; ++cp)
  {
    dynamic_cast<vtkUnsignedCharArray*>(&(*scalars))->InsertNextTupleValue (scalar);
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Obtain the actual color for the input dataset as vtk scalars.
  * \param scalars the resultant scalars containing the color for the input dataset
  */
void
  pcl_visualization::PointCloudColorHandlerRandom<sensor_msgs::PointCloud2>::getColor (vtkSmartPointer<vtkDataArray> &scalars) const
{
  if (!capable_)
    return;

  if (!scalars)
    scalars = vtkSmartPointer<vtkUnsignedCharArray>::New ();
  scalars->SetNumberOfComponents (3);
  
  // Get a random color
  double r, g, b;
  pcl_visualization::getRandomColors (r, g, b);

  unsigned char scalar[3] = {lrint(r*255.0), lrint(g*255.0), lrint(b*255.0)};

  // Color every point
  for (size_t cp = 0; cp < cloud_->width * cloud_->height; ++cp)
  {
    dynamic_cast<vtkUnsignedCharArray*>(&(*scalars))->InsertNextTupleValue (scalar);
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Constructor. */
pcl_visualization::PointCloudColorHandlerRGBField<sensor_msgs::PointCloud2>::PointCloudColorHandlerRGBField (
    const pcl_visualization::PointCloudColorHandler<sensor_msgs::PointCloud2>::PointCloud &cloud) : 
  pcl_visualization::PointCloudColorHandler<sensor_msgs::PointCloud2>::PointCloudColorHandler (cloud)
{
  // Handle the 24-bit packed RGB values
  field_idx_ = pcl::getFieldIndex (cloud, "rgb");
  if (field_idx_ != -1)
    capable_ = true;
  else
    capable_ = false;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Obtain the actual color for the input dataset as vtk scalars.
  * \param scalars the resultant scalars containing the color for the input dataset
  */
void 
  pcl_visualization::PointCloudColorHandlerRGBField<sensor_msgs::PointCloud2>::getColor (vtkSmartPointer<vtkDataArray> &scalars) const
{
  if (!capable_)
    return;

  if (!scalars)
    scalars = vtkSmartPointer<vtkUnsignedCharArray>::New ();
  scalars->SetNumberOfComponents (3);

  unsigned char scalar[3];
  float rgb_data;
  // Color every point
  for (size_t cp = 0; cp < cloud_->width * cloud_->height; ++cp)
  {
    int point_offset = cp * cloud_->point_step;
    // Copy the value at the specified field
    memcpy (&rgb_data, &cloud_->data[point_offset + cloud_->fields[field_idx_].offset], sizeof (float));
    if (!std::isfinite (rgb_data))
    {
      scalar[0] = scalar[1] = scalar[2] = 0;
      dynamic_cast<vtkUnsignedCharArray*>(&(*scalars))->InsertNextTupleValue (scalar);
      continue;
    }

    int rgb = *reinterpret_cast<int*>(&rgb_data);
    scalar[0] = ((rgb >> 16) & 0xff);
    scalar[1] = ((rgb >> 8) & 0xff);
    scalar[2] = (rgb & 0xff);
    dynamic_cast<vtkUnsignedCharArray*>(&(*scalars))->InsertNextTupleValue (scalar);
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Constructor. */
pcl_visualization::PointCloudColorHandlerGenericField<sensor_msgs::PointCloud2>::PointCloudColorHandlerGenericField (
    const pcl_visualization::PointCloudColorHandler<sensor_msgs::PointCloud2>::PointCloud &cloud, 
    const std::string &field_name) : pcl_visualization::PointCloudColorHandler<sensor_msgs::PointCloud2>::PointCloudColorHandler (cloud)
{
  field_name_ = field_name;
  field_idx_  = pcl::getFieldIndex (cloud, field_name);
  if (field_idx_ != -1)
    capable_ = true;
  else
    capable_ = false;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Obtain the actual color for the input dataset as vtk scalars.
  * \param scalars the resultant scalars containing the color for the input dataset
  */
void 
  pcl_visualization::PointCloudColorHandlerGenericField<sensor_msgs::PointCloud2>::getColor (vtkSmartPointer<vtkDataArray> &scalars) const
{
  if (!capable_)
    return;

  if (!scalars)
    scalars = vtkSmartPointer<vtkFloatArray>::New ();
  scalars->SetNumberOfComponents (1);

  float scalar[1];
  float field_data;
  for (size_t cp = 0; cp < cloud_->width * cloud_->height; ++cp)
  {
    int point_offset = cp * cloud_->point_step;
    // Copy the value at the specified field
    memcpy (&field_data, &cloud_->data[point_offset + cloud_->fields[field_idx_].offset], sizeof (float));
    if (!std::isfinite (field_data))
    {
      scalar[0] = 0;
      dynamic_cast<vtkFloatArray*>(&(*scalars))->InsertNextTupleValue (scalar);
      continue;
    }
    scalar[0] = field_data;
    dynamic_cast<vtkFloatArray*>(&(*scalars))->InsertNextTupleValue (scalar);
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Obtain the actual point geometry for the input dataset as a vtk pointset.
  * \param points the resultant geometry 
  */
void 
  pcl_visualization::PointCloudGeometryHandler<sensor_msgs::PointCloud2>::getGeometry (vtkSmartPointer<vtkPoints> &points) const
{
  if (!capable_)
    return;

  if (!points)
    points = vtkSmartPointer<vtkPoints>::New ();
  points->SetDataTypeToFloat ();
  int nr_points = cloud_->width * cloud_->height;
  points->SetNumberOfPoints (nr_points);

  // Add all points
  double p[3];
  float data;

  for (vtkIdType cp = 0; cp < nr_points; ++cp)
  {
    int point_offset = cp * cloud_->point_step;
    // Copy the value at the specified field
    memcpy (&data, &cloud_->data[point_offset + cloud_->fields[field_x_idx_].offset], sizeof (float));
    p[0] = data;

    memcpy (&data, &cloud_->data[point_offset + cloud_->fields[field_y_idx_].offset], sizeof (float));
    p[1] = data;

    memcpy (&data, &cloud_->data[point_offset + cloud_->fields[field_z_idx_].offset], sizeof (float));
    p[2] = data;

    points->SetPoint (cp, p);
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Constructor. */
pcl_visualization::PointCloudGeometryHandlerXYZ<sensor_msgs::PointCloud2>::PointCloudGeometryHandlerXYZ (const PointCloud &cloud)
: pcl_visualization::PointCloudGeometryHandler<sensor_msgs::PointCloud2>::PointCloudGeometryHandler (cloud)
{
  field_x_idx_ = pcl::getFieldIndex (cloud, "x");
  if (field_x_idx_ == -1)
    return;
  field_y_idx_ = pcl::getFieldIndex (cloud, "y");
  if (field_y_idx_ == -1)
    return;
  field_z_idx_ = pcl::getFieldIndex (cloud, "z");
  if (field_z_idx_ == -1)
    return;
  capable_ = true;
}



//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Constructor. */
pcl_visualization::PointCloudGeometryHandlerSurfaceNormal<sensor_msgs::PointCloud2>::PointCloudGeometryHandlerSurfaceNormal (const PointCloud &cloud) 
: pcl_visualization::PointCloudGeometryHandler<sensor_msgs::PointCloud2>::PointCloudGeometryHandler (cloud)
{
  field_x_idx_ = pcl::getFieldIndex (cloud, "normal_x");
  if (field_x_idx_ == -1)
    return;
  field_y_idx_ = pcl::getFieldIndex (cloud, "normal_y");
  if (field_y_idx_ == -1)
    return;
  field_z_idx_ = pcl::getFieldIndex (cloud, "normal_z");
  if (field_z_idx_ == -1)
    return;
  capable_ = true;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Constructor. */
pcl_visualization::PointCloudGeometryHandlerCustom<sensor_msgs::PointCloud2>::PointCloudGeometryHandlerCustom (
    const PointCloud &cloud, const std::string &x_field_name, const std::string &y_field_name, const std::string &z_field_name) 
: pcl_visualization::PointCloudGeometryHandler<sensor_msgs::PointCloud2>::PointCloudGeometryHandler (cloud)
{
  field_x_idx_ = pcl::getFieldIndex (cloud, x_field_name);
  if (field_x_idx_ == -1)
    return;
  field_y_idx_ = pcl::getFieldIndex (cloud, y_field_name);
  if (field_y_idx_ == -1)
    return;
  field_z_idx_ = pcl::getFieldIndex (cloud, z_field_name);
  if (field_z_idx_ == -1)
    return;
  field_name_ = x_field_name + y_field_name + z_field_name;
  capable_ = true;
}

