/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: pcl_point_cloud_handlers.hpp 32257 2010-09-01 05:25:25Z rusu $
 *
 */

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Obtain the actual color for the input dataset as vtk scalars.
  * \param scalars the resultant scalars containing the color for the input dataset
  */
template <typename PointT> void
  pcl_visualization::PointCloudColorHandlerCustom<PointT>::getColor (vtkSmartPointer<vtkDataArray> &scalars) const
{
  if (!capable_)
    return;

  if (!scalars)
    scalars = vtkSmartPointer<vtkUnsignedCharArray>::New ();
  scalars->SetNumberOfComponents (3);
  
  // Get a random color
  unsigned char scalar[3] = {r_, g_, b_};

  // Color every point
  for (size_t cp = 0; cp < cloud_->points.size (); ++cp)
  {
    dynamic_cast<vtkUnsignedCharArray*>(&(*scalars))->InsertNextTupleValue (scalar);
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Obtain the actual color for the input dataset as vtk scalars.
  * \param scalars the resultant scalars containing the color for the input dataset
  */
template <typename PointT> void
  pcl_visualization::PointCloudColorHandlerRandom<PointT>::getColor (vtkSmartPointer<vtkDataArray> &scalars) const
{
  if (!capable_)
    return;

  if (!scalars)
    scalars = vtkSmartPointer<vtkUnsignedCharArray>::New ();
  scalars->SetNumberOfComponents (3);
  
  // Get a random color
  double r, g, b;
  pcl_visualization::getRandomColors (r, g, b);

  unsigned char scalar[3] = {lrint(r*255.0), lrint(g*255.0), lrint(b*255.0)};

  // Color every point
  for (size_t cp = 0; cp < cloud_->points.size (); ++cp)
  {
    dynamic_cast<vtkUnsignedCharArray*>(&(*scalars))->InsertNextTupleValue (scalar);
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Constructor. */
template <typename PointT>
pcl_visualization::PointCloudColorHandlerRGBField<PointT>::PointCloudColorHandlerRGBField (const typename pcl_visualization::PointCloudColorHandler<PointT>::PointCloud &cloud) : 
  pcl_visualization::PointCloudColorHandler<PointT>::PointCloudColorHandler (cloud)
{
  // Handle the 24-bit packed RGB values
  field_idx_ = pcl::getFieldIndex (cloud, "rgb", fields_);
  if (field_idx_ != -1)
    capable_ = true;
  else
    capable_ = false;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Obtain the actual color for the input dataset as vtk scalars.
  * \param scalars the resultant scalars containing the color for the input dataset
  */
template <typename PointT> void 
  pcl_visualization::PointCloudColorHandlerRGBField<PointT>::getColor (vtkSmartPointer<vtkDataArray> &scalars) const
{
  if (!capable_)
    return;

  if (!scalars)
    scalars = vtkSmartPointer<vtkUnsignedCharArray>::New ();
  scalars->SetNumberOfComponents (3);

  unsigned char scalar[3];
  typedef typename pcl::traits::fieldList<PointT>::type FieldList;

  float rgb_data;
  // Color every point
  for (size_t cp = 0; cp < cloud_->points.size (); ++cp)
  {
    // Copy the value at the specified field
    uint8_t* pt_data = (uint8_t*)&cloud_->points[cp];
    memcpy (&rgb_data, pt_data + fields_[field_idx_].offset, sizeof (float));

    if (!std::isfinite (rgb_data))
    {
      scalar[0] = scalar[1] = scalar[2] = 0;
      dynamic_cast<vtkUnsignedCharArray*>(&(*scalars))->InsertNextTupleValue (scalar);
      continue;
    }

    int rgb = *reinterpret_cast<int*>(&rgb_data);
    scalar[0] = ((rgb >> 16) & 0xff);
    scalar[1] = ((rgb >> 8) & 0xff);
    scalar[2] = (rgb & 0xff);
    dynamic_cast<vtkUnsignedCharArray*>(&(*scalars))->InsertNextTupleValue (scalar);
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Constructor. */
template <typename PointT>
pcl_visualization::PointCloudColorHandlerGenericField<PointT>::PointCloudColorHandlerGenericField (const typename pcl_visualization::PointCloudColorHandler<PointT>::PointCloud &cloud, 
    const std::string &field_name) : pcl_visualization::PointCloudColorHandler<PointT>::PointCloudColorHandler (cloud)
{
  field_name_ = field_name;
  field_idx_  = pcl::getFieldIndex (cloud, field_name, fields_);
  if (field_idx_ != -1)
    capable_ = true;
  else
    capable_ = false;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Obtain the actual color for the input dataset as vtk scalars.
  * \param scalars the resultant scalars containing the color for the input dataset
  */
template <typename PointT> void 
  pcl_visualization::PointCloudColorHandlerGenericField<PointT>::getColor (vtkSmartPointer<vtkDataArray> &scalars) const
{
  if (!capable_)
    return;

  if (!scalars)
    scalars = vtkSmartPointer<vtkFloatArray>::New ();
  scalars->SetNumberOfComponents (1);

  float scalar[1];
  typedef typename pcl::traits::fieldList<PointT>::type FieldList;

  float field_data;
  for (size_t cp = 0; cp < cloud_->points.size (); ++cp)
  {
    uint8_t* pt_data = (uint8_t*)&cloud_->points[cp];
    memcpy (&field_data, pt_data + fields_[field_idx_].offset, sizeof (float));

    if (!std::isfinite (field_data))
    {
      scalar[0] = 0;
      dynamic_cast<vtkFloatArray*>(&(*scalars))->InsertNextTupleValue (scalar);
      continue;
    }

    scalar[0] = field_data;
    dynamic_cast<vtkFloatArray*>(&(*scalars))->InsertNextTupleValue (scalar);
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Constructor. */
template <typename PointT>
pcl_visualization::PointCloudGeometryHandlerXYZ<PointT>::PointCloudGeometryHandlerXYZ (const typename pcl_visualization::PointCloudGeometryHandler<PointT>::PointCloud &cloud) 
  : pcl_visualization::PointCloudGeometryHandler<PointT>::PointCloudGeometryHandler (cloud)
{
  field_x_idx_ = pcl::getFieldIndex (cloud, "x", fields_);
  if (field_x_idx_ == -1)
    return;
  field_y_idx_ = pcl::getFieldIndex (cloud, "y", fields_);
  if (field_y_idx_ == -1)
    return;
  field_z_idx_ = pcl::getFieldIndex (cloud, "z", fields_);
  if (field_z_idx_ == -1)
    return;
  capable_ = true;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Obtain the actual point geometry for the input dataset as a vtk pointset.
  * \param points the resultant geometry 
  */
template <typename PointT> void 
  pcl_visualization::PointCloudGeometryHandlerXYZ<PointT>::getGeometry (vtkSmartPointer<vtkPoints> &points) const
{
  if (!capable_)
    return;

  if (!points)
    points = vtkSmartPointer<vtkPoints>::New ();
  points->SetDataTypeToFloat ();
  points->SetNumberOfPoints (cloud_->points.size ());

  // Add all points
  double p[3];
  for (vtkIdType i = 0; i < (int)cloud_->points.size (); ++i)
  {
    p[0] = cloud_->points[i].x;
    p[1] = cloud_->points[i].y;
    p[2] = cloud_->points[i].z;

    points->SetPoint (i, p);
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Constructor. */
template <typename PointT>
pcl_visualization::PointCloudGeometryHandlerSurfaceNormal<PointT>::PointCloudGeometryHandlerSurfaceNormal (const typename pcl_visualization::PointCloudGeometryHandler<PointT>::PointCloud &cloud) 
  : pcl_visualization::PointCloudGeometryHandler<PointT>::PointCloudGeometryHandler (cloud)
{
  field_x_idx_ = pcl::getFieldIndex (cloud, "normal_x", fields_);
  if (field_x_idx_ == -1)
    return;
  field_y_idx_ = pcl::getFieldIndex (cloud, "normal_y", fields_);
  if (field_y_idx_ == -1)
    return;
  field_z_idx_ = pcl::getFieldIndex (cloud, "normal_z", fields_);
  if (field_z_idx_ == -1)
    return;
  capable_ = true;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Obtain the actual point geometry for the input dataset as a vtk pointset.
  * \param points the resultant geometry 
  */
template <typename PointT> void 
  pcl_visualization::PointCloudGeometryHandlerSurfaceNormal<PointT>::getGeometry (vtkSmartPointer<vtkPoints> &points) const
{
  if (!capable_)
    return;

  if (!points)
    points = vtkSmartPointer<vtkPoints>::New ();
  points->SetDataTypeToFloat ();
  points->SetNumberOfPoints (cloud_->points.size ());

  // Add all points
  double p[3];
  for (vtkIdType i = 0; i < (int)cloud_->points.size (); ++i)
  {
    p[0] = cloud_->points[i].normal[0];
    p[1] = cloud_->points[i].normal[1];
    p[2] = cloud_->points[i].normal[2];

    points->SetPoint (i, p);
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Constructor. */
template <typename PointT>
pcl_visualization::PointCloudGeometryHandlerCustom<PointT>::PointCloudGeometryHandlerCustom (const typename pcl_visualization::PointCloudGeometryHandler<PointT>::PointCloud &cloud,
    const std::string &x_field_name, const std::string &y_field_name, const std::string &z_field_name) : pcl_visualization::PointCloudGeometryHandler<PointT>::PointCloudGeometryHandler (cloud)
{
  field_x_idx_ = pcl::getFieldIndex (cloud, x_field_name, fields_);
  if (field_x_idx_ == -1)
    return;
  field_y_idx_ = pcl::getFieldIndex (cloud, y_field_name, fields_);
  if (field_y_idx_ == -1)
    return;
  field_z_idx_ = pcl::getFieldIndex (cloud, z_field_name, fields_);
  if (field_z_idx_ == -1)
    return;
  field_name_ = x_field_name + y_field_name + z_field_name;
  capable_ = true;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/** \brief Obtain the actual point geometry for the input dataset as a vtk pointset.
  * \param points the resultant geometry 
  */
template <typename PointT> void 
  pcl_visualization::PointCloudGeometryHandlerCustom<PointT>::getGeometry (vtkSmartPointer<vtkPoints> &points) const
{
  if (!capable_)
    return;

  if (!points)
    points = vtkSmartPointer<vtkPoints>::New ();
  points->SetDataTypeToFloat ();
  points->SetNumberOfPoints (cloud_->points.size ());

  float data;
  // Add all points
  double p[3];
  for (vtkIdType i = 0; i < (int)cloud_->points.size (); ++i)
  {
    // Copy the value at the specified field
    uint8_t* pt_data = (uint8_t*)&cloud_->points[i];
    memcpy (&data, pt_data + fields_[field_x_idx_].offset, sizeof (float));
    p[0] = data;

    memcpy (&data, pt_data + fields_[field_y_idx_].offset, sizeof (float));
    p[1] = data;

    memcpy (&data, pt_data + fields_[field_z_idx_].offset, sizeof (float));
    p[2] = data;

    points->SetPoint (i, p);
  }
}

